/* ============================= C MeatAxe ==================================
   File:        $Id: args.c,v 1.7 1998/06/25 19:02:32 mringe Exp $
   Comment:     Command line options and arguments.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <ctype.h>
#include <string.h>
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

char opt_char;
char opt_text[50];
char *opt_text_ptr;
int opt_ind = 0;

#if !defined(MTXBIN)
#define MTXBIN "/usr/local/mtx/bin"
#endif
#if !defined(MTXLIB)
#define MTXLIB "/usr/local/mtx/lib"
#endif

char MeatAxeBin[250] = MTXBIN;	/* The MeatAxe binary directory */
char MeatAxeLib[250] = MTXLIB;	/* The MeatAxe library directory */



/* --------------------------------------------------------------------------
  Local data
  --------------------------------------------------------------------------- */

static int origargc;
static char **origargv;
static const proginfo_t *pinfo;



/*{{{ --------------------------------------------------------------------------
!text general.cmdline
The {\MeatAxe} library provides a minimal framework for applications. All
{\MeatAxe} applications should use this framework to achieve a consistent
behaviour. The functions described above enable an application to process 
options and arguments on the command line. They implement also a rudimentary
on-line help and automatic processing of some common options like |-V|. 
Here is an example:
\begin{verbatim}
    static char *helptext[] = {
      "SYNTAX",
      "    dummy [-ab] [-x <First>-<Last>] <File> ...",
      "",
      "DESCRIPTION",
      "    This program does nothing.",
      NULL};
    static proginfo_t pinfo =
      { "dummy", "Do Nothing", "$Revision 1.16$", helptext };

    int main(int argc, char *argv[])
    {
	init_args(argc, argv, &pinfo);
	while (MtxGetOpt("abx:") != OPT_END)
	{   switch (opt_char)
	    {   
		case 'a': ignore_a(); break;
		case 'b': ignore_b(); break;
		case 'x':
		    first = MtxGetInt();
		    if (*opt_text_ptr++ != '-') error();
		    last = MtxGetInt();
		    break;
	    }
	}
	...
    }
\end{verbatim}
In this example, the \verb"-x" option takes an argument of the
form {\it first}--{\it last}. After the first \verb"MtxGetInt()"
|opt_text_ptr| points to the `-' and must be incremented
for the next |MtxGetInt()| to work.
-------------------------------------------------------------------------- }}}*/



static void SetOption(char *opt)
{
    char *name, *value;

    for (name = strtok(opt,","); name != NULL; name = strtok(NULL,","))
    {
	value = strchr(name,'=');
	if (value != NULL)
	    *value++ = 0;
	if (!strcmp(name,"new-word-generator"))
	{
	    extern int UseNewWordGenerator;
	    UseNewWordGenerator = 1;
	}
	else
	    MESSAGE(0,("Ignoring unknown option '%s'\n",name));
    }
}




/*{{{ --------------------------------------------------------------------------
!structure proginfo_t "Program information structure"
!section general.cmdline
!synopsis 
    typedef struct 
    { 
	char *name; 
	char *shortdesc; 
	char *rcsrev;
	char **helptext; 
    } proginfo_t;
!description
    This data structure is used to store information about the program.
    It is used by the command line parser, e.g., to display the help
    text.

    The meanings of the fields are as follows. \verb"name" is the program
    name, usually the same name as the executable file. \verb"shortdesc"
    is a one-line description of the program which will be included in the
    help text. \verb"rcsrev" should contain the RCS version number. If
    you don't use RCS, you can insert the version number by hand, e.g.
    \begin{verbatim}
	rcsversion = "$Revision: 2.11";
    \end{verbatim}
    \verb"helptext" is the help text, which will be displayed if the user
    uses the \verb"-help" option. See \verb"zgetopt()" for an example.
-------------------------------------------------------------------------- }}}*/



/*{{{ --------------------------------------------------------------------------
!function MtxSetCommandLine "Initialize the command line parser"
!section general.cmdline
!synopsis 
    void MtxSetCommandLine(int argc, char *argv[], const proginfo_t *pi);
!arg argc
    Argument count.
!arg argv
    Command line arguments.
!arg pi
    Pointer to |proginfo_t| structure with program information.
!description
    This function initializes the command line parser. This is usually done
    done at the beginning of the program's |main()| function. |argc| is the 
    number of arguments, and |argv| is an array of pointers to the arguments. 
    Note that |argv[0]| is always the name of the program itself. So, the 
    first argument is in |argv[1]|. |pi| is a pointer to a |proginfo_t| 
    structure containing general information about the program. This data 
    must be set up by the caller.

    |MtxSetCommandLine()| performs the following tasks:
    \begin{enumerate}
    \item If the first argument is `-help', the help text contained in |pi| is
	printed and the program terminates.
    \item The |MTXLIB| and |MTXBIN| environment variables are evaluated. If set,
	their values overwrite the default directory names that were chosen at 
	compile time.
    \item The command line parser is initialized. This is required for later
	calls of |MtxGetOpt()|.
    \end{enumerate}
!seealso proginfo_t @cmdline
-------------------------------------------------------------------------- }}}*/

void MtxSetCommandLine(int argc, char *argv[], const proginfo_t *pi)

{
    char hrev[200];
    char *c;

    origargc = argc;
    origargv = argv;
    pinfo = pi;
    sscanf(pi->rcsrev,"$Revision: %s",hrev);

    /* Set directories
       --------------- */
    if ((c = getenv("MTXBIN")) != NULL)
	strcpy(MeatAxeBin,c);
    if ((c = getenv("MTXLIB")) != NULL)
	strcpy(MeatAxeLib,c);

    /* Display help text
       ----------------- */
    if ((argc == 2 && !strcmp(argv[1],"help"))
	|| (argc >= 2 && !strcmp(argv[1],"-help")))
    {
	/* BUG: compiler warning without cast */
	const char **ht = (const char **) pi->helptext;

	printf("NAME\n    %s - %s\n    Revision %s\n\n",
	  pi->name,pi->shortdesc,hrev);
	if (ht != NULL)
	    for (; *ht != NULL; ++ht)
	   	 printf("%s\n",*ht);
	else
	    printf("Sorry, no help available\n");
	exit(EXIT_OK);
    }
}



/*{{{ --------------------------------------------------------------------------
!function MtxGetOpt "Get next option"
!section general.cmdline
!synopsis 
    extern char opt_char;
    extern char opt_text[];
    extern char *opt_text_ptr;
    extern int opt_ind;
    int MtxGetOpt(const char *pattern);
!arg pattern
    List of valid options (see below).
!return
    Next option or |OPT_END|.
!description
    This function gets the next option from the command line. Before using
    |MtxGetOpt()|, the command line parser must be initialized with 
    |MtxSetCommandLine()|.
    Then, the application may call |MtxGetOpt()| successively until the return
    value is |OPT_END|, indicating that there are no more command line options.

    \verb"pattern" is a list of valid options. Each option consists of one 
    character, optionally followed by a colon indicating that the option takes 
    an argument. The options |-Q|, |-V|, |-B|, and |-L| are handled 
    automatically. 
    These options need not be specified in |pattern|. In order to process these
    common options, the application should always call |MtxGetOpt()| even if 
    there are no options specific to the program.

    |MtxGetOpt()| returns the option character or the special value |OPT_END| 
    if there are no more options. In addition, some global variables are set 
    on return:
    \begin{itemize}
    \item
	|opt_char| contains the option character.
    \item
	|opt_ind| contains the index of the argument to be read
	next. For example, if |MtxGetOpt()| return |OPT_END|,
	|opt_ind| is the index of the first non-optional argument.
    \item
	|opt_text|, an array of characters, contains the argument
	(if present).
    \item
	|opt_text_ptr| points to |opt_text[0]|.
    \end{itemize}
    This setting enables \verb"MtxGetInt()" to access the argument
    immediately after \verb"MtxGetOpt()" returns.
!seealso @cmdline
-------------------------------------------------------------------------- }}}*/

int MtxGetOpt(const char *pattern)

{
    static char pat[50];
    static const char *c = NULL;
    char *d;

    strcpy(pat,pattern);
    strcat(pat,"o:QVT:L:B:");

    while (1)
    {
    	opt_char = 0;
    	opt_text[0] = 0;
    	opt_text_ptr = opt_text;
    	if (c == NULL)
    	{
	    if (++opt_ind >= origargc || *(c = origargv[opt_ind]) !=
		'-' || *c == 0)
	    	return OPT_END;		/* No more options */
	    c = origargv[opt_ind] + 1;
    	}
    	opt_char = *c;
    	for (d = pat; *d != 0 && *d != opt_char; ++d);
    	if (*d != opt_char)
	    MtxErrorExit("%s: %E\n",pinfo->name,ERR_BADUSAGE);
    	if (d[1] == ':')		/* Option needs an argument */
    	{
	    if (*++c == 0)
	    {
		if (++opt_ind >= origargc)
		    MtxErrorExit("%s: %E\n",pinfo->name,ERR_BADUSAGE);
		c = origargv[opt_ind];
	    }
	    strcpy(opt_text,c);
	    c = NULL;
    	}
    	else
    	{
	    if (*++c == 0) c = NULL;
    	}
    
        /* Check standard options
           ---------------------- */
        if (opt_char == 'Q')
	   msg_level = -1000;
        else if (opt_char == 'V')
	   ++msg_level;
        else if (opt_char == 'B')
	    strcpy(MeatAxeBin,opt_text);
        else if (opt_char == 'L')
	    strcpy(MeatAxeLib,opt_text);
        else if (opt_char == 'o')
	    SetOption(opt_text);
        else if (opt_char == 'T')
        {
	    long tl = MtxGetInt();
	    if (tl < 0)
		MtxErrorExit("-T: %E",ERR_OPTION);
	    STimeLimit(tl);
        }
        else
            return opt_char;
    }
    return OPT_END;
}


/*{{{ --------------------------------------------------------------------------
!function MtxGetInt "Get integer argument"
!section general.cmdline
!synopsis 
    long MtxGetInt();
!description
    This function is to read an integer argument which follows a command line 
    option. |MtxGetInt()| should be used immediately after |MtxGetOpt()|. 
    |opt_text_ptr| must point to a string of decimal digits. |MtxGetInt()| scans 
    this string, returns the integer value, and increments |opt_text_ptr|
    to the character after the last digit. If the argument is not a number in 
    decimal notation |MtxGetInt()| returns the value |GETINT_ERR|.
!seealso @cmdline
-------------------------------------------------------------------------- }}}*/

long MtxGetInt()

{   long k;

    if (!isdigit(*opt_text_ptr))
	return GETINT_ERR;
    for (k = 0; isdigit(*opt_text_ptr); ++opt_text_ptr)
    {   switch (*opt_text_ptr)
	{   case '0': k = k * 10 + 0; break;
	    case '1': k = k * 10 + 1; break;
	    case '2': k = k * 10 + 2; break;
	    case '3': k = k * 10 + 3; break;
	    case '4': k = k * 10 + 4; break;
	    case '5': k = k * 10 + 5; break;
	    case '6': k = k * 10 + 6; break;
	    case '7': k = k * 10 + 7; break;
	    case '8': k = k * 10 + 8; break;
	    case '9': k = k * 10 + 9; break;
	}
    }
    return k;
}


