/* ============================= C MeatAxe ==================================
   File:        $Id:  $
   Comment:     Program to build a basis reflecting semisimplicity.
   --------------------------------------------------------------------------
   Written by Markus Wiegelmann.
   Revised by Michael Ringe.
   (C) Copyright 1998  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"



/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

static char *helptext[] = {
"SYNTAX",
"    basbypeak <gens> <basis> [<stdgens>]",
"",
"FILES",
"    <gens>.1 ... <gens>.nbgen   i  generators of a semisimple representation",
"    <gens>.cfinfo               i  info-file after running 'mkpeak'",
"    <gens>1a.std.1 ...          i  standard generators of each composition",
"                                   factor produced by 'pwkond'",
"    <basis>                     o  basis reflecting the semisimplicity",
"    <basis>i                    o  inverse of <basis>",
"    <stdgens>.1 ...             o  generators in the new basis",
"",
"DESCRIPTION",
"    This is one of two programs for calculating a block-diagonal basis for a",
"    semisimple representation. It uses peakwords. The other one is 'symbas'",
"    and works by calculating primitive idempotents. If you are not able to",
"    run 'mkpeak' try to work with 'symbas'.", 
NULL};

static proginfo_t pinfo =
   { "basbypeak", "semisimplicity basis  - 1st way", "$Revision: 1.1$",
     helptext };

static Lat_Info Info;		    /* Data from .cfinfo file */
static char *Name = NULL;	    /* Module base name */
static char *BasisName = NULL;	    /* Basis name */
static char *StdGenName = NULL;	    /* Base name for transposed generators */
matrix_t *Gen[MAXGEN];		    /* Generators */




/* --------------------------------------------------------------------------
   Init() - Initialize everything

   Description:
     This function initializes global data and processes command line
     options and arguments.

   Arguments:
     <argc>: Number of arguments
     <argv>: Argument list
   -------------------------------------------------------------------------- */

static void Init(int argc, char *argv[])
{
    int i;

    MtxInit();

    /* Process command line options
       ---------------------------- */
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("")) != OPT_END)
    {
	switch (i)
	{
	    default:
		MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
	}
    }

    /* Process any remaining arguments
       ------------------------------- */
    if (argc - opt_ind < 2 || argc - opt_ind > 3)
	MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);

    Name = argv[opt_ind++];
    BasisName = argv[opt_ind++];
    if (opt_ind < argc)
    	StdGenName = argv[opt_ind++];

}


/* --------------------------------------------------------------------------
   ReadFiles() - Read input files

   Description:
     This function reads the .cfinfo file and the generators.

   Arguments:
     <argc>: Number of arguments
     <argv>: Argument list
   -------------------------------------------------------------------------- */

static void ReadFiles()

{
   int i;

    /* Read the .cfinfo file
       --------------------- */
    if (Lat_ReadInfo(&Info,Name) != 0)
	MtxErrorExit("Error reading %s.cfinfo",Name);

    /* Read the generators
       ------------------- */
    MESSAGE(1,("Reading generators\n"));
    for (i = 0; i < Info.NGen; ++i)
    {
	char fn[LAT_MAXBASENAME + 20];
        sprintf(fn,"%s.%d",Name,i+1);
        if ((Gen[i] = MatLoad(fn)) == NULL)
            MtxErrorExit("Cannot read %s: %E",fn,MtxErrNo);
    }
}



int main(int argc, char *argv[])
{
    MTXCONTEXT(main);
    long i, j;
    matrix_t  *word, *seed, *bas, *partbas;
    wgdata_t *rep;
    long bascount;

   /* Initializtion
      ------------- */
    Init(argc,argv);
    ReadFiles();


    rep = WgInit(Info.NGen,(const matrix_t **)Gen);
    bas = MatAlloc(Gen[0]->fl,Gen[0]->nor,Gen[0]->noc);
    bascount = 0;

    /* Main loop: for each constituent
       ------------------------------- */
    for (j = 0; j < Info.NCf; ++j)
    {
	/* Make the peak word and calculate its null-space
	   ----------------------------------------------- */
        if (Info.Cf[j].peakword == 0)
	    MtxErrorExit("0 is definitly not a peakword! Did you run mkpeak?");
        word = WgMakeWord(rep, Info.Cf[j].peakword);
        seed = MatNullSpace__(MatInsert(word,Info.Cf[j].peakpol));
        MatFree(word);

	/* Spin up the null-space
	   ---------------------- */
        partbas = SpinUp(seed,Info.NGen,(const matrix_t **)Gen,
	    SF_EACH|SF_COMBINE|SF_STD);
        MatFree(seed);
	MESSAGE(1,("%s%s: %ld basis vectors\n",Info.BaseName,
	    Lat_CfName(&Info,j),partbas->nor));


        /* Append the result to the basis
	   ------------------------------ */
	MatCopyRegion(bas,bascount+1,1,partbas,1,1,-1,-1);
	bascount += partbas->nor;
        MatFree(partbas);
    }
    MTX_VERIFY(bascount == bas->nor);

    /* Clean up and write output files
       ------------------------------- */
    WgFree(rep);
    MatSave(bas,BasisName);
    if (StdGenName != NULL)
    {
        matrix_t *basi;
	char fn[LAT_MAXBASENAME+20];
        MESSAGE(1,("Transposing generators\n"));
        basi = MatInv(bas);
        sprintf(fn,"%si", BasisName);
        MatSave(basi,fn);
        for( i = 0; i < Info.NGen; i++ )
        {
            matrix_t *stgen = MatDup(bas);
            MatMul(stgen, Gen[i]);
            MatMul(stgen, basi);
            sprintf(fn, "%s.%ld", StdGenName, i+1);
            MatSave(stgen, fn);
            MatFree(stgen);
        }
        MatFree(basi);
    }

    MatFree(bas);
    for (i = 0; i < Info.NGen; ++i)
        MatFree(Gen[i]);
    MtxCleanup();
    return 0;	
}
    

/*----------------------------------------------------------------------------*/



