/* ============================= C MeatAxe ==================================
   File:        $Id: berlekmp.c,v 1.8 1998/06/25 19:02:32 mringe Exp $
   Comment:     Berlekamp factorization
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


typedef struct { poly_t *p; long n; } factor_t;

/* ------------------------------------------------------------------
   factorsquarefree() - Squarefree factorization.

   This function factors a polynomial into square-free, but not
   necessarily irreducible, factors. The factors are stored in a
   list of <factor_t> structures. A pointer to this list is passed
   back to the caller. The list is terminated by a ``factor'' with
   p=NULL, n=0.

   Return: List of factors.
   ------------------------------------------------------------------ */

static factor_t *factorsquarefree(const poly_t *pol)

{
    long int e,j,k,ltmp,tdeg,exp;
    poly_t *t0, *t, *w, *v;
    FEL *tbuf, *t0buf;
    factor_t *list;
    int nfactors = 0;

    /* Initialize variables
       -------------------- */
    zsetfield(pol->fl);
    for (exp = 0, ltmp = zfl; ltmp % zchar == 0; ++exp, ltmp /= zchar);
    t0 = PolDup(pol);
    e = 1;

    /* Allocate the list of factors. The can be at most <pol->deg>
       factors, but we need one extra entry to terminate the list.
       ----------------------------------------------------------- */
    list = NALLOC(factor_t,pol->deg+1);


    /* Main loop
       --------- */
    while ( t0->deg > 0 )
    {
	poly_t *der = PolDerive(PolDup(t0));
        t = PolGcd(t0,der);
	PolFree(der);
        v = PolDivMod(t0,t);
        PolFree(t0); 
        for (k = 0; v->deg > 0; )
	{
	    poly_t *tmp;
	    if ( ++k % zchar == 0 )
	    {
		tmp = PolDivMod(t,v);
              	PolFree(t);
              	t = tmp;
	      	k++;
	    }
	    w = PolGcd(t,v);
	    list[nfactors].p = PolDivMod(v,w);
	    list[nfactors].n = e * k;
	    if (list[nfactors].p->deg > 0)
	       	++nfactors;			/* add to output */
	    else
	    	PolFree(list[nfactors].p);	/* discard const. */
            PolFree(v);
	    v = w;
	    tmp = PolDivMod(t,v);
            PolFree(t); 
            t = tmp; 
	} 
	PolFree(v);

	/* shrink the polynomial */
      	tdeg = t->deg;
      	e *= zchar;
      	if ( tdeg % zchar != 0 )
	    printf("error in t, degree not div. by prime \n");
      	t0 = PolAlloc(zfl,tdeg/zchar);
      	t0buf = t0->buf;
      	tbuf = t->buf;
      	for (j = t0->deg; j >= 0; --j)
	{
	    FEL el, el1;
	    long lexp;

	    el = *tbuf;
	    tbuf += zchar;
	    el1 = el;
	    /* this is a bug, it should apply the e-1 power of the
	        frobenius K. Lux 16th 11 1996 */

	    /* replace el1 by el1^(zchar^(exp-1)) */
	    for ( lexp = 0; lexp < exp-1; lexp++ )
	    {
		long n;
		/* replace el1 by el1^zchar */
		el = el1;
		for ( n = zchar-1;  0 < n;  n-- ) {
		    el1 = zmul(el1,el);
		}
	    }
	    *t0buf = el1;
	    ++t0buf;
	}
        PolFree(t); 
    }
    PolFree(t0);

    /* Terminate the list
       ------------------ */
    list[nfactors].p = NULL;
    list[nfactors].n = 0;
    return list;
}
		      


/* ------------------------------------------------------------------
   makekernel() - Determines the matrix of the frobenius and returns
   its nullspace.
   ------------------------------------------------------------------ */

static matrix_t *makekernel(const poly_t *pol)

{
    matrix_t *materg;
    PTR rowptr;
    FEL *xbuf, *pbuf = pol->buf;
    long pdeg = pol->deg;
    int k, xshift;
    long fl = pol->fl;

    materg = MatAlloc(fl,pdeg,pdeg);
    rowptr = materg->d;

    xbuf = NALLOC(FEL,pdeg+1);
    for (k = 0; k <= pdeg; ++k) 
	xbuf[k] = F_ZERO;
    xbuf[0] = F_ONE;

    for (k = 0; k < pdeg; ++k)
    {
	int l;
	for (l = 0; l < pdeg; ++l) 
	    zinsert(rowptr,l+1,xbuf[l]);
	zinsert(rowptr,k+1,zsub(xbuf[k],F_ONE));
	zadvance(&rowptr,(long)1);
        for (xshift = (int) fl; xshift > 0; )
	{
	    FEL f;
	    int d;

	    /* Find leading pos */
	    for (l = pdeg-1; xbuf[l] == F_ZERO && l >= 0; --l);

	    /* Shift left as much as possible */
	    if ((d = pdeg - l) > xshift) d = xshift;
	    for (; l >= 0; l--) xbuf[l+d] = xbuf[l];
	    for (l = d-1; l >= 0; --l) xbuf[l] = F_ZERO;
	    xshift -= d;
	    if (xbuf[pdeg] == F_ZERO) continue;

	    /* Reduce with pol */
	    f = zneg(zdiv(xbuf[pdeg],pbuf[pdeg]));
	    for (l = pdeg-1; l >= 0; --l)
		xbuf[l] = zadd(xbuf[l],zmul(pbuf[l],f));
	    xbuf[pdeg] = F_ZERO;
	}
    }
    free(xbuf);
    return MatNullSpace__(materg);
 } 


/* ------------------------------------------------------------------
   berlekamp() - Find the irreducible factors of a squarefree
   polynomial.
   ------------------------------------------------------------------ */

static poly_t **berlekamp(const poly_t *pol, const matrix_t  *kernel)

{
    MTXCONTEXT(berlekamp);
    poly_t **list, **list2;
    int nfactors;
    PTR vec = kernel->d;
    int i, j;
    poly_t *t;

    list = NALLOC(poly_t *,kernel->nor+1);
    list2 = NALLOC(poly_t *,kernel->nor+1);
    list[0] = PolDup(pol);
    nfactors = 1;
    t = PolAlloc(kernel->fl,kernel->noc-1);
    zsetlen(kernel->noc);

    /* Loop through all kernel vectors */
    for (j = 2; j <= kernel->nor; ++j)
    {
	int ngcd = 0;
	zadvance(&vec,1);			/* Next vector */
	if (nfactors == kernel->nor) 
	    break;	/* Done? */
	for (i = 1; i <= kernel->noc; ++i)
	    t->buf[i-1] = zextract(vec,i);
	for (i = kernel->noc-1; t->buf[i] == F_ZERO; --i);
	t->deg = i;
	for (i = 0; i < nfactors; )
	{
	    long s;
	    if (list[i]->deg <= 1) {++i; continue;}
	    for (s = 0; s < zfl; ++s)
	    {
		poly_t *gcd;

		t->buf[0] = zitof(s);
		gcd = PolGcd(list[i],t);
		if (gcd->deg >= 1)
		{
		    list2[ngcd++] = gcd;
		}
		else
		    PolFree(gcd);
		if (nfactors == kernel->nor) break;	/* Done? */
	    }
	    if (ngcd > 0)
	    {
		int p;
		PolFree(list[i]);
		for (p = i; p < nfactors -1; ++p)
		    list[p] = list[p+1];
		--nfactors;
	    }
	    else
		++i;
	    if (nfactors == kernel->nor) break;		/* Done? */
	}
	if (ngcd > 0)
	{
	    int p;
	    for (p = 0; p < ngcd; ++p)
	    {
	        list[nfactors++] = list2[p];
	    }
	}
    }
    PolFree(t);
    free(list2);

    list[nfactors] = NULL;	/* Terminate the list */
    MTX_VERIFY(nfactors == kernel->nor);
    return list;
}




   
/*{{{ --------------------------------------------------------------------------
!function PolFactorization() "Factor a polynomial"
!section pol.op
!synopsis 
    fpoly_t *PolFactorization(const poly_t *pol);
!arg pol
    Polynomial to factor.
!return
    The factorization of |pol| or |NULL| on error.
!description
    This function decomposes a polynomial into irreducible factors using the
    Berlekamp algorithm.
!seealso
-------------------------------------------------------------------------- }}}*/

fpoly_t *PolFactorization(const poly_t *pol)

{
    MTXCONTEXT(PolFactorization);
    factor_t *list, *l;
    fpoly_t *factors;    

    /* Allocate result
       --------------- */
    if ((factors = FpolAlloc()) == NULL)
    {
	MTXERROR("Cannot allocate result");
	return NULL;
    }

    /* Step 1: Squarefree factorization
       -------------------------------- */
    if ((list = factorsquarefree(pol)) == NULL)
    {
	MTXERROR("Squarefree factorization failed");
	return NULL;
    }

    /* Step 2: Decompose the squarefree factors using Berlekamp's algorithm
       -------------------------------------------------------------------- */
    for (l = list; l->p != NULL; ++l)
    {
	matrix_t *kernel;
	poly_t **irr, **i;

	kernel = makekernel(l->p);
	if ((irr = berlekamp(l->p,kernel)) == NULL)
	{
	    MTXERROR("Berlekamp factorization failed");
	    return NULL;
	}
	MatFree(kernel);
	for (i = irr; *i != NULL; ++i)
	{
	    FpolMulP(factors,*i,l->n);
	    PolFree(*i);
	}

	/* Clean up
	   -------- */
	free(irr);
	PolFree(l->p);
    }

    /* Clean up
       -------- */
    free(list);
    return factors;
}








#if 0

poly_t *p, *q, *r;
factor_t *f, *ff;

int main(void)
{
    int i;
    int count = 0;

    p = PolAlloc(5,20);
    while (1)
    {
	if (++count % 100 == 0)
	{
	    printf("%d\n",count);
	    fflush(stdout);
	}
	memset(p->buf,0,21);
	p->buf[20] = F_ONE;
	for (i = MtxRandomInt(20); i > 0; --i)
	    p->buf[RandInt(20)] = zitof(MtxRandomInt(5));
	/*polprint("p",p);*/
	ff = f = PolFactorization(p);
	q = PolAlloc(5,0);
	for (; f->n != 0; ++f)
	{
    	    int i;
    	    /*polprint("factor",f->p);*/
    	    for (i = 0; i < f->n; ++i) 
		PolMul(q,f->p);
    	    /*printf("exp = %ld\n",f->n);*/
	    PolFree(f->p);
	}
	free(ff);
	/*polprint("q",q);*/
	if (PolCompare(p,q)) break;
	PolFree(q);
    }
    polprint("p",p);
    polprint("q",q);
    return 0;
}

#endif
