/* ============================= C MeatAxe ==================================
   File:        $Id: cfinfo.c,v 1.6 1998/04/28 17:35:58 mringe Exp $
   Comment:     Functions for reading and writing the .cfinfo file.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>





/*{{{ --------------------------------------------------------------------------
!structure CFInfo "Constituent data structure"
!section lattice
!synopsis 
    typedef struct
    {
	long dim, num, mult;
	long idword;
	poly_t *idpol;
	long peakword;
	poly_t *peakpol;
	long nmount;
	long ndotl;
	long spl;
    }
    CfInfo;
!description
    The |CFInfo| data structure contains all information about one 
    irreducible constituent of a module. Most of this information is
    useful only in the context of a |Lat_Info| structure.

    |dim| is the dimension of the constituent, and |num| enumerates
    constituents of same dimension. |mult| ist the multiplicity of this
    constituent in the module.
    |idword| and |idpol| specify the identifying word for this module.
    The identifying word is used by the CHOP program.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!structure Lat_Info "Lattice data structure"
!section lattice
!synopsis 
    typedef struct
    {
	char BaseName[LAT_MAXBASENAME];
	int Field;
	int NGen;
	int NCf;
	CfInfo Cf[LAT_MAXCF];
    } Lat_Info;
!description
    The |Lat_Info| data structure contains all information about a 
    module, which is gathered during the submodule lattice calculation.
-------------------------------------------------------------------------- }}}*/



/* ------------------------------------------------------------------
   writeword(), readword() - Write and read words
   ------------------------------------------------------------------ */

static void WriteWord(StfData *f, long w, poly_t *p)

{
    int i;

    StfPut(f,"[");
    StfPutInt(f,w);
    StfPut(f,",");
    StfPutInt(f,zfl);
    if (p == NULL)
    {
	StfPut(f,",-1");
    }
    else
    {
    StfPut(f,",");
    StfPutInt(f,p->deg);
	for (i = 0; i <= p->deg; ++i)
	{
	    StfPut(f,",");
	    StfPutInt(f,zftoi(p->buf[i]));
	}
    }
    StfPut(f,"]");
}


static int ReadWord(StfData *f, long *w, poly_t **p, char *fn)

{
    int fl, deg;  
    int i;

    if (StfMatch(f," [")) 
	MtxErrorExit("%s: missing '['",fn);
    StfGetInt(f,&i);
    *w = i;
    if (StfMatch(f,",")) 
	MtxErrorExit("%s: Missing ','",fn);
    StfGetInt(f,&fl);
    if (StfMatch(f,",")) 
	MtxErrorExit("%s: Missing ','",fn);
    StfGetInt(f,&deg);
    if (deg == -1)
	*p = NULL;
    else
    {
        *p = PolAlloc(fl,deg);
    	for (i = 0; i <= deg; ++i)
    	{
	    int coeff;
            if (StfMatch(f,",")) 
		MtxErrorExit("%s: Missing ','",fn);
	    StfGetInt(f,&coeff);
	    (*p)->buf[i] = zitof(coeff);
    	}
    }
    if (StfMatch(f,"]")) 
	MtxErrorExit("%s: Missing ']'",fn);
    return 1;
}


/* --------------------------------------------------------------------------
   Lat_ReadInfo() - Read a lattice information file
  
   Description:
     This funktion reads a .cfinfo file and stores the data into a <Lat_Info> 
     structure.
  
   Arguments:
     <li>: Pointer to the data structure.
     <basename>: Base name.
  
   Remarks:
     <basename> must not contain ".cfinfo".
  
   Return:
     0 = success, -1 = error
   -------------------------------------------------------------------------- */

#define RDVEC(name,fld)\
	else if (!strcmp(c,name))\
	{\
	    if (StfMatch(f," ["))\
		MtxErrorExit("Error in cfinfo file: Missing '['");\
	    for (i = 0; i < li->NCf; ++i) \
	    {\
		int val = 0;\
		if (i > 0)\
		    StfMatch(f,",");\
		StfGetInt(f,&val);\
		li->Cf[i].fld = val;\
	    }\
	    if (StfMatch(f,"]"))\
	    	MtxErrorExit("Error in cfinfo file: Missing ']'");\
	}

int Lat_ReadInfo(Lat_Info *li, const char *basename)

{
    MTXCONTEXT(Lat_ReadInfo);
    int i;
    StfData *f;
    char fn[LAT_MAXBASENAME + 20];

    /* Check arguments
       --------------- */
    MTX_VERIFY(li != NULL);
    MTX_VERIFY(basename != NULL);
    MTX_VERIFY(strlen(basename) < LAT_MAXBASENAME - 1);

    /* Initialize the data structure
       ----------------------------- */
    memset(li,0,sizeof(Lat_Info));
    strcpy(li->BaseName,basename);
    li->NCf = 0;
    li->NGen = 2;

    /* Open the file
       ------------- */
    sprintf(fn,"%s.cfinfo",basename);
    if ((f = StfOpen(fn,"rt")) == NULL)
	MtxErrorExit("Cannot open %s",fn);

    /* Read header
       ----------- */
    if (StfReadLine(f) || strcmp(StfGetName(f),"CFInfo"))
	MtxErrorExit("%s: %E",fn,ERR_FILEFMT);

    /* Read data
       --------- */
    while (StfReadLine(f) == 0)
    {
	const char *c = StfGetName(f);
	if (!strcmp(c,"CFInfo.NCF"))
	    StfGetInt(f,&li->NCf);
	else if (!strcmp(c,"CFInfo.Field"))
	{   
	    StfGetInt(f,&li->Field);
	    zsetfield((long)li->Field);
	}
	else if (!strcmp(c,"CFInfo.NGen"))
	    StfGetInt(f,&li->NGen);
	RDVEC("CFInfo.Dimension",dim)
	RDVEC("CFInfo.Number",num)
	RDVEC("CFInfo.Multiplicity",mult)
	RDVEC("CFInfo.SplittingField",spl)
	RDVEC("CFInfo.NMountains",nmount)
	RDVEC("CFInfo.NDottedLines",ndotl)
	else if (!strcmp(c,"CFInfo.IdWord"))
	{
	    if (StfMatch(f," [") != 0) 
		MtxErrorExit("%s: Missing '['",fn);
	    for (i = 0; i < li->NCf; ++i)
	    {
		ReadWord(f,&(li->Cf[i].idword),&(li->Cf[i].idpol),fn);
		if (StfMatch(f,i < li->NCf - 1 ? "," : "];") != 0)
		    MtxErrorExit("%s: %E",fn,ERR_FILEFMT);
	    }
	}
	else if (!strcmp(c,"CFInfo.PeakWord"))
	{
	    if (StfMatch(f," [") != 0) 
		MtxErrorExit("%s: %E",fn,ERR_FILEFMT);
	    for (i = 0; i < li->NCf; ++i)
	    {
		ReadWord(f,&(li->Cf[i].peakword),&(li->Cf[i].peakpol),fn);
		if (StfMatch(f,i < li->NCf - 1 ? "," : "];") != 0)
		    MtxErrorExit("%s: %E",fn,ERR_FILEFMT);
	    }
	}
	else
	    MtxErrorExit("%s: %E",fn,ERR_FILEFMT);
    }
    StfClose(f);
    MESSAGE(1,("Read %s: %d composition factors\n",fn,li->NCf));
    return 0;
}


/* --------------------------------------------------------------------------
   Lat_WriteInfo() - Write a lattice information
  
   Description:
     This funktion writes the contents of a <Lat_Info> structure into a file.
  
   Arguments:
     <li>: Pointer to the data structure.
  
   Remarks:
     The file name is constructed from the <BaseName> field of the structure
     by appending ".cfinfo".
  
   Return:
     0 = success, -1 = error
   -------------------------------------------------------------------------- */

int Lat_WriteInfo(Lat_Info *li)

{
    MTXCONTEXT(Lat_WriteInfo);
    StfData *f;
    int i;
    int tmp[LAT_MAXCF];
    char fn[LAT_MAXBASENAME + 20];

    /* Check arguments
       --------------- */
    MTX_VERIFY(li != NULL);

    /* Open the file
       ------------- */
    strcpy(fn,li->BaseName);
    strcat(fn,".cfinfo");
    f = StfOpen(fn,"w");
    if (f == NULL)
	return -1;

    /* Write data
       ---------- */
    StfWriteValue(f,"CFInfo","rec()");
    StfWriteInt(f,"CFInfo.NGen",li->NGen);
    StfWriteInt(f,"CFInfo.Field",li->Field);
    StfWriteInt(f,"CFInfo.NCF",li->NCf);

#define WRVEC(name,field)\
    for (i = 0; i < li->NCf; ++i)\
        tmp[i] = li->Cf[i].field;\
    StfWriteVector(f,"CFInfo." #name,li->NCf,tmp);

    WRVEC(Dimension,dim);
    WRVEC(Number,num);
    WRVEC(Multiplicity,mult);
    WRVEC(SplittingField,spl);
    WRVEC(NMountains,nmount);
    WRVEC(NDottedLines,ndotl);

    StfBeginEntry(f,"CFInfo.PeakWord");
    StfPut(f,"[");
    for (i = 0; i < li->NCf; ++i)
    {
        WriteWord(f,li->Cf[i].peakword,li->Cf[i].peakpol);
	if (i < li->NCf-1) StfPut(f,",");
    }
    StfPut(f,"]");
    StfEndEntry(f);

    StfBeginEntry(f,"CFInfo.IdWord");
    StfPut(f,"[");
    for (i = 0; i < li->NCf; ++i)
    {
        WriteWord(f,li->Cf[i].idword,li->Cf[i].idpol);
	if (i < li->NCf-1) StfPut(f,",");
    }
    StfPut(f,"]");
    StfEndEntry(f);
    StfClose(f);
    MESSAGE(1,("Wrote %s: %d composition factors\n",fn,li->NCf));
    return 0;
}




/* --------------------------------------------------------------------------
   Lat_CfName() - Build name of constituent
  
   Description:
     
   Arguments:
     <li>: Pointer to the lattice data structure.
     <cf>: Index of the constituent.
  
   Remarks:
     The return value points to a static buffer which is overwritten at each 
     call to <Lat_CFInfo()>.

     The constituent data inside <li> must have been set up properly, i.e.,
     the module must have been chopped.

     The constituent name consists of the dimension and an appendix which is
     built from the <num> field in the constituent's data structure. Usually
     the appendix is a single letter ('a', 'b', ...). If there are more than 
     26 constituents with the same dimension, a two-letter appendix is used. 
     If there are more than 26*26 constituents with the same dimension...
     IF YOU REALLY WANT TO KNOWN THIS, LOOK BELOW!
  
   Return:
     Constituent name
   -------------------------------------------------------------------------- */

char *Lat_CfName(Lat_Info *li, int cf)

{
    MTXCONTEXT(Lat_CfName);
    static char buf[20];
    int num, dim;

    /* Check arguments
       --------------- */
    MTX_VERIFY(li != NULL);
    MTX_VERIFY(cf >= 0 && cf < li->NCf);

    /* Get dimension and number of the constituent
       ------------------------------------------- */
    num = li->Cf[cf].num;
    dim = li->Cf[cf].dim;

    /* Build the constituent's name
       ---------------------------- */
    if (num < 26)
	sprintf(buf,"%d%c",dim,(char)num+'a');
    else if (num < 26*26)
	sprintf(buf,"%d%c%c",dim,(char)(num%26)+'a',(char)(num/26-1)+'a');
    else
	sprintf(buf,"%dcf%d",dim,num);

    return buf;
}

