/* ============================= C MeatAxe ==================================
   File:        $Id: charpol.c,v 1.8 1998/07/29 16:51:57 mringe Exp $
   Comment:     Characteristic polynomial of a matrix
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"




/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

/*{{{ --------------------------------------------------------------------------
!variable CharPolSeed "Seed for characteristic polynomial"
!section mat.op
!synopsis 
    long CharPolSeed = 1;
!description
    This variable is used by \verb"CharPolFactor()" to select the first
    seed vector. By default, \verb"CharPolSeed" has the value 1, i.e., the
    first seed vector is $(1,0,\ldots,0)$. Assigning the value 2 selects
    the start vector $(0,1,\ldots,0)$ in all subsequent calls to
    \verb"CharPolFactor()".

    If |CharPolSeed| is out of bounds, |CharPolFactor()| will reset it to 1.
!seealso @charpol
-------------------------------------------------------------------------- }}}*/

long CharPolSeed = 1;		/* Seed */


/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

static long fl, nor;		/* Field and size */
static long *piv = NULL;	/* Pivot table */
static char *ispiv = NULL;	/* Pivot flags */
static PTR mat = NULL,		/* The matrix */
    A = NULL,			/* Work space (for spin-up) */
    B = NULL;			/* Work space II (coefficients) */
static long dim;		/* Dimension reached so far */
static long n;			/* Dimension of cyclic subspace */



/* ------------------------------------------------------------------
   cleanup() - Free memory
   ------------------------------------------------------------------ */

static void cleanup()

{
    if (mat != NULL) free(mat);
    if (A != NULL) free(A);
    if (B != NULL) free(B);
    if (piv != NULL) free(piv);
    if (ispiv != NULL) free(ispiv);
    mat = A = B = NULL;
    piv = NULL;
    ispiv = NULL;
}


/* ------------------------------------------------------------------
   init()
   ------------------------------------------------------------------ */

static int init(const matrix_t *matrix)

{	
    MTXCONTEXT(init);

    /* Set some global variables
       ------------------------- */
    if (matrix->nor != matrix->noc) 
    {
	MTXERROR1("%E",ERR_NOTSQUARE);
	return -1;
    }
    fl = matrix->fl;
    nor = matrix->nor;
    dim = 0;
    if (CharPolSeed < 1 || CharPolSeed > nor)
	CharPolSeed = 1;

    /* Allocate memory
       --------------- */
    cleanup();
    zsetfield(fl);
    zsetlen(nor);
    if ((mat = zalloc(nor)) == NULL
	|| (A = zalloc(nor+1)) == NULL
	|| (B = zalloc(nor+1)) == NULL
	|| (piv = NALLOC(long,nor+2)) == NULL
	|| (ispiv = NALLOC(char,nor+2)) == NULL
       )
	return -1;

    /* Initialize memory
       ----------------- */
    memcpy(mat,matrix->d,zsize(nor));
    memset(ispiv,0,(size_t)(nor+2));

    return 0;
}


/* ------------------------------------------------------------------
   mkpoly() - Make polynomial for the latest cyclic subspace
   ------------------------------------------------------------------ */

static poly_t *mkpoly()

{
    int k;
    PTR x;
    poly_t *pol;
    
    pol = PolAlloc(fl,nor);
    x = B;
    zadvance(&x,n);
    pol->deg = n;
    for (k = 1; k <= n; ++k)
	pol->buf[k-1] = zextract(x,k);
    pol->buf[n] = F_ONE;
    return pol;
}


/* ------------------------------------------------------------------
   spinup_cyclic() - Spin up one cyclic subaspace
   ------------------------------------------------------------------ */

static void spinup_cyclic()

{   PTR a, b;
    long pv, k;
    FEL f;

    a = A;
    zadvance(&a,dim);
    b = B;
    zmulrow(b,F_ZERO);
    n = 0;
    while ((pv = zfindpiv(a,&f)) != 0)
    {	
	PTR x, y;

    	/* Add new vector to basis
	   ----------------------- */
	piv[dim+1+n] = pv;
/*fprintf(stderr,"piv[%ld]=%ld\n",dim+1+n,pv);*/
	ispiv[pv] = 1;
	++n;
	zinsert(b,n,F_ONE);

	/* Calculate the next vector
	   ------------------------- */
	x = a;
	zadvance(&a,(long)1);
	zmaprow(x,mat,nor,a);
/*
{long ii; fprintf(stderr," x ="); for (ii=1; ii < nor; ++ii)
 fprintf(stderr,"%ld",zftoi(zextract(x,ii))); fprintf(stderr,"\n xA=");
 for (ii=1; ii < nor; ++ii) fprintf(stderr,"%ld",zftoi(zextract(a,ii)));
 fprintf(stderr,"\n");
}
*/
	y = b;
	zadvance(&b,(long)1);
	zmulrow(b,F_ZERO);
	for (k = 1; k < nor; ++k)
	    zinsert(b,k+1,zextract(y,k));
    
	/* Clean with existing basis vectors
	   --------------------------------- */
	x = A;
	y = B;
	for (k = 1; k <= dim+n; ++k)
	{   f = zdiv(zextract(a,piv[k]),zextract(x,piv[k]));
	    zaddmulrow(a,x,zneg(f));
	    if (k > dim)
	    {	zaddmulrow(b,y,zneg(f));
		zadvance(&y,(long)1);
	    }
	    zadvance(&x,(long)1);
	}
    }
    dim += n;
}




/*{{{ --------------------------------------------------------------------------
!function CharPolFactor "Characteristic polynomial"
!section mat.op
!synopsis 
    poly_t *CharPolFactor(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    One factor of the characteristic polynomial of |mat|,
    or |NULL| if there are no more factors.
!description
    This function returns one factor of the characteristic polynomial of
    a given matrix. Further calls with a \verb"NULL" argument return
    more factors or \verb"NULL", if there are no more factors. 
    Note that the factors obtained in this way are in general not irreducible.

    Here is how \verb"CharPolFactor()" works: If |mat| is different from 
    |NULL|, |CharPolFactor()| initializes its internal data and starts
    computing one cyclic subspace. The choice of starting vector for this
    first subspace depends on the global variable \verb"CharPolSeed".
    Usually, this variable has a value of 1, corresponding to the vector
    $(1,0,\ldots,0)$. Then, the polynomial of the matrix restricted to 
    that cyclic subspace is constructed and returned to the caller.

    If |mat| is |NULL| on the next call, |CharPolFactor()| resumes at 
    the point where it returned the last time, calculates the next cyclic 
    subspace and so on, until the complete space is exhausted.

    {\em Note:} Since the function uses static variables to store
    information across multiple calls, your program must not use
    |CharPolFactor()| on more than one matrix at the same time.
   
!seealso @charpol
-------------------------------------------------------------------------- }}}*/

poly_t *CharPolFactor(const matrix_t *mat)

{
    MTXCONTEXT(CharPolFactor);
    PTR seed;
    int i;

    /* If called with mat != NULL, initialize everything
       ------------------------------------------------- */
    if (mat != NULL)
    {
	if (!MATRIX_IS_VALID(mat))
	    return NULL;
    	if (init(mat) != 0)
	{
    	    MTXERROR("Initialization failed");
	    return NULL;
	}
    }

    /* If there is nothing left to do, return NULL
       ------------------------------------------- */
    /* we could call cleanup() here... */
    if (dim >= nor) 
	return NULL;

    /* Prepare the next seed vector
       ---------------------------- */
    zsetfield(fl);
    zsetlen(nor);
    seed = A;
    zadvance(&seed,dim);
    if (dim == 0)
	i = CharPolSeed;
    else
    	for (i = 1; i <= nor && ispiv[i] != 0; ++i);
    zmulrow(seed,F_ZERO);
    zinsert(seed,i,F_ONE);

    /* Spin up and return the polynomial 
       --------------------------------- */
    spinup_cyclic();
    return mkpoly();    
}




/*{{{ --------------------------------------------------------------------------
!function CharPol "Characteristic polynomial"
!section mat.op
!synopsis 
    fpoly_t *CharPol(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    The characteristic polynomial of |mat| or |NULL| on error.
!description
    This function calculates the characteristic polynomial of a matrix in
    factored form. The return value is a pointer to a |fpoly_t| structure
    containing the irreducible factors of the characteristic polynomial.
!seealso @charpol
-------------------------------------------------------------------------- }}}*/

fpoly_t *CharPol(const matrix_t *mat)

{
    MTXCONTEXT(CharPol);
    fpoly_t *cpol = FpolAlloc();
    poly_t *p;

    if (cpol == NULL)
    {
	MTXERROR("Cannot allocate result");
	return NULL;
    }

    for (p = CharPolFactor(mat); p != NULL; p = CharPolFactor(NULL))
    {
	fpoly_t *fac = PolFactorization(p);
	if (fac == NULL)
	{
	    MTXERROR("Factorization failed");
	    return NULL;
	}
    	PolFree(p);
	FpolMul(cpol,fac);
	FpolFree(fac);
    }
    return cpol;
}
