/* ============================= C MeatAxe ==================================
   File:        $Id: chbasis.c,v 1.6 1998/04/28 17:36:00 mringe Exp $
   Comment:     Change basis.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"


/*{{{ --------------------------------------------------------------------------
!function ChangeBasis "Change basis"
!section reps
!synopsis 
    int ChangeBasis(const matrix_t *M, int ngen, const matrix_t *gen[],
	matrix_t *newgen[])
!arg M
    New basis.
!arg ngen
    Number of matrices in |gen|.
!arg gen
    Array of matrices to transform.
!arg newgen
    Pointer to a buffer for the transformed generators.
!return
    The function returns |dest| or |NULL| on error.
!description
    This function performs a change of basis on a set of square matrices.
    Typically, the matrices are the generators for an algebra with respect
    to a basis $B$. |ChangeBasis()| calculates the matrix representation of
    the generators with respect to another basis, $C$.
    The rows of |M| must contain the new basis vectors, expressed in the
    old basis. Then, the new generators are given by
    \[
	g_i' = M g_i M^{-1}
    \]
    If |newgen| is different from |gen|, |newgen| must point to an array
    of at least |ngen| elements of type |matrix_t *|. The elements of 
    |newgen| must not be in use, because they are overwritten by |ChangeBasis()| 
    with newly allocated matrices.
    If |newgen| is euqal to |gen|, the new generators replace the original
    generators. In this case, |ChangeBasis()| performs the necessary clean-up
    before overwriting the generators.
!seealso 
-------------------------------------------------------------------------- }}}*/

int ChangeBasis(const matrix_t *M, int ngen, const matrix_t *gen[],
	matrix_t *newgen[])

{
    MTXCONTEXT(ChangeBasis);
    matrix_t *bi, *tmp;
    int i;

    MTX_VERIFY(M != NULL && M->id == T_MATRIX);
    MTX_VERIFY(ngen >= 0);
    if ((bi = MatInv(M)) == NULL) 
	return MTXERROR("Matrix is singular"), -1;
    for (i = 0; i < ngen; ++i)
    {
	tmp = MatDup(M);
	MatMul(tmp,gen[i]);
	MatMul(tmp,bi);
	if ((const matrix_t **)newgen == gen)
	    MatFree(newgen[i]);
	newgen[i] = tmp;
    }
    MatFree(bi);
    return 0;
}


