/* ============================= C MeatAxe ==================================
   File:        $Id: chop.c,v 1.13 1998/09/17 17:03:52 mringe Exp $
   Comment:     Chop a representation (Calculate composition series)
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>


/* --------------------------------------------------------------------------
   Some constants
   -------------------------------------------------------------------------- */

#define MAXTRIES 10000	    /* Number of tries before FindIdWord() fails */


/* --------------------------------------------------------------------------
   typedefs
   -------------------------------------------------------------------------- */

typedef struct nodestruct
{
    struct nodestruct *sub, *quot, *parent;
    long dim, num;		/* Dimension and number */
    matrix_t *gen[MAXGEN];	/* Generators */
    long mult;			/* Multiplicity */
    long spl;			/* Degree of splitting field */
    long idword;		/* Word used for std basis */
    poly_t *idpol;		/* Polynomial "  "    "    */
    poly_t *f1, *f2;		/* characteristic Polynomial c=f1*f2*/
    long fprint[MAXFP];		/* Fingerprint */
    set_t *badwords;
    matrix_t *nsp, *nsptr;	/* Null space */
    matrix_t *subsp;		/* Subspace */
    matrix_t **gentr;		/* Transposed generators */
    long ggt;			/* g.c.d. of nullities */
    wgdata_t *wg;		/* Used by the word generator */
    long wnum;
    matrix_t *word;
} node_t;


/* --------------------------------------------------------------------------
   Function prototypes
   -------------------------------------------------------------------------- */

static int Chop(node_t *n);


/* --------------------------------------------------------------------------
   Global variables
   -------------------------------------------------------------------------- */

node_t *root;
long opt_deglimit = -1;
long opt_nullimit;
node_t *irred[LAT_MAXCF];		/* List of irred. constituents */
long firstword = -1;
int opt_G = 0;				/* GAP output */
set_t *goodwords;			/* List of `good' words */
Lat_Info LI;				/* Data for .cfinfo */


static long stat_svsplit = 0;
static long stat_cpirred = 0;
static long stat_nssplit = 0;
static long stat_dlsplit = 0;
static long stat_irred = 0;
static long stat_exsplit = 0;



static char *helptext[] = {
"SYNTAX",
"    chop [<Options>] <Name>",
"",
"OPTIONS",
"    -G               GAP output (implies -Q).",
"    -Q               Quiet, no messages.",
"    -V               Verbose, more messages.",
"    -g <NGen>        Set number of generators (default is 2).",
"    -n <MaxNul>      Set limit on nullity",
"    -d <MaxDeg>      Set limit on degrees of polynomials",
"    -T <MaxTime>     Set CPU time limit",
"",
"FILES",
"    The input must be <NGen> square matrices. They are read from",
"    <Name>.1, <Name>.2, ... <Name>.<NGen>.",
"",
"    Output is written to various files:",
"      <Name>.cfinfo        Information about the composition factors.",
"      <Name><Dim><a>.<N>   Generators on the composition factors.",
"                           <Dim> is the dimension, <a> is a letter",
"                           discriminating different factors, and <N>",
"                           runs from 1 to <NGen>.",
NULL};

static proginfo_t pinfo =
   { "chop", "Composition Series", "$Revision: 1.13 $", helptext };


/* ------------------------------------------------------------------
   newnode() - Create a new node
   ------------------------------------------------------------------ */

static node_t *newnode(matrix_t **gen, node_t *parent)

{
    node_t *n = ALLOC(node_t);
    n->parent = parent;
    n->sub = n->quot = NULL;
    n->dim = gen[0]->nor;
    n->num = -1;
    n->mult = -1;
    memcpy(n->gen,gen,(size_t)LI.NGen*sizeof(matrix_t *));
    n->spl = -1;
    n->idword = -1;
    n->idpol = NULL;
    n->badwords = SetAlloc();
    n->nsp = n->nsptr = NULL;
    n->subsp = NULL;
    n->ggt = 0;
    n->f1 = n->f2 = NULL;

    n->wg = WgInit(LI.NGen,(const matrix_t **)gen);
    if (n->wg == NULL) 
	MtxErrorExit("WgInit(): %E",MtxErrNo);
    n->wnum = -1;
    n->word = NULL;
    n->gentr = NULL;

    return n;
}

#define MATFREE(x) { if ((x) != NULL) { MatFree(x); (x)=NULL; } }


/* ------------------------------------------------------------------
   makeword() - Make a word
   ------------------------------------------------------------------ */

static void makeword(node_t *n,long w)

{
    MATFREE(n->word);
    n->word = WgMakeWord(n->wg,w);
    n->wnum = w;
}


/* ------------------------------------------------------------------
   insertword() - Insert the word into a polynomial and calculate the
   null-space (nsp) and the null-space of p(A)^T (nsptr).
   ------------------------------------------------------------------ */

static void insertword(node_t *n, poly_t *p)

{
    matrix_t *m, *mt;

    m = MatInsert(n->word,p);
    mt = MatTr(m);
    MATFREE(n->nsp);
    n->nsp = MatNullSpace__(m);
    MATFREE(n->nsptr);
    n->nsptr = MatNullSpace__(mt);
}


/* ------------------------------------------------------------------
   init () - Read generators
   ------------------------------------------------------------------ */

static void init()

{
    char fn[100];
    char fn0[200];
    int i;
    matrix_t *x;
    matrix_t *gen[MAXGEN];

    for (i = 0; i < LI.NGen; ++i)
    {
	sprintf(fn,"%s.%d",LI.BaseName,i+1);
	x = gen[i] = MatLoad(fn);
	if (i == 0) 
	{
	    strcpy(fn0,fn);
	    LI.Field = (int) zfl;
	}
	if (x->nor != x->noc)
	    MtxErrorExit("%s: %E",fn,ERR_NOTSQUARE);
	if (i > 0 && (x->nor != gen[0]->nor || x->fl != gen[0]->fl))
	    MtxErrorExit("%s and %s: %E",fn0,fn,ERR_INCOMPAT);
    }
    root = newnode(gen,NULL);
}


/* ------------------------------------------------------------------
   isbadword() - Look up a word in the 'badwords' list
   ------------------------------------------------------------------ */

static int isbadword(long w,node_t *n)

{
    if (n == NULL) 
	return 0;
    if (n->badwords != NULL && SetContains(n->badwords,w))
	return 1;
    return isbadword(w,n->parent);
}



/* ------------------------------------------------------------------
   writetree() - Write the composition series to stdout
   ------------------------------------------------------------------ */

static void writetree(node_t *n)

{
    static int count = 0;	/* Characters written in one line */
    static int first = 1;
    int i;

    if (n->sub == NULL)		/* Irreducible */
    {
	if (count >= 20)
	{   
	    printf("\n");
	    count = 0;
	}
	for (i = 0; LI.Cf[i].dim != n->dim || LI.Cf[i].num != n->num;
	    ++i);
	if (opt_G)
	{
	    if (!first) printf(","); else { printf(" "); first = 0; }
	    printf("%d",i+1);
	}
	else
	    printf("%s ",Lat_CfName(&LI,i));
	++count;
    }
    else
    {
	writetree(n->sub);
	writetree(n->quot);
    }
}


/* ------------------------------------------------------------------
   writeinfo() - Write some information to stdout and create the
	.cfinfo file.
   ------------------------------------------------------------------ */

static void writeinfo(node_t *root)

{
    int i, k;

    MESSAGE(0,("\n\nChopping completed: %d different composition"
	" factors\n",LI.NCf));

    /* Write the cfinfo file
       --------------------- */
    MESSAGE(0,("Writing %s.cfinfo\n",LI.BaseName));
    for (i = 0; i < LI.NCf; ++i)
    {
    	LI.Cf[i].dim = irred[i]->dim;
	LI.Cf[i].num = irred[i]->num;
	LI.Cf[i].mult = irred[i]->mult;
	LI.Cf[i].spl = irred[i]->spl;
	LI.Cf[i].idword = irred[i]->idword;
	LI.Cf[i].idpol = irred[i]->idpol;
    }
    Lat_WriteInfo(&LI);


    /* Write composition factors
       ------------------------- */
    if (opt_G)
    {
	printf("MeatAxe.CompositionFactors := [\n");
    	for (i = 0; i < LI.NCf; ++i)
    	{
	    printf("  [ \"%s\", %ld, %ld ]",Lat_CfName(&LI,i),
	      irred[i]->mult,irred[i]->spl);
	    if (i < LI.NCf-1) printf(",");
	    printf("\n");
	}
	printf("\n];\n");
    }
    else if (MSG0)
    {
    	printf("\nName   Mult  SF  Fingerprint\n");
    	for (i = 0; i < LI.NCf; ++i)
    	{
	    printf("%-6s %4ld  %2ld  ",Lat_CfName(&LI,i),
	      irred[i]->mult,irred[i]->spl);
	    for (k = 0; k < MAXFP; ++k)
		printf("%ld%s",irred[i]->fprint[k],k==MAXFP-1?"\n":",");
	}
    }

    /* Write the composition series
       ---------------------------- */
    if (opt_G)
    {
	printf("MeatAxe.CompositionSeries := [\n");
    	writetree(root);
	printf("];\n");
    }
    else if (MSG0)
    {
	printf("\nAscending composition series:\n");
    	writetree(root);
	printf("\n");
    }

    /* Write statistics
       ---------------- */
    if (MSG3)
    {
	printf("\nStatistics\n----------\n");

	printf("Saved vectors split: %4ld\n",stat_svsplit);
	printf("c(x) irreducible:    %4ld\n",stat_cpirred);
	printf("Normal split:        %4ld\n",stat_nssplit);
	printf("Dual split:          %4ld\n",stat_dlsplit);
	printf("Exceptional split:   %4ld\n",stat_exsplit);
	printf("Irreducible:         %4ld\n",stat_irred);
    }
}


/* --------------------------------------------------------------------------
   CleanUpNode() - Clean up a node_t structure

   Arguments:
     <n>: Pointer to the <node_t> structure

   Description:
     This function releases all memory allocated internally in a <node_t>
     structure. Note that the memory for the <node_t> structure itself is
     n o t  free()'d here!
   -------------------------------------------------------------------------- */

static void CleanUpNode(node_t *n)

{   int i;
    if (n->gentr != NULL)
    {
        for (i = 0; i < LI.NGen; ++i)
	    MatFree(n->gentr[i]);
	free(n->gentr);
	n->gentr = NULL;
    }
    if (n->subsp != NULL) { MatFree(n->subsp); n->subsp = NULL; }
    if (n->wg != NULL) { WgFree(n->wg); n->wg = NULL; }
    if (n->nsp != NULL) { MatFree(n->nsp); n->nsp = NULL; }
    if (n->nsptr != NULL) { MatFree(n->nsptr); n->nsptr = NULL; }
    if (n->word != NULL) { MatFree(n->word); n->word = NULL; }
    if (n->f1 != NULL) { PolFree(n->f1); n->f1 = NULL; }
    if (n->f2 != NULL) { PolFree(n->f2); n->f2 = NULL; }
}



/* --------------------------------------------------------------------------
   splitnode() - Split a node

   Arguments:
     <n>: Pointer to the node to split.
     <tr>: Indicates that it was a `dual split'

   Description:
     This function is called after a proper submodule has been found. It
     calculates the action of the generators on both submodule and quotient,
     and creates two new <note_t> structures for the two parts.
     The original module, <n>, is cleaned up and <Chop()> is called with
     both submodule and quotient.
   -------------------------------------------------------------------------- */

static void splitnode(node_t *n, int tr)

{	
    matrix_t **sub, **quot;
    int i;

    /* Split the module
       ---------------- */
    MESSAGE(0,("Split: Subspace=%ld, Quotient=%ld\n",
	n->subsp->nor,n->dim - n->subsp->nor));
    sub = NALLOC(matrix_t *,LI.NGen);
    quot = NALLOC(matrix_t *,LI.NGen);
    Split(n->subsp,LI.NGen,(const matrix_t **)(tr ? n->gentr : n->gen),
	sub,quot);

    /* If it was a dual split, subspace and quotient have been
       calculated in the dual module. To get back to the original
       module, transpose again and exchange sub and quot.
       ---------------------------------------------------------- */
    if (tr)
    {	int i;
	matrix_t *x, *y;
	for (i = 0; i < LI.NGen; ++i)
	{
	    x = MatTr(sub[i]);
	    MatFree(sub[i]);
	    y = MatTr(quot[i]);
	    MatFree(quot[i]);
	    sub[i] = y;
	    quot[i] = x;
	}
    }
    
    /* Make new nodes for subspace and quotient
       ---------------------------------------- */
    n->sub = newnode(sub,n);
    n->quot = newnode(quot,n);
    free(sub);
    free(quot);

    /* Project saved vectors on the quotient
       ------------------------------------- */
    if (!tr && n->nsp != NULL)
	n->quot->nsp = QuotProj(n->subsp,n->nsp);

    /* Clean up
       -------- */
    CleanUpNode(n);
    for (i = 0; i < LI.NGen; ++i)
    {
	MatFree(n->gen[i]);
	n->gen[i] = NULL;
    }

    /* Chop the subspace and quotient
       ------------------------------ */
    Chop(n->sub);
    Chop(n->quot);
}


/* ------------------------------------------------------------------
   extendbasis() - Find a vector in 'space' which is not in the
	(linear) span of 'basis'. 'basis' must be linearly
	independent.
   ------------------------------------------------------------------ */

static matrix_t *extendbasis(matrix_t *basis, matrix_t *space)

{
    long i, j, piv;
    long dimb = basis->nor;
    long dims = space->nor;
    PTR tmp, x, y;
    FEL f;


    /* Concatenate basis and space
       --------------------------- */
    zsetlen(basis->noc);
    tmp = zalloc(dimb+dims);
    memcpy(tmp,basis->d,zsize(dimb));
    x = tmp;
    zadvance(&x,dimb);
    memcpy(x,space->d,zsize(dims));

    /* Clean with basis
       ---------------- */
    for (i = 1, x = tmp; i <= dimb; zadvance(&x,(long)1), ++i)
    {
	piv = zfindpiv(x,&f);
	if (piv == 0) 
	    MtxErrorExit("extendbasis(): zero vector in basis");
	y = x;
	for (j = i+1; j <= dimb+dims; ++j)
	{
	    zadvance(&y,(long)1);
	    zaddmulrow(y,x,zsub(F_ZERO,zdiv(zextract(y,piv),f)));
	}
    }


    /* Find the first non-zero row
       --------------------------- */
    x = tmp;
    zadvance(&x,dimb);
    for (j = 1; zfindpiv(x,&f) == 0; ++j, zadvance(&x,(long)1));
    free(tmp);
    if (j > dims)
    {
	MtxErrorExit("extendbasis() failed");
    }
    return MatCut(space,j,1,1,-1);
}



/* ------------------------------------------------------------------
   checkspl() - Checks if a given representation's splitting field
   has degree [E:F] = dim(V), where V is a given subspace (usually
   the kernel of an algebra element).
   
   Returns 1 if [E:F]=dim(V) or zero otherwise.
   ------------------------------------------------------------------ */

#define MAXENDO 10	/* Max. dimension of endomorphism ring */


static int checkspl(const matrix_t *gen[], matrix_t *nsp)

{
    MTXCONTEXT(checkspl);
    matrix_t *sb1, *sb2;	/* Standard bases */
    matrix_t *g1[MAXGEN];	/* Generators in standard basis sb1 */
    matrix_t *g2[MAXGEN];	/* Generators in standard basis sb2 */
    matrix_t *endo[MAXENDO];	/* Endomorphisms */
    int nendo = 0;		/* # of endomorphisms obtained so far */
    long dim = gen[0]->nor;
    int i, result;

    /* Take the first vector from nsp and change to standard basis.
       ------------------------------------------------------------ */
    sb1 = SpinUp(nsp,LI.NGen,(const matrix_t **)gen,SF_FIRST|SF_CYCLIC|SF_STD);
    MTX_VERIFY(sb1 != NULL && sb1->nor == sb1->noc);
    ChangeBasis(sb1,LI.NGen,gen,g1);

    sb2 = NULL;	/* Mark as unused */
    while (1)
    {  
	matrix_t *v2, *subsp;

	/* Spin up v1 under all endomorphisms found so far. If this
	   yields the whole null-space, we know that the endomorphism
	   ring has at least dimension dim(nsp).
	   ---------------------------------------------------------- */
	subsp = SpinUp(nsp,nendo,(const matrix_t **)endo,SF_FIRST|SF_SUB);
	MTX_VERIFY(subsp != NULL);
	if (subsp->nor == nsp->nor)
	{
	    MatFree(subsp);
	    result = 1;	/* Successfull! */
	    break;
	}

	/* Take a vector which is not in span(v1)
	   and make again the standard basis.
	   --------------------------------------- */
	if (sb2 != NULL)	/* Clean up first */
	{
	    int j;
	    MatFree(sb2);
	    for (j = 0; j < LI.NGen; ++j) MatFree(g2[j]);
	}
	v2 = extendbasis(subsp,nsp);	/* Get vector */
	MatFree(subsp);
	sb2 = SpinUp(v2,LI.NGen,(const matrix_t **)gen,
	    SF_FIRST|SF_CYCLIC|SF_STD);
	MTX_VERIFY(sb2 != NULL && sb2->nor == sb2->noc);
	MatFree(v2);
	ChangeBasis(sb2,LI.NGen,gen,g2);

	/* Compare the two representations. If they are different,
	   we know that the splitting field degree must be smaller
	   than dim(nsp).
	   -------------------------------------------------------- */
	result = 1;
	for (i = 0; result && i < LI.NGen; ++i)
	{
	    if (memcmp(g1[i]->d,g2[i]->d,zsize(dim)))
	        result = 0;
	}
	if (result == 0) break;	/* Not successfull */

	/* They are identical, i.e., we have found an endomorphism.
	   Put it into the list and try the next vector.
	   -------------------------------------------------------- */
	if (nendo >= MAXENDO) MtxErrorExit("Too many endomorphisms");
	endo[nendo] = MatInv(sb2);
	MatMul(endo[nendo],sb1);
	++nendo;
    }

    /* Clean up
       -------- */
    MatFree(sb1); 
    for (i = 0; i < LI.NGen; ++i) MatFree(g1[i]);
    if (sb2 != NULL)
    {
	MatFree(sb2);
        for (i = 0; i < LI.NGen; ++i) MatFree(g2[i]);
    }
    while (nendo > 0)
	MatFree(endo[--nendo]);
    MESSAGE(3,("checkspl(): result = %d\n",result));
    return result;
}



/* --------------------------------------------------------------------------
   FindIdWord() - Find an identifying word for a module
   
   Description:
     This function finds an identifying word for a given irreducible module. 
     This is a word in the generators with minimal nullity, i.e., the nullity 
     equals the splitting field degree [E:F]. The word is stored in n->idword, 
     the polynomial in n->idpol, and its null-space in n->nsp.

   Arguments:
     <n>: The module

   Return:
    0 = Ok, -1 = Error.
   -------------------------------------------------------------------------- */

static int FindIdWord(node_t *n)

{
    long i;
    int k;
    long count = 0;
    static fpoly_t *cpol = NULL;

    /* Main loop: Try all words
       ------------------------ */
    MESSAGE(1,("Searching idword, dim=%ld\n",n->dim));
    for (i = 1; count <= MAXTRIES; ++i)
    {
	if (isbadword(i,n)) 
	    continue;

	/* Make the word and its characteristic polynomial
	   ----------------------------------------------- */
	MESSAGE(2,("  Word %ld  gcd=%ld\n",i,n->ggt));
	makeword(n,i);
	if (cpol != NULL) FpolFree(cpol);
	cpol = CharPol(n->word);
	if (MSG3) FpolPrint("  c(x)",cpol);
	for (k = 0; k < (int) cpol->len; ++k)
	    n->ggt = gcd(n->ggt,cpol->e[k] * cpol->p[k]->deg);

	/* Try all factors of c(x) with degree<=gcd
	   ---------------------------------------- */
	for (k = 0; k < (int) cpol->len; ++k)
	{
	    if (cpol->p[k]->deg > n->ggt) 
		continue;
	    ++count;
	    if (MSG3) PolPrint("  factor",cpol->p[k]);
	    insertword(n,cpol->p[k]);
	    n->ggt = gcd(n->ggt,n->nsp->nor);
	    if (checkspl((const matrix_t **)n->gen,n->nsp))
	    {
		MESSAGE(2,("  idword = %ld\n",i));
		n->idword = i;
		n->idpol = PolDup(cpol->p[k]);
		return 0;
	    }
	}
    }

    /* Failed...
       --------- */
    MtxErrorExit("FindIdWord() failed");
    return -1;
}




/* ------------------------------------------------------------------
   newirred() - Check if a given irreducible module is already
	contained in the list of composition factors. If yes, return
	the index. If not, insert the new irreducible and return its
	index.
   ------------------------------------------------------------------ */

static void newirred(node_t *n)

{
    MTXCONTEXT(newirred);
    int i, k;
    matrix_t *b;
    
    /* Check if the module is already in the list
       ------------------------------------------ */
    WgMakeFingerPrint(n->wg,n->fprint);
    for (i = 0; i < LI.NCf && n->dim >= irred[i]->dim; ++i)
    {
	/* Compare dimensions and fingerprints
	   ----------------------------------- */
	if (n->dim != irred[i]->dim ||
	    memcmp(n->fprint,irred[i]->fprint,sizeof(n->fprint)))
	    continue;

	if (IsIsomorphic(LI.NGen,(const matrix_t **)irred[i]->gen,LI.Cf + i,
	    (const matrix_t **)n->gen,NULL,0))
	{
	    ++irred[i]->mult;
	    n->num = irred[i]->num;
	    MESSAGE(0,("Irreducible (%s)\n",Lat_CfName(&LI,i)));
	    CleanUpNode(n);
	    return;
	}
    }

    /* It's a new irreducible!
       ----------------------- */
    if (LI.NCf >= LAT_MAXCF)
	MtxErrorExit("TOO MANY CONSTITUENTS");
    for (k = LI.NCf-1; k >= i; --k)
    {
	irred[k+1] = irred[k];
	LI.Cf[k+1] = LI.Cf[k];
    }
    irred[i] = n;
    ++LI.NCf;

    /* Set the fields
       -------------- */
    if (i == 0 || irred[i]->dim != irred[i-1]->dim)
	irred[i]->num = 0;
    else
	irred[i]->num = irred[i-1]->num + 1;
    irred[i]->mult = 1;
    LI.Cf[i].dim = irred[i]->dim;  /* cfname() needs this */
    LI.Cf[i].num = irred[i]->num;
    MESSAGE(0,("Irreducible (%s)\n",Lat_CfName(&LI,i)));

    /* Make idword and change to std basis 
       ----------------------------------- */
    MATFREE(n->nsp);
    if (n->idword == -1)
	FindIdWord(n);
    else
    {
	matrix_t *m;
    	n->word = WgMakeWord(n->wg,n->idword);
    	m = MatInsert(n->word,irred[i]->idpol);
    	n->nsp = MatNullSpace__(m);
    }
    LI.Cf[i].idword = n->idword;
    LI.Cf[i].idpol = n->idpol;
    LI.Cf[i].spl = n->spl = n->nsp->nor;
    b = SpinUp(n->nsp,LI.NGen,(const matrix_t **)n->gen,
	SF_FIRST|SF_CYCLIC|SF_STD);
    MTX_VERIFY(b != NULL && b->nor == b->noc);
    ChangeBasis(b,LI.NGen,(const matrix_t **)n->gen,n->gen);
    MatFree(b);

    /* Write out the generators
       ------------------------ */
    if (irred[i]->spl > 1)
	MESSAGE(1,("Splitting field has degree %ld\n",irred[i]->spl));
    for (k = 0; k < LI.NGen; ++k)
    {
	char fn[200];
	sprintf(fn,"%s%s.%d",LI.BaseName,Lat_CfName(&LI,i),k+1);
   	MatSave(irred[i]->gen[k],fn);
    }
    CleanUpNode(n);
}



/* ------------------------------------------------------------------
   split_with_saved_vectors() - Try to split a module using saved
   vectors.
   Returns 1 on success, 0 otherwise.
   ------------------------------------------------------------------ */

static int split_with_saved_vectors(node_t *n)

{
    matrix_t *span;

    if (n->nsp == NULL || n->nsp->nor == 0) 
	return 0;
    MESSAGE(1,("Trying saved vectors..."));

    if (n->subsp != NULL) 
	MatFree(n->subsp);
    n->subsp = NULL;
    span = SpinUp(n->nsp,LI.NGen,(const matrix_t **)n->gen,SF_EACH|SF_SUB);
    if (span->nor > 0 && span->nor < span->noc)
    {
	n->subsp = span;
	MESSAGE(1,("\n"));
	++stat_svsplit;
	splitnode(n,0);
	return 1;
    }

    MatFree(span);
    MESSAGE(1,("failed\n"));
    return 0;
}


/* --------------------------------------------------------------------------
   polymap() - Calculate the image of a single vector -- or a set
   of vectors -- under p(A).
   -------------------------------------------------------------------------- */

matrix_t *polymap(matrix_t *v, matrix_t *m, poly_t *p)

{
    matrix_t *result, *tmp;
    int i;

    result = MatAlloc(v->fl,v->nor,v->noc);
    tmp = MatDup(v);
    for (i = 0; i <= p->deg; ++i)
    {
	FEL f = p->buf[i];
	PTR x = result->d, y = tmp->d;
	long i;
	for (i = v->nor; i > 0; --i)
	{   zaddmulrow(x,y,f);
	    zadvance(&x,1);
	    zadvance(&y,1);
	}
	MatMul(tmp,m);
    }
    MatFree(tmp);
    return result;
}




/* --------------------------------------------------------------------------
   make_kern() - Calculate a vector in the null-space of p(A). p(x)
   is assumed to be a factor of f1(x), i.e., p(x) must occur in the
   first cyclic subspace.
   -------------------------------------------------------------------------- */

static void make_kern(node_t *n, poly_t *p)

{
    MTXCONTEXT(make_kern);
    poly_t *cof, *f;
    matrix_t *seed;

    f = PolDup(n->f1);
    cof = PolDivMod(f,p);
    MTX_VERIFY(f->deg == -1);
    PolFree(f);

    if (n->nsp != NULL) 
	MatFree(n->nsp);
    seed = MatAlloc(zfl,1,n->dim);
    zinsert(seed->d,CharPolSeed,F_ONE);
    n->nsp = polymap(seed,n->word,cof);
    MatFree(seed);
    PolFree(cof);
}


/* ------------------------------------------------------------------
   make_trkern() - Calculate a vector in the null-space of p(A^T). We
   consider only the first cyclic subspace, so make_trkern() may not
   find a vector.

   Returns 0 if a vector has been found, -1 else.
   ------------------------------------------------------------------ */

int make_trkern(node_t *n, poly_t *p)

{
    matrix_t *mt;
    poly_t *pt, *cof;
    int result = -1;

    mt = MatTr(n->word);
    pt = CharPolFactor(mt);
    cof = PolDivMod(pt,p);
    if (pt->deg == -1)		/* Glueck gehabt! */
    {
	matrix_t *seed = MatAlloc(zfl,1,n->dim);
    	zinsert(seed->d,CharPolSeed,F_ONE);
    	if (n->nsptr != NULL) 
	    MatFree(n->nsptr);
    	n->nsptr = polymap(seed,mt,cof);
	MatFree(seed);
	result = 0;
    }
    PolFree(cof);
    PolFree(pt);
    MatFree(mt);
    return result;
}


/* ------------------------------------------------------------------
   make_f1() - Calculate the characteristic polynomial on the first
   cyclic subspace.
   ------------------------------------------------------------------ */

static fpoly_t *make_f1(node_t *n)

{
    fpoly_t *cpol;
    int i, k;

    if (n->f2 != NULL) PolFree(n->f2);
    n->f2 = NULL;
    if (n->f1 != NULL) PolFree(n->f1);
    n->f1 = CharPolFactor(n->word);
    cpol = PolFactorization(n->f1);

    /* Sort the factors by ascending multiplicity
       ------------------------------------------ */
    for (i = 0; i < (int)cpol->len; ++i)
    {
    	for (k = i+1; k < (int)cpol->len; ++k)
    	{
	    long val1 = cpol->e[i];
	    long val2 = cpol->e[k];
	    if (val1 > val2)
	    {
	    	poly_t *p = cpol->p[i];
	    	long e = cpol->e[i];
	    	cpol->p[i] = cpol->p[k];
	    	cpol->e[i] = cpol->e[k];
	    	cpol->p[k] = p;
	    	cpol->e[k] = e;
	    }
    	}
    }
    if (MSG3) FpolPrint("  f1(x)",cpol);
    return cpol;
}


/* --------------------------------------------------------------------------
   make_f2() - Complete the characteristic polynomial

   Description:
     This function completes the calculation of the characteristic
     polynomial. We assume that the first factor (i.e. the characteristic
     polynomial on the first cyclic subspace) has already be calculated.
     The remaining factors are stored in <f2>.
   --------------------------------------------------------------------------- */

static void make_f2(node_t *n)	/* Make f2 */

{
    poly_t *f;

    if (n->f2 != NULL) return;
    n->f2 = PolAlloc(n->f1->fl,0);
    while ((f = CharPolFactor(NULL)) != NULL)
    {
	PolMul(n->f2,f);
	PolFree(f);
    }
    if (MSG3)
    { fpoly_t *x = PolFactorization(n->f2);
      FpolPrint("  f2(x)",x);
      FpolFree(x);
    }
}


/* ------------------------------------------------------------------
   try_poly() - Tries to split with one irreducible factor of c(x).
   
    Return: 1 = success, 0 = failed
   ------------------------------------------------------------------ */

static int try_poly(node_t *n, poly_t *pol, long vfh)

{
    long nul;
    int chkirred = 1;
    matrix_t *sub;

    make_kern(n,pol);
    nul = n->nsp->nor;

    /* Try to split
       ------------ */
    if (n->subsp != NULL) 
	MatFree(n->subsp);
    n->subsp = NULL;
    sub = SpinUp(n->nsp,LI.NGen,(const matrix_t **)n->gen,SF_FIRST|SF_SUB);
    if (sub->nor > 0 && sub->nor < sub->noc)
    {
	/* Module was split */
	n->subsp = sub;
	++stat_nssplit;
	SetInsert(goodwords,n->wnum);
	splitnode(n,0);
	return 1;
    }
    MatFree(sub);
    MESSAGE(3,("Split failed\n"));

    if (n->f2 == NULL) 
	make_f2(n);

    /* Now find out if we can prove irreducibility with Norton's
       criterion. We need that the factor we just tried occurs
       with multiplicity 1 in the characteristic polynomial.
       --------------------------------------------------------- */
    {
	/* Calculate the multiplicity of <pol>
	   in the characteristic polynomial.
	   ----------------------------------- */
	poly_t *f2 = PolDup(n->f2);
	long mult = 0;
	while (f2->deg > 0)
	{
	    poly_t *q = PolDivMod(f2,pol);
	    if (f2->deg == -1)
	    {
		++mult;
		PolFree(f2);
		f2 = q;
	    }
	    else
	    {
		PolFree(q);
		break;
	    }
	}
	PolFree(f2);
	mult += vfh;	/* Add multiplicity of <pol> in <f1> */

	/* If the multiplicity is not one, spin up every vector in
	   the kernel, but only if the kernel is small enough.
	   Otherwise give up, i.e. do not try to prove irreducibility 
	   ----------------------------------------------------------- */
	if (mult > 1)
	{
	    if (mult * pol->deg > opt_nullimit)
	    {
		chkirred = 0;
		MESSAGE(3,("Cannot check for irreducibility, mult=%ld\n",mult));
	    }
	    else
	    {
		/* Spin up all vectors */
		MatFree(n->nsp);
		n->nsp = MatNullSpace__(MatInsert(n->word,pol));
		if (n->subsp != NULL) 
		    MatFree(n->subsp);
		n->subsp = NULL;
		MESSAGE(3,("2nd spin-up, seed space dimemsion = %ld\n",
		    n->nsp->nor));
		sub = SpinUp(n->nsp,LI.NGen,(const matrix_t **)n->gen,
		    SF_MAKE|SF_SUB);
		if (sub->nor > 0 && sub->nor < sub->noc)
		{
		    /* Module was split */
		    n->subsp = sub;
		    ++stat_nssplit;
		    SetInsert(goodwords,n->wnum);
		    splitnode(n,0);
		    return 1;
		}
	    }
	}
    }


    /* Not split, try dual split
       ------------------------- */
    MESSAGE(3,("Try dual split...\n"));
    if (make_trkern(n,pol) != 0)
    {
	MESSAGE(3,(" No seed vector found, dual split skipped\n"));
	return 0;
    }
    if (n->gentr == NULL)	/* Transpose generators */
    {   int i;
	n->gentr = NALLOC(matrix_t *,LI.NGen);
	for (i = 0; i < LI.NGen; ++i)
	    n->gentr[i] = MatTr(n->gen[i]);
    }


    sub = SpinUp(n->nsptr,LI.NGen,(const matrix_t **)n->gentr,SF_FIRST|SF_SUB);
    if (sub->nor > 0 && sub->nor < sub->noc)
    {
	/* Dual split */
	n->subsp = sub;
	++stat_dlsplit;
	splitnode(n,1);
	return 1;
    }
    MatFree(sub);

    /* Dual split failed*/
    MESSAGE(3,("Dual split failed\n"));
    if (!chkirred)
	return 0;
    newirred(n);
    SetInsert(goodwords,n->wnum);
    ++stat_irred;
    return 1;
}



/* --------------------------------------------------------------------------
   try_ex_factor() - Try to split in exceptional cases

   Arguments:
     <n>: Pointer to the module
     <cp>: Characteristic polynomial
     <cpf>: Characteristic polynomial in factored form
     <factor>: Index of the factor to try

   Return:
     1 = success (module was split), 0 = failed

   Description:
     This function is called from <try_exceptional()> for each irreducible
     factor of the characteristic polynomial of degree >= 2.
   -------------------------------------------------------------------------- */

static int try_ex_factor(node_t *n, poly_t *cp, fpoly_t *cpf, int factor)

{
    MTXCONTEXT(try_ex_factor);
    int k;
    long rndword;
    poly_t *p, *q, *tmp, *gcd[3];
    matrix_t *B, *iA, *v;

    if (MSG3) 
    {
	printf("Trying factor (");
	PolPrint(NULL,cpf->p[factor]);
	printf(")^%ld\n",cpf->e[factor]);
    }

    /* Calculate p(x) = maximal power of the irred factor in c(x)
       ---------------------------------------------------------- */
    p = PolDup(cpf->p[factor]);
    for (k = 1; k < cpf->e[factor]; ++k)
	PolMul(p,cpf->p[factor]);

    /* Calculate the complement q(x) with q(x)p(x)=c(x)
       ------------------------------------------------ */
    tmp = PolDup(cp);
    q = PolDivMod(tmp,p);
    MTX_VERIFY(tmp->deg == -1);
    PolFree(tmp);

    /* Calculate i(x):=b(x)q(x) with a(x)p(x)+b(x)q(x) = 1.
       ---------------------------------------------------- */
    PolGcdEx(p,q,gcd);
    MTX_VERIFY(gcd[0]->deg == 0);
    PolMul(q,gcd[2]);

    /* Insert the word into i(x) and clean up polynomials.
       --------------------------------------------------- */
    iA = MatInsert(n->word,q);
    PolFree(gcd[0]);
    PolFree(gcd[1]);
    PolFree(gcd[2]);
    PolFree(p);
    PolFree(q);

    /* Choose a second random word, B, and calculate [A,i(A)Bi(A)]
       ----------------------------------------------------------- */
    rndword = n->wnum + MtxRandomInt(42);
    MESSAGE(3,("Choosing random word %ld\n",rndword));
    B = WgMakeWord(n->wg,rndword);
    MatMul(B,iA);
    MatMul(iA,B);
    MatFree(B);
    B = MatDup(n->word);
    MatMul(B,iA);
    MatMul(iA,n->word);
    MatMulScalar(iA,zsub(F_ZERO,F_ONE));
    MatAdd(B,iA);
    MatFree(iA);

    /* Select a random vector in the image of the commutator
       ----------------------------------------------------- */
    MESSAGE(3,("Choosing random vector\n"));
    v = MatAlloc(B->fl,1,B->noc);
    for (k = 1; k <= v->noc; ++k)
	zinsert(v->d,k,zitof(MtxRandomInt(zfl)));
    MatMul(v,B);
    MatFree(B);

    /* Try to split with this vector
       ----------------------------- */
    MESSAGE(3,("Spinning up:"));
    if (n->subsp != NULL) 
	MatFree(n->subsp);
    n->subsp = SpinUp(v,LI.NGen,(const matrix_t **)n->gen,SF_FIRST|SF_SUB);
    MatFree(v);
    if (n->subsp->nor > 0 && n->subsp->nor < n->subsp->noc)
    {
	MESSAGE(3,("SUCCESS!\n"));
	++stat_exsplit;
    	splitnode(n,0);
	return 1;
    }
    MESSAGE(3,("failed\n"));
    return 0;
}


/* --------------------------------------------------------------------------
   try_exceptional() - Try to split in exceptional cases

   Arguments:
     <n>: Pointer to the module

   Return:
     1 = success (module was split), 0 = failed

   Description:
     This function tries to split a module. It uses an algorithm developed by
     G. Ivanyos and K. Lux which is specifically designed for the `exceptional'
     cases where the standard methods fail.
   -------------------------------------------------------------------------- */

static int try_exceptional(node_t *n)

{
    MTXCONTEXT(try_exceptional);
    poly_t *cp;
    fpoly_t *cpf;
    int factor;
    int success = 0;
    MTX_VERIFY(n->f1 != NULL && n->f2 != NULL);

    MESSAGE(2,("Trying exceptional cases\n"));

    /* Calculate the complete characteristic polynomial c(x) 
       and its irreducible factors.
       ----------------------------------------------------- */
    cp = PolDup(n->f1);
    PolMul(cp,n->f2);
    cpf = PolFactorization(cp);

    /* Try all factors of degree >= 2
       ------------------------------ */
    for (factor = 0; factor < (int)cpf->len && !success; ++factor)
    {
	if (cpf->p[factor]->deg < 2)
	    continue;
	success = try_ex_factor(n,cp,cpf,factor);
    }

    FpolFree(cpf);
    PolFree(cp);
    return success;
}




/* ------------------------------------------------------------------
   chop_word() - Chop a module using a given word. Returns 1 on
   success, 0 otherwise.
   ------------------------------------------------------------------ */

static int chop_word(node_t *n, long wn)

{
    long dlimit = opt_deglimit;		/* Limit on degree */
    int pi, done;
    fpoly_t *cpol;

    CharPolSeed = (CharPolSeed + 1) % n->dim + 1;
    makeword(n,wn);
    cpol = make_f1(n);		/* Make first part of c(x) */

    /* If c(x) is irreducible, then the module is irreducible
       ------------------------------------------------------ */
    if (cpol->p[0]->deg == n->dim)
    {
	MESSAGE(2,("c(x) is irreducible\n"));
	newirred(n);
	++stat_cpirred;
	SetInsert(goodwords,wn);
	FpolFree(cpol);
	return 1;
    }

    /* Try all factors of c(x)
       ----------------------- */
    for (done = pi = 0; !done && pi < (int) cpol->len; ++pi)
    {    
	if (MSG3) {
	    printf("Next factor: ");
	    PolPrint(NULL,cpol->p[pi]);
	    printf("^%ld\n",cpol->e[pi]);
	}
	if (dlimit > 0 && cpol->p[pi]->deg > dlimit)
	{
	   MESSAGE(3,("deg > %ld -- discarded\n",dlimit));
	   continue;
	}
	done = try_poly(n,cpol->p[pi],cpol->e[pi]);
	MESSAGE(3,("try_poly()=%d\n",done));

    }
    FpolFree(cpol);

    if (!done)
        done = try_exceptional(n);
        
    return done;
}




/* ------------------------------------------------------------------
   Chop() - Chop a module
   ------------------------------------------------------------------ */

static int Chop(node_t *n)

{
    long w;

    MESSAGE(0,("Chop: Dim=%ld\n",n->dim));

    /* Check if the module is one-dimensional
       -------------------------------------- */
    if (n->dim == 1)
    {
	MESSAGE(1,("Dimension is one -- irreducible!\n"));
	newirred(n);
	++stat_irred;
	return 0;
    }

    /* Try to split with saved vectors
       ------------------------------- */
    if (split_with_saved_vectors(n))
	return 0;

    /* Try words
       --------- */
    for (w = 0; ; )
    {
	long i;

	/* Select the next word. First we try all 'good' words
	   then all words starting with 1. Known 'bad' words are
           filtered out.
           ----------------------------------------------------- */
	if ((w <= 0) && (int)-w < (int)goodwords->len)
	    i = goodwords->buf[(int) -(w--)];
	else 
	{
	    if (w <= 0)
		w = i = 1;	/* Start with word 1 */
	    else
		i = ++w;
	    if (SetContains(goodwords,w))  /* Do not try again */
		continue;
	}
	if (isbadword(i,n)) 
	    continue;

	MESSAGE(2,("Next word is %ld (=%s)\n",i,WgSymbolicName(n->wg,i)));
	if (chop_word(n,i))
	    break;
	/*SetInsert(n->badwords,i);*/
    }
    return 0;
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{   
    int i;

    MtxInit();
    goodwords = SetAlloc();
    opt_nullimit = zfl > 10 ? 2 : 3;

    memset(&LI,0,sizeof(Lat_Info));
    LI.NGen = 2;

    /* Parse command line
       ------------------ */
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("Gs:g:n:d:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 's':
		firstword = MtxGetInt();
		if (firstword < 1 || *opt_text_ptr != 0)
		    MtxErrorExit("-s: %E",ERR_OPTION);
		break;
	    case 'g':
		LI.NGen = MtxGetInt();
		if (LI.NGen < 1 || LI.NGen > MAXGEN || *opt_text_ptr != 0)
		    MtxErrorExit("-g: %E",ERR_OPTION);
		break;
	    case 'd':
		opt_deglimit = MtxGetInt();
		if (opt_deglimit < 0 || *opt_text_ptr != 0)
		    MtxErrorExit("-d: %E",ERR_OPTION);
		break;
	    case 'n':
		opt_nullimit = MtxGetInt();
		if (opt_nullimit < 1 || *opt_text_ptr != 0)
		    MtxErrorExit("-n: %E",ERR_OPTION);
		break;
	}
    }

    if (opt_ind != argc-1) 
	MtxErrorExit("%s: %E",pinfo.name,ERR_NARGS);

    /* Initialize the lattice data structure
       ------------------------------------- */
    strcpy(LI.BaseName,argv[argc-1]);

    MESSAGE(0,("*** CHOP MODULE ***\n\n"));
    init();
    Chop(root);
    writeinfo(root);
    MtxCleanup();
    return 0;
}


