/* ============================= C MeatAxe ==================================
   File:        $Id: error.c,v 1.1 1998/04/28 17:36:07 mringe Exp $
   Comment:     Error handling
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>



static void (*ErrorHandler)(const MtxErrorRecord *err) = NULL;
static FILE *LogFile = stderr;
static MtxErrorRecord Error = { NULL,0,NULL,"" };




static void DefaultHandler(const MtxErrorRecord *e)

{
    if (e->FunctionName != NULL)
    {
	fprintf(LogFile,"%s(%d), %s(): %s\n",e->FileName,e->LineNo,
	    e->FunctionName,e->Text);
    }
    else
	fprintf(LogFile,"%s(%d): %s\n",e->FileName,e->LineNo,e->Text);
    exit(255);
}


/*{{{ --------------------------------------------------------------------------
!function MtxSetErrorHandler "Define an application error handler"
!section general.err
!synopsis 
    typedef struct { ... } MtxErrorRecord;
    typedef void MtxErrorHandler(const MtxErrorRecord *err);
    MtxErrorHandler *MtxSetErrorHandler(MtxErrorHandler *h);
!arg h
    Address of the new error handler or |NULL| to restore the default handler.
!return
    Returns the previous error handler.
!description
    This function defines an application error handler which is called evera
    time an error occurs inside the {\MeatAxe} library.
!seealso @errmsg MtxErrorRecord MtxErrorHandler
-------------------------------------------------------------------------- }}}*/

MtxErrorHandler *MtxSetErrorHandler(MtxErrorHandler *h)

{
    MtxErrorHandler *old = ErrorHandler;
    ErrorHandler = h;
    return old;
}


/*{{{ --------------------------------------------------------------------------
!function MtxError "Signal an error"
!section general.err
!synopsis 
    void MtxError(const char *file, int line, const char *func, 
	const char *msg, ...)
!arg file
    Name of the source file where the error occured.
!arg line
    Line number where the error occured.
!arg func
    Name of the Function where the error occured.
!arg msg
    Error description.
!arg ...
    Optional arguments to be inserted into the error description.
!return
    The return value is always 0.
!description
    This function is mainly used internally by the {\MeatAxe} library to
    report errors. If an application error handler has been installed, it 
    is executed. Otherwise an error message is written to the error log 
    file and the program is terminmated.

    The error message, |msg|, is formatted with |MtxFormatMessage()|.
!seealso @errmsg
-------------------------------------------------------------------------- }}}*/

int MtxError(const char *file, int line, const char *func, const char *msg, ...)

{
    const char *fn;
    va_list al;

    /* Remove directory names from file name
       ------------------------------------- */
    for (fn = file; *fn != 0; ++fn);
    while (fn != file && fn[-1] != '/' && fn[-1] != '\\')
	--fn;

    /* Create error record
       ------------------- */
    Error.FileName = fn;
    Error.LineNo = line;
    Error.FunctionName = func;
    va_start(al,msg);
    MtxFormatMessage(Error.Text,sizeof(Error.Text),msg,al);
    va_end(al);

    /* Call the error handler
       ---------------------- */
    if (ErrorHandler != NULL)
	ErrorHandler(&Error);
    else
	DefaultHandler(&Error);

    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function MtxErrorExit "Exit program with error message"
!section general.err
!synopsis 
    int MtxErrorExit(const char *msg, ...);
!arg msg
    Message to print or |NULL|.
!arg ...
    Optional arguments to be inserted into the message.
!description
    This function prints an error message to the error log file (|stderr| by 
    default) and terminates the calling process. |msg| is the error message 
    and may contain special format characters. See |MtxFormatMessage()| for 
    a full description. In particular, |%E| takes an integer argument and 
    produces the corresponding {\MeatAxe} error message.
!seealso @errmsg
-------------------------------------------------------------------------- }}}*/

int MtxErrorExit(const char *fmt, ...)

{
    char tmp[MTX_MAXERRORTEXT];
    va_list vl;

    va_start(vl,fmt);
    MtxFormatMessage(tmp,sizeof(tmp),fmt,vl);
    fputs(tmp,stderr);
    putc('\n',stderr);
    va_end(vl);
    exit(1);
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!text general.err
Errors can occur inside the {\MeatAxe} library for several reasons including
\begin{itemize}
\item insufficient system resources (memory, disk space),
\item device or operating system failure,
\item invalid parameters passed to a {\MeatAxe} library function,
\item invalid usage of {\MeatAxe} library functions, for example missing
    initialization.
\end{itemize}
Note that invalid parameters are not always detected by the {\MeatAxe} library.
For example, most kernel functions such as |zadd()| do not check their 
arguments for the sake of performance. Thus, calling these function with 
invalid arguments may produce random results or even crash the program. 
On the other hand, most higher-level functions like |MatAdd()| do some
plausibility checks on their arguments before processing them.

When an error is detected, the default action is to terminate the program
immediately with an appropriate error message. While this minimizes the chance 
of not noticing an error, it may be undesirable in some situations. For example,
the application may be able to recover from the error. In order to prevent 
the {\MeatAxe} library from terminating the program, the application can
define an application error handler. This is a function which is called on
each error.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!structure MtxErrorRecord "Error data structure"
!section general.err
!synopsis 
typedef struct
{
    const char *FileName;
    int LineNo;
    const char *FunctionName;
    char Text[MTX_MAXERRORTEXT];
} MtxErrorRecord;

!description
    This data structure contains detailed information on an error that occured
    inside the {\MeatAxe} library.
!seealso MtxSetErrorHandler
-------------------------------------------------------------------------- }}}*/

/*{{{ --------------------------------------------------------------------------
!structure MtxErrorHandler "Error handler function type"
!section general.err
!synopsis 
    typedef void MtxErrorHandler(const MtxErrorRecord *err);
!description
    This is the type of an application error handler.
!seealso MtxSetErrorHandler
-------------------------------------------------------------------------- }}}*/

