/* ============================= C MeatAxe ==================================
   File:        $Id: fpolmul.c,v 1.2 1998/07/29 16:52:00 mringe Exp $
   Comment:     Factored polynomials: multiplication
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

#define BLOCKSIZE 5	/* Granularity */



/*{{{ --------------------------------------------------------------------------
!function FpolMulP "Multiply with an irreducible polynomial"
!section pol.factor
!synopsis 
    fpoly_t *FpolMulP(fpoly_t *dest, const poly_t *src, long pwr);
!arg dest
    Factored polynomial to modify.
!arg src
    Irreducible polynomial.
!arg pwr
    Power of the irreducible polynomial.
!return
    The function returns |dest| or |NULL| on error.
!description
    This function multiplies a factored polynomial with the power of an
    an irreducible factor.
!seealso @fpolmul
-------------------------------------------------------------------------- }}}*/

fpoly_t *FpolMulP(fpoly_t *dest, const poly_t *src, long pwr)

{
    MTXCONTEXT(FpolMulP);
    int i;

    /* Check the arguments
       ------------------- */
    if (!POLYNOMIAL_IS_VALID(src) || !FPOLY_IS_VALID(dest))
	return NULL;

    for (i = 0; i < (int) dest->len && PolCompare(dest->p[i],src) < 0; ++i);
    if (i >= (int) dest->len || PolCompare(dest->p[i],src))
    {
	int k;
	if (dest->len >= dest->max)	/* extend */
	{
	    size_t newsize = dest->max + BLOCKSIZE;
	    poly_t **x = NREALLOC(dest->p,poly_t *,newsize);
	    long *e = NREALLOC(dest->e,long,newsize);
	    if (e == NULL || x == NULL) 
	    {
		MTXERROR("Cannot grow: %S");
		return NULL;
	    }
	    dest->p = x;
	    dest->e = e;
	    dest->max = newsize;
	}

	/* Make room for the new factor
	   ---------------------------- */
	for (k = dest->len; k > i; --k)
	{
	    dest->p[k] = dest->p[k-1];
	    dest->e[k] = dest->e[k-1];
	}
	++dest->len;

	/* Insert new factor
	   ----------------- */
	dest->p[i] = PolDup(src);
	dest->e[i] = pwr;
    }
    else
	dest->e[i] += pwr;
    return dest;
}



/*{{{ --------------------------------------------------------------------------
!function FpolMul "Multiply factored polynomials"
!section pol.factor
!synopsis 
    fpoly_t *FpolMul(fpoly_t *dest, const fpoly_t *src);
!arg dest
    Factored polynomial to modify.
!arg src
    Factored polynomial.
!return
    The function returns |dest| or |NULL| on error.
!description
    Multiplies the factored polynomial |dest| by the factored 
    polynomial |src|.
!seealso @fpolmul
-------------------------------------------------------------------------- }}}*/

fpoly_t *FpolMul(fpoly_t *dest, const fpoly_t *src)

{
    MTXCONTEXT(FpolMul);
    size_t i;

    /* Check the arguments
       ------------------- */
    if (!FPOLY_IS_VALID(src) || !FPOLY_IS_VALID(dest))
	return NULL;

    for (i = 0; i < src->len; ++i)
    {
	if (FpolMulP(dest,src->p[i],src->e[i]) == NULL)
	{
	    MTXERROR("Cannot multiply");
	    return NULL;
	}
    }
    return dest;
}

