/* ============================= C MeatAxe ==================================
   File:        $Id: fpoly.c,v 1.7 1998/04/30 19:18:55 mringe Exp $
   Comment:     Factored polynomials
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>
#include <stdlib.h>


#define BLOCKSIZE 5	/* Granularity */



/*{{{ --------------------------------------------------------------------------
!function FpolPrint "Print a factored polynomial"
!section pol.factor
!synopsis 
    void FpolPrint(const char *name, const fpoly_t *p);
!arg name
    Name of the polynomial, or |NULL|.
!arg p
    Pointer to the factored polynomial.
!description
    This function prints a factored polynomial in human-readable format
    on the standard output.
    If |name| is not |NULL|, ``name $=$'' is printed before the polynomial.
!seealso 
-------------------------------------------------------------------------- }}}*/

void FpolPrint(const char *name, const fpoly_t *p)

{
    size_t i;

    if (name != NULL) 
	printf("%s =",name);
    for (i = 0; i < p->len; ++i)
    {   
	if (i > 0) printf("      * ");
	printf("(");
	PolPrint(NULL,p->p[i]);
	printf(")^%ld\n",p->e[i]);
    }
}




/*{{{ --------------------------------------------------------------------------
!function FpolAlloc "Allocate a factored polynomial"
!section pol.factor
!synopsis 
    fpoly_t *FpolAlloc();
!return
    Pointer to a newly allocated |fpoly_t| structure, or |NULL| on error.
!description
    This function allocates and initializes a factored polynomial 
    (|fpoly_t|) data structure.
!seealso @fpolalloc
-------------------------------------------------------------------------- }}}*/

fpoly_t *FpolAlloc()

{
    MTXCONTEXT(FpolAlloc);
    fpoly_t *x;

    x = ALLOC(fpoly_t);
    if (x != NULL)
    {
    	x->id = T_FPOLY;
    	x->len = 0;
    	x->max = BLOCKSIZE;
    	x->p = NALLOC(poly_t *,BLOCKSIZE);
   	if (x->p != NULL)
	{
	    x->e = NALLOC(long,BLOCKSIZE);
	    if (x->e != NULL)
	    {
		return x;
	    }
	    free(x->p);
	}
	free(x);
    }
    MTXERROR("Cannot allocate fpoly_t: %S");
    return NULL;
}





/*{{{ --------------------------------------------------------------------------
!function FpolFree "Free a factored polynomial"
!section pol.factor
!synopsis 
    void FpolFree(fpoly_t *x);
!arg x
    Pointer to the factored polynomial.
!description
    This function frees a factored polynomial (|fpoly_t|) data structure,
    including any internal data structures.
!seealso @fpolalloc
-------------------------------------------------------------------------- }}}*/

void FpolFree(fpoly_t *x)

{
    MTXCONTEXT(FpolFree);
    int i;

    /* Check the argument
       ------------------ */
    if (!FPOLY_IS_VALID(x))
	return;

    for (i = 0; i < (int) x->len; ++i)
	PolFree(x->p[i]);
    free(x->p);
    free(x->e);
    free(x);
}



/*{{{ --------------------------------------------------------------------------
!function FpolDup "Duplicate a factored polynomial"
!section pol.factor
!synopsis 
    fpoly_t *FpolDup(const fpoly_t *s);
!arg s
    Pointer to a factored polynomial.
!return
    The function returns a pointer to a copy of |s|, or |NULL| on error.
!description
    This function creates a copy of a factored polynomial.
!seealso @fpolalloc
-------------------------------------------------------------------------- }}}*/

fpoly_t *FpolDup(const fpoly_t *s)

{
    MTXCONTEXT(FpolDup);
    fpoly_t *x;

    /* Check the argument
       ------------------ */
    if (!FPOLY_IS_VALID(s))
	return NULL;

    x = ALLOC(fpoly_t);
    if (x != NULL)
    {
    	x->id = T_FPOLY;
    	x->len = x->max = s->len;
    	x->p = NALLOC(poly_t *,s->len);
   	if (x->p != NULL)
	{
	    x->e = NALLOC(long,s->len);
	    if (x->e != NULL)
	    {
		int i;
		for (i = 0; i < (int) s->len; ++i)
		{
		    x->e[i] = s->e[i];
		    if ((x->p[i] = PolDup(s->p[i])) == NULL)
			break;
		}
		if (i >= (int) s->len)
		    return x;		/* Done */

		/* Clean up after error */
		while (--i > 0) 
		    PolFree(x->p[i]);
	    }
	    free(x->p);
	}
	free(x);
    }

    MTXERROR("Cannot allocate fpoly_t: %S");
    return NULL;
}


