/* ============================= C MeatAxe ==================================
   File:        $Id: gauss.c,v 1.7 1998/07/29 16:52:02 mringe Exp $
   Comment:     Null-space, echelon form etc.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */



#include "meataxe.h"
#include <string.h>
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static long *piv = NULL;		/* Pivot table */
static long maxnoc = -1;		/* Table size */


/* ------------------------------------------------------------------
   pivalloc() - Allocate pivot table. All functions in this module
	use one global pivot table (piv). pivalloc() checks first
	if the currently allocated table is large enough.
   ------------------------------------------------------------------ */

static void pivalloc(long noc)

{
    if (noc <= maxnoc)
	return;
    maxnoc = noc;
    if (piv != NULL)
	free(piv);
    piv = NALLOC(long,noc+1);
    if (piv == NULL)
	MtxErrorExit("pivalloc(): out of memory");
}


/*{{{ --------------------------------------------------------------------------
!function MatEchelonize "Reduce to echelon form"
!section mat.op
!synopsis 
    matrix_t *MatEchelonize(matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    The function returns |mat|, or |NULL| on error.
!description
    This function performs a Gaussian elimination on the matrix |mat|. On
    return, |mat| is in echelon form. If the rank of |mat| was smaller than
    the number of rows, some rows are removed during the process.

    Note that this function modifies the original matrix. If you need a
    non-destructive Gaussian elimination use |MatEchelon()|.
!seealso @ech
-------------------------------------------------------------------------- }}}*/

matrix_t *MatEchelonize(matrix_t *mat)

{
    MTXCONTEXT(MatEchelonize);
    long rank;

    if (!MATRIX_IS_VALID(mat))
	return NULL;
    pivalloc(mat->nor);
    zsetfield(mat->fl);
    zsetlen(mat->noc);
    rank = zmkechelon(mat->d,mat->nor,piv);
    mat->nor = rank;
    mat->d = (PTR) Srealloc(mat->d,zsize(rank));
    return mat;
}




/*{{{ --------------------------------------------------------------------------
!function MatEchelon "Echelon form of a matrix"
!section mat.op
!synopsis 
    matrix_t *MatEchelon(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Echelon form of |mat|, or |NULL| on error.
!description
    This function calculates the echelon form of the matrix |mat|. Unlike
    |MatEchelonize()|, this function does not modify the original matrix. The result
    is a newly allocated matrix which must be freed by the caller.
!seealso @ech
-------------------------------------------------------------------------- }}}*/

matrix_t *MatEchelon(const matrix_t *mat)

{
    MTXCONTEXT(MatEchelon);
    matrix_t *tmp;

    if (!MATRIX_IS_VALID(mat))
	return NULL;
    if ((tmp = MatDup(mat)) == NULL)
    {
	MTXERROR("Cannot duplicate matrix");
	return NULL;
    }
    return MatEchelonize(tmp);
}



/*{{{ --------------------------------------------------------------------------
!function MatNullity "Nullity of a matrix"
!section mat.op
!synopsis 
    long MatNullity(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Nullity of |mat|, or $-1$ on error.
!description
    This function calculates the dimension of the null-space of a matrix.
    Unline |MatNullity__()|, this function does not modify the matrix.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

long MatNullity(const matrix_t *mat)

{
    MTXCONTEXT(MatNullity);
    matrix_t *ech;
    long nul;

    if (!MATRIX_IS_VALID(mat))
	return -1;
    ech = MatEchelon(mat);
    if (ech == NULL) 
	return MTXERROR("Cannot echelonize matrix"), -1;
    nul = ech->noc - ech->nor;
    MatFree(ech);
    return nul;
}



/*{{{ --------------------------------------------------------------------------
!function MatNullity__ "Nullity of a matrix"
!section mat.op
!synopsis 
    long MatNullity__(matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Nullity of |mat|, or $-1$ on error.
!description
    This function calculates the dimension of the null-space of a matrix
    and deletes the matrix.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

long MatNullity__(matrix_t *mat)

{
    long nul;
    MatEchelonize(mat);
    nul = mat->noc - mat->nor;
    MatFree(mat);
    return nul;
}




/*{{{ --------------------------------------------------------------------------
!function MatNullSpace_ "Null-space of a matrix"
!section mat.op
!synopsis 
    matrix_t *MatNullSpace_(matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Pointer to the null-space of |mat|, or |NULL| on error.
!description
    This function reduces a matrix to echelon form and returns its null-space. 
    Unlike |MatNullspace()|, |MatNullSpace_()| modifies the orginal matrix, 
    but uses less memory since no temporary workspace is allocated.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

matrix_t *MatNullSpace_(matrix_t *mat)

{
    MTXCONTEXT(MatNullspace_);
    long dim;
    matrix_t *nsp;


    /* Check arguments
       --------------- */
    if (!MATRIX_IS_VALID(mat))
	return NULL;

    /* Allocate workspace
       ------------------ */
    pivalloc(mat->nor);
    nsp = MatAlloc(mat->fl,mat->nor,mat->nor);	/* worst case */
    if (nsp == NULL) 
	return NULL;

    /* Calculate the null-space
       ------------------------ */
    zsetlen(mat->noc);
    dim = znullsp(mat->d,mat->nor,piv,nsp->d);

    /* Resize the result buffer to its actual size
       ------------------------------------------- */
    zsetlen(nsp->noc);
    nsp->nor = dim;
    nsp->d = (PTR) Srealloc(nsp->d,zsize(dim));

    /* There may be some zero rows at the end of <mat> now. Remove them.
       ----------------------------------------------------------------- */
    zsetlen(mat->noc);
    mat->nor = mat->nor - dim;
    mat->d = (PTR) Srealloc(mat->d,zsize(mat->nor));

    return nsp;
}




/*{{{ --------------------------------------------------------------------------
!function MatNullspace "Null-space of a matrix"
!section mat.op
!synopsis 
    matrix_t *MatNullspace(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Pointer to the null-space of |mat|, or |NULL| on error.
!description
    This function calculates the null-space of a matrix. Unlike |MatNullSpace_()|,
    |MatNullspace()| does not change the original matrix, but it temporarily 
    allocates a copy of the matrix.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

matrix_t *MatNullspace(const matrix_t *mat)

{
    MTXCONTEXT(MatNullspace);
    matrix_t *tmp, *nsp;

    /* Check arguments
       --------------- */
    if (!MATRIX_IS_VALID(mat))
	return NULL;

    /* Non-destructive null-space
       -------------------------- */
    if ((tmp = MatDup(mat)) == NULL)
    {
	MTXERROR("Cannot duplicate matrix");
	return NULL;
    }
    nsp = MatNullSpace_(tmp);
    MatFree(tmp);
    return nsp;
}



/*{{{ --------------------------------------------------------------------------
!function MatNullSpace__ "Null-space of a matrix"
!section mat.op
!synopsis 
    matrix_t *MatNullSpace__(matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Pointer to the null-space of |mat|, or |NULL| on error.
!description
    This function calculates the null-space of a matrix and deletes the
    matrix.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

matrix_t *MatNullSpace__(matrix_t *mat)

{
    matrix_t *nsp;
    nsp = MatNullSpace_(mat);
    MatFree(mat);
    return nsp;
}


