/* ============================= C MeatAxe ==================================
   File:        $Id: genmod.c,v 1.6 1998/04/28 17:36:12 mringe Exp $
   Comment:     This program calculates a basis for any submodule listet in
		the .out file.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

matrix_t *gen[MAXGEN];		/* Generators for the algebra */
matrix_t *mountains;		/* Genrators for all mountains */
int nmount;			/* Number of mountains */
int modnum;
bitstring_t *bs;		/* Bit string read from .sub file */
int opt_m = 0;			/* Option -m used */
static Lat_Info LI;		/* Data from .cfinfo file */

static char *helptext[] = {
"SYNTAX",
"    genmod [-QVm] <Name> <Number>",
"",
"OPTIONS",
"    -m    Make mountain (works after mkinc)",
"    -V    Verbose",
"    -Q    Quiet, no messages",
"",
"FILES",
"    <Name>.sub       i  Submodule information (generated by mksub)",
"    <Name>.v         i  Mountains",
"    <Name>.s<Number> o  The result (without -m)",
"    <Name>.m<Number> o  The result (with -m)",
NULL};

static proginfo_t pinfo =
   { "genmod", "Make Submodule", "$Revision: 1.6 $", helptext };




/* -----------------------------------------------------------------
   init() - Read generators and mountains
   ----------------------------------------------------------------- */

static void init(const char *basename, char *arg)

{
    char fn[200];
    FILE *f;
    int i;

    MtxInit();
    if (Lat_ReadInfo(&LI,basename) != 0)
	MtxErrorExit("Error reading %s.cfinfo",basename);

    /* Read the generators
       ------------------- */
    for (i = 0; i < LI.NGen; ++i)
    {
    	sprintf(fn,"%s.%d",LI.BaseName,i+1);
	gen[i] = MatLoad(fn);
    }

    /* Read the mountains
       ------------------ */
    mountains = MatLoad(strcat(strcpy(fn,LI.BaseName),".v"));
    nmount = (int) mountains->nor;
    bs_setlen(nmount);
    MESSAGE(1,("%d mountains\n",nmount));
    fflush(stdout);
    

    /* Read the bit string from xxx.sub or set up the bit string
       to contain only the requested mountain (-m)
      ------------------------ --------------------------------- */
    modnum = atoi(arg);
    if (opt_m)
    {
	bs = bs_alloc();
	bs_reset(bs);
	bs_set(bs,modnum);
    }
    else
    {
    	f = SFOpen(strcat(strcpy(fn,LI.BaseName),".sub"),FM_READ);
    	if (f == NULL)
	    MtxErrorExit("CANNOT OPEN .sub FILE");
    	SFSeek(f,modnum*bs_size);
    	bs = bs_read(f);
    	if (MSG1)
    	{
	    printf("Mountains: ");
	    for (i = 0; i < nmount; ++i)
	    	if (bs_test(bs,i)) printf("%d ",i);
	    printf("\n");
    	}
	fclose(f);
    }
}



/* -----------------------------------------------------------------
   sp()
   ----------------------------------------------------------------- */

static void sp()

{
    int i;
    matrix_t *m, *subsp;
    PTR p;
    char fn[200];

    m = MatAlloc(gen[0]->fl,(long)nmount,gen[0]->noc);
    p = m->d;
    for (i = 0; i < nmount; ++i)
    {
	if (bs_test(bs,i))
	{
	    PTR q = mountains->d;
	    zadvance(&q,i);
	    zmoverow(p,q);
	    zadvance(&p,1);
	}
    }
    MatEchelonize(m);
    MESSAGE(0,("Seed space has dimension %ld\n",m->nor));
    subsp = SpinUp(m,LI.NGen,(const matrix_t **)gen,SF_EACH|SF_COMBINE);
    MESSAGE(0,("Submodule has dimension %ld\n",subsp->nor));
    sprintf(fn,"%s.%c%d",LI.BaseName,opt_m ? 'm' : 's',modnum);
    MatSave(subsp,fn);
    MESSAGE(0,("Module written to `%s'\n",fn));
}


/* -----------------------------------------------------------------
   main()
   ----------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    int i;

    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("m")) != OPT_END)
    {
	switch (i)
	{
	    case 'm': opt_m = 1; break;
	}
    }
    if (opt_ind != argc-2)
	MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
    init(argv[opt_ind],argv[opt_ind+1]);
    sp();
    MtxCleanup();
    return 0;
}


