/* ============================= C MeatAxe ==================================
   File:        $Id: matcopy.c,v 1.2 1998/06/25 19:02:33 mringe Exp $
   Comment:     Cut a rectangular piece out of a matrix.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"



/*{{{ --------------------------------------------------------------------------
!function MatCopyRegion "Copy a rectanglular region of a matrix"
!section mat.op
!synopsis 
    int MatCopyRegion(matrix_t *dest, int destrow, int destcol, 
	const matrix_t *src, int row1, int col1, int nrows, int ncols)
!arg dest
    Pointer to the destination matrix.
!arg destrow
    Destination row.
!arg destcol
    Destination column.
!arg src
    Pointer to the source matrix.
!arg row1
    First row in region.
!arg col1
    First column in region.
!arg nrows
    Number of rows to cut. $-1$ means as many rows as possible.
!arg ncols
    Number of columns to cut. $-1$ means as many columns as possible.
!return
    Returns $0$ on success, or $-1$ on error.
!description
    This function copies a rectangular region of a matrix into another matrix.
    |row1| and |col1| define the upper left corner of the region. |nrows| and 
    |ncols| are the height (number of 
    rows) and width (number of columns) of the region. Both |nrows| and |ncols| 
    can be $-1$. In this case the region extends up to the last row or last 
    column, respectivly. |destrow| and |destcol| specify where the rectangle
    is placed inside |dest|. As uusual, all row and column indexes start with
    1, not 0.

    |src| and |dest| muts be matrices over the same field. Both source and 
    destination rectangle must not exceed the size of the matrix. In particular,
    it is not possible to extend |dest|'s size with this function.
!seealso MatCut
-------------------------------------------------------------------------- }}}*/

int MatCopyRegion(matrix_t *dest, int destrow, int destcol, 
    const matrix_t *src, int row1, int col1, int nrows, int ncols)
{
    MTXCONTEXT(MatCopyRegion);
    PTR s, d;
    int i, k;

    /* Check the arguments
       ------------------- */
    if (!MATRIX_IS_VALID(src) || !MATRIX_IS_VALID(dest))
	return -1;
    if (src->fl != dest->fl)
	return MTXERROR1("%E",ERR_INCOMPAT), -1;
    if (nrows == -1)
	nrows = src->nor - row1 + 1;
    if (ncols == -1)
	ncols = src->noc - col1 + 1;
    if (row1 < 1 || nrows < 0 || row1 + nrows - 1 > src->nor)
	return MTXERROR("Source row index out of bounds"), -1;
    if (col1 < 1 || ncols < 0 || col1 + ncols - 1 > src->noc)
	return MTXERROR("Source column index out of bounds"), -1;
    if (destrow < 1 || destrow + nrows - 1 > dest->nor)
	return MTXERROR("Destination row index out of bounds"), -1;
    if (destcol < 1 || destcol + ncols - 1 > dest->noc)
	return MTXERROR("Destination column index out of bounds"), -1;

    /* Initialize data pointers
       ------------------------ */
    zsetfield(src->fl);
    zsetlen(src->noc);
    s = src->d;
    zadvance(&s,row1-1);
    zsetlen(dest->noc);
    d = dest->d;
    zadvance(&d,destrow-1);

    /* Copy the rectangle
       ------------------ */
    for (i = row1; i < row1 + nrows; ++i)
    {
	for (k = col1; k < col1 + ncols; ++k)
	    zinsert(d,destcol+k-col1,zextract(s,k));
	zsetlen(src->noc);
	zadvance(&s,1);
	zsetlen(dest->noc);
	zadvance(&d,1);
    }

    return 0;
}




