/* ============================= C MeatAxe ==================================
   File:        $Id: matcut.c,v 1.3 1998/07/29 16:52:04 mringe Exp $
   Comment:     Cut a rectangular piece out of a matrix.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"




/*{{{ --------------------------------------------------------------------------
!function MatCut "Cut a rectangle out of a matrix"
!section mat.op
!synopsis 
    matrix_t *MatCut(const matrix_t *src, int row1, int col1, int nrows,
	int ncols);
!arg src
    Pointer to the matrix.
!arg row1
    First row in region.
!arg col1
    First column in region.
!arg nrows
    Number of rows to cut. $-1$ means as many rows as possible.
!arg ncols
    Number of columns to cut. $-1$ means as many columns as possible.
!return
    Pointer to a new matrix containing the specified region of |src|, or 
    |NULL| on error.
!description
    This function cuts a rectangular region out of a matrix. |row1| and |col1|
    define the upper left corner of the region. As usual, indexes start with 
    1, not 0. |nrows| and |ncols| are the height (number of rows) and width 
    (number of columns) of the region. Both |nrows| and |ncols| can be $-1$.
    In this case the region extends up to the last row or last column, 
    respectivly. For example, to extract the first 10 rows from a matrix 
    independently of the number of columns, you could say 
    |MatCut(mat,1,1,10,-1)|.
    The rectangle defined by |row1|, |col1|, |nrows| and |ncols| must not
    exceed the size of the matrix.

    The function returns a new matrix containing a copy of the specified
    rectangle. In case of error, the return value is |NULL|.
!seealso MatCopyRegion
-------------------------------------------------------------------------- }}}*/

matrix_t *MatCut(const matrix_t *src, int row1, int col1, int nrows, int ncols)
{
    MTXCONTEXT(MatCut);
    matrix_t *result;
    PTR s, d;
    int n;
    
    /* Check arguments
       --------------- */
    if (!MATRIX_IS_VALID(src))
	return NULL;
    if (nrows == -1)
	nrows = src->nor - row1 + 1;
    if (ncols == -1)
	ncols = src->noc - col1 + 1;
    if (row1 < 1 || nrows < 0 || row1 + nrows - 1 > src->nor)
    {
	MTXERROR("Source row index out of bounds");
	return NULL;
    }
    if (col1 < 1 || ncols < 0 || col1 + ncols - 1 > src->noc)
    {
	MTXERROR("Source column index out of bounds");
	return NULL;
    }

    /* Allocate a new matrix for the result
       ------------------------------------ */
    result = MatAlloc(src->fl,nrows,ncols);
    if (result == NULL)
	return NULL;

    /* Initialize pointers to the source and destination matrix 
       -------------------------------------------------------- */
    s = src->d;
    zsetlen(src->noc);
    zadvance(&s,row1-1);
    d = result->d;

    /* Copy the requested data
       ----------------------- */
    for (n = nrows; n > 0; --n)
    {
	zsetlen(ncols);
	if (col1 == 1)
	    zmoverow(d,s);
	else
	{
	    register int k;
	    for (k = 1; k <= ncols; ++k)
		zinsert(d,k,zextract(s,col1+k-1));
	}
	zadvance(&d,1);
	zsetlen(src->noc);
	zadvance(&s,1);
    }

    return result;
}




