/* ============================= C MeatAxe ==================================
   File:        $Id: matins.c,v 1.6 1998/07/29 16:52:05 mringe Exp $
   Comment:     Insert a matrix into a polynomial
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"

long tmatinsert = 0;


/*{{{ --------------------------------------------------------------------------
!function MatInsert_ "Insert a matrix into a polynomial"
!section mat.op pol.op
!synopsis 
    matrix_t *MatInsert_(matrix_t *mat, const poly_t *pol);
!arg mat
    Pointer to the matrix.
!arg pol
    Pointer to the polynomial.
!return
    The function returns |mat|, or |NULL| on error.
!description
    This function inserts a matrix into a polynomial. This function is 
    destructive. It uses |mat| for the result of the insertion, and does not
    allocate a new matrix for the result.

    |mat| must be a square matrix, and |pol| must be over the same field 
    as |mat|.
!seealso @matinsert
-------------------------------------------------------------------------- }}}*/

matrix_t *MatInsert_(matrix_t *mat, const poly_t *pol)

{
    MTXCONTEXT(MatInsert_);
    matrix_t *x = NULL;
    int i;
    long nor;
    long l;
    PTR v;
    FEL f;

    /* Check the arguments
       ------------------- */
    if (!MATRIX_IS_VALID(mat))
	return NULL;
    if (!POLYNOMIAL_IS_VALID(pol))
	return NULL;
    if ((nor = mat->nor) != mat->noc) 
    {
	MTXERROR1("%E",ERR_NOTSQUARE);
	return NULL;
    }
    if (mat->fl != pol->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    zsetfield(mat->fl);
    zsetlen(mat->noc);

    /* Special case: p(x) = 0
       ---------------------- */
    if (pol->deg == -1)
    {
	for (l = 1, v = mat->d; l <= mat->nor; ++l, zadvance(&v,(long)1))
	    zmulrow(v,F_ZERO);
	return mat;
    }

    /* Special case: deg(p) = 0
       ------------------------ */
    if (pol->deg == 0)
    {
	for (l = 1, v = mat->d; l <= mat->nor; ++l, zadvance(&v,(long)1))
	{
	    zmulrow(v,F_ZERO);
	    zinsert(v,l,pol->buf[0]);
	}
	return mat;
    }

    /* Evaluate p(A)
       ------------- */
    if (pol->deg > 1) 
	x = MatDup(mat);
    if ((f = pol->buf[pol->deg]) != F_ONE)
    {
	for (l = nor, v = mat->d; l > 0; --l, zadvance(&v,(long)1))
	    zmulrow(v,f);
    }
    for (i = pol->deg - 1; i >= 0; --i)
    {
        if ((f = pol->buf[i]) != F_ZERO)
        {
	    for (l = 1, v = mat->d; l <= nor; ++l, zadvance(&v,(long)1))
		zinsert(v,l,zadd(zextract(v,l),f));
        }
	if (i > 0) MatMul(mat,x);
    }
    if (pol->deg > 1) MatFree(x);
    return mat;
}



/*{{{ --------------------------------------------------------------------------
!function MatInsert "Insert a matrix into a polynomial"
!section mat.op pol.op
!synopsis 
    matrix_t *MatInsert(const matrix_t *mat, const poly_t *pol);
!arg mat
    Pointer to the matrix.
!arg pol
    Pointer to the polynomial.
!return
    The function returns |mat|, or |NULL| on error.
!description
    This function inserts a matrix into a polynomial. Unlike |MatInsert_()|,
    this function is not destructive. It allocates a new matrix for the result
    and does not modify the original matrix.

    |mat| must be a square matrix, and |pol| must be over the same field 
    as |mat|.
!seealso @matinsert
-------------------------------------------------------------------------- }}}*/

matrix_t *MatInsert(const matrix_t *mat, const poly_t *pol)

{
    MTXCONTEXT(MatInsert);
    matrix_t *x;
    int i;
    long nor = mat->nor;
    long l;
    PTR v;
    FEL f;

    if (!MATRIX_IS_VALID(mat))
	return NULL;
    if (!POLYNOMIAL_IS_VALID(pol))
	return NULL;
    if (nor != mat->noc) 
    {
	MTXERROR1("%E",ERR_NOTSQUARE);
	return NULL;
    }
    if (mat->fl != pol->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Special case: p = 0
       ------------------- */
    if (pol->deg == -1)
    {
	x = MatAlloc(mat->fl,mat->nor,mat->nor);
	return x;
    }

    /* Special case: deg(p) = 0
       ------------------------ */
    if (pol->deg == 0)
    {
	x = MatAlloc(mat->fl,mat->nor,mat->nor);
	for (l = 1, v = x->d; l <=mat->nor; ++l, zadvance(&v,(long)1))
	    zinsert(v,l,pol->buf[0]);
	return x;
    }

    /* Evaluate p(A)
       ------------- */
    x = MatDup(mat);
    if ((f = pol->buf[pol->deg]) != F_ONE)
    {
	for (l = nor, v = x->d; l > 0; --l, zadvance(&v,(long)1))
	    zmulrow(v,f);
    }
    for (i = pol->deg - 1; i >= 0; --i)
    {
        if ((f = pol->buf[i]) != F_ZERO)
        {
	    for (l = 1, v = x->d; l <= nor; ++l, zadvance(&v,(long)1))
		zinsert(v,l,zadd(zextract(v,l),f));
        }
	if (i > 0) 
	    MatMul(x,mat);
    }
    return x;
}

