/* ============================= C MeatAxe ==================================
   File:        $Id: matorder.c,v 1.1 1998/05/26 17:42:01 mringe Exp $
   Comment:     Order of a matrix
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>
#include <string.h>


/*{{{ --------------------------------------------------------------------------
!function MatOrder "Order of a matrix"
!section mat.op
!synopsis 
    long MatOrder(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    The order of |mat|, or $-1$ on error.
!description
    This function calculates the order of a matrix. |mat| must be a 
    non-singular, square matrix. 
    
    Even if |mat| is non-singular, the function may fail. This happens if
    the order is greater than 1000000, or if the order on any cyclic
    subspace is grater than 1000.

!seealso
-------------------------------------------------------------------------- }}}*/

long MatOrder(const matrix_t *mat)

{
    MTXCONTEXT(MatOrder);
    PTR m1, v1, v2, v3;
    PTR basis, bend, bptr;
    long *piv;
    char *done;
    long dim;
    long j1, i;
    FEL f1;
    long tord;
    long ord;
    int flag;

    /* Check the matrix
       ---------------- */
    if (!MATRIX_IS_VALID(mat))
	return -1;
    if (mat->nor != mat->noc)
	return MTXERROR1("%E",ERR_NOTSQUARE), -1;

    zsetfield(mat->fl);
    zsetlen(mat->noc);
    m1 = zalloc(mat->nor);
    memcpy(m1,mat->d,zsize(mat->nor));
    bend = basis = zalloc(mat->nor+1);

    piv = NALLOC(long,mat->nor+2);
    done = NALLOC(char,mat->nor+1);
    memset(done,0,(size_t)((mat->nor+1)));
    v1 = zalloc((long)1);
    v2 = zalloc((long)1);
    v3 = zalloc((long)1);
    tord = ord = 1;
    dim = 0;
    j1 = 1;
    while (dim < mat->nor && tord <= 1000 && ord <= 1000000)
    {
	/* Get next start vector
	   --------------------- */
	for (j1 = 1; j1 <= mat->nor && done[j1]; ++j1);
	if (j1 > mat->nor) break;	/* Done! */
	zmulrow(v1,F_ZERO);
	zinsert(v1,j1,F_ONE);

	/* Calculate order on cyclic subspace
	   ---------------------------------- */
	tord = 0;
	flag = 1;
	zmoverow(v3,v1);
	do
	{   zmoverow(v2,v3);
	    if (flag)
	    {   zmoverow(bend,v3);
		bptr = basis;
		for (i = 1; i <= dim; ++i)
		{   f1 = zextract(bend,piv[i]);
		    if (f1 != 0)
		    {
			zaddmulrow(bend,bptr,zneg(zdiv(f1,
			    zextract(bptr,piv[i]))));
		    }
		    zadvance(&bptr,(long)1);
		}
		if ((piv[dim+1] = zfindpiv(bend,&f1)) != 0)
		{   done[piv[++dim]] = 1;
		    zadvance(&bend,(long)1);
		}
		else
		    flag = 0;
	    }
	    zmaprow(v2,m1,mat->nor,v3);
	    ++tord;
	}
	while (tord <= 1000 && zcmprow(v3,v1) != 0);

	/* Order = lcm(orders on cyclic subspaces)
	   --------------------------------------- */
	ord = lcm(ord,tord);
    }

    /* Clean up
       -------- */
    free(done);
    free(v1);
    free(v2);
    free(v3);
    free(m1);
    free(basis);
    free(piv);

    if (tord > 1000 || ord > 1000000)
	return -1;
    return ord;
}
