/* ============================= C MeatAxe ==================================
   File:        $Id: meataxe.h,v 1.18 1998/09/16 15:39:26 mringe Exp $
   Comment:     MeatAxe library external interfaces.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#ifndef _MEATAXE_H_
#define _MEATAXE_H_

#include <stdio.h>
#include <stdarg.h>



/* ------------------------------------------------------------------
   System-dependent Functions (os.c)
   ------------------------------------------------------------------ */

void SInit(void);
long STimeUsed(void);
void STimeLimit(long nsecs);

void *Smalloc(size_t nbytes);
void *Srealloc(void *buf, size_t nbytes);

/* Files */
#define FM_READ	    0x01
#define FM_CREATE   0x02
#define FM_APPEND   0x03
#define FM_TEXT	    0x10
#define FM_LIB	    0x20
#define FM_NOERROR  0x40

FILE *SFOpen(const char *name, int mode);
int SFSeek(FILE *f,long pos);


/* ------------------------------------------------------------------
   malloc Macros
   ------------------------------------------------------------------ */

#define ALLOC(type) ((type *) Smalloc(sizeof(type)))
#define NALLOC(type,n) ((type *) Smalloc((size_t)(n) * sizeof(type)))
#define NREALLOC(x,type,n)\
	((type *) Srealloc(x,(size_t)(n) * sizeof(type)))
#define FREE(x) free(x)


/* ------------------------------------------------------------------
   MeatAxe data types
   ------------------------------------------------------------------ */

typedef enum {
    T_INVALID = 0,	/* Not a valid type */
    T_MATRIX = 1,	/* Matrix over a finite field */
    T_PERM = 2,		/* Permutation */
    T_POLY = 3,		/* Polynomial over a finite field */
    T_SET = 4,		/* Set of integers */
    T_BITS = 5,	        /* Bit string */
    T_IMAT = 8,		/* Integer Matrix */
    T_FPOLY = 21	/* Factored polynomial */
} tMtxObjType;


/* ------------------------------------------------------------------
   Return codes
   ------------------------------------------------------------------ */

#define	EXIT_OK		0	/* Exit code: normal end */
#define EXIT_ERR	1	/*            error */


/* MeatAxe error codes */

#define ERR_NULL	0	/* */
#define ERR_NOMEM	1	/* Not enough memory */
#define ERR_GAMEOVER	2	/* Time limit exceeded */
#define ERR_DIV0	8	/* Division by 0 or singular Matrix */

#define ERR_FILE	20	/* File errors */
#define ERR_FILEOPEN	21	/* Could not open */
#define ERR_FILEREAD	22	/* Read error */
#define ERR_FILEWRITE	23	/* Write error */
#define ERR_FILEFMT	24	/* Bad format */
#define ERR_NFILES	25	/* Too many files */
#define ERR_EOF		26	/* Unexpected EOF */

#define ERR_ARGS	30	/* Arguments */
#define ERR_BADARG	31	/* Bad argument */
#define ERR_BADTYPE	32	/* Bad type */
#define ERR_RANGE	33	/* Out of range */
#define ERR_NOTECH	34	/* Matrix not in chelon form */
#define ERR_NOTSQUARE	35	/* Matrix not square */
#define ERR_INCOMPAT	36	/* Arguments are incompatible */

#define ERR_USAGE	40
#define ERR_BADUSAGE	41	/* Bad command line */
#define ERR_OPTION	42	/* Bad usage of option */
#define ERR_NARGS	43	/* Bad number of arguments */

#define ERR_NOTMATRIX	51	/* Not a matrix */
#define ERR_NOTPOLY	52	/* Not a polynomial */
#define ERR_NOTPERM	53	/* Not a permutation */

#define ERR_UNKNOWN	61	/* Unkown symbol */


/* ------------------------------------------------------------------
   Kernel specific part
   ------------------------------------------------------------------ */

#if !defined(ZZZ)		/* Assume default if undefined */
#define ZZZ zzz
#endif

/* ZZZ kernel for small fields (default)
   ------------------------------------- */
#if ZZZ==zzz

typedef unsigned char FEL;
typedef FEL *PTR;
#define F_ZERO ((FEL)0)
#define F_ONE ((FEL)1)
#define ZZZVERSION 6
extern FEL tmult[256][256];
extern FEL tadd[256][256];
extern FEL taddinv[256], tmultinv[256];
#define zadd(a,b) ((FEL)tadd[(unsigned char)a][(unsigned char)b])
#define zmul(a,b) ((FEL)tmult[(unsigned char)a][(unsigned char)b])
#define zneg(a) (taddinv[(unsigned char)a])
#define zinv(a) (tmultinv[(unsigned char)a])


/* BIGZZZ kernel for field orders up to 2^{16}
   ------------------------------------------- */ 

#elif ZZZ==bigzzz

typedef unsigned short FEL;
typedef unsigned short *PTR;
#define F_ZERO ((FEL)0xFFFF)
#define F_ONE ((FEL)0)
#define ZZZVERSION 0x101

FEL zadd(FEL a, FEL b);
FEL zneg(FEL a);
FEL zmul(FEL a, FEL b);
FEL zinv(FEL a);

#endif



/* ------------------------------------------------------------------
   Common definitinos for all kernels
   ------------------------------------------------------------------ */

extern long zfl, zchar;		/* Field order and characteristic */
extern FEL zgen;		/* Generator */
extern long znoc;		/* Number of columns for row ops */
extern size_t zrowsize;		/* Row size in memory */
extern size_t zrowsize_io;	/* Row size in files */
extern char zzzversion[];


/* Function prototypes. Some of these may be defined
   as macros in the kernel specific part.
   ------------------------------------------------- */

int zsetfield(long field);
FEL zembed(FEL a, long subfield);	/* Embed from subfield */
FEL zrestrict(FEL a, long subfield);	/* Restrict to subfield */
FEL zitof(long l);
long zftoi(FEL f);
int zsetlen(long ncols);		/* Set row size */
void zinsert(PTR row, long col, FEL mark);
FEL zextract(PTR row, long col);

long zfindpiv(PTR row, FEL *mark);
PTR zextractcol(PTR mat, long nor, long col, PTR result);
PTR zaddrow(PTR dest, PTR src);
PTR zmulrow(PTR row, FEL mark);
PTR zaddmulrow(PTR dest, PTR src, FEL f);
PTR zmaprow(PTR row, PTR matrix, long nor, PTR result);
PTR zpermrow(PTR row, long *perm, PTR result);


/* ------------------------------------------------------------------
   Other low-level functions (zzz2.c)
   ------------------------------------------------------------------ */

#define zsub(a,b) zadd((a),zneg(b))
#define zdiv(a,b) zmul((a),zinv(b))

void zmoverow(PTR dest, PTR src);
void zswaprow(PTR dest, PTR src);
int zcmprow(PTR dest, PTR src);
size_t zsize(long nrows);
PTR zalloc(long NROWS);
void zadvance(PTR *ptr, long nrows);

void MtxCleanup(void);
int MtxInit(void);


/* ------------------------------------------------------------------
   zzz <--> GAP conversion (zgap.c)
   ------------------------------------------------------------------ */

char *zftogap(FEL f);


/* ------------------------------------------------------------------
   Bitstring operations (zbitstr.c)
   ------------------------------------------------------------------ */

typedef struct
{
    tMtxObjType id;	    /* Always = T_BITS */
    long pad;		    /* Make buf aligned for type long */
    unsigned char buf[1];
} bitstring_t;

extern size_t bs_size;

void bs_setlen(int l);
void bs_reset(bitstring_t *x);
bitstring_t *bs_alloc(void);
bitstring_t *bs_dup(const bitstring_t *s);
void bs_free(bitstring_t *b);
void bs_set(bitstring_t *b, int i);
void bs_clear(bitstring_t *b, int i);
int bs_test(const bitstring_t *b, int i);
bitstring_t *bs_read(FILE *f);
int bs_write(FILE *f, const bitstring_t *b);
void bs_and(bitstring_t *dest, const bitstring_t *src);
void bs_minus(bitstring_t *dest, const bitstring_t *src);
void bs_or(bitstring_t *dest, const bitstring_t *src);
int bs_match(const bitstring_t *x, const bitstring_t *y);
int bs_issub(const bitstring_t *a, const bitstring_t *b);
int bs_cmp(const bitstring_t *a, const bitstring_t *b);
void bs_cpy(bitstring_t *a, const bitstring_t *b);


/* ------------------------------------------------------------------
   File i/o (zfile.c)
   ------------------------------------------------------------------ */

size_t zreadlong(FILE *f, long *buf, size_t n);
size_t zwritelong(FILE *f, const long *buf, size_t n);
size_t zreadvec(FILE *f, PTR buf, size_t n);
size_t zwritevec(FILE *f, PTR buf, size_t n);
int zseek(FILE *f, long pos);
FILE *zreadhdr(const char *name, long *fld, long *nor, long *noc);
FILE *zwritehdr(const char *name, long fld, long nor, long noc);



/* ------------------------------------------------------------------
   Gauss elimination and related functions (zgauss.c, znullsp.c,
   zmatinv.c)
   ------------------------------------------------------------------ */

int zmkpivot(PTR matrix, long nor, long *piv);
long zmkechelon(PTR matrix, long nor, long *piv);
int zcleanrow(PTR row, PTR matrix, long nor, const long *piv);
int zcleanrow2(PTR row, PTR matrix, long nor, const long *piv, PTR row2);
long znullsp(PTR matrix, long nor, long *piv, PTR nsp);
int zmatinv(PTR mat, PTR result);


/* ------------------------------------------------------------------
   Projection on the quotient (zquot.c)
   ------------------------------------------------------------------ */

int zquotinit(PTR subspace, long dim, long *piv);
int zquot(PTR space, long dim, PTR quot);
int zquotop(PTR matrix, PTR quot);


/* ------------------------------------------------------------------
   Seed vector generation (zpseed.c)
   ------------------------------------------------------------------ */

extern PTR zpseed_vec;		/* The seed vector */

void zpseed_free(void);
int zpseed_init(long sdim, PTR seedbasis);
long zpseed_make(long num);
long zpseed_next(void);

/* ------------------------------------------------------------------
   Command line parsing (args.h)
   ------------------------------------------------------------------ */

typedef struct { char *name; char *shortdesc; char *rcsrev;
	char **helptext; } proginfo_t;

#define GETINT_ERR -12345
#define OPT_END -1

extern char opt_char;		/* Current option */
extern char opt_text[50];	/* Option text */
extern char *opt_text_ptr;	/* Current position in option text */
extern int opt_ind;		/* Index in argv[] list */
extern char MeatAxeLib[250];	/* MeatAxe library directory */
extern char MeatAxeBin[250];	/* MeatAxe program directory */

void MtxSetCommandLine(int argc, char *argv[], const proginfo_t *pi);
int MtxGetOpt(const char *pattern);
long MtxGetInt(void);



/* ------------------------------------------------------------------
   Messages and debugging (message.c, errstack.c)
   ------------------------------------------------------------------ */

#define MTX_MAXERRORTEXT    500

typedef struct
{
    const char *FileName;		/* Source file name */
    int LineNo;				/* Line number */
    const char *FunctionName;		/* Function name */
    char Text[MTX_MAXERRORTEXT];	/* Error description */
} MtxErrorRecord;
typedef void MtxErrorHandler(const MtxErrorRecord *err);
extern int MtxErrNo;

int MtxErrorExit(const char *fmt, ...);
int MtxError(const char *file, int line, const char *func,
    const char *msg, ...);
int MtxFormatMessage(char *buf, int bufsize, const char *msg, va_list al);
int MtxPrintMessage(FILE *f, const char *fmt, ...);
MtxErrorHandler *MtxSetErrorHandler(MtxErrorHandler *h);


#define MTXCONTEXT(f) static const char MtxFunctionName[] = #f
#define MTXERROR(msg) MtxError(__FILE__,__LINE__,MtxFunctionName,msg)
#define MTXERROR1(msg,a) MtxError(__FILE__,__LINE__,MtxFunctionName,msg,a)
#define MTXERROR2(msg,a,b) MtxError(__FILE__,__LINE__,MtxFunctionName,msg,a,b)
#define MTXERROR3(msg,a,b,c) MtxError(__FILE__,__LINE__,MtxFunctionName,msg,a,b,c)

extern int msg_level;
#define MSG0 (msg_level >= 0)
#define MSG1 (msg_level >= 1)
#define MSG2 (msg_level >= 2)
#define MSG3 (msg_level >= 3)
#define MESSAGE(level,args)\
  (msg_level>=(level) ? ( printf args , fflush(stdout), 1) : 0 )


/* --------------------------------------------------------------------------
   MTX_ASSERT(), MTX_VERIFY() - Debugging macros

   Both macros take a boolean expression as argument. If the expression
   evaluates to 0, an error message is printed and the program is stopped
   immedeately. Unlinke MTX_VERIFY, MTX_ASSERT is removed in release builds 
   (DEBUG not defined).
   -------------------------------------------------------------------------- */

#define MTX_VERIFY(e) ((void)((e) ||\
    MTXERROR1("Assertion failed: %s",#e)))

#ifdef DEBUG
#define MTX_ASSERT(e) MTX_VERIFY(e)
#else
#define MTX_ASSERT(e)
#endif


 
/* ------------------------------------------------------------------
   Random numbers (random.c)
   ------------------------------------------------------------------ */

void MtxRandomInit(unsigned seed);
long int MtxRandom(void);
#define MtxRandomInt(max) ((unsigned int) (MtxRandom() % (max)))



/* ------------------------------------------------------------------
   Miscellaneous
   ------------------------------------------------------------------ */

long gcd(long a, long b);
long lcm(long a, long b);



/* ------------------------------------------------------------------
   Matrices
   ------------------------------------------------------------------ */

typedef struct
{
    tMtxObjType id;
    long fl, nor, noc;
    PTR d;
} matrix_t;

#define MATRIX_IS_VALID(m) (((m) != NULL && (m)->id == T_MATRIX) ||\
    (MTXERROR("Invalid matrix '" #m "'"), 0))

matrix_t *MatAdd(matrix_t *dest, const matrix_t *src);
matrix_t *MatAddMul(matrix_t *dest, const matrix_t *src, FEL coeff);
matrix_t *MatAlloc(long fl, long nor, long noc);
int MatCopyRegion(matrix_t *dest, int destrow, int destcol, 
    const matrix_t *src, int row1, int col1, int nrows, int ncols);
matrix_t *MatCut(const matrix_t *src, int row1, int col1, int nrows, int ncols);
matrix_t *MatDup(const matrix_t *src);
int MatFree(matrix_t *m);
matrix_t *MatInv(const matrix_t *src);
matrix_t *MatId(long fl, long nor);
int MatIsEchelon(const matrix_t *mat);
matrix_t *MatLoad(const char *fn);
matrix_t *MatMul(matrix_t *dest, const matrix_t *src);
matrix_t *MatMulScalar(matrix_t *dest, FEL coeff);
long MatOrder(const matrix_t *mat);
matrix_t *MatPower(const matrix_t *mat, long n);
void MatPrint(char *name, const matrix_t *p);
matrix_t *MatRead(FILE *f);
int MatSave(const matrix_t *mat, const char *fn);
matrix_t *MatTensor(const matrix_t *m1, const matrix_t *m2);
matrix_t *MatTr(const matrix_t *src);
FEL MatTrace(const matrix_t *mat);
int MatWrite(FILE *f, const matrix_t *mat);


/* ------------------------------------------------------------------
   Polynomials
   ------------------------------------------------------------------ */

typedef struct
{
    tMtxObjType id;	/* Always = T_POLY */
    long fl;
    long deg;
    size_t size;	/* size of buffer, deg < size */
    FEL *buf;
}
poly_t;

#define POLYNOMIAL_IS_VALID(p) (((p) != NULL && (p)->id == T_POLY) ||\
    (MTXERROR("Invalid polynomial '" #p "'"), 0))


poly_t *PolAdd(poly_t *dest, const poly_t *src);
poly_t *PolAlloc(long fl, long degree);
int PolCompare(const poly_t *a, const poly_t *b);
poly_t *PolDerive(poly_t *pol);
poly_t *PolDivMod(poly_t *a, const poly_t *b);
poly_t *PolDup(const poly_t *x);
void PolFree(poly_t *x);
poly_t *PolGcd(const poly_t *a, const poly_t *b);
int PolGcdEx(const poly_t *a, const poly_t *b, poly_t **result);
poly_t *PolMul(poly_t *dest, const poly_t *src);
poly_t *PolMod(poly_t *a, const poly_t *b);
void PolPrint(char *name, const poly_t *p);
poly_t *PolRead(FILE *f);
int PolWrite(FILE *f, const poly_t *p);



/* ------------------------------------------------------------------
   Higher-level gauss functions (gauss.c)
   ------------------------------------------------------------------ */

matrix_t *MatEchelonize(matrix_t *mat);

long MatNullity(const matrix_t *mat);
long MatNullity__(matrix_t *mat);

matrix_t *MatNullspace(const matrix_t *mat);
matrix_t *MatNullSpace_(matrix_t *mat);
matrix_t *MatNullSpace__(matrix_t *mat);




/* ------------------------------------------------------------------
   The lattice package
   ------------------------------------------------------------------ */

#define MAXGEN 50	/* Max. number of generators */
#define LAT_MAXCF 200	/* Max. number of composition factors */
#define MAXCYCL 30000	/* Max. number of cyclic submodules */
#define MAXDOTL 90000	/* Max. number of dotted lines */
#define MAXNSUB 20000	/* Max. number of submodules */

#define LAT_MAXBASENAME 100


typedef struct
{
    long dim, num, mult;
    long idword;		/* Identifying word */
    poly_t *idpol;
    long peakword;		/* Peak word */
    poly_t *peakpol;
    long nmount;		/* Number of mountains */
    long ndotl;			/* Number of dotted lines */
    long spl;			/* Degree of splitting field */
}
CfInfo;

typedef struct

{
    char BaseName[LAT_MAXBASENAME];	/* Base name */
    int Field;				/* Field order */
    int NGen;				/* Number of generators */
    int NCf;				/* Number of irred. constituents */
    CfInfo Cf[LAT_MAXCF];		/* Data for irred. constituents */
} Lat_Info;



int Lat_ReadInfo(Lat_Info *li, const char *basename);
int Lat_WriteInfo(Lat_Info *li);
char *Lat_CfName(Lat_Info *li, int cf);


#define LAT_RG_INVERT		0x0001	/* Invert generators */
#define LAT_RG_TRANSPOSE	0x0002	/* Transpose generators */
#define LAT_RG_STD		0x0004	/* Use standard form */

void Lat_ReadCfGens(Lat_Info *info, int cf, matrix_t *gen[], int flags);



/* ------------------------------------------------------------------
   Function operating on representations and vector spaces
   ------------------------------------------------------------------ */

#define SF_FIRST	0x0001	/* Try only the first seed vector */
#define SF_EACH		0x0002	/* Try each seed vector */
#define SF_MAKE		0x0004	/* Try all 1-dimensional subspaces */
#define SF_SUB		0x0010	/* Try until finding a proper subspace */
#define SF_CYCLIC	0x0020	/* Try until finding a cyclic vector */
#define SF_COMBINE	0x0040	/* Combine the spans */
#define SF_SEED_MASK	0x000F
#define SF_MODE_MASK	0x00F0
#define SF_STD		0x0100	/* Spin up 'canonically' */

int ChangeBasis(const matrix_t *M, int ngen, const matrix_t **gen,
    matrix_t **newgen);
matrix_t *EndoSpin(const matrix_t *seed, int ngen, const matrix_t *gen[],
	int nendo, const matrix_t *endo[]);
int IsSubspace(const matrix_t *sub, const matrix_t *space, long ngen);
int IsIsomorphic(int ngen, const matrix_t *gen1[], const CfInfo *info1,
    const matrix_t *gen2[],matrix_t **trans, int use_pw);
int MakeEndomorphisms(int ngen, const matrix_t *gen[], const matrix_t *nsp,
    matrix_t *endo[]);
matrix_t *QuotProj(matrix_t *subspace, matrix_t *vectors);
matrix_t *SpinUp(const matrix_t *seed,int ngen,const matrix_t **gen,int flags);
int Split(matrix_t *subspace, int ngen, const matrix_t *gen[],
    matrix_t **sub, matrix_t **quot);




/* ------------------------------------------------------------------
   The word generator (words.h)
   ------------------------------------------------------------------ */

#define MAXFP 6            /* Standard fingerprint size */


/*typedef struct { int ngen; size_t len, max; matrix_t **b; } wgdata_t;*/

typedef struct 
{
    int NGen;
    int len, max;	    /* Nur von der alte Version benutzt */
    const matrix_t **Gen;   /* Entspricht <b> in der alten Version */
    matrix_t *Basis[8];	    /* Von der alten Version nicht benutzt */
    int N2[8];		    /* Von der alten Version nicht benutzt */
} wgdata_t;



wgdata_t *WgInit(int ngen, const matrix_t *gen[]);
void WgFree(wgdata_t *b);
matrix_t *WgMakeWord(wgdata_t *b, long n);
void WgMakeFingerPrint(wgdata_t *b, long fp[]);
const char *WgSymbolicName(const wgdata_t *b, long n);


/* ------------------------------------------------------------------
   Permutations
   ------------------------------------------------------------------ */

typedef struct
{
    tMtxObjType id;		/* Always = T_PERM */
    long deg;
    long *d;		/* d[0] contains P(1), and so on */
} perm_t;

#define PERMUTATION_IS_VALID(p) (((p) != NULL && (p)->id == T_PERM) ||\
    (MTXERROR("Invalid permuatation '" #p "'"), 0))

int PermPrint(const char *name, const perm_t *x);
perm_t *PermAlloc(long deg);
int PermFree(perm_t *m);
perm_t *PermDup(const perm_t *src);
perm_t *PermRead(FILE *f);
int PermWrite(FILE *f, const perm_t *perm);
perm_t *PermLoad(char *fn);
int PermSave(const perm_t *perm, char *fn);
perm_t *PermMul(perm_t *dest, const perm_t *src);
long PermOrder(const perm_t *perm);
perm_t *PermPower(const perm_t *p, long n);


/* ------------------------------------------------------------------
   Insertion (matins.c)
   ------------------------------------------------------------------ */

matrix_t *MatInsert(const matrix_t *mat, const poly_t *pol);
matrix_t *MatInsert_(matrix_t *mat, const poly_t *pol);


/* ------------------------------------------------------------------
   Factored polynomials (fpoly.c)
   ------------------------------------------------------------------ */

typedef struct {
    tMtxObjType id;			/* Always T_FPOLY */
    size_t len, max;
    poly_t **p;			/* Irreducible factors */
    long *e;			/* Multiplicities */
} fpoly_t;

#define FPOLY_IS_VALID(p) (((p) != NULL && (p)->id == T_FPOLY) ||\
    (MTXERROR("Invalid factored polynomial '" #p "'"), 0))


void FpolPrint(const char *name, const fpoly_t *p);
fpoly_t *FpolAlloc();
fpoly_t *FpolDup(const fpoly_t *s);
void FpolFree(fpoly_t *p);
fpoly_t *FpolMulP(fpoly_t *dest, const poly_t *src, long pwr);
fpoly_t *FpolMul(fpoly_t *dest, const fpoly_t *src);


/* ------------------------------------------------------------------
   Characteristic and minimal polynomials (charpol.c, minpol.c)
   ------------------------------------------------------------------ */

extern long CharPolSeed;
poly_t *CharPolFactor(const matrix_t *mat);
fpoly_t *CharPol(const matrix_t *mat);
poly_t *MinPolFactor(matrix_t *mat);
fpoly_t *MinPol(matrix_t *mat);


/* ------------------------------------------------------------------
   Berlekamp factorization (berlekmp.c)
   ------------------------------------------------------------------ */

fpoly_t *PolFactorization(const poly_t *pol);


/* ------------------------------------------------------------------
   Symmetric power (matsym.c)
   ------------------------------------------------------------------ */

matrix_t *MatSymmetricPower(matrix_t *mat, int n);



/* ------------------------------------------------------------------
   Sets (ysets.c)
   ------------------------------------------------------------------ */

typedef struct
{
    tMtxObjType id;			/* Always = T_SET */
    size_t len, max;
    long *buf;
} set_t;

set_t *SetAlloc(void);
int SetContains(const set_t *set, long elem);
set_t *SetDup(const set_t *s);
void SetFree(set_t *x);
int SetInsert(set_t *set, long elem);
void SetPrint(char *name, const set_t *x);
set_t *SetRead(FILE *f);
int SetWrite(FILE *f, const set_t *set);




/* ------------------------------------------------------------------
   Structured text files (stfXXX.c)
   ------------------------------------------------------------------ */

typedef struct
{
    FILE *File;		/* The stream we're using */
    char *LineBuf;	/* Buffers one 'line' */
    char *GetPtr;	/* Current input position */
    int LineBufSize;	/* Current buffer size */
    int OutPos;		/* Number of chars in current line (writing only) */
    int LineNo;		/* Current line number (reading and writing) */
} StfData;

int StfClose(StfData *f);
StfData *StfOpen(const char *name, const char *mode);

int StfPut(StfData *f, const char *text);
int StfPutInt(StfData *f, int value);
int StfPutVector(StfData *f, int size, int *value);
int StfBeginEntry(StfData *f, const char *name);
int StfEndEntry(StfData *f);
int StfWriteValue(StfData *f, const char *name, const char *value);
int StfWriteInt(StfData *f, const char *name, int value);
int StfWriteVector(StfData *f, const char *name, int size, int *value);

int StfReadLine(StfData *f);
const char *StfGetName(StfData *f);
int StfGetInt(StfData *f, int *buf);
int StfMatch(StfData *f, const char *pattern);
int StfGetVector(StfData *f, int *bufsize, int *buf);


/* ------------------------------------------------------------------
   Tensor condensation package
   ------------------------------------------------------------------ */

#define TK_MAXORDER 2000    /* Maximal order of condensation subgroup */
#define TK_MAXGEN 20	    /* Maximal number of generators for subgroup */

#define TK_BASISTYPE_PEAKWORD	1   /* BASBYPEAK */
#define TK_BASISTYPE_GOPERATOR	2   /* SYMBAS */
#define TK_BASISTYPE_MANUAL	3   /* Basis was contructed manually */

typedef struct
{
    int Order;			/* Kondesation subgroup order */
    int NGen;			/* Number of generators */
    int Dim;			/* Dimension of condensed module */
    int Source[TK_MAXORDER];	/* Element tree vertex (1-based) */
    int Gen[TK_MAXORDER];	/* Element tree edge (1-based) */
    int NCf;			/* Number of relevant constituents */
    int CfIndex[2][LAT_MAXCF];	/* Constituent number */
    int BasisMode[2][LAT_MAXCF];/* See TK_BASISTYPE_xxx */
} TK_Info;

int TK_ReadInfo(TK_Info *tki, const char *name);
int TK_WriteInfo(TK_Info *tki, const char *name);

void TK_MakePrimCoeff(TK_Info *info,const matrix_t *gen[],FEL coeff[]);
matrix_t *TK_GOperator(TK_Info *info,const matrix_t *gen[],FEL *coeff,
    int normalize);
matrix_t *TK_TensorGOperator(TK_Info *info, const matrix_t *gen1[],
    const matrix_t *gen2[]);
int TK_MultiGOperator(TK_Info *info, const matrix_t *gen[], 
    int ncf, FEL coeff[][TK_MAXORDER], int degree, matrix_t **result);


/* ------------------------------------------------------------------
   Compatibilty with previous versions
   ------------------------------------------------------------------ */

#ifndef _WIN32

#define echelon(a) MatEchelon(a)
#define matadd(dest,src) MatAdd(src,dest)
#define matinsert_(a,b) MatInsert_(a,b)
#define matinsert(a,b) MatInsert(a,b)
#define matmul(dest,src) MatMul(src,dest)
#define matinv(m) MatInv(m)
#define mattr(m) MatTrace(m)
#define nullity(a) MatNullity(a)
#define nullity__(a) MatNullity__(a)
#define nullspace(a) MatNullspace(a)
#define nullspace_(a) MatNullspace_(a)
#define nullspace__(a) MatNullspace__(a)
#define tensormat(a,b) MatTensor(a,b)
#define matalloc(a,b,c) MatAlloc(a,b,c)
#define matfree(a) MatFree(a)
#define matdup(a) MatDup(a)
#define matpower(a,b) MatPower(a,b)
#define matload(a) MatLoad(a)
#define matsave(a) MatSave(a)
#define matread(a) MatRead(a)
#define matwrite(a,b) MatWrite(a,b)
#define matid(a,b) MatId(a,b)

#define spcequal(a,b) ((a)->nor == (b)->nor && IsSubspace(a,b,0))
#define spccontains(a,b) (IsSubspace(b,a,0))

#define initargs(a,b,c) MtxSetCommandLine(a,b,c)
#define errexit(a,b) MtxErrorExit("%s: %E",b,a)
#define mtxinit() MtxInit()
#define zgetopt(pat) MtxGetOpt(pat)
#define zgetint() MtxGetInt()

#define polalloc(a,b) PolAlloc(a,b)
#define polfree(a,b) PolFree(a,b)
#define poldup(a) PolDup(a)
#define poladd(a,b) PolAdd(a,b)
#define polmul(a,b) PolMul(a,b)
#define polmod(a,b) PolMod(a,b)
#define poldivmod(a,b) PolDivMod(a,b)
#define polprint(a,b) PolPrint(a,b)
#define polgcd(a,b) PolGcd(a,b)
#define polderive(a) PolDerive(a)
#define polcmp(a,b) PolCompare(a,b)

#define fpolprint(a,b) FpolPrint(a,b)
#define fpolalloc() FpolAlloc()
#define fpolfree(a) FpolFree(a)
#define fpoldup(a) FpolDup(a)

#define permprint(a,b) PermPrint(a,b)
#define permalloc(a) PermAlloc(a)
#define permpower(a,b) PermPower(a,b)
#define permorder(a) PermOrder(a)
#define permmul(a,b) PermMul(a,b)
#define permfree(a) PermFree(a)
#define permdup(a) PermDup(a)
#define permread(a) PermRead(a)
#define permwrite(a,b) PermWrite(a,b)
#define permload(a) PermLoad(a)
#define permsave(a,b) PermSave(a,b)

#define charpolfqcator(a) CharPolFactor(a)
#define charpol(a) CharPol(a)
#define minpolfactor(a) MinPolFactor(a)

#define WGInit(a,b) WgInit(a,b)
#define WGFree(a) WgFree(a)
#define MakeWord(a,b) WgMakeWord(a,b)

#define mtxerrno MtxErrNo
#define errexit(a,b) MtxErrorExit("%s: %E",b,a)
#define MTXFAIL(errcode,retval) \
  { MTXERROR1("MeatAxe error: %E",errcode); return (retval); }
#define FATAL(msg) (MtxErrorExit("FATAL ERROR: %s",msg), exit(1), 1)

#endif



#endif	/* !defined(__MEATAXE_INCLUDED) */


