/* ============================= C MeatAxe ==================================
   File:        $Id: message.c,v 1.11 1998/09/17 17:03:54 mringe Exp $
   Comment:     Messages.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>

extern char *sys_errlist[];

/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

int msg_level = 0;
int MtxErrNo = 0;


static struct msg_struct { int Errno; char *smsg, *lmsg; } msgs[] =
  {
    { ERR_NOMEM, "Not enough memory", NULL },
    { ERR_GAMEOVER, "Time limit exceeded", NULL },
    { ERR_DIV0, "Division by zero", NULL },

    { ERR_FILEOPEN, "Could not open file", NULL },
    { ERR_FILEREAD, "File read error", NULL },
    { ERR_FILEWRITE, "File write error", NULL },
    { ERR_FILEFMT, "Bad file format", NULL },
    { ERR_NFILES, "Too many files", NULL },

    { ERR_BADARG, "Bad argument", NULL },
    { ERR_BADTYPE, "Bad type", NULL },
    { ERR_RANGE, "Argument out of range", NULL },
    { ERR_BADARG, "Bad argument", NULL },
    { ERR_NOTECH, "Matrix not in echelon form", NULL },
    { ERR_NOTSQUARE, "Matrix not square", NULL },
    { ERR_INCOMPAT, "Incompatible objects", NULL },

    { ERR_BADUSAGE, "Bad syntax, try `-help'", NULL},
    { ERR_OPTION, "Bad usage of option, try `-help'", NULL},
    { ERR_NARGS, "Bad number of arguments, try `-help'", NULL},

    { ERR_NOTMATRIX, "Not a matrix", NULL},
    { ERR_NOTPERM, "Not a permutation", NULL},
    { ERR_NOTPOLY, "Not a polynomial", NULL},
    
    { ERR_UNKNOWN, "Unknown symbol", NULL},
    { -1, NULL, NULL }
  };


/* ------------------------------------------------------------------
   errmsg()
   ------------------------------------------------------------------ */

static char *errmsg(int en)

{
    static char buf[30];
    struct msg_struct *x;

    for (x = msgs; x->Errno >= 0; ++x)
        if (x->Errno == en)
	    return x->smsg;

    sprintf(buf,"Unknown error %d",en);
    return buf;
}




/* ------------------------------------------------------------------
   FormatString() - Copy string to message buffer

   Description:
       This function is used internally by MtxFormatMessage().
   ------------------------------------------------------------------ */

static void FormatString(char **buf, int *count, int bufsize, const char *c)

{
    while (*c != 0 && *count < bufsize)
    {
	*(*buf)++ = *c++;
	++*count;
    }
}


/* ------------------------------------------------------------------
   FormatInteger() - Write decimal integer to message buffer

   Description:
       This function is used internally by MtxFormatMessage().
   ------------------------------------------------------------------ */

static void FormatInteger(char **buf, int *count, int bufsize, int val)
{
    char tmp[20];
    sprintf(tmp,"%d",val);
    FormatString(buf,count,bufsize,tmp);
}





/*{{{ --------------------------------------------------------------------------
!function MtxFormatMessage "Format a message"
!section general.msg
!synopsis 
    int MtxFormatMessage(char *buf, int bufsize, const char *msg, va_list al);
!arg buf
    Buffer for the message.
!arg bufsize
    Size of buffer.
!arg msg
    The message text.
!arg al
    Optional arguments for the message (see below).
!description
    This function formats a message using a |printf|-like syntax.
    The following table contains all format codes that may be used within
    |msg|:
    \begin{center}\begin{tabular}{|l|l|p{0.6\linewidth}|}
    \hline
    Code & Argument type & Description \\
    \hline\hline
    \verb"%d" & |int| & Decimal integer \\
    \hline
    \verb"%s" & |char *| & Text \\
    \hline
    \verb"%E" & |int| & A {\MeatAxe} error message. The argument must be
	one of the |MTX_ERR_xxx| constants defined in |meataxe.h|.\\
    \hline
    \verb"%S" & none & System error message corresponding to the
	current value of |errno|.\\
    \hline
    \end{tabular}\end{center}

!seealso @msg
-------------------------------------------------------------------------- }}}*/

int MtxFormatMessage(char *buf, int bufsize, const char *msg, va_list al)

{
    char *d;
    int count = 0;

    d = buf;
    while (*msg != 0 && count < bufsize - 1)
    {
	if (*msg == '%')
	{
	    switch (*++msg)
	    {
		case 'd':
		    FormatInteger(&d,&count,bufsize,va_arg(al,int));
		    break;
		case 's':
		    FormatString(&d,&count,bufsize,va_arg(al,char *));
		    break;
		case 'E':
		    FormatString(&d,&count,bufsize,errmsg(va_arg(al,int)));
		    break;
		case 'S':
		    FormatString(&d,&count,bufsize,sys_errlist[errno]);
		    break;
	    }
	    ++msg;
	}
	else
	{
	    if (count < bufsize - 1)
		*d++ = *msg++;
	}
    }

    /* Terminate message and return number of characters written
       --------------------------------------------------------- */
    *d = 0;
    return count;
}



/*{{{ --------------------------------------------------------------------------
!function MtxPrintMessage "Print a message"
!section general.msg
!synopsis 
    int MtxPrintMessage(FILE *f, const char *fmt, ...);
!arg f
    File to write to.
!arg fmt
    Message to write
!arg ...
    Optional arguments for the message (see below).
!description
    This function writes a message to a file using a |printf|-like syntax.
    See |MtxFormatMessage()| for details.
!seealso @msg
-------------------------------------------------------------------------- }}}*/

int MtxPrintMessage(FILE *f, const char *fmt, ...)

{
    va_list vl;
    char tmp[2000];

    va_start(vl,fmt);
    MtxFormatMessage(tmp,sizeof(tmp),fmt,vl);
    fputs(tmp,f);
    va_end(vl);
    return 0;
}

