/* ============================= C MeatAxe ==================================
   File:        $Id: minpol.c,v 1.8 1998/06/25 19:02:33 mringe Exp $
   Comment:     Minimal polynomial of a matrix
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"




/* ------------------------------------------------------------------
   Function prototypes
   ------------------------------------------------------------------ */

static void cleanup(void);
static int init(matrix_t *matrix);
static void spinup_cyclic(void);
static poly_t *mkpoly(void);

/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static long fl, nor;		/* Field and size */
static long *APiv = NULL;	/* Pivot table (A) */
static long *CPiv = NULL;	/* Pivot table (C) */
static char *CIsPiv = NULL;	/* Pivot flags */
static PTR mat = NULL,		/* The matrix */
    A = NULL,			/* Work space (spin-up) */
    B = NULL,			/* Work space II (coefficients) */
    C = NULL,			/* Work space III (basis) */
    seed = NULL;
static long CDim;		/* Dimension reached so far */
static long ADim;
static poly_t *mpol = NULL;

/* ------------------------------------------------------------------
   cleanup() - Free memory
   ------------------------------------------------------------------ */

static void cleanup()

{
    if (mat != NULL) free(mat);
    if (A != NULL) free(A);
    if (B != NULL) free(B);
    if (C != NULL) free(C);
    if (seed != NULL) free(seed);
    if (APiv != NULL) free(APiv);
    if (CPiv != NULL) free(CPiv);
    if (CIsPiv != NULL) free(CIsPiv);
    if (mpol != NULL) PolFree(mpol);
    mpol = NULL;
    mat = A = B = C = NULL;
    CPiv = APiv = NULL;
    CIsPiv = NULL;
}


/* ------------------------------------------------------------------
   init()
   ------------------------------------------------------------------ */

static int init(matrix_t *matrix)

{	
    MTXCONTEXT(init);

    /* Set some global variables
       ------------------------- */
    if (matrix->nor != matrix->noc) 
	return MTXERROR1("%E",ERR_NOTSQUARE), -1;
    fl = matrix->fl;
    nor = matrix->nor;
    CDim = 0;

    /* Allocate memory
       --------------- */
    cleanup();
    zsetfield(fl);
    zsetlen(nor);
    if ((mat = zalloc(nor)) == NULL
	|| (A = zalloc(nor+1)) == NULL		/* We need 1 extra row !! */
	|| (B = zalloc(nor+1)) == NULL
	|| (C = zalloc(nor+1)) == NULL
	|| (seed = zalloc(1)) == NULL
	|| (mpol = PolAlloc(fl,0)) == NULL
	|| (APiv = NALLOC(long,nor+2)) == NULL
	|| (CPiv = NALLOC(long,nor+2)) == NULL
	|| (CIsPiv = NALLOC(char,nor+2)) == NULL
       )
	return -1;

    /* Initialize memory
       ----------------- */
    memcpy(mat,matrix->d,zsize(nor));
    memset(CIsPiv,0,(size_t)(nor+2));

    return 0;
}


/* ------------------------------------------------------------------
   mkpoly() - Make polynomial for the latest cyclic subspace
   ------------------------------------------------------------------ */

static poly_t *mkpoly()

{
    int k;
    PTR x;
    poly_t *pol;
    
    pol = PolAlloc(fl,ADim);
    x = B;
    zadvance(&x,ADim);
    for (k = 1; k <= ADim; ++k)
	pol->buf[k-1] = zextract(x,k);
    pol->buf[ADim] = F_ONE;
    return pol;
}


/* ------------------------------------------------------------------
   spinup_cyclic() - Spin up one cyclic subaspace
   ------------------------------------------------------------------ */

static void spinup_cyclic()

{   PTR a, b, c;
    long k, pv;
    FEL f;

    a = A;
    b = B;
    c = C;
    zadvance(&c,CDim);
    zmoverow(a,seed);
    zmulrow(b,F_ZERO);
    zinsert(b,1,F_ONE);
    ADim = 0;
    while ((pv = zfindpiv(a,&f)) != 0)
    {	
	PTR x, y;

    	/* Add new vector to basis A
	   ------------------------- */
	zmoverow(c,a);
	APiv[++ADim] = pv;
	zadvance(&a,(long)1);
	zadvance(&b,(long)1);

    	/* Add new vector to basis C
	   ------------------------- */
	zcleanrow(c,C,CDim,CPiv);
	if ((pv = zfindpiv(c,&f)) != 0)
	{
	    CPiv[++CDim] = pv;
	    CIsPiv[pv] = 1;
	    zadvance(&c,(long)1);
	}

	/* Calculate the next vector
	   ------------------------- */
	zmaprow(seed,mat,nor,a);
	zmoverow(seed,a);
	zmulrow(b,F_ZERO);
	zinsert(b,ADim+1,F_ONE);

	/* Clean with existing basis vectors
	   --------------------------------- */
	x = A; y = B;
	for (k = 1; k <= ADim; ++k)
	{   f = zneg(zdiv(zextract(a,APiv[k]),
		zextract(x,APiv[k])));
	    zaddmulrow(a,x,f);
	    zaddmulrow(b,y,f);
	    zadvance(&x,(long)1);
	    zadvance(&y,(long)1);
	}
    }
}



/*{{{ --------------------------------------------------------------------------
!function MinPolFactor "Minimal polynomial"
!section mat.op
!synopsis 
    poly_t *MinPolFactor(matrix_t *mat);
!arg mat
    Pointer to the matrix, or |NULL|.
!return
    One factor of the minmal polynomial of |mat|,
    or |NULL| if there are no more factors.
!description
    This function returns one factor of the minimal polynomial of
    a given matrix. Further calls with a \verb"NULL" argument return
    more factors or \verb"NULL", if there are no more factors. 
    Note that the factors obtained in this way are in general not irreducible.

    If |mat| is different from |NULL|, |MinPolFactor()| initializes its 
    internal data and starts computing one cyclic subspace. Then, the polynomial 
    of the matrix restricted to that cyclic subspace is constructed and returned 
    to the caller.

    If |mat| is |NULL| on the next call, |MinPolFactor()| resumes at 
    the point where it returned the last time, calculates the next cyclic 
    subspace and so on, until the complete space is exhausted.

    {\em Note:} Since the function uses static variables to store
    information across multiple calls, your program must not use
    |MinPolFactor()| on more than one matrix at the same time.
   
!seealso @minpol 
-------------------------------------------------------------------------- }}}*/

poly_t *MinPolFactor(matrix_t *mat)

{
    MTXCONTEXT(MinPolFactor);
    int i;
    poly_t *ggt, *l, *h;

    /* If called with mat != NULL, initialize everything
       ------------------------------------------------- */
    if (mat != NULL)
    {
	if (!MATRIX_IS_VALID(mat))
	    return NULL;
    	if (init(mat) != 0)
	{
    	    MTXERROR("Cannot initialize");
	    return NULL;
	}
    }

    /* If there is nothing left to do, return NULL
       ------------------------------------------- */
    /* we could call cleanup() here... */
    if (CDim >= nor) 
	return NULL;

    /* Prepare the next seed vector
       ---------------------------- */
    zsetfield(fl);
    zsetlen(nor);
    for (i = 1; i <= nor && CIsPiv[i] != 0; ++i);
    MTX_VERIFY(i <= nor);
    zmulrow(seed,F_ZERO);
    zinsert(seed,i,F_ONE);

    /* Spin up and return the polynomial 
       --------------------------------- */
    spinup_cyclic();
    h = mkpoly();    
    ggt = PolGcd(h,mpol);
    l = PolDivMod(h,ggt);
    PolFree(h);
    PolFree(ggt);
    PolMul(mpol,l);
    return l;
}


/*{{{ --------------------------------------------------------------------------
!function MinPol "Minimal polynomial"
!section mat.op
!synopsis 
    fpoly_t *MinPol(matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Minimal polynomial of |mat|, or |NULL| on error.
!description
    This function calculates the minimal polynomial of a matrix in
    factored form. The return value is a pointer to a |fpoly_t| structure
    containing the irreducible factors of the minimal polynomial.
!seealso @minpol
-------------------------------------------------------------------------- }}}*/

fpoly_t *MinPol(matrix_t *mat)

{
    MTXCONTEXT(MinPol);
    poly_t *f;
    fpoly_t *mp;

    /* Check the argument
       ------------------ */
    if (!MATRIX_IS_VALID(mat))
	return NULL;

    f = MinPolFactor(mat);
    mp = FpolAlloc();
    while (f != NULL)
    {
    	fpoly_t *ff = PolFactorization(f);
    	FpolMul(mp,ff);
    	PolFree(f);
    	FpolFree(ff);
	f = MinPolFactor(NULL);
    }
    return mp;
}
