/* ============================= C MeatAxe ==================================
   File:        $Id: mkcycl.c,v 1.6 1998/04/28 17:36:37 mringe Exp $
   Comment:     This program calculates a representative for each cyclic
		submodule of the condensed modules.h
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static char *helptext[] = {
"SYNTAX",
"    mkcycl [<Options>] <Name>",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -V            Verbose, more messages.",
"    -G            GAP output (implies -Q).",
"    -T <Seconds>  Set CPU time limit",
"",
"FILES",
"    <Name><Dim><a>.<N>k    i    Generators in condensed modules",
"    <Name><Dim><a>.np      i    Condensed peak word",
"    <Name><Dim><a>.v       o    Cyclic submodules",
NULL,
};

static proginfo_t pinfo =
   { "mkcycl", "Find Cyclic Submodules",
     "$Revision: 1.6 $", helptext };


int opt_G = 0;

matrix_t *gen[MAXGEN+1];	/* Generators and peak word*/
long kdim;			/* Dimension of condensed module */
int ncycl;			/* Number of cyclic submodules found */
long count;			/* Number of vectors tried */
matrix_t *cmod[MAXCYCL];	/* Cyclic submodules */
Lat_Info LI;			/* Data from .cfinfo file */


static void init(int argc, char *argv[])

{
    int i;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("G")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	}
    }
    if (opt_ind != argc-1) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    if (Lat_ReadInfo(&LI,argv[argc-1]) != 0)
	MtxErrorExit("Error reading %s",argv[argc-1]);
}



static void Spinup(matrix_t *seed)

{
    matrix_t *sub;
    int i;

    ++count;
    sub = SpinUp(seed,LI.NGen+1,(const matrix_t **)gen,SF_FIRST|SF_SUB);
    for (i = 0; i < ncycl; ++i)
    {
	int issub = IsSubspace(sub,cmod[i],1);
	if (issub == -1)
	    MtxErrorExit("Subspace comparison failed");
	if (issub && sub->nor == cmod[i]->nor)
	{
	    MatFree(sub);	/* Module already in list */
	    return;
	}
    }
    if (ncycl >= MAXCYCL)
	MtxErrorExit("TOO MANY CYCLIC SUBMODULES");
    cmod[ncycl] = sub;
    ++ncycl;
}


void writeresult(int cf)

{
    int i;
    char fn[100];
    FILE *f;

    MESSAGE(0,("%d cyclic submodule%s (%ld vectors tried)\n",
	ncycl,ncycl == 1 ? " " : "s",count));
    sprintf(fn,"%s%s.v",LI.BaseName,Lat_CfName(&LI,cf));
    zsetlen(kdim);
    if ((f = zwritehdr(fn,zfl,(long)ncycl,kdim)) == NULL)
	MtxErrorExit("Cannot open %s for output: %S",fn);
    for (i = 0; i < ncycl; ++i)
	zwritevec(f,cmod[i]->d,1);
    fclose(f);

    /* Clean up
       -------- */
    for (i = 0; i < ncycl; ++i)
	MatFree(cmod[i]);
    for (i = 0; i <= LI.NGen; ++i)
	MatFree(gen[i]);
}



void cyclic(int cf)

{
    matrix_t *seed;
    char fn[200], fn1[200];
    long k, maxk;
    FEL f;
    int i;

    MESSAGE(0,("%s%s: ",LI.BaseName,Lat_CfName(&LI,cf)));
    fflush(stdout);

    /* Read generators
       --------------- */
    for (i = 0; i < LI.NGen; ++i)
    {
	sprintf(fn,"%s%s.%dk",LI.BaseName,Lat_CfName(&LI,cf),i+1);
	gen[i] = MatLoad(fn);
	if (gen[i]->nor != gen[i]->noc)
	    MtxErrorExit("%s: %E",fn,ERR_NOTSQUARE);
	if (i == 0)
	{
	    kdim = gen[0]->nor;
	    strcpy(fn1,fn);
	}
	else
	{
	    if (gen[i]->nor != kdim)
		MtxErrorExit("%s and %s: %E",fn,fn1,ERR_INCOMPAT);
	}
    }

    /* Read the condensed peak word
       ---------------------------- */
    sprintf(fn,"%s%s.np",LI.BaseName,Lat_CfName(&LI,cf));
    gen[LI.NGen] = MatLoad(fn);

    /* Initialize the seed vector generator
       ------------------------------------ */
    seed = MatAlloc(zfl,1,kdim);
    zmulrow(seed->d,(FEL)0);
    ncycl = 0;
    count = 0;
    maxk = 1;

    /* Spin up each seed vector
       ------------------------ */
    while (1)
    {
	for (k=1; 1; ++k)
	{
	    /* This won't work with bigzzz! */

	    f = zextract(seed->d,k) + 1;
	    if (k == maxk && f == 2)
	    {
		if (k == kdim)
		{	writeresult(cf);
			MatFree(seed);
			return;
		}
		else
		{	f = 0;
			++maxk;
		}
	    }
	    if (f == (FEL) zfl)
		f = F_ZERO;
	    zinsert(seed->d,k,f);
	    if (f != 0)
	    {
		Spinup(seed);
		break;
	    }
	}
    }
}



int main(int argc, char *argv[])

{
    int i;

    init(argc,argv);
    MESSAGE(0,("\n*** FIND CYCLIC SUBMODULES ***\n\n"));
    for (i = 0; i < LI.NCf; ++i)
	cyclic(i);
    if (msg_level >= 0)
    {
        printf("\n");
    }
    Lat_WriteInfo(&LI);
    MtxCleanup();
    return 0;
}





