/* ============================= C MeatAxe ==================================
   File:        $Id: mkdotl.c,v 1.6 1998/04/28 17:36:38 mringe Exp $
   Comment:     This program calculates the dotted lines.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>


/* -----------------------------------------------------------------
   Global data
   ----------------------------------------------------------------- */

matrix_t *gen[MAXGEN] = {NULL};	/* Generators */
matrix_t *cycl = NULL;		/* List of cyclic submodules */
long *class[MAXCYCL];		/* Classes of vectors */
long nmountains = 0;		/* Number of mountains */
matrix_t *mountlist[MAXCYCL];	/* Mountains */
bitstring_t *subof[MAXCYCL];	/* Incidence matrix */
int cfstart[LAT_MAXCF+1];	/* First mountain of each c.f. */
char lck[MAXCYCL];
char lck2[MAXCYCL];
bitstring_t *dotl[MAXDOTL];	/* Dotted lines */
int ndotl = 0;			/* Number of dotted-lines in <dotl> */
int firstdotl = 0;		/* Used for locking */
int firstm, nextm;		/* First and last+1 mountain for the */
				/* current constituent */

long *sumdim[MAXCYCL];
    /* sumdim[i][j] contains the dimension of mountain[i] + mountain[j].
       Most of the program's work constists in comparing sums of pairs of
       mountains and looking if they are equal. Thus, a given sum may be 
       needed many times. Since we don't keep the sums in memory we have
       to recalculate them each time. But comparing the dimensions we can 
       avoid many unnecessary calculations. */

int dotlen;
    /* This is the length of a dotted line for the current constituent.
       For theoretical reasions the is always Q+1, where GF(Q) is the
       splitting field for the constituent. */

int opt_G = 0;			/* GAP output */
static Lat_Info LI;		/* Data from .cfinfo file */

static char *helptext[] = {

"SYNTAX",
"    mkdotl [Options] <Name>",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -G            GAP output (implies -Q).",
"    -T <Limit>    Set CPU time limit",
"",
"FILES",
"    <Name>.cfinfo         i/o   Composition factor data ",
"    <Name><Dim><N>.v      i     Cyclic submodules generated by MKCYCL",
"    <Name>.inc            i     Incidence matrix generated by MKINC",
"    <Name>.mnt            i     Mountain data (from MKINC)",
"    <Name>.dot            o     Dotted lines",
"",
NULL};

static proginfo_t pinfo =
   { "mkdotl", "Find Dotted Lines", "$Revision: 1.6 $", helptext };



/* -----------------------------------------------------------------
   init()
   ----------------------------------------------------------------- */

static void init(const char *basename)

{
    FILE *f;
    char fn[200];
    int i;

    MtxInit();
    if (Lat_ReadInfo(&LI,basename) != 0)
	MtxErrorExit("Error reading %s.cfinfo",basename);

    cfstart[0] = 0;
    for (i = 1; i <= LI.NCf; ++i)
	cfstart[i] = cfstart[i-1] + (int)(LI.Cf[i-1].nmount);

    /* Read the incidence matrix
       ------------------------- */
    sprintf(fn,"%s.inc",LI.BaseName);
    f = SFOpen(fn,FM_READ);
    if (f == NULL) 
	MtxErrorExit("ERROR READING .inc FILE");
    zreadlong(f,&nmountains,1);
    if (nmountains != cfstart[LI.NCf]) 
	MtxErrorExit("Bad number of mountains in .inc file");
    MESSAGE(1,("Reading incidence matrix (%ld mountains)\n",
	nmountains));
    fflush(stdout);
    bs_setlen((int) nmountains);
    for (i = 0; i < (int) nmountains; ++i)
	subof[i] = bs_read(f);
    fclose(f);

    for (i = 0; i < (int) nmountains; ++i)
    {
	sumdim[i] = NALLOC(long,nmountains);
	memset(sumdim[i],0,(size_t)nmountains * sizeof(long));
    }

    /* Read classes
       ------------ */
    sprintf(fn,"%s.mnt",LI.BaseName);
    MESSAGE(1,("Reading classes (%s)\n",fn));
    f = SFOpen(fn,FM_READ);
    if (f == NULL) MtxErrorExit("ERROR OPENING .mnt FILE");
    for (i = 0; i < nmountains; ++i)
    {
	long mno, mdim, nvec, *p;
	int k;
	if (fscanf(f,"%ld%ld%ld",&mno,&mdim,&nvec) != 3 ||
	    mno != i || nvec < 1  || mdim < 1)
	    MtxErrorExit("ERROR IN .mnt FILE");
	p = class[i] = NALLOC(long,nvec+2);
	*p++ = nvec;
	for (k = 0; k < nvec; ++k, ++p)
	{
	    if (fscanf(f,"%ld",p) != 1 || *p < 1)
		MtxErrorExit("ERROR IN .mnt FILE");
	}
	if (fscanf(f,"%ld",p) != 1 || *p != -1)
		MtxErrorExit("ERROR IN .mnt FILE");
    }

}




/* -----------------------------------------------------------------
   mkmount() - Make mountain
   ----------------------------------------------------------------- */

static void mkmount(int i)

{
    matrix_t *seed;
    PTR x, y;
    long *p;

    seed = MatAlloc(cycl->fl,class[i][0],cycl->noc);
    x = seed->d;
    zsetlen(cycl->noc);
    for (p = class[i] + 1; *p > 0; ++p)
    {
    	y = cycl->d;
	if (*p > cycl->nor)
	    MtxErrorExit("BAD VECTOR IN CLASS");
	zadvance(&y,*p - 1);
	zmoverow(x,y);
	zadvance(&x,1);
    }

    mountlist[i] = SpinUp(seed,LI.NGen,(const matrix_t **)gen,
	SF_EACH|SF_COMBINE);
    if (mountlist[i] == NULL)
	MtxErrorExit("Cannot spin up mountain");
    MatFree(seed);
}

/* -----------------------------------------------------------------
   nextcf() - Initialize everything for the next composition
	factor: Read generators and vectors, calculate mountains...
   ----------------------------------------------------------------- */

static void nextcf(int cf)

{
    char fn[200];
    int j;

    if (cycl != NULL)
    {
	MatFree(cycl);
	for (j = 0; j < LI.NGen; ++j)
	    MatFree(gen[j]);
    }

    /* Read the generators of the condensed module
       ------------------------------------------- */
    for (j = 0; j < LI.NGen; ++j)
    {
	sprintf(fn,"%s%s.%dk",LI.BaseName,Lat_CfName(&LI,cf),j+1);
	gen[j] = MatLoad(fn);
    }

    /* Read the cyclic submodules
       -------------------------- */
    sprintf(fn,"%s%s.v",LI.BaseName,Lat_CfName(&LI,cf));
    cycl = MatLoad(fn);

    /* Calculate the length of dotted lines, i.e.,
       the order of the splitting field + 1.
       ------------------------------------------- */
    dotlen = (int) zfl;
    for (j = LI.Cf[cf].spl - 1; j > 0; --j)
    	dotlen *= (int) zfl;
    ++dotlen;
    MESSAGE(1,("Length of dotted-lines is %d\n",dotlen));

    /* Calculate the mountains
       ----------------------- */
    for (j = cfstart[cf]; j < cfstart[cf+1]; ++j)
	mkmount(j);

    
}


/* -----------------------------------------------------------------
   sum() - Sum of two mountains
   ----------------------------------------------------------------- */

static matrix_t *sum(int i, int k)

{
    matrix_t *s;
    PTR x;

    s = MatAlloc(mountlist[i]->fl,mountlist[i]->nor+mountlist[k]->nor,
	    mountlist[k]->noc);
    memcpy(s->d,mountlist[i]->d,zsize(mountlist[i]->nor));
    x = s->d;
    zadvance(&x,mountlist[i]->nor);
    memcpy(x,mountlist[k]->d,zsize(mountlist[k]->nor));
    MatEchelonize(s);

    /* Remember the dimension of the sum. We use this information 
       later to avoid unnecessary recalculations.
       ----------------------------------------------------------- */
    sumdim[i][k] = sumdim[k][i] = s->nor;
    return s;
}



/* -----------------------------------------------------------------
   lock()
   ----------------------------------------------------------------- */

static void lock(int i, char *c)

{	
    int l, m;
    bitstring_t *b;

    memset(c,0,sizeof(lck));
    for (m = firstm; m < nextm; ++m)
    {	
	if (bs_test(subof[i],m) || bs_test(subof[m],i))
	    c[m] = 1;
    }
    for (l = firstdotl; l < ndotl; ++l)
    {	
	b = dotl[l];
	if (!bs_test(b,i))
	    continue;
	for (m = firstm; m < nextm; ++m)
	{
	    if (bs_test(b,m))
		c[m] = 1;
	}
    }
}



/* -----------------------------------------------------------------
   trydot() - Find out if mountains #i and #k generate a dotted
	line.
   ----------------------------------------------------------------- */

static void trydot(int i, int k, int beg, int next)

{
    matrix_t *span, *sp;
    bitstring_t *dot;
    int count, l, m, abort = 0;

    lock(k,lck2);
    dot = bs_alloc();
    bs_set(dot,i);
    bs_set(dot,k);
    span = sum(i,k);
    count = 2;
    for (l = beg; l < next && count < dotlen; ++l)
    {
	if (lck[l] || lck2[l]) 
	    continue;
	for (m = i; !abort && m < l; ++m)
	{
	    if (!bs_test(dot,m)) 
		continue;
	    if (sumdim[l][m] != 0 && sumdim[l][m] != span->nor)
	    	abort = 1;
	    else
	    {	
		sp = sum(l,m);
		abort = (sp->nor != span->nor) || !IsSubspace(span,sp,0);
		MatFree(sp);
	    }
	}
	if (!abort)
	{	
	    bs_set(dot,l);
	    ++count;
	    lck[l] = 1;
	}
    }
    MatFree(span);
    if (count == dotlen)	/* We have found a dotted line */
    {	if (ndotl >= MAXDOTL)
	    MtxErrorExit("TOO MANY DOTTED LINES");
	dotl[ndotl++] = dot;
    }
    else
	free(dot);
}


/* --------------------------------------------------------------------------
   mkdot() - Find dotted-lines in one constituent

   Arguments:
     <cf>: Index of the constituent.

   Description:
     This function finds all dotted-lines corresponding to a given 
     constituent.
   -------------------------------------------------------------------------- */

static void mkdot(int cf)

{
    int i, k;

    firstm = cfstart[cf];
    nextm = cfstart[cf+1];
    firstdotl = ndotl;
    for (i = firstm; i < nextm; ++i)
    {
	lock(i,lck);
    	for (k = i+1; k < nextm; ++k)
    	{	if (lck[k]) continue;
    		trydot(i,k,k+1,nextm);
    	}
    }
}



/* -----------------------------------------------------------------
   WriteResult() - Write dotted lines.
   ----------------------------------------------------------------- */

static void WriteResult()

{
    FILE *f;
    char fn[200];
    int i;
    long l;

    strcat(strcpy(fn,LI.BaseName),".dot");
    MESSAGE(1,("Writing %s (%d dotted line%s)\n",
	fn,ndotl,ndotl!=1 ? "s" : ""));
    f = SFOpen(fn,FM_CREATE);
    if (f == NULL) MtxErrorExit("ERROR OPENING .dot FILE");
    l = (long) ndotl;
    zwritelong(f,&l,1);
    for (i = 0; i < ndotl; ++i)
	bs_write(f,dotl[i]);
    fclose(f);
    Lat_WriteInfo(&LI);
}


/* -----------------------------------------------------------------
   WriteResultGAP() - Write dotted lines in GAP format.
   ----------------------------------------------------------------- */

static void WriteResultGAP()

{
    int i,j ;

    printf ("MeatAxe.DottedLines := [\n") ;
    for (i = 0; i < ndotl; ++i)
    {
	printf( "BlistList([" );
	for (j = 0 ; j < nmountains ; j++)
        printf( j < (nmountains - 1) ? "%s," : "%s], [1])" , 
	       bs_test(dotl[i],j) ? "1" : "0" );
        printf( ",\n" );
    }
    printf( "];\n" );
}








/* -----------------------------------------------------------------
   main()
   ----------------------------------------------------------------- */


int main(int argc, char *argv[])

{
    int i, nn = 0;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("G")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	}
    }

    if (opt_ind != argc-1) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    MESSAGE(0,("\n*** CALCULATE DOTTED LINES ***\n\n"));
    init(argv[opt_ind]);

    for (i = 0; i < LI.NCf; ++i)
    {
	nextcf(i);
	mkdot(i);
	LI.Cf[i].ndotl = ndotl - nn;

	MESSAGE(0,("%s%s: %ld vectors, %ld mountains, %ld dotted line%s\n",
	    LI.BaseName,Lat_CfName(&LI,i),  cycl->nor,LI.Cf[i].nmount,
	    LI.Cf[i].ndotl, LI.Cf[i].ndotl != 1 ? "s": ""));
	nn = ndotl;
    }
    WriteResult();
    if (opt_G)
         WriteResultGAP();
    MtxCleanup();
    return 0;
}


