/* ============================= C MeatAxe ==================================
   File:        $Id: mkendo.c,v 1.1 1998/05/07 18:58:27 mringe Exp $
   Comment:     Find endomorphisms
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"





static matrix_t *MakeEndo(int ngen, const matrix_t *gen[],
    const matrix_t *sb1, const matrix_t *vec)

{
    MTXCONTEXT(MkEndo);
    matrix_t *sb2;
    matrix_t *endo;

    /* Make standard basis from <vec>
       ------------------------------ */
    sb2 = SpinUp(vec,ngen,(const matrix_t **)gen,SF_FIRST|SF_CYCLIC|SF_STD);
    MTX_VERIFY(sb2 != NULL && sb2->nor == sb2->noc);

    /* The linear mapping that maps <sb1> on <sb2> is the endomorphism
       we are looking for!
       --------------------------------------------------------------- */
    endo = MatInv(sb1);
    MatMul(endo,sb2);
    MatFree(sb2);
    return endo;
}




/*{{{ --------------------------------------------------------------------------
!function MakeEndomoprhisms "Find endomorphism ring"
!section reps
!synopsis 
    int MakeEndomorphisms(int ngen, const matrix_t *gen[], const matrix_t *nsp,
	matrix_t *endo[])
!arg ngen
	Number of generators in |gen|.
!arg gen
	List of generators.
!arg nsp
	Idword kernel (see below).
!arg endo
	Pointer to result buffer.
!return
	$0$ on success, $-1$ on error.
!description
    This function calculates a basis for the endomorphism ring $E$ of an
    irreducible module. The dimension of $E$ is equal to the degree $[F':F]$
    of the splitting field extension for the module.

    The function takes three input arguments: |ngen| is the number of generators
    of the algebra, and |gen| points to an array of |ngen| matrices containing
    the representations of the generators. |nsp| is the kernel of an idword
    (identifying word)
    for the module, i.e., an algebra element $a$ with $\dim(\ker(a_V))=[F`:F]$.

    On success, basis of $E$ (a set of matrices) is stored into |endo|.
    If the function fails, no matrices are stored in |endo|. 
!seealso 
-------------------------------------------------------------------------- }}}*/

int MakeEndomorphisms(int ngen, const matrix_t *gen[], const matrix_t *nsp,
	matrix_t *endo[])

{
    MTXCONTEXT(MakeEndomorphisms);
    matrix_t *sb1;		/* Standard bases */
    int nendo;			/* # of endomorphisms obtained so far */

    MTX_ASSERT(nsp->nor > 0);
    MTX_ASSERT(ngen > 0);

    /* Take the first vector from <nsp> and make the standard basis.
       ------------------------------------------------------------- */
    sb1 = SpinUp(nsp,ngen,(const matrix_t **)gen,SF_FIRST|SF_CYCLIC|SF_STD);
    MTX_VERIFY(sb1 != NULL && sb1->nor == sb1->noc);

    /* Take the identity as the first basis element for E
       -------------------------------------------------- */
    endo[0] = MatId(gen[0]->fl,gen[0]->nor);
    nendo = 1;

    /* For each of the remaining vectors v_2,..v_d in <nsp>, construct the
       endomorphism that maps v_1 to v_j.
       -------------------------------------------------------------------- */
    while (nendo < nsp->nor)
    {
	matrix_t *vec = MatCut(nsp,nendo+1,1,1,-1);
	endo[nendo] = MakeEndo(ngen,gen,sb1,vec);
	MatFree(vec);
	if (endo[nendo] == NULL)	/* Error */
	    break;
	++nendo;
    }

    /* Clean up
       -------- */
    if (nendo < nsp->nor)
    {
	while (nendo > 0)
	    MatFree(endo[--nendo]);
    }
    MatFree(sb1); 

    /* Return error code
       ----------------- */
    return nendo == nsp->nor ? 0 : -1;
}

