/* ============================= C MeatAxe ==================================
   File:        $Id: mkinc.c,v 1.7 1998/04/28 17:36:41 mringe Exp $
   Comment:     Calculate the incidences between mountains.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>
#include <stdlib.h>



/* -------------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------------- */

matrix_t *gen[MAXGEN];			/* Generators */
matrix_t *bild[LAT_MAXCF];		/* Image of peak word (gkond) */
long *piv;				/* Pivot table */
int nmount = 0;				/* Number of mountains */
matrix_t * mountlist[MAXCYCL];		/* List of mountains */
long mountdim[MAXCYCL];			/* Dim. of mountains */
matrix_t **proj[MAXCYCL];		/* Projections of mountains */
int moffset[LAT_MAXCF];			/* Number of first mountain */
long *class[MAXCYCL];			/* Classes of vectors */
bitstring_t * subof[MAXCYCL];		/* Incidence matrix */
static Lat_Info LI;			/* Data from .cfinfo file */
					
static char *helptext[] = {
"SYNTAX",
"    mkinc [<Options>] <Name>",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -V            Verbose, more messages.",
"    -G            GAP output (implies -Q).",
"    -T <Seconds>  Set CPU time limit",
"",
"FILES",
"    <Name><Dim><a>.<N>   i    Generators",
"    <Name><Dim><a>.<N>k  i    Condensed generators",
"    <Name><Dim><a>.v     i    Cyclic submodules",
"    <Name><Dim><a>.im    i    Image used for condensation",
"    <Name><Dim><a>.k     i    Uncondense matrix",
"    <Name>.v             o    Mountains",
"    <Name>.mnt           o    Dimensions of mountains and classes",
"                              of cyclic subspaces",
"    <Name>.inc           o    Incidence matrix",
NULL};

static proginfo_t pinfo =
   { "mkinc", "Mountains And Incidence Matrix",
     "$Revision: 1.7 $", helptext };


int opt_G = 0;


/* -----------------------------------------------------------------
   init() - Read generators and images of peak words
   ----------------------------------------------------------------- */

static void init(const char *basename)

{
    char fn[200];
    int i;

    if (Lat_ReadInfo(&LI,basename) != 0)
	MtxErrorExit("Error reading %s.cfinfo",basename);
    for (i = 0; i < LI.NGen; ++i)
    {
	sprintf(fn,"%s.%d",LI.BaseName,i+1);
	gen[i] = MatLoad(fn);
	if (gen[i] == NULL)
	    MtxErrorExit(NULL);
    }
    for (i = 0; i < LI.NCf; ++i)
    {
	sprintf(fn,"%s%s.im",LI.BaseName,Lat_CfName(&LI,i));
	bild[i] = MatLoad(fn);
    }
    piv = NALLOC(long,gen[0]->nor+1);
}


/* -----------------------------------------------------------------
   writemountains() - Write mountains, dimensions and classes.
   ----------------------------------------------------------------- */

static void writemountains()

{
    FILE *f;
    char fn[200];
    int i;
    long *p;

    /* Write dimensions and classes
       ---------------------------- */
    f= SFOpen(strcat(strcpy(fn,LI.BaseName),".mnt"),FM_CREATE|FM_TEXT);
    if (f == NULL) 
	MtxErrorExit("Cannot open %s: %S",fn);
    MESSAGE(0,("Writing dimensions and classes to %s\n",fn));
    for (i = 0; i < nmount; ++i)
    {
	fprintf(f,"%4d %4ld  ",i,mountdim[i]);
	for (p = class[i]; *p > 0; ++p)
	    fprintf(f,"%ld ",*p);
	fprintf(f,"-1\n");
    }
    fclose(f);

    /* Write mountains
       --------------- */
    strcat(strcpy(fn,LI.BaseName),".v");
    MESSAGE(0,("Writing mountains to %s\n",fn));
    zsetlen(gen[0]->noc);
    if ((f = zwritehdr(fn,zfl,nmount,gen[0]->noc)) == NULL)
	MtxErrorExit("Cannot open %s for output: %S",fn);
    for (i = 0; i < nmount; ++i)
    {
	zwritevec(f,mountlist[i]->d,(long)1);
	MatFree(mountlist[i]);	/* We don't need them for step 2*/
    }
    fclose(f);
}


/* -----------------------------------------------------------------
   newmountain() - Take a vector, spin it up and check if it is a
	new mountain. If yes, add it to the mountain list and
	calculate the projection on each condensed module.
	Returns 1 if a new mountain has been found, 0 if not.
   ----------------------------------------------------------------- */

static int newmountain(matrix_t *vec, int cf)

{
    matrix_t *span, *backproj;
    int i;

    /* Spin up the vector and project back onto
       the condensed module where it came from.
       ----------------------------------------- */
    span = SpinUp(vec,LI.NGen,(const matrix_t **)gen,SF_FIRST|SF_SUB);
    MESSAGE(2,("Next vector spins up to %ld\n",span->nor));
    backproj = QuotProj(bild[cf],span);

    /* Check if it is a new mountain
       ----------------------------- */
    for (i = moffset[cf]; i < nmount; ++i)
    {
	if (backproj->nor == proj[i][cf]->nor)
	{
	    int issub = IsSubspace(proj[i][cf],backproj,0);
	    if (issub == -1)
		MtxErrorExit("IsSubspace() failed");
	    if (issub != 0)
		break;
	}
    }

    /* If it is new, add it to the list and calculate
       the other projections. Otherwise just forget it.
       ------------------------------------------------ */
    if (i == nmount)
    {
	int k;

	if (nmount >= MAXCYCL)
	    MtxErrorExit("TOO MANY MOUNTAINS, INCREASE MAXCYCL");
	proj[nmount] = NALLOC(matrix_t *,LI.NCf);

    	MESSAGE(2,("New Mountain %d\n",(int)i));
	for (k = 0; k < LI.NCf; ++k)
	{
    	    MESSAGE(3,("Projecting on %d\n",k));
	    if (k == cf)
	    	proj[nmount][cf] = backproj;
	    else
		proj[nmount][k] = QuotProj(bild[k],span);
	}
	mountlist[nmount] = vec;
	mountdim[nmount] = span->nor;
	++nmount;
	MatFree(span);
	return 1;
    }
    else
    {
	MatFree(backproj);
	MatFree(span);
	return 0;
    }
}


/* -----------------------------------------------------------------
   makeclass() - Find all cyclic subspaces of the projection of a
   mountain onto `its' irreducible.
   ----------------------------------------------------------------- */

static void makeclass(int mnt, int cf, matrix_t *vectors)

{
    char *tmp = NALLOC(char,vectors->nor+2);
    matrix_t *vec;
    long k, *p;
    size_t nvec;

    nvec = 0;
    MESSAGE(2,("Making equivalence class\n"));
    for (k = 1; k <= vectors->nor; ++k)
    {
	vec = MatCut(vectors,k,1,1,-1);
	tmp[k] = 0;
        MESSAGE(3,(" %ld",k));
	if (IsSubspace(vec,proj[mnt][cf],1))
	{
	    tmp[k] = 1;
	    ++nvec;
	}
	MatFree(vec);
    }
    MESSAGE(3,("\n"));

    p = class[mnt] = NALLOC(long,nvec+2);
    *p++ = nvec;
    for (k = 1; k <= vectors->nor; ++k)
	if (tmp[k]) *p++ = k;
    *p = -1;
    free(tmp);
}


/* -----------------------------------------------------------------
   step1() - Make all mountains and calculate the projections of
	mountains on condensed modules.
   ----------------------------------------------------------------- */

static void step1()

{
    matrix_t *vectors, *vec, *uk;
    char fn[200];
    int cf;
    long i;

    MESSAGE(0,("Step 1 (Mountains)\n"));
    nmount = 0;
    for (cf = 0; cf < LI.NCf; ++cf)	/* For each irreducible */
    {
	MESSAGE(0,("  %s%s: ",LI.BaseName,Lat_CfName(&LI,cf)));

	/* Read the vectors and the uncondense matrix
	   ------------------------------------------ */
	sprintf(fn,"%s%s.v",LI.BaseName,Lat_CfName(&LI,cf));
	vectors = MatLoad(fn);
	sprintf(fn,"%s%s.k",LI.BaseName,Lat_CfName(&LI,cf));
	uk = MatLoad(fn);

	/* Try each vector
	   --------------- */
	moffset[cf] = nmount;
	for (i = 1; i <= vectors->nor; ++i)
	{
	    if (i % 50 == 0)
		MESSAGE(1,("[%ld] ",i));
	    vec = MatCut(vectors,i,1,1,-1);
	    MatMul(vec,uk);	/* Uncondense */
	    if (newmountain(vec,cf))
		makeclass(nmount-1,cf,vectors);
	    else
	        MatFree(vec);
	}
	LI.Cf[cf].nmount = nmount - moffset[cf];

	MatFree(vectors);
	MatFree(uk);
	MESSAGE(0,("%ld mountain%s\n",LI.Cf[cf].nmount,
		LI.Cf[cf].nmount != 1 ? "s" : ""));

    }
    MESSAGE(0,("Total: %d mountain%s\n",nmount,nmount != 1 ? "s" : ""));
}




/* -----------------------------------------------------------------
   writeresult() - Write the incidence matrix
   ----------------------------------------------------------------- */

static void writeresult()

{
    FILE *f;
    char fn[200];
    int i;
    long l;

    /* Write the incidence matrix
       -------------------------- */
    f = SFOpen(strcat(strcpy(fn,LI.BaseName),".inc"),FM_CREATE);
    if (f == NULL) 
	MtxErrorExit("Cannot open %s: %S",fn);
    MESSAGE(0,("Writing incidence matrix (%s)\n",fn));
    l = (long) nmount;
    zwritelong(f,&l,1);
    for (i = 0; i < nmount; ++i)
	bs_write(f,subof[i]);
    fclose(f);

    /* Write the .cfinfo file
       ---------------------- */
    Lat_WriteInfo(&LI);
}



/* -----------------------------------------------------------------
   step2() - Calculate all incidences
   ----------------------------------------------------------------- */

static void step2()

{
    int i, k;
    int cfi, cfk;	/* Comp. factor corresponding to mountain i,j */

    MESSAGE(0,("Step 2 (Incidences)\n"));

    /* Allocate memory for the incidence matrix
       ---------------------------------------- */
    bs_setlen(nmount);
    for (i = 0; i < nmount; ++i)
	subof[i] = bs_alloc();

    /* Calculate the incidences
       ------------------------ */
    for (cfi = i = 0; i < nmount; ++i)
    {
	if (i == moffset[cfi])
	    MESSAGE(0,("  %s%s\n",LI.BaseName,Lat_CfName(&LI,cfi)));
	for (cfk=0, k = 0; k < nmount; ++k)
	{
	    int isubk, ksubi;
	    ksubi = IsSubspace(proj[k][cfk],proj[i][cfk],0);
	    isubk = IsSubspace(proj[i][cfi],proj[k][cfi],0);
	    if (ksubi < 0 || isubk < 0)
		MtxErrorExit("Subspace comparison failed");
	    if (ksubi )
		bs_set(subof[k],i);
	    if (isubk )
		bs_set(subof[i],k);
	    if (cfk < LI.NCf && k == moffset[cfk+1]-1)
		++cfk;
    	}
    	if (cfi < LI.NCf && i == moffset[cfi+1]-1)
	    ++cfi;
    }
}


/* -----------------------------------------------------------------
   writemountainsGAP() - Write mountains, dimensions and classes.
   ----------------------------------------------------------------- */

static void writemountainsGAP()

{
    int i;
    long *p;

    /* Write dimensions and classes
       ---------------------------- */
    printf( "MeatAxe.Dimensions := [" ) ;
    for (i = 0; i < nmount-1 ; ++i)
	printf( "%ld,", mountdim[i] );
    printf( "%ld] ;\n", mountdim[nmount-1] ) ;

    printf( "MeatAxe.Classes := [\n" ) ;
    for (i = 0; i < nmount; ++i)
    {
        printf( "[" ) ;
	for (p = class[i] ; *p > 0 ; ++p)
	{
	    printf( "%ld", *p ) ;
	    printf( p[1] > 0 ? "," : "]" ) ;
	}
	printf( i < nmount-1 ? ",\n" : "] ;\n");
    }
}


/* -----------------------------------------------------------------
   writeresultGAP() - Write the incidence matrix
   ----------------------------------------------------------------- */

static void writeresultGAP()

{	int i,  j;

	/* Write the incidence matrix
	   -------------------------- */
	printf( "MeatAxe.NMount := %ld;\n", (long) nmount );
	printf( "MeatAxe.Incidences := [\n" );
	for (i = 0; i < nmount; ++i)
        {
	    printf( "BlistList([" );
	    for (j = 0 ; j < nmount ; j++)
      	        printf( j < (nmount - 1) ? "%s," : "%s], [1])" , 
		       bs_test(subof[i],j) ? "1" : "0" ) ;

	    printf( i < nmount-1 ? ",\n" : "] ;\n" ) ;
	}
}




/* -----------------------------------------------------------------
   main()
   ----------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    int i;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("G")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	}
    }
    if (opt_ind != argc-1) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    MESSAGE(0,("\n*** INCIDENCE MATRIX ***\n\n"));
    init(argv[argc-1]);
    step1();
    writemountains();
    step2();
    writeresult();
    if (opt_G) 
    {
	writeresultGAP();
	writemountainsGAP();
    }
    if (msg_level >= 0)
    {
    	printf("\n");
    }
    MtxCleanup();
    return 0;
}


