/* ============================= C MeatAxe ==================================
   File:        $Id: mktree.c,v 1.6 1998/04/28 17:36:48 mringe Exp $
   Comment:     Program to calculate an element tree of a group.
   --------------------------------------------------------------------------
   Written by Markus Wiegelmann.
   Revised by Michael Ringe.
   (C) Copyright 1997  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>





/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

static matrix_t *elms[TK_MAXORDER];		/* The element list */
static TK_Info TKI;				/* Other global data */
static char *GenName;				/* Generator file name */
static char *OutName;				/* Output file name */


static char *helptext[] = {
    "SYNTAX",
    "    mktree [-g <ngen>] <gen> [<output>]",
    "",
    "OPTIONS",
    "    -g <ngen> ........ Set number of generators (default is 2)",
    "",
    "FILES",
    "    <gen>.1 ... <gen>.<ngen>  i  generators",
    "    <output>                  o  element tree",
    NULL};

static proginfo_t pinfo =
    { "mktree", "Calculate an element tree","$Revision: 1.6 $",helptext};



/* --------------------------------------------------------------------------
   IsInList() - Check if a matrix is in our list

   Description:
     This function checks if a given matrix, is already contained in our
	 element list, <elms>.

   Arguments:
     <mat>: The matrix to be checked.

   Remarks:
     The number of matrices in <elms> is taken from <TKI.Order>.

   Return:
     1 if the matrix is in <elms>, 0 otherwise
   -------------------------------------------------------------------------- */

static int IsInList(matrix_t *mat)

{
    long i, j;
    long test = 1;

    for (i = 0; i < TKI.Order; i++)
    {
        PTR matptr = mat->d;
	PTR compptr = elms[i]->d;
	for (j = 0; j < mat->nor; j++)
	{
	    if ((test = zcmprow(matptr,compptr)) != 0)
		break;
            zadvance(&matptr,(long)1);
            zadvance(&compptr,(long)1);
        }
        if (test == 0) 
	    return 1;
    }
    return 0;
}


/* --------------------------------------------------------------------------
   Init() - Initialize

   Description:
     This function initializes all global variables and processes command
	 line options and arguments.

   Arguments:
     <argc>: Number of arguments
	 <argv>: Argument list.
   -------------------------------------------------------------------------- */

static void Init(int argc, char *argv[])

{
    int i;
    MtxInit();
    memset(&TKI,0,sizeof(TKI));
    TKI.NGen = 2;
    MtxSetCommandLine(argc,argv,&pinfo);
    while ((i = MtxGetOpt("g:")) != OPT_END)
    {
	switch (i)
	{
	    case 'g':
		TKI.NGen = MtxGetInt();
		if (TKI.NGen < 1 || TKI.NGen > TK_MAXGEN)
		{
		    fprintf(stderr,"Illegal number of generators (%d)\n",
			TKI.NGen);
		    exit(1);
		}
		break;
	    default:
		MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
	}
    }

    /* Process any remaining arguments
       ------------------------------- */
    if (opt_ind >= argc) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    OutName = GenName = argv[opt_ind++];
    if (opt_ind < argc)
	OutName = argv[opt_ind++];
}




/* --------------------------------------------------------------------------
   AddToList() - Add one element to the element tree

   Description:
     This function adds one element to the element tree.

   Arguments:
	 <mat>: The new element.
	 <src>: Index of element <mat> was calculated from-
	 <gen>: Number of generator (1,2,3,...) used to calculate <mat> from
		<elem[src]>.

   Remarks:
     Both <src> and <gen> start counting with 1, not 0!
   -------------------------------------------------------------------------- */

static void AddToList(matrix_t *mat, int src, int gen)

{
    if (TKI.Order >= TK_MAXORDER)
    {
	    fprintf(stderr,"ERROR: Too many elements (>%d)\n",TK_MAXORDER);
	    exit(1);
    }
    elms[TKI.Order] = mat;
    TKI.Source[TKI.Order] = src;
    TKI.Gen[TKI.Order] = gen;
    ++TKI.Order;
}




/* --------------------------------------------------------------------------
   main() - Program entry point
   -------------------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    long i, src;
    matrix_t *gens[TK_MAXGEN];

    Init(argc,argv);

    /* Read generators
       --------------- */
    for (i = 0; i < TKI.NGen; i++)
    {
	char name[500];
        sprintf(name,"%s.%ld",GenName,i+1);
        if ((gens[i] = MatLoad(name)) == NULL)
        {
            fprintf(stderr,"Cannot read matrix %s.\n", name);
            return -1;
        }
    }

    /* Start tree with root vertex (that's the unit matrix)
       ---------------------------------------------------- */
    AddToList(MatId(gens[0]->fl,gens[0]->nor),-1,-1);

    /* Calculate all elements
       ---------------------- */
    for (src = 0; src < TKI.Order; ++src)
    {
	int g;
	for (g = 0; g < TKI.NGen; ++g)
	{
	    /* Calculate next element
	       ---------------------- */
	    matrix_t *newelem = MatDup(elms[src]);
	    MatMul(newelem,gens[g]);

	    /* If it is new, add to tree, else discard
	       --------------------------------------- */
	    if (IsInList(newelem))
	    {
		MatFree(newelem);
		if (src == 0)
		    fprintf(stderr,"Warning: generator %d is redundant\n",g+1);
	    }
	    else
	    {
		AddToList(newelem,src+1,g+1);
		if (TKI.Order % 50 == 0)
		{
		    printf("%d elements\n",TKI.Order);
		    fflush(stdout);
		}
	    }
	}
    }
    printf("Done. The group has %d elements.\n",TKI.Order);

    /* Write output file
       ----------------- */
    TK_WriteInfo(&TKI,OutName);

    MtxCleanup();
    return 0;
}
