/* ============================= C MeatAxe ==================================
   File:        $Id: mtensor.c,v 1.5 1998/06/25 19:02:33 mringe Exp $
   Comment:     Kronecker product of two matrices
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <stdlib.h>


/*{{{ --------------------------------------------------------------------------
!function MatTensor "Tensor product"
!section mat.op
!synopsis 
    matrix_t *MatTensor(const matrix_t *m1, const matrix_t *m2);
!arg m1
    Pointer to the first matrix.
!arg m2
    Pointer to the second matrix.
!return
    The tensor product of |m1| and |m2|, or |NULL| on error.
!description
    This function calculates the (Kronecker) tensor product $|m1|\otimes|m2|$.
    Both matrices must be over the same field.
!seealso 
-------------------------------------------------------------------------- }}}*/

matrix_t *MatTensor(const matrix_t *m1, const matrix_t *m2)

{
    MTXCONTEXT(MatTensor);
    matrix_t *temat;	    /* The result */
    PTR  x2;		    /* Row pointer for m2 */
    long i2;		    /* Row counter for m2 */
    FEL *rowbuf;	    /* Holds one row of <m2> in unpacked form */

    /* Check arguments
       --------------- */
    if (!MATRIX_IS_VALID(m1) || !MATRIX_IS_VALID(m2))
	return NULL;
    if (m1->fl != m2->fl)
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Allocate the result matrix and workspace
       ---------------------------------------- */
    temat = MatAlloc(m1->fl,m1->nor * m2->nor,m1->noc * m2->noc);
    if (temat == NULL)
	return NULL;
    rowbuf = NALLOC(FEL,m2->noc);
    if (rowbuf == NULL)
    {
	MatFree(temat);
	return NULL;
    }

    /* Calculate the Kronecker product
       ------------------------------- */
    x2 = m2->d;
    for (i2 = 1; i2 <= m2->nor; ++i2)
    {
	int k;
	PTR x1, x3;
	int i1;

	/* Unpack the row
	   -------------- */
	for (k = 0; k < m2->noc; ++k)
	    rowbuf[k] = zextract(x2,k+1);

	/* Initialize everything for the inner loop
	   ---------------------------------------- */
	x1 = m1->d;
	x3 = temat->d;
	zsetlen(temat->noc);
	zadvance(&x3,i2-1);

	/* Loop through all rows of <m1>
	   ----------------------------- */
	for (i1 = 1; i1 <= m1->nor; ++i1)
	{
	    int k1, k3;

	    /* Loop through all columns of <m1>
	      --------------------------------- */
	    for (k1 = k3 = 1; k1 <= m1->noc; ++k1)
	    {
		int k2;
		FEL f = zextract(x1,k1);
		if (f == F_ZERO)
		    k3 += m2->noc;
		else if (f == F_ONE)
		{
		    register FEL *rp = rowbuf;
		    for (k2 = 1; k2 <= m2->noc; ++k2, ++k3, ++rp)
		    {
			if (*rp != F_ZERO)
			    zinsert(x3,k3,*rp);
		    }
		}
		else
		{
		    register FEL *rp = rowbuf;
		    for (k2 = 1; k2 <= m2->noc; ++k2, ++k3, ++rp)
		    {
			if (*rp != F_ZERO)
			    zinsert(x3,k3,zmul(f,*rp));
		    }
		}
	    }

	    /* Next row of <m1>
	       ---------------- */
	    zsetlen(m1->noc);
	    zadvance(&x1,1);

	    zsetlen(temat->noc);
	    zadvance(&x3,m2->nor);
	}

	/* Next row of <m2>
	   ---------------- */
	zsetlen(m2->noc);
	zadvance(&x2,1);
    }


    /* Clean up
       -------- */
    FREE(rowbuf);
    return temat;
}
