/* ============================= C MeatAxe ==================================
   File:        $Id: os.c,v 1.7 1998/06/25 19:02:33 mringe Exp $
   Comment:     OS dependent stuff
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */



/* --------------------------------------------------------------------------
   These sybols can be defined:

   OS_NO_CPU_TIME .......... no CPU times are available
   OS_TIMES_AND_SYSCONF	.... use times() and sysconf() instead of getrusage()
   OS_NO_ITIMER ............ no interval timers
   -------------------------------------------------------------------------- */


/* --------------------------------------------------------------------------
   Include files
   -------------------------------------------------------------------------- */

#include "meataxe.h"

#if defined (_WIN32)

#include <windows.h>

#else

#include <sys/time.h>
#include <sys/times.h>
#include <signal.h>
#include <unistd.h>
#include <sys/resource.h>

#endif


#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>


/* ------------------------------------------------------------------
   Add some missing declarations
   ------------------------------------------------------------------ */


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

#if defined(OS_NO_CPU_TIME)
time_t zinittime = 0;
#endif


/* ------------------------------------------------------------------
   fopen() modes
   ------------------------------------------------------------------ */

static char *fmodes[6] = { "at","rt","wt","ab","rb","wb" };



/*{{{ --------------------------------------------------------------------------
!function SInit "Initialize OS dependent stuff"
!section kernel.os
!synopsis 
    void SInit();
!description
    This function is called during library initialization. It performs any
    OS-dependent actions. Applications never call this function directly.
-------------------------------------------------------------------------- }}}*/

void SInit()

{
#if defined(OS_NO_CPU_TIME)
    zinittime = time(NULL);
#endif
}




/*{{{ --------------------------------------------------------------------------
!function STimeUsed "CPU time"
!section kernel.os
!synopsis 
    long STimeUsed();
!return
    CPU time.
!description
    This function returns the CPU time used by the calling process in 
    units of $1/10$ seconds.
!seealso STimeLimit
-------------------------------------------------------------------------- }}}*/


long STimeUsed(void)
{

#if defined(_WIN32)

    FILETIME cr, ex, krnl, usr;
    LARGE_INTEGER k, u;
    GetProcessTimes(GetCurrentProcess(),&cr,&ex,&krnl,&usr);
    memcpy(&k,&krnl,sizeof(k));
    memcpy(&u,&usr,sizeof(u));
    return (long)((k.QuadPart + u.QuadPart) / 1000000);

#elif defined(OS_TIMES_AND_SYSCONF)

    struct tms t;
    static long clk_tck = 0;
    if (clk_tck == 0) 
	clk_tck = sysconf(_SC_CLK_TCK);
    times(&t);
    return ((long)((t.tms_utime + t.tms_stime) * 10 / clk_tck ));

#elif defined(OS_NO_CPU_TIME)

    return (time(NULL) - zinittime) * 10;

#else

    static struct rusage ru;
    getrusage(RUSAGE_SELF,&ru);
    return (ru.ru_utime.tv_sec * 10 + ru.ru_utime.tv_usec / 100000);

#endif


}




/*{{{ --------------------------------------------------------------------------
!function STimeLimit "Set CPU time limit"
!section kernel.os
!synopsis 
    void STimeLimit(long nsecs);
!arg nsecs
    CPU time limit in seconds.
!description
    This function sets a CPU time limit for the calling process. If the
    process does not exit within the specified time, it it killed.

    {\em Note:} This function is not available on all platforms.
!seealso STimeUsed
-------------------------------------------------------------------------- }}}*/

#ifdef _WIN32

DWORD CALLBACK Killer(void *x)
{
    DWORD ms = (long) x;
    Sleep(ms);
    MtxErrorExit("*****************\n*** GAME OVER ***\n*****************\n");
    return 0;
}

void STimeLimit(long nsecs)
{
    DWORD id;
    CreateThread(NULL,0,Killer,nsecs * 1000,0,&id);
}

#elif !defined(OS_NO_ITIMER)

void vtalarm(int i)

{
    MtxErrorExit("*****************\n*** GAME OVER ***\n*****************\n");
}

void STimeLimit(long nsecs)

{
    struct itimerval tv;

    tv.it_interval.tv_sec = 0;
    tv.it_interval.tv_usec = 0;
    tv.it_value.tv_sec = nsecs;
    tv.it_value.tv_usec = 0;
    setitimer(ITIMER_VIRTUAL,&tv,NULL);
    signal(SIGVTALRM,vtalarm);
}

#else /* No interval timer, sorry... */

void STimeLimit(long nsecs)

{
}

#endif




/*{{{ --------------------------------------------------------------------------
!function SFOpen "Open file"
!section kernel.os
!synopsis 
    FILE *SFOpen(const char *name, int mode);
!arg name
    File name.
!arg mode
    Open mode and additional flags (see below).
!return
    Pointer to an open file or |NULL| on error.
!description
    This function opens a file. |mode| must be any of |FM_READ| (open 
    for reading), |FM_CREAT| (create a new file) or |FM_APPEND| (append 
    to existing file or create a new file).  Additional flags may be or'ed
    to the mode:
    \begin{center}
    \begin{tabular}{lp{0.7\linewidth}}
    \hline
    Mode & Meaning \\
    \hline
    |FM_LIB| & If the file does not exist in the current directory,
	    look in the library directory. The library directory is
	    defined either by the environment variable \verb"MTXLIB" or
	    at compile-time by the symbol \verb"MTXLIB".\\
    |FM_TEXT| & Open in text mode. This flag must be used on some
	    systems (e.g., MS-DOS) to open text files. By default, files
	    are assumed to contain binary data.\\
    |FM_NOERROR| & Do not generate an error if the file does not exist.\\
    \hline
    \end{tabular}
    \end{center}
    The return value is a pointer to the open file or NULL on error.
    Possible error codes are (in |MtxErrNo|) are |ERR_BADARG|
    indicating an invalid mode, or |ERR_FILEOPEN| if the file could
    not be opened for whatever reason.
!seealso @OsIo zreadhdr zwritehdr
-------------------------------------------------------------------------- }}}*/

FILE *SFOpen(const char *name, int mode)

{
    char buf[300];
    int m;
    FILE *f;
    MTXCONTEXT(SFOpen);

    m = mode & 0x0F;			/* Append, read or create? */
    if ((mode & FM_TEXT) == 0) m += 3;	/* Binary mode */
    if (m < 0 || m >= 6)
    {
	MTXERROR1("Invalid file mode %d",mode);
	return NULL;
    }
    f = fopen(name,fmodes[m]);
    if (f != NULL) 
	return f;

    /* Search library directory
       ------------------------ */
    if ((mode & FM_LIB) != 0) 
    {
	strcpy(buf,MeatAxeLib);
	strcat(buf,"/");
	strcat(buf,name);
	f = fopen(buf,fmodes[m]);
    }

    /* Error handling
       -------------- */
    if (f == NULL && (mode & FM_NOERROR) == 0)
	MTXERROR1("%s: %S",name);

    return f;
}



/*{{{ --------------------------------------------------------------------------
!function SFSeek "Set file pointer"
!section kernel.os
!synopsis
    int SFSeek(FILE *file, long pos);
!arg file
    Pointer to |FILE| structure.
!arg pos
    New position of file pointer.
!return
    The function returns 0 on success. Any other value indicates an error.
!description
    This function sets the file pointer to a given position. if |pos| is greater
    or equal to zero, it is interpreted as an absolute position (relative to
    start of file). If |pos| is negative, the file pointer is moved to the end 
    of file.
!seealso @OsIo zseek
-------------------------------------------------------------------------- }}}*/

int SFSeek(FILE *file, long pos)

{
    if (pos < 0)
	return fseek(file,(long) 0,SEEK_END);
    else
	return fseek(file,pos,SEEK_SET);
}


/*{{{ --------------------------------------------------------------------------
!function Smalloc "Allocate memory"
!section kernel.os
!synopsis 
    void *Smalloc(size_t nbytes);
!arg nbytes
    Size of memory block to allocate.
!return
    Pointer to memory block or |NULL| on error.
!description
   This function works like |malloc()|. Unlike some implementations of 
   |malloc()|, it returns a valid pointer if |nbytes| is zero.
!seealso @OsMem
-------------------------------------------------------------------------- }}}*/

void *Smalloc(size_t nbytes)

{
    MTXCONTEXT(Smalloc);
    void *x;

    if (nbytes == 0) 
	nbytes = 1;
    x = malloc(nbytes);
    if (x == NULL)
	MTXERROR1("Cannot allocate %d bytes: %S",(int) nbytes);
    return x;
}



/*{{{ --------------------------------------------------------------------------
!function Srealloc "Resize memory block"
!section kernel.os
!synopsis 
    void *Srealloc(size_t nbytes);
!arg buf
    Pointer to memory block.
!arg nbytes
    Desired size.
!return
    Pointer to resized memory block or NULL on error.
!description
   This function works like |relloc()| but handles zero-length blocks 
   correctly.
!seealso @OsMem
-------------------------------------------------------------------------- }}}*/

void *Srealloc(void *buf, size_t nbytes)

{
    MTXCONTEXT(Srealloc);
    void *x;

    if (nbytes == 0) 
	nbytes = 1;
    x = realloc(buf,nbytes);
    if (x == NULL)
	MTXERROR1("Cannot reallocate %d bytes: %S",(int) nbytes);
    return x;
}


