/* ============================= C MeatAxe ==================================
   File:        $Id: polgcd.c,v 1.2 1998/06/25 19:02:33 mringe Exp $
   Comment:     Polynomials g.c.d functions
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"




static void normalize(poly_t *p, FEL f)
{
    register FEL *buf;
    register int i;

    if (f == F_ONE)
	return;
    for (buf = p->buf, i = (int) p->deg; i >= 0; --i, ++buf)
    {
        if (*buf != F_ZERO) 
	    *buf = zdiv(*buf,f);
    }
}



/*{{{ --------------------------------------------------------------------------
!function PolGcd "G.c.d. of polynomials"
!section pol.op
!synopsis 
    poly_t *PolGcd(const poly_t *a, const poly_t *b);
!arg a
    First polynomial.
!arg b
    Second polynomial.
!return
    Greates common divisor of |a| and |b|, or |NULL| on error.
!description
    This function calculates the gratest common divisor of two poynomials.
    The polynomials must be over the same field, and at least one of
    them must be different from zero.

    Unlike most polynomial functions, |PolGcd()| normalizes the result, 
    i.e.\ the leading coefficient of the g.c.d.\ is always one.
!seealso PolGcdEx
-------------------------------------------------------------------------- }}}*/

poly_t *PolGcd(const poly_t *a, const poly_t *b)

{
    MTXCONTEXT(PolGcd);
    poly_t *p,*q,*tmp;
    FEL f;

    /* Check arguments
       --------------- */
    if (!POLYNOMIAL_IS_VALID(a) || !POLYNOMIAL_IS_VALID(b))
	return NULL;
    if (a->fl != b->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Handle special cases
       -------------------- */
    if (a->deg == -1)
    {
	if (b->deg == -1) 
	{
	    MTXERROR1("%E",ERR_DIV0);
	    return NULL;
	}
	return PolDup(b);
    }
    else if (b->deg == -1)
	return PolDup(a);

    /* Both |a| and |b| are not null: Calculate gcd
       -------------------------------------------- */
    zsetfield(a->fl);
    if (a->deg < b->deg)
    {
	p = PolDup(b);
	q = PolDup(a);
    }
    else
    {
	p = PolDup(a);
	q = PolDup(b);
    }
    while (q->deg >= 0)
    {
	if (PolMod(p,q) == NULL)
	    return NULL;
	tmp = p;
	p = q;
	q = tmp;
    }
    PolFree(q);

    /* Normalize the result
       -------------------- */
    if ((f = p->buf[p->deg]) != F_ONE)
	normalize(p,f);

    return p;
}




/*{{{ --------------------------------------------------------------------------
!function PolGcdEx "G.c.d. of polynomials"
!section pol.op
!synopsis 
    int PolGcdEx(const poly_t *a, const poly_t *b, poly_t **result);
!arg a
    First polynomial.
!arg b
    Second polynomial.
!arg result
    Result buffer for the g.c.d. and coefficients (see below).
!return
    $0$ on sucess, $-1$ on error.
!description
    Given two polynomials $a,b\in\GF{q}[x]$, this function calculates the
    greatest common divisor $g:=\gcd(a,b)$ and two polynomial $p,q\in\GF{q}[x]$
    such that $g=pa+qb$. As with |PolGcd()|, $a$ and $b$ must not both be 
    zero, and the {g.c.d.} is always normalized.

    |result| must be a pointer to an array of three |poly_t *| items. If the 
    function is successful, pointers to $g$, $p$, and $q$ are stored in
    |result[0]|, |result[1]|, and |result[2]|, respectively. The caller is
    responsible for destroying these polynomials when they are no longer
    needed. In particular, you must not use the same |result| buffer again
    with |PolGcdEx()| before the contents have been freed.
!seealso PolGcd
-------------------------------------------------------------------------- }}}*/

int PolGcdEx(const poly_t *a, const poly_t *b, poly_t **result)

{
    MTXCONTEXT(PolGcd);
    poly_t *p, *q, *pa, *pb, *qa, *qb;
    FEL f;
    int alessb;

    /* Check arguments
       --------------- */
    if (!POLYNOMIAL_IS_VALID(a) || !POLYNOMIAL_IS_VALID(b))
	return -1;
    if (result == NULL)
	return MTXERROR1("result: %E",ERR_BADARG), -1;
    if (a->fl != b->fl) 
	return MTXERROR1("%E",ERR_INCOMPAT), -1;

    alessb = a->deg < b->deg;
    p = PolDup(alessb ? b : a);
    q = PolDup(alessb ? a : b);
    pb = PolAlloc(a->fl,alessb ? 0 : -1);
    qa = PolAlloc(a->fl,alessb ? 0 : -1);
    pa = PolAlloc(a->fl,alessb ? -1 : 0);
    qb = PolAlloc(a->fl,alessb ? -1 : 0);

    while (q->deg >= 0)
    {
	int i;
	poly_t *tmp, *atmp, *btmp;
	poly_t *m = PolDivMod(p,q);
	tmp = p;
	p = q;
	q = tmp;

	atmp = PolDup(qa);
	btmp = PolDup(qb);
	for (i = 0; i <= m->deg; ++i)
	    m->buf[i] = zsub(F_ZERO,m->buf[i]);
	PolMul(atmp,m);
	PolMul(btmp,m);
	PolAdd(atmp,pa);
	PolAdd(btmp,pb);
	
	PolFree(pa);
	PolFree(pb);

	pa = qa;
	pb = qb;

	qa = atmp;
	qb = btmp;
	PolFree(m);
    }

    /* Normalize the result
       -------------------- */
    if ((f = p->buf[p->deg]) != F_ONE)
    {
	normalize(p,f);
	normalize(pa,f);
	normalize(pb,f);
    }

    result[0] = p;
    result[1] = pa;
    result[2] = pb;
    PolFree(q);
    PolFree(qa);
    PolFree(qb);

    return 0;
}



