/* ============================= C MeatAxe ==================================
   File:        $Id: pwkond.c,v 1.12 1998/07/29 16:52:10 mringe Exp $
   Comment:     This program finds the peak words and does the generalized
                condensation for each composition factor.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>




/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

long dim;
long quotdim;
static long *piv;		/* Pivot table */
static matrix_t *gen[MAXGEN];	/* Generators */
static wgdata_t *basis;
int opt_G = 0;
int opt_n = 0;			/* No condensation, PW only */
int opt_p = 0;			/* Use full polynomials in PW search */
int opt_t = 0;			/* Transform generators to std basis  */
matrix_t *peakword;
Lat_Info LI;			/* Data from .cfinfo file */

static char *helptext[] = {
"SYNTAX",
"    pwkond [<Options>] <Name>",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -V            Verbose, more messages.",
"    -G            GAP output (implies -Q).",
"    -T <Seconds>  Set CPU time limit",
"    -n            Find peak words only, do not condense",
"    -p            Use full polynomials in peak word search",
"    -e <List>     Exclude words from search. <List> is a",
"                  comma-separated list of numbers or ranges (X-Y).",
"    -t            Transform generators into standard basis"
"",
"FILES",
"    Input files are produced by CHOP (see `chop -help').",
"    Output is written to the following files:",
"       <Name><Dim><a>.<N>k        Condensed generators",
"       <Name><Dim><a>.<N>.std.k   Condensed generators in std basis (with -t)",
"       <Name><Dim><a>.np          Condensed peak word",
"       <Name><Dim><a>.im          Image used for condenstion",
"       <Name><Dim><a>.k           Uncondense matrix (used by mkinc)",
NULL};

static proginfo_t pinfo =
   { "pwkond", "Peakword Condensation",
     "$Revision: 1.12 $", helptext };



#define MAXLOCK 100
wgdata_t *cfbasis[LAT_MAXCF];
matrix_t *PWNullSpace[LAT_MAXCF];	/* Peak word null space */
long exclude[MAXLOCK][2];
int nexclude=0;
static int PeakWordsMissing;	    /* Number of missing peak words */


/* ------------------------------------------------------------------
   init2() - Program Initialization, Step 2
   ------------------------------------------------------------------ */

static void init2()

{
    char fn[LAT_MAXBASENAME+10];
    int i;
    long fl;

    for (i = 0; i < LI.NGen; ++i)
    {
	sprintf(fn,"%s.%d",LI.BaseName,i+1);
	gen[i] = MatLoad(fn);
    }
    fl = gen[0]->fl;
    basis = WgInit(LI.NGen,(const matrix_t **)gen);
    dim = gen[0]->nor;

    /* Allocate the pivot table
       ------------------------ */
    piv = NALLOC(long,dim+1);
    if (piv == NULL)
	MtxErrorExit("Cannot allocate pivot table");

}


/* ------------------------------------------------------------------
   power() - Find the stable power of the peak word
   ------------------------------------------------------------------ */

static matrix_t *power(int i)

{
    long nul1, nul2, k0;
    matrix_t *tmp;

    k0 = 1;
    tmp = MatDup(peakword);
    do
    {   matrix_t *tmp2 = MatMul(MatDup(tmp),tmp);
	nul1 = MatNullity__(tmp);
	tmp = MatDup(tmp2);
	nul2 = MatNullity__(tmp2);
	k0 *= 2;
    }
    while (nul2 != nul1);

    MESSAGE(0,("pwr=%ld, nul=%ld, mult=%ld, ",k0/1,nul2,
	LI.Cf[i].mult));
    if (nul2 != LI.Cf[i].mult * LI.Cf[i].spl)
	MtxErrorExit("Something is wrong...");
    quotdim = nul2;
    return tmp;
}



/* ------------------------------------------------------------------
   gkond() - Generalized condensation of one matrix
   ------------------------------------------------------------------ */

static void gkond(int i,matrix_t *k,matrix_t *w, const char *name)

{
    char fn[LAT_MAXBASENAME+10];
    matrix_t *x1, *x2;

    x1 = MatDup(k);
    MatMul(x1,w);

    x2 = MatAlloc(x1->fl,x1->nor,quotdim);
    zquot(x1->d,x1->nor,x2->d);

    sprintf(fn,"%s%s.%s",LI.BaseName,Lat_CfName(&LI,i),name);
    MatSave(x2,fn);
    MatFree(x1);
    MatFree(x2);
}

/* ------------------------------------------------------------------
   Standardize() - Bringe Kompositionsfaktor auf Standardform.
   Schreibt die neuen Erzeuger in XXX.std.n und die Operationen in
   XXX.op.
   ------------------------------------------------------------------ */

static void Standardize(int i)

{
    matrix_t *sb;
    int k;
    char fn[200];
    FILE *f;
    matrix_t *std[MAXGEN];
    extern long *SBasisOp;

    MESSAGE(1,("Make standard basis\n"));
    sb = SpinUp(PWNullSpace[i],cfbasis[i]->NGen,
	(const matrix_t **)cfbasis[i]->Gen,SF_FIRST|SF_CYCLIC|SF_STD);

    /* Schreibe .op-Datei
       ------------------ */
    sprintf(fn,"%s%s.op",LI.BaseName,Lat_CfName(&LI,i));
    MESSAGE(2,("Write operations to %s\n",fn));
    f = zwritehdr(fn,-T_IMAT,sb->nor,2);
    if (f == NULL)
	MtxErrorExit("%s: %E",fn,MtxErrNo);
    zwritelong(f,SBasisOp+2,sb->nor * 2);
    fclose(f);

    /* Bringe Erzeuger auf Standardform
       -------------------------------- */
    MESSAGE(2,("Change to standard basis:"));
    ChangeBasis(sb,cfbasis[i]->NGen,(const matrix_t **)cfbasis[i]->Gen,std);
    MatFree(sb);
    for (k = 0; k < cfbasis[i]->NGen; ++k)
    {
    	sprintf(fn,"%s%s.std.%d",LI.BaseName,Lat_CfName(&LI,i),k+1);
    	MESSAGE(2,(" %s",fn));
	MatSave(std[k],fn);
	MatFree(std[k]);
    }
    MESSAGE(2,("\n"));
}



/* ------------------------------------------------------------------
   kond() - Generalized condensation for one irreducible
   ------------------------------------------------------------------ */

static void kond(int cf)

{
    char fn[LAT_MAXBASENAME+10];
    matrix_t *kern, *bild, *m, *k;
    int j;
		
    /* Make the peak word, find its stable power,
       calculate both kernel and image.
       ------------------------------------------- */
    MESSAGE(0,("Condensing %s%s: word=%ld, ",LI.BaseName,Lat_CfName(&LI,cf),
    	LI.Cf[cf].peakword));
    peakword = WgMakeWord(basis,LI.Cf[cf].peakword);
    MatInsert_(peakword,LI.Cf[cf].peakpol);
    bild = power(cf);
    kern = MatNullSpace_(bild);
    /* OPT: The pivot table is calculated twice */
    MatEchelonize(bild);
    if (zmkpivot(bild->d,bild->nor,piv) != 0)	
	MtxErrorExit("INTERNAL ERROR: Basis is not in echelon form");
    zquotinit(bild->d,bild->nor,piv);

    /* Write out the image
       ------------------- */
    sprintf(fn,"%s%s.im",LI.BaseName,Lat_CfName(&LI,cf));
    MatSave(bild,fn);

    /* Write out the `uncondense matrix'
       --------------------------------- */
    m = MatAlloc(kern->fl,kern->nor,quotdim);
    zquot(kern->d,kern->nor,m->d);
    k = MatInv(m);
    MatMul(k,kern);
    sprintf(fn,"%s%s.k",LI.BaseName,Lat_CfName(&LI,cf));
    MatSave(k,fn);

    /* Condense all generators
       ----------------------- */
    MESSAGE(0,("("));
    for (j = 0; j < LI.NGen; ++j)
    {
	sprintf(fn,"%dk",j+1);
	gkond(cf,k,gen[j],fn);
	MESSAGE(0,(" %d",j+1));
    }
    MESSAGE(0,(" )\n"));

    /* Condense the peak word
       ---------------------- */
    gkond(cf,k,peakword,"np");

    MatFree(m); MatFree(k);
    MatFree(kern);
    MatFree(bild);
    MatFree(peakword);
}


/* --------------------------------------------------------------------------
   PeakWordFound() - A peak word has been found

   Arguments:
     <w>: The word number

   Description:
     This function is called each time a peak word is found. Depending on the
     command line options we condense the peak word, and transform the
     generators to standard basis.
     We also write the .cfinfo file each time. This may be useful if the
     program does not terminate or dies unexpectedly.
   -------------------------------------------------------------------------- */

static void PeakWordFound(int i)

{
    if (MSG0)
    {
	printf("Peak word for %s%s is %ld (%s), pol=",LI.BaseName,
	    Lat_CfName(&LI,i), LI.Cf[i].peakword,
	    WgSymbolicName(cfbasis[i],LI.Cf[i].peakword));
	PolPrint(NULL,LI.Cf[i].peakpol);
	printf(", nullity=%ld\n",LI.Cf[i].spl);
	fflush(stdout);
    }
    if (!opt_n)
	kond(i);		/* Condense */
    if (opt_t)
	Standardize(i);		/* Transform to std basis */
    Lat_WriteInfo(&LI);		/* Write the .cfinfo file */
    --PeakWordsMissing;
}




static int isexcluded(long w)

{	
    int i;

    for (i = 0; i < nexclude; ++i)
	if (w >= exclude[i][0] && w <= exclude[i][1])
	    return 1;
    return 0;
}


static void parselist(char *c)

{	
    long a, b;

    while (*c != 0)
    {	
	a = b = 0;
	while (*c >= '0' && *c <= '9')
	    a = a * 10 + (*c++ - '0');
	if (*c == '-')
	{	
	    ++c;
	    while (*c >= '0' && *c <= '9')
		b = b * 10 + (*c++ - '0');
	}
	else
	b = a;
	if (a == 0 || b == 0 || a > b) 
		MtxErrorExit("BAD ARGUMENTS");
	exclude[nexclude][0] = a;
	exclude[nexclude][1] = b;
	++nexclude;
	if (*c == ',') 
	    ++c;
    }
}


static void init(int argc, char *argv[])

{
    char fn[50];
    int i;
    matrix_t *cfgen[MAXGEN];

    /* Read the .cfinfo file
       --------------------- */
    Lat_ReadInfo(&LI,argv[argc-1]);
    PeakWordsMissing = LI.NCf;

    /* Read the generators for each composition factor
       ----------------------------------------------- */
    for (i = 0; i < LI.NCf; ++i)
    {
	int k;
	MESSAGE(1,("Reading %s%s",LI.BaseName,Lat_CfName(&LI,i)));
	for (k = 0; k < LI.NGen; ++k)
	{
	    sprintf(fn,"%s%s.%d",LI.BaseName,Lat_CfName(&LI,i),k+1);
	    MESSAGE(1,(" %d",k+1));
	    cfgen[k] = MatLoad(fn);
	}
	MESSAGE(1,("\n"));
	cfbasis[i] = WgInit(LI.NGen,(const matrix_t **)cfgen);
	LI.Cf[i].peakword = -1;
    }
}




/* ------------------------------------------------------------------
   try() - Try another word
   ------------------------------------------------------------------ */

static void addid(matrix_t *m, FEL f)
{
    PTR x;
    long i;
    if (f == F_ZERO)
	return;
    zsetlen(m->noc);
    for (i = 1, x = m->d; i <= m->nor; ++i, zadvance(&x,1))
	zinsert(x,i,zadd(zextract(x,i),f));
}


static int try2(long w, FEL f)

{
    int i, ppos = -1;
    long nul;

    if (isexcluded(w))
	return -1;
    MESSAGE(3,("Word %ld+%ldI:",w,zftoi(f)));
    for (i = 0; i < LI.NCf; ++i)  /* For each composition factor... */
    {
	matrix_t *word;
	
	word = WgMakeWord(cfbasis[i],w);
	addid(word,f);
	nul = MatNullity__(MatDup(word));
        MESSAGE(3,(" %ld",nul));
	if (nul != 0 && nul != LI.Cf[i].spl)
	{
	    MatFree(word);
	    MESSAGE(3,("failed\n"));
	    return -1;
	}
	if (nul == LI.Cf[i].spl)
	{
	    if (ppos >= 0 || LI.Cf[i].peakword > 0)
	    {
		MatFree(word);
	    	MESSAGE(3,("failed\n"));
		return -1;  /* Nullity should be 0 */
	    }
	    nul = MatNullity__(MatMul(MatDup(word),word));
	    if (nul != LI.Cf[i].spl)
	    {
		MatFree(word);
	    	MESSAGE(3,("failed (nullity not stable)\n"));
		return -1;  /* Nullity is not stable */
	    }
	    ppos = i;
	}
        MatFree(word);
    }
    MESSAGE(3,("\n"));

    if (ppos > -1) /* we have found a new peak word */
    {
	poly_t *pp;
	matrix_t *word;
	LI.Cf[ppos].peakword = w;

	/* Rechne nochmals den Nullraum aus (wir brauchen
	   ihn spaeter fuer die Standardform.
	   ----------------------------------------------- */
	word = WgMakeWord(cfbasis[ppos],w);
	addid(word,f);
	PWNullSpace[ppos] = MatNullSpace__(word);

	/* Erzeuge das Peakpolynom (hier immer vom Grad 1)
	   ----------------------------------------------- */
	pp = PolAlloc(zfl,1);
	pp->buf[0] = f;
	LI.Cf[ppos].peakpol = pp;
        PeakWordFound(ppos);
    }
    return ppos;
}





/* --------------------------------------------------------------------------
   Try_Linear() - Find peak word using linear polynomials

   Arguments:
     <w>: The word number

   Description:
     This function checks if there is one or more peak words $W_w+\lambda 1$
     with $\lambda\in F$.
   -------------------------------------------------------------------------- */

static void Try_Linear(long w)

{
    long f;
    for (f = 0; f < zfl && PeakWordsMissing > 0; ++f)
	try2(w,zitof(f));
}



/* ------------------------------------------------------------------
   tryp2()
   ------------------------------------------------------------------ */

static int tryp2(long w, int cf, poly_t *pol)

{
    int i;

    for (i = 0; i < LI.NCf; ++i)
    {
	matrix_t *word, *wordp;
	long nul;

	if (i == cf)
	    continue;
	word = WgMakeWord(cfbasis[i],w);
	wordp = MatInsert(word,pol);
	MatFree(word);
	nul = MatNullity__(wordp);
	if (nul != 0) 
	    return -1;
    }
    return 0;
}





/* ------------------------------------------------------------------
   try_p() - Try another word using polynomials
   ------------------------------------------------------------------ */

static int try_p(long w)

{
    int i;

    if (isexcluded(w)) 
	return -1;

    for (i = 0; i < LI.NCf; ++i)
    {
	matrix_t *word;
	fpoly_t *mp;
	int k;

	if (LI.Cf[i].peakword > 0)  
	    continue;			/* We already have a peak word */
	word = WgMakeWord(cfbasis[i],w);
	mp = MinPol(word);
        if (MSG3)
	{
	    int j;
	    printf("%s, minpol =\n",Lat_CfName(&LI,i));
	    for (j = 0; j < (int) mp->len; ++j)
	    {
		PolPrint(NULL,mp->p[j]);
		printf(" ^ %ld\n",mp->e[j]);
	    }
        }
	for (k = 0; k < (int) mp->len; ++k)
	{
	    if (mp->p[k]->deg * mp->e[k] == LI.Cf[i].spl)
	    {
	       matrix_t *wp, *wp2;
	       long nul;

	       if (MSG3)
	       {
		   printf("%s, ",Lat_CfName(&LI,i));
		   PolPrint("factor",mp->p[k]);
	       }
	       if (tryp2(w,i,mp->p[k]) == -1) 
		   continue;

	       /* Check if the nullity is stable
	          ------------------------------ */
	       wp = MatInsert(word,mp->p[k]);
	       wp2 = MatMul(MatDup(wp),wp);
	       MatFree(wp);
	       nul = MatNullity__(wp2);
	       if (nul != LI.Cf[i].spl) 
		   continue;
	       break;
	    }
	}

	if (k < (int) mp->len)
	{
	    LI.Cf[i].peakword = w;
	    LI.Cf[i].peakpol = PolDup(mp->p[k]);
	    PWNullSpace[i] = MatNullSpace__(MatInsert(word,mp->p[k]));
            PeakWordFound(i);
	}
	else
	    k = -1;	/* Not found */
	FpolFree(mp);
	MatFree(word);
	if (k >= 0)
	    return i;
    }
    return -1;
}



/* --------------------------------------------------------------------------
   main()
   -------------------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    int i;
    long w;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("Gptne:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 'n': opt_n = 1; break;
	    case 'p': opt_p = 1; break;
	    case 't': opt_t = 1; break;
	    case 'e':
		parselist(opt_text);
		break;
	}
    }
    if (opt_ind != argc-1) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    MESSAGE(0,("\n*** PEAK WORD CONDENSATION ***\n\n"));
    init(argc,argv);
    init2();

    /* Find peak words
       --------------- */
    for (w = 1; PeakWordsMissing > 0; ++w)
    {
	if ((MSG1 && (w % 50 == 0)) || MSG2)
	{
	    printf("Word %ld\n",w);
	    fflush(stdout);
	}
	if (opt_p) try_p(w); else Try_Linear(w);
    }
    Lat_WriteInfo(&LI);
    if (opt_G)
    {
	printf("MeatAxe.PeakWords := [");
	for (i = 0; i < LI.NCf; ++i)
	{
	    if (i > 0) printf(",");
	    printf("%ld",LI.Cf[i].peakword);
	}
	printf("];\n");
    }
    MtxCleanup();
    return 0;
}


