/* ============================= C MeatAxe ==================================
   File:        $Id: rdcfgen.c,v 1.7 1998/04/28 17:37:08 mringe Exp $
   Comment:     Read generators for constituents.
   --------------------------------------------------------------------------
   (C) Copyright 1997  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"


/* --------------------------------------------------------------------------
   Lat_ReadCfGens() - Read generators for one constituent

   Description:
     This function reads the generators of one constituent and optionally
     performs some basic operations on the genrators (invert, transpose).

   Arguments:
     <info>: Pointer to a <Lat_Info> data structure for the module.
     <cf>: Index of the constituent to be read.
     <gen>: Pointer to a user-supplied array where the genrators are stored.
     <flags>: Optional flags. This may be any combination of LAT_RF_xxx
	constants (see below).

   Remarks:
     If any error occurs, the function terminates the program with an
     appropriate error message.

     The following table lists all available flags:

	 Name              Description
	 LAT_RG_STD        Read generators in standard basis (file XY.std.N).
			   These files are created by pwkond. Default is to
			   read the generators in 'random' basis as they are
			   produced by chop.
	 LAT_RG_INVERT     Invert the generators
	 LAT_RG_TRANSPOSE  Transpose the generators
   -------------------------------------------------------------------------- */

void Lat_ReadCfGens(Lat_Info *info, int cf, matrix_t *gen[], int flags)

{
    int i;
    for (i = 0; i < info->NGen; ++i)
    {
	char fn[LAT_MAXBASENAME + 20];

	/* Make file name
	   -------------- */
	if (flags & LAT_RG_STD)
	    sprintf(fn,"%s%s.std.%d",info->BaseName,Lat_CfName(info,cf),i+1);
	else
	    sprintf(fn,"%s%s.%d",info->BaseName,Lat_CfName(info,cf),i+1);

	/* Load the matrix
	   --------------- */
	gen[i] = MatLoad(fn);
	if (gen[i] == NULL)
	    MtxErrorExit("Error reading %s",fn);

	/* Perform any additional transformations
	   -------------------------------------- */
	if (flags & LAT_RG_INVERT)
	{
	    matrix_t *mat2 = MatInv(gen[i]);
	    if (mat2 == NULL)
		continue;
	    MatFree(gen[i]);
	    gen[i] = mat2;
	}
	if (flags & LAT_RG_TRANSPOSE)
	{
	    matrix_t *mat2 = MatTr(gen[i]);
	    if (mat2 == NULL)
		continue;
	    MatFree(gen[i]);
	    gen[i] = mat2;
	}
    }
}

