/* ============================= C MeatAxe ==================================
   File:        $Id: spinup.c,v 1.8 1998/06/25 19:02:34 mringe Exp $
   Comment:     Spin-up and split.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"


long *SBasisOp = NULL;

#define OPVEC(i) SBasisOp[2*(i)]
#define OPGEN(i) SBasisOp[2*(i)+1]


/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

static long Dim = 0;			/* Dimension of the whole space */
static long *Piv = NULL;		/* Pivot table */
static const matrix_t *Seed = NULL;	/* Seed space */
static matrix_t *Span = NULL;		/* Span */
static long SpanDim = 0;		/* Dimension of the span */
static int Flags = 0;			/* Flags */
static int NGen;			/* Number of generators */
static const matrix_t **Gen;		/* Generators */
static matrix_t *StdSpan = NULL;	/* Span */


/* --------------------------------------------------------------------------
   Spin1() - Spin up one seed vector

   Arguments:
     <seed>: The seed vector.
     <seedno>: The seed vector number.

   Description:
     This function spins up the next seed vector.

   Return:
     0 = success, 1 = no success, -1 = error
   -------------------------------------------------------------------------- */

static int Spin1(PTR seed, int seedno)
{
    MTXCONTEXT(Spin1);
    int igen;	    /* Generator to apply next */
    PTR get;	    /* Vector to map */
    PTR put;	    /* New vectors go here */
    PTR stdget, stdput;
    FEL f;
    int iget;

    /* If we are not in 'combine' mode, start with an empty space.
       In 'combine' mode, keep what has been found so far.
       ------------------------------------------------------------ */
    if ((Flags & SF_MODE_MASK) != SF_COMBINE)
    {
	SpanDim = 0;
    }

    /* Initialize <get> and <put> to point to the first free row in <Span>
       ------------------------------------------------------------------- */
    get = Span->d;
    zadvance(&get,SpanDim);
    iget = SpanDim + 1;
    put = get;
    if (Flags & SF_STD)
    {
	stdget = StdSpan->d;
	zadvance(&stdget,SpanDim);
	stdput = stdget;
    }

    /* Copy the seed vector to <put>, extending the pivot table
       -------------------------------------------------------- */
    zmoverow(put,seed);
    zcleanrow(put,Span->d,SpanDim,Piv);
    OPVEC(SpanDim+1) = seedno;
    OPGEN(SpanDim+1) = 0;	/* 0 means it's a seed vector */
    if ((Piv[SpanDim + 1] = zfindpiv(put,&f)) != 0)
    {
	++SpanDim;
	zadvance(&put,1);
	if (Flags & SF_STD)
	{
    	    zmoverow(stdput,seed);
	    zadvance(&stdput,1);
	}
    }

    /* Spin up
       ------- */
    igen = 0;
    while (get != put && SpanDim < Dim && NGen > 0)
    {
	/* Apply the next generator to <get>. IF this was 
	   the last generator, advance <get> to the next vector.
	   ----------------------------------------------------- */
	if (Flags & SF_STD)
	{
	    zmaprow(stdget,Gen[igen]->d,Dim,stdput);
	    zmoverow(put,stdput);
	}
	else
	    zmaprow(get,Gen[igen]->d,Dim,put);
    	OPVEC(SpanDim+1) = iget;
	OPGEN(SpanDim+1) = igen+1;
        if (++igen >= NGen)
	{
	    igen = 0;
	    zadvance(&get,1);
	    zadvance(&stdget,1);
	    ++iget;
	}

	/* Clean the result with the existing basis
	   ---------------------------------------- */
	zcleanrow(put,Span->d,SpanDim,Piv);
	if ((Piv[SpanDim + 1] = zfindpiv(put,&f)) != 0)
	{
	    ++SpanDim;
	    zadvance(&put,1);
	    zadvance(&stdput,1);
	}
    }


    /* Return success code, depending on the mode and the result 
       of the spin up 
       --------------------------------------------------------- */
    switch (Flags & SF_MODE_MASK)
    {
	case SF_SUB:
	    return (SpanDim > 0 && SpanDim < Dim) ? 0 : 1;
	case SF_CYCLIC:
	case SF_COMBINE:
	    return SpanDim < Dim ? 1 : 0;
    }
    MTXERROR1("Invalid search mode %d",Flags & SF_MODE_MASK);
    return -1;
}



/* --------------------------------------------------------------------------
   CheckArgs() - Check arguments
   -------------------------------------------------------------------------- */

static int CheckArgs(const matrix_t *seed, int ngen, const matrix_t **gen,
	int flags)
{
    MTXCONTEXT(CheckArgs);
    int i;

    if (!MATRIX_IS_VALID(seed))
	return -1;
    if (gen == NULL)
	return MTXERROR("Invalid gen=NULL"), -1;
    if (ngen < 0)
	return MTXERROR1("Invalid number of generators (%d)",(int)ngen), -1;
    if (seed->nor < 1) 
	return MTXERROR("Empty seed space"), -1;
    for (i = 0; i < ngen; ++i)
    {
        if (!MATRIX_IS_VALID(gen[i]))
	    return -1;
	if (gen[i]->fl != seed->fl || gen[i]->noc != seed->noc)
	    return MTXERROR1("%E",ERR_INCOMPAT), -1;
	if (gen[i]->noc != gen[i]->nor)
	    return MTXERROR2("gen[%d]: %E",i,ERR_NOTSQUARE), -1;
    }

    if (flags == -1)
	return MTXERROR("Invalid flags"), -1;

    return 0;
}


/* --------------------------------------------------------------------------
   Init() - Initialize for spin-up
   -------------------------------------------------------------------------- */

static int Init(const matrix_t *seed, int ngen, const matrix_t **gen,
	int flags)

{
    MTXCONTEXT(Init);

    zsetfield(seed->fl);
    zsetlen(seed->noc);
    Flags = flags;
    Gen = gen;
    NGen = ngen;
    Dim = seed->noc;
    Seed = seed;
    Span = MatAlloc(seed->fl,Dim+1,Dim);
    if (Span == NULL)
	return MTXERROR("Cannot allocate result buffer"), -1;
    SpanDim = 0;
    Piv = NREALLOC(Piv,long,Dim+2);
    if (Piv == NULL)
	return MTXERROR("Cannot allocate pivot table"), -1;
    SBasisOp = NREALLOC(SBasisOp,long,2 * (Dim+2));
    if (SBasisOp == NULL)
	return MTXERROR("Cannot allocate optable"), -1;
    if (Flags & SF_STD)
    {
	StdSpan = MatAlloc(seed->fl,Dim+1,Dim);
	if (StdSpan == NULL)
	{
	    MatFree(Span);
	    return MTXERROR("Cannot allocate workspace"), -1;
	}
    }
    return 0;
}



/* --------------------------------------------------------------------------
   DoSpinup() - Spin up

   Return:
     0 = Ok, 1 = Not found, -1 = Error
   -------------------------------------------------------------------------- */

static int DoSpinup()

{
    MTXCONTEXT(DoSpinup);
    long n;
    PTR seed;

    switch (Flags & SF_SEED_MASK)
    {
    	case SF_FIRST:
	    /* Try the first seed vector only
	       ------------------------------ */
	    return Spin1(Seed->d,1);

    	case SF_EACH:
	    /* Try each seed vector until successful
	       ------------------------------------- */
	    for (seed = Seed->d, n = 1; n <= Seed->nor; zadvance(&seed,1), ++n)
	    {
	    	if (Spin1(seed,n) == 0)
		    return 0;
	    }
	    return 1;

    	case SF_MAKE:
	    /* Try each 1-dimensional subspace until successfull
	       ------------------------------------------------- */
	    zpseed_init(Seed->nor,Seed->d);
	    for (n = zpseed_next(); n >= 0; n = zpseed_next())
	    {
	    	if (Spin1(zpseed_vec,1) == 0)
		    return 0;
	    }
	    return 1;
   
	default:
	    MTXERROR("Invalid seed mode");
	    return -1;
    }

    MTX_VERIFY(0);
    return -1;
}


/*{{{ --------------------------------------------------------------------------
!function SpinUp "Spin up"
!section reps
!synopsis 
    matrix_t *SpinUp(const matrix_t *seed, int ngen, const matrix_t **gen,
	int flags);
!arg seed
	Matrix with seed vectors.
!arg ngen
	Number of generators in |gen|.
!arg gen
	List of generators.
!arg flags
	Flags, a combination of |SF_xxx| constants (see below).
!return
    Span of the seed vector(s) under the action of the generators, or |NULL|
    on error.
!description
    This function calculates the submodule generated by one or more `seed' 
    vectors under the action of a set of matrices. |gen| must be an array of
    |ngen| (pointers to) square matrices, and seed must be a matrix with the
    same number of columns and any number of rows. Of course, all matrices must
    be over the same field.

    |flags| must be a combination of the following values:
    \begin{list}{}{\leftmargin 35mm\labelwidth 25mm\labelsep 2mm
	\itemsep 0mm\parsep 0mm}
    \item[{\tt SF\_FIRST}\dotfill]
	Only the first row of |seed| is taken as seed vector.
    \item[{\tt SF\_EACH}\dotfill]
	Each row of |seed| is taken as seed vector.
    \item[{\tt SF\_MAKE}\dotfill]
	One vector from each 1-dimensional subspace of the row space of |seed| 
	is taken as seed vector.
    \item[{\tt SF\_SUB}\dotfill]
	Find a submodule: spin up seed vectors one-by-one until a seed 
	vector generates a proper submodule.
    \item[{\tt SF\_CYCLIC}\dotfill]
	Find a cyclic vector: spin up vectors one-by-one until a seed 
	vector generates the whole space.
    \item[{\tt SF\_COMBINE}\dotfill]
	Calculate the submodule generated by the set of all seed vectors.
	This ist typically used with |SF_EACH| to calculate the submodule
	generate by the row space of |seed|.
    \item[{\tt SF\_STD}\dotfill]
	Spin up `canonically'. Note: This increases both computation time and
	memory usage.
    \end{list}
    Note that the seed modes, |SF_FIRST|, |SF_EACH| and |SF_MAKE|, and the
    search modes, |SF_SUB|, |SF_CYCLIC|, |SF_COMBINE|, are mutually exclusive.
    If, in mode |SF_SUB| or |SF_CYCLIC|, no seed vector generates a proper
    submodule or the whole space, respectively, this is {\em not} considered
    an error, and the return value is not |NULL|. The rows of the matrix 
    returned by |SpinUp()| always form a basis of an invariant subspace, 
    but you must examine the number of rows of that matrix to find out if it
    is a proper subspace, or null, or the whole space.
!seealso 
-------------------------------------------------------------------------- }}}*/

matrix_t *SpinUp(const matrix_t *seed, int ngen, const matrix_t **gen,
	int flags)

{
    MTXCONTEXT(SpinUp);

    if (CheckArgs(seed,ngen,gen,flags) != 0)
    {
	MTXERROR("Invalid arguments");
	return NULL;
    }
    if (Init(seed,ngen,gen,flags) != 0)
    {
	MTXERROR("Initialization failed");
	return NULL;
    }
    if (DoSpinup() < 0)
    {
	MatFree(Span);
	if (Flags & SF_STD)
	    MatFree(StdSpan);
	MTXERROR("Spin-up failed");
	return NULL;
    }

    /* Adjust the result size
       ----------------------- */
    if (Flags & SF_STD)
    {
	MatFree(Span);
	Span = StdSpan;
    }
    Span->nor = SpanDim;
    Span->d = (PTR) Srealloc(Span->d,zsize(SpanDim));
    return Span;
}



/*{{{ --------------------------------------------------------------------------
!text reps
The functions descriped in this section operate on matrix representations.
There is no special data type for matrix representations, but all functions
expect that the representation is given as an array of $n$ pointers to square 
matrices representing the generators of the algebra.

\paragraph{Spin-up and standard basis}
%-------------------------------------
\index{spin-up}
\index{standard basis}

The spin-up algorith calculates, for a given `seed' vector $v$ the submodule
generated by $v$, i.e., the smallest subspace which is invariant under the 
generators and contains $v$.

Normally, the basis vectors computed during the spin-up process are chosen
randomly. However, the spin-up algorithm can run in `canonical' or `standard 
basis' mode. In this mode the result is invariant under a change of basis.
More precisely, if $v$ (seed vector) and $g_1,\ldots,g_n$ (generators) produce
the basis $(b_1,\ldots,b_m)$, and $A$ is any nonsingular matrix, then
$vA$ and $A^{-1}g_1A,\ldots,A^{-1}g_nA$ produce the basis $(b_1A,\ldots,b_mA)$.

-------------------------------------------------------------------------- }}}*/

