/* ============================= C MeatAxe ==================================
   File:        $Id: split.c,v 1.8 1998/07/29 16:52:12 mringe Exp $
   Comment:     Split a representation.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <stdlib.h>


static long *split_pivot = NULL;	/* Pivot table */


/*{{{ --------------------------------------------------------------------------
!function Split "Split a module"
!section reps
!synopsis 
    int Split(matrix_t *subspace, int ngen, const matrix_t *gen[], 
	matrix_t **sub, matrix_t **quot)
!arg subspace
    Pointer to an invariant subspace.
!arg ngen 
    Number of generators.
!arg gen
    Generators.
!arg sub
    Buffer for action on the subspace.
!arg quot
    Buffer for action on the quotient.
!return
    The function returns $0$ on success. Any other return value indicates 
    an error.
!description
    Given a matrix representation of an Algebra $A$ and an $A$-invariant 
    subspace, this function calculates matrix representations corrsponding 
    to both subspace and quotient.

    |subspace| is a basis for the invariant subspace. This matrix will be 
    converted to echelon if necessary. |ngen| is the number of generators,
    and |gen| is an array of (pointers) to matrices containing the 
    representation of the generators. The result consists of |ngen| 
    matrices for both subspace and quotient. Pointers to these matrices 
    are stored in |sub| and |quot|, respectively. Both |sub| and |quot| 
    may be |NULL|, indicating that the corresponding representation should
    not be calculated.
  
    \medskip\noindent
    Here is an example for \verb"SpinUp()" and \verb"Split()":
    \begin{verbatim}
    matrix_t *gens[5];
    matrix_t *sub[5], quot[5];
    matrix_t *seed;
    matrix_t *subspace;
    ...
    subspace = SpinUp(seed,5,gens,SF_FIRST|SF_SUB);
    if (subspace->nor > 0 && subspace->nor < subspace->noc)
    {
	printf("Split!\n");
	Split(subspace,5,gens,sub,quot);
    }
    else
	printf("Not split!\n");
    \end{verbatim}
!seealso SpinUp QuotProj
-------------------------------------------------------------------------- }}}*/

int Split(matrix_t *subspace, int ngen, const matrix_t *gen[], matrix_t **sub, 
    matrix_t **quot)

{
    MTXCONTEXT(Split);
    int g;
    long i;
    PTR xi, yi;
    long fl;
    long dim, sdim, qdim;
    PTR mat,tmp;

    /* Check Arguments 
       --------------- */
    MTX_VERIFY(ngen >= 0);
    if (!MATRIX_IS_VALID(subspace))
	return -1;
    for (i = 0; i < ngen; ++i)
    {
	if (!MATRIX_IS_VALID(gen[i]))
	    return -1;
	if (gen[i]->fl != subspace->fl || gen[i]->noc != subspace->noc)
	    return MTXERROR2("gen[%d]: %E",i,ERR_INCOMPAT), -1;
	if (gen[i]->nor != gen[i]->noc)
	    return MTXERROR2("gen[%d]: %E",i,ERR_NOTSQUARE), -1;
    }

    fl = subspace->fl;
    dim = subspace->noc;
    sdim = subspace->nor;
    qdim = dim - sdim;
    mat = subspace->d;
    zsetfield(fl);
    zsetlen(dim);
    tmp = zalloc((long)1);

    /* Make pivot table. Echelonize the subspace if necessary.
       ------------------------------------------------------- */
    split_pivot = NREALLOC(split_pivot,long,dim+1);
    MTX_VERIFY(split_pivot != NULL);
    if (zmkpivot(subspace->d,subspace->nor,split_pivot) != 0)
    {
	sdim = zmkechelon(subspace->d,sdim,split_pivot);
	subspace->nor = sdim;
	qdim = dim - sdim;
    }

    /* Subspace
       -------- */
    for (g = 0; g < ngen; ++g)
    {
	sub[g] = MatAlloc(fl,sdim,sdim);
	xi = mat;
	yi = sub[g]->d;
	for (i = 1; i <= sdim; ++i)
	{
	    FEL f;
	    zsetlen(sdim);
	    zmulrow(yi,F_ZERO);
	    zsetlen(dim);
	    zmaprow(xi,gen[g]->d,dim,tmp);
	    zadvance(&xi,1);
	    zcleanrow2(tmp,mat,sdim,split_pivot,yi);
	    if (zfindpiv(tmp,&f) != 0)
		MtxErrorExit("Split(): Subspace not invariant");
	    zsetlen(sdim);
	    zadvance(&yi,1);
	}
    }

    /* Quotient
       -------- */
    if (quot != NULL)
    {
    	zsetlen(dim);
    	zquotinit(mat,sdim,split_pivot);
    	for (g = 0; g < ngen; ++g)
    	{
    	    quot[g] = MatAlloc(fl,qdim,qdim);
    	    zquotop(gen[g]->d,quot[g]->d);
    	}
    }

    zsetlen(dim);
    free(tmp);
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function QuotProj "Projection on quotient"
!section reps
!synopsis 
    matrix_t *QuotProj(matrix_t *subspace, matrix_t *vectors):
!return
    Projection of |vectors| on the quotient by |subspace|, or |NULL| on error.
!description
    This function calculates the projection of a matrix onto
    the quotient by a subspace. As with \verb"Split()", the subspace need
    not be in echelon form, but it will be converted to echelon form during
    the projection.
    The return value is a pointer to the projection, which is always in
    echelon form.
!seealso SpinUp Split
-------------------------------------------------------------------------- }}}*/

matrix_t *QuotProj(matrix_t *subspace, matrix_t *vectors)

{
    MTXCONTEXT(QuotProj);
    long fl = subspace->fl;
    long dim = subspace->noc;
    long sdim = subspace->nor;
    long qdim = dim - sdim;
    PTR mat = subspace->d;
    matrix_t *result;


    if (!MATRIX_IS_VALID(subspace) || !MATRIX_IS_VALID(vectors))
	return NULL;
    if (subspace->fl != vectors->fl || subspace->noc != vectors->noc)
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Initialize the zquot routines
       ----------------------------- */
    split_pivot = NREALLOC(split_pivot,long,dim+1);
    zsetfield(fl);
    zsetlen(dim);
    if (zmkpivot(subspace->d,subspace->nor,split_pivot) < 0)
    {
	sdim = zmkechelon(subspace->d,subspace->nor,split_pivot);
	subspace->nor = sdim;
	qdim = dim - sdim;
    }
    zquotinit(mat,sdim,split_pivot);
    
    /* Calculate the projection and reduce to echelon form
       --------------------------------------------------- */
    result = MatAlloc(fl,vectors->nor,qdim);
    zsetlen(dim);
    zquot(vectors->d,vectors->nor,result->d);
    zsetlen(qdim);
    result->nor = zmkechelon(result->d,result->nor,split_pivot);
    zsetlen(dim);
    return result;
}




/*{{{ --------------------------------------------------------------------------
!text reps
\paragraph{Splitting a representation}
%-------------------------------------
\index{splitting a representation}

If a proper invariant subspace $U<V$ has been found for a matrix representation 
$M$, the restriction of $M$ to $U$ as well as the representation on $V/U$ can
be calculated. This is called splitting the representation. Note that the
representation on $V/U$ depends on a randomly chosen basis.
-------------------------------------------------------------------------- }}}*/

