/* ============================= C MeatAxe ==================================
   File:        $Id:  $
   Comment:     Program to build a basis reflecting semisimplicity.
   --------------------------------------------------------------------------
   Written by Markus Wiegelmann.
   Revised by Michael Ringe.
   (C) Copyright 1998  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"



/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

static char *helptext[] = {
"SYNTAX",
"    symbas <info> <gens> <basis> [<stdgens>]",
"",
"FILES",
"    <info>.tki                  i  Info file produced by mktree",
"    <gens>.1 ... <gens>.nbgen   i  generators of a semisimple representation",
"    <gens>.cfinfo               i  info-file after running 'chop'",
"    <gens>1a.1 ...              i  generators of each costituent",
"    <basis>                     o  basis reflecting the semisimplicity",
"    <basis>i                    o  inverse of <basis>",
"    <stdgens>.1 ...             o  generators in the new basis",
"",
"DESCRIPTION",
"    This is one of two programs for calculating a block-diagonal basis for a",
"    semisimple representation. It uses peakwords. The other one is 'basbypeak'",
"    and works by calculating peak words. If you are able to run 'mkpeak' use",
"    'basbypeak' since it is faster.", 
NULL};

static proginfo_t pinfo =
   { "symbas", "semisimplicity basis  - 2nd way", "$Revision: 1.1$",
     helptext };

static TK_Info TKI;		    /* Data from .tki file */
static Lat_Info Info;		    /* Data from .cfinfo file */
static char *TkiName = NULL;	    /* .tki file base name */
static char *Name = NULL;	    /* Module base name */
static char *BasisName = NULL;	    /* Basis name */
static char *StdGenName = NULL;	    /* Base name for transposed generators */
matrix_t *Gen[MAXGEN];		    /* Generators */




/* --------------------------------------------------------------------------
   Init() - Initialize everything

   Description:
     This function initializes global data and processes command line
     options and arguments.

   Arguments:
     <argc>: Number of arguments
     <argv>: Argument list
   -------------------------------------------------------------------------- */

static void Init(int argc, char *argv[])
{
    int i;

    MtxInit();

    /* Process command line options
       ---------------------------- */
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("")) != OPT_END)
    {
	switch (i)
	{
	    default:
		MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
	}
    }

    /* Process any remaining arguments
       ------------------------------- */
    if (argc - opt_ind < 3 || argc - opt_ind > 4)
	MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);

    TkiName = argv[opt_ind++];
    Name = argv[opt_ind++];
    BasisName = argv[opt_ind++];
    if (opt_ind < argc)
    	StdGenName = argv[opt_ind++];

}


/* --------------------------------------------------------------------------
   ReadFiles() - Read input files

   Description:
     This function reads the .cfinfo file and the generators.

   Arguments:
     <argc>: Number of arguments
     <argv>: Argument list
   -------------------------------------------------------------------------- */

static void ReadFiles()

{
    int i;

    /* Read the .tki file
       ------------------ */
    if (TK_ReadInfo(&TKI,TkiName) != 0)
	MtxErrorExit("Error reading %s.tki",TkiName);

    /* Read the .cfinfo file
       --------------------- */
    if (Lat_ReadInfo(&Info,Name) != 0)
	MtxErrorExit("Error reading %s.cfinfo",Name);
    if (Info.NGen != TKI.NGen)
    {
	MtxErrorExit("Different number of generators in %s.cfinfo and %s.tki",
	    Name,TkiName);
    }

    /* Read the generators
       ------------------- */
    MESSAGE(1,("Reading generators\n"));
    for (i = 0; i < Info.NGen; ++i)
    {
	char fn[LAT_MAXBASENAME + 20];
        sprintf(fn,"%s.%d",Name,i+1);
        if ((Gen[i] = MatLoad(fn)) == NULL)
            MtxErrorExit("Cannot read %s: %E",fn,MtxErrNo);
    }
}



int main(int argc, char *argv[])

{
    int cf, i;
    matrix_t *bas;
    PTR basptr;
    long baslen;
    FEL coeff[LAT_MAXCF][TK_MAXORDER];
    matrix_t *idems[LAT_MAXCF];

   /* Initializtion
      ------------- */
    Init(argc,argv);
    ReadFiles();
    bas = MatAlloc(Gen[0]->fl,Gen[0]->nor,Gen[0]->noc);
    basptr = bas->d;
    baslen = 0;

    /* For each contituent, calculate the coefficients 
       for the primitive idempotent
       ----------------------------------------------- */
    MESSAGE(1,("Step 1: Calculating coefficients for idempotents\n"));
    for (cf = 0; cf < Info.NCf; ++cf)
    {
	matrix_t *cfgen[MAXGEN];
	Lat_ReadCfGens(&Info,cf,cfgen,LAT_RG_INVERT);
	TK_MakePrimCoeff(&TKI,(const matrix_t **)cfgen,coeff[cf]);
        for (i = 0; i < TKI.NGen; ++i)
            MatFree(cfgen[i]);
    }

    /* Calculate all idempotents. Treating all constituents simultaneously
       enables us to delete intermediate matrices as early as possible. This
       should save a lot of memory when the dimension and/or the condensation
       subgroup order are large.
       ---------------------------------------------------------------------- */
    MESSAGE(1,("Step 2: Calculating idempotents\n"));
    TK_MultiGOperator(&TKI,(const matrix_t **)Gen,Info.NCf,coeff,0,idems);

    /* Calculate the basis
       ------------------- */
    MESSAGE(1,("Step 3: Calculating basis\n"));
    for (cf = 0; cf < Info.NCf; ++cf)
    {
	matrix_t *partbas;
	PTR partbasptr;
        
	/* Calculate the partial basis
	   --------------------------- */
	partbas = SpinUp(idems[cf],TKI.NGen,(const matrix_t **)Gen,
		SF_EACH|SF_COMBINE|SF_STD);
        MatFree(idems[cf]);
	MESSAGE(2,("%s%s: %ld basis vectors\n",Info.BaseName,
	    Lat_CfName(&Info,cf),(long) partbas->nor));

        /* Append the result to the basis
	   ------------------------------ */
        partbasptr = partbas->d;
        zsetlen(bas->noc);
	if (baslen + partbas->nor > bas->nor)
	    MtxErrorExit("Internal error: Basis too long");
	baslen += partbas->nor;
        for (i = 0; i < partbas->nor; ++i)
        {
            zmoverow(basptr,partbasptr);
            zadvance(&basptr,1);
            zadvance(&partbasptr,1);
        }
        MatFree(partbas);
    }
    if (baslen != bas->nor)
        MtxErrorExit("Internal error: Basis too short");

    /* Clean up and write output files
       ------------------------------- */
    MatSave(bas,BasisName);
    if (StdGenName != NULL)
    {
        matrix_t *basi;
	char fn[LAT_MAXBASENAME+20];
        MESSAGE(1,("Transposing generators\n"));
        basi = MatInv(bas);
        sprintf(fn,"%si", BasisName);
        MatSave(basi,fn);
        for (i = 0; i < Info.NGen; ++i)
        {
            matrix_t *stgen = MatDup(bas);
            MatMul(stgen, Gen[i]);
            MatMul(stgen, basi);
            sprintf(fn, "%s.%d", StdGenName, i+1);
            MatSave(stgen, fn);
            MatFree(stgen);
        }
        MatFree(basi);
    }

    MatFree(bas);
    for (i = 0; i < Info.NGen; ++i)
        MatFree(Gen[i]);
    MtxCleanup();
    return 0;	
}
    




