/* ============================= C MeatAxe ==================================
   File:        $Id: tkcoeff.c,v 1.2 1998/06/25 19:02:34 mringe Exp $
   Comment:     Coefficients for idempotents.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"


/* -----------------------------------------------------------------------------
   CalcElements() - Calculate all elements of the condensation group

   Description:
     This function calculates all elements of the condensation group given
     by the generators and the element tree.

   Arguments:
     <info>: Pointer to <TKinfo> structure
     <gen>: Array of (inverse) generators
     <elem>: Group elements are stored here

   Remarks:
     The generators passed in <gen> must be inverted. All multiplications in
     the elemet tree are performed in reverse order, i.e., we calculate the
     inverse elements.
   -------------------------------------------------------------------------- */

static void CalcElems(TK_Info *info, const matrix_t *gen[], matrix_t *elem[])

{
    int i;

    /* Begin the element tree with the unit element
       -------------------------------------------- */
    elem[0] = MatId(gen[0]->fl,gen[0]->nor);

    /* Calculate all elements. Actually we're calculating the inverse
       elements because we are using inverted generators. Consequently,
       we perform all multiplications in reverse order.
       ---------------------------------------------------------------- */
    for (i = 1; i < info->Order; ++i)
    {
	if (info->Gen[i] < 1 || info->Gen[i] > info->NGen)
	    MtxErrorExit("Illegal generator element tree (i=%d)\n",i);
	if (info->Source[i] < 1 || info->Source[i]-1 >= i)
	    MtxErrorExit("Illegal source node in element tree (i=%d)\n",i);
	elem[i] = MatDup(gen[info->Gen[i]-1]);
	MatMul(elem[i],elem[info->Source[i] - 1]);
    }
}





/* -----------------------------------------------------------------------------
   TK_MakePrimCoeff() - Calculate coefficients for the primitive idempotent

   Description:
     This function calculates the coefficients for the primitive idempotent.

   Arguments:
     <info>: Pointer to <TKinfo> structure
     <gen>: Array of inverse generators
     <coeff>: Coefficients are stored here.

   Remarks:
     The matrices passed in <gen> must be the  i n v e r s e  generators!

     <coeff> must be large enough to hold at least <info->Order>
     coefficients. The order will never be greater than TK_MAXORDER.
   -------------------------------------------------------------------------- */

void TK_MakePrimCoeff(TK_Info *info,const matrix_t *gen[], FEL coeff[])

{
    int i;
    matrix_t *elem[TK_MAXORDER];

    /* Calculate all elements
       ---------------------- */
    CalcElems(info,gen,elem);

    /* Calculate the coefficients
       -------------------------- */
    for (i = 0; i < info->Order; ++i)
    {
	coeff[i] = zextract(elem[i]->d,1);
    }

    /* Clean up
       -------- */
    for (i = 0; i < info->Order; ++i)
	MatFree(elem[i]);
}


