/* ============================= C MeatAxe ==================================
   File:        $Id: tkgop.c,v 1.8 1998/04/28 17:37:19 mringe Exp $
   Comment:     Functions for calculationg idempotents in tensor products.
   --------------------------------------------------------------------------
   Originally written by M. Wiegelmann, J. Mueller and C. Jansen.
   Revised by M. Ringe.
   (C) Copyright 1998 Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>



/* --------------------------------------------------------------------------
   TK_GOperator() - Calculate idempotent

   Description:

   Arguments:
     <info>: Pointer to <TK_Info> structure.
     <gen>: Generators
     <coeff>: Array of coefficients, or NULL
     <degree>: Degree used for normalization (may be 0)

   Remarks:
      <coeff>, if not NULL, must be a pointer to an array of <info->Order>
      field elements. These values are used as coefficients in the
      calculation. If <coeff> is NULL, elements are summed up without 
      coefficients.

      If <degree> is different from zero, the result is multiplied by
      <degree>/<info->order>.
   -------------------------------------------------------------------------- */

int TK_MultiGOperator(TK_Info *info, const matrix_t *gen[], 
    int ncf, FEL coeff[][TK_MAXORDER], int degree, matrix_t **result)

{
    MTXCONTEXT(TK_MultiGOperator);
    matrix_t *elms[TK_MAXORDER];
    int edgecount[TK_MAXORDER];
    int cf;
    int i;

    /* For each node of the element tree, count the number
       of downward edges. We use this information below to
       decide when we can safely delete an element.
       --------------------------------------------------- */
    memset(edgecount,0,sizeof(edgecount));
    memset(elms,0,sizeof(elms));
    for (i = 1; i < info->Order; ++i)
	++edgecount[info->Source[i] - 1];

    /* Start with unit matrix
       ---------------------- */
    elms[0] = MatId(gen[0]->fl, gen[0]->nor);
    for (cf = 0; cf < ncf; ++cf)
    {
	result[cf] = MatId(gen[0]->fl, gen[0]->nor);
	if (coeff != NULL)
	    MatMulScalar(result[cf],coeff[cf][0]);
    }


    /* Calculate all remaining elements and sum up the idempotent
       ---------------------------------------------------------- */
    for (i = 1; i < info->Order; ++i)
    {
	int srcno = info->Source[i] - 1;
	int genno = info->Gen[i] - 1;

	/* Calculate the element
	   --------------------- */
	MTX_VERIFY(elms[srcno] != NULL);
	MTX_VERIFY(gen[genno] != NULL);
	elms[i] = MatDup(elms[srcno]);
	MatMul(elms[i],gen[genno]);

	/* Add to the idempotent
	   --------------------- */
	for (cf = 0; cf < ncf; ++cf)
	    MatAddMul(result[cf],elms[i],coeff != NULL ? coeff[cf][i] : F_ONE);

	/* Delete any elements that are no longer needed.
	   ---------------------------------------------- */
	MTX_VERIFY(edgecount[srcno] > 0);
	if (--edgecount[srcno] <= 0)
	{
	    MatFree(elms[srcno]);
	    elms[srcno] = NULL;
	}
	if (edgecount[i] <= 0)
	{
	    MatFree(elms[i]);
	    elms[i] = NULL;
	}
    }

    /* Clean up
       -------- */
    for (i = 0; i < info->Order; ++i)
    {
	if (elms[i] != NULL)	    /* If not yet deleted */
	    MatFree(elms[i]);
    }

    /* Normalization, if requested
       --------------------------- */
    if (degree != 0)
    {
	FEL f1;
	if ((info->Order % zchar) == 0)
	    MtxErrorExit("Division by zero while calculating idempotent");
	f1 = zdiv(zitof(degree % zchar),zitof(info->Order % zchar));
	for (cf = 0; cf < ncf; ++cf)
	    MatMulScalar(result[cf],f1);
    }
    return 0;
}


matrix_t *TK_GOperator(TK_Info *info, const matrix_t *gen[], FEL *coeff, 
    int degree)

{
    matrix_t *result;
    TK_MultiGOperator(info,(const matrix_t **)gen, 1,(FEL (*)[2000])coeff,
    	degree, &result);
    return result;
}


/* --------------------------------------------------------------------------
   TK_TensorGOperator() - Calculate idempotent in MxN

   Description:
     This function calculates (M1xN1 + ... + NrxNr)/r, where r = |H| is the
     condensation subgroup order, M1,...Mr are the group elements in N,
     N1,...Nr are the group elements in N, and 'x' denotes the Kronecker
     product.

   Arguments:
     <info>: Pointer to <TK_Info> structure.
     <gen1>: Generators in M
     <gen2>: Generators in N

   Remarks:
      The generators are supposed to be in standard basis.
      The condensation subgroup order must not be divisible by the
      characteristic.
   -------------------------------------------------------------------------- */

matrix_t *TK_TensorGOperator(TK_Info *info, const matrix_t *gen1[],
    const matrix_t *gen2[])

{
    matrix_t *idem, *elms1[TK_MAXORDER], *elms2[TK_MAXORDER];
    FEL f1;
    long i;

    /* Start with unit matrix
       ---------------------- */
    elms1[0] = MatId(gen1[0]->fl, gen1[0]->nor);
    elms2[0] = MatId(gen2[0]->fl, gen2[0]->nor);
    idem = MatTensor(elms1[0],elms2[0]);

    /* Add the remaining elements
       -------------------------- */
    for (i = 1; i < info->Order; ++i)
    {
	matrix_t *elm;

	/* Calculate the i-th element in both factors
	   ------------------------------------------ */
	elms1[i] = MatDup(elms1[info->Source[i] - 1]);
	MatMul(elms1[i],gen1[info->Gen[i] - 1]);
	elms2[i] = MatDup(elms2[info->Source[i] - 1]);
	MatMul(elms2[i],gen2[info->Gen[i] - 1]);

	/* Calculate the Kronecker product and sum up
	   ------------------------------------------ */
	elm = MatTensor(elms1[i],elms2[i]);
	MatAdd(idem,elm);
	MatFree(elm);
    }

    /* Clean up
       -------- */
    for (i = 0; i < info->Order; ++i)
    {
	MatFree(elms1[i]);
	MatFree(elms2[i]);
    }

    /* Divide by normalization factor
       ------------------------------ */
    if ((info->Order % zchar) == 0)
    {
	MtxErrorExit("Cannot calculate idempotent, Order=%d=0 in characteristic %d",
	    (int)info->Order,(int)zchar);
    }
    f1 = zdiv(F_ONE,zitof(info->Order % zchar));
    MatMulScalar(idem,f1);

    return idem;
}



