/* ============================= C MeatAxe ==================================
   File:        $Id: tkinfo.c,v 1.13 1998/07/29 16:52:13 mringe Exp $
   Comment:     Functions for reading and writing the tensor condense
		information (.tki) file.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


static int ReadVector(StfData *f, const char *c,
    const char *name, int size, int *vec)
{
    MTXCONTEXT(ReadVector);
    int len = size;
    if (!strcmp(c,name))
    {
	StfGetVector(f,&len,vec);
	if (len != size)
	    MTXERROR1("Illegal %s in .tki file",c);
	return 1;
    }
    return 0;
}



/* -----------------------------------------------------------------------------
   ParseTKInfoFile() - Parse the contents of a .tki file
   
   Description:
     This function reads the contents of a .tki file and puts the data into a
	 <TK_Info> structure.
   
   Arguments:
     <f>: Data file. Must be opened for reading.
     <tki>: Pointer to a <TK_Info> structure where data is stored.
   
   Return:
     0 = ok, -1 = Error
   -------------------------------------------------------------------------- */

static int ParseTKInfoFile(StfData *f, TK_Info *tki)

{

    /* Fill the data structure with zeros
       --------------------------------- */
    memset(tki,0,sizeof(TK_Info));

    /* Read header
       ----------- */
    if (StfReadLine(f) || strcmp(StfGetName(f),"TKInfo"))
	MtxErrorExit("File header not found in .tki file");

    /* Read data
       --------- */
    while (StfReadLine(f) == 0)
    {
	const char *c = StfGetName(f);
	if (c == NULL) continue;
        if (!strcmp(c,"TKInfo.NGen"))
	{
	    StfGetInt(f,&tki->NGen);
	    if (tki->NGen < 1 || tki->NGen > TK_MAXGEN)
	    {
		fprintf(stderr,"Illegal number of generators in .tki file\n");
		return -1;
	    }
	}
	else if (!strcmp(c,"TKInfo.Order"))
	{
	    StfGetInt(f,&tki->Order);
	    if (tki->Order < 1 || tki->Order > TK_MAXORDER)
	    {
		fprintf(stderr,"Illegal subgroup order in .tki file\n");
		return -1;
	    }
	}
	else if (!strcmp(c,"TKInfo.Dim"))
	{
	    StfGetInt(f,&tki->Dim);
	    if (tki->Dim < 0 || tki->Dim > 1000000)
	    {
		fprintf(stderr,"Illegal dimension in .tki file\n");
		return -1;
	    }
	}
	else if (!strcmp(c,"TKInfo.Source"))
	{
	    int len = tki->Order;
	    if (StfGetVector(f,&len,tki->Source))
	    {
		fprintf(stderr,"Cannot read 'TKInfo.Source'\n");
		return -1;
	    }
	}
        else if (!strcmp(c,"TKInfo.Gen"))
	{
	    int len = tki->Order;
	    if (StfGetVector(f,&len,tki->Gen))
	    {
		fprintf(stderr,"Cannot read 'TKInfo.Gen'\n");
		return -1;
	    }
	}
	else if (!strcmp(c,"TKInfo.NCf"))
	{
	    StfGetInt(f,&tki->NCf);
	    if (tki->NCf < 1 || tki->NCf > LAT_MAXCF)
	    {
		fprintf(stderr,"Illegal number of constituents in .tki file");
		return -1;
	    }
	}
	else if (ReadVector(f,c,"TKInfo.CfIndexM",tki->NCf,tki->CfIndex[0]))
	    continue;
	else if (ReadVector(f,c,"TKInfo.CfIndexN",tki->NCf,tki->CfIndex[1]))
	    continue;
	else if (ReadVector(f,c,"TKInfo.BasisModeM",tki->NCf,tki->BasisMode[0]))
	    continue;
	else if (ReadVector(f,c,"TKInfo.BasisModeN",tki->NCf,tki->BasisMode[1]))
	    continue;
    }

    /* TODO: check cf mapping */

    return 0;
}



/* -----------------------------------------------------------------------------
   TK_ReadInfo() - Read a .tki file
   
   Description:
     This function reads the contents of a .tki file and puts the data into a
     <TK_Info> structure.
   
   Arguments:
     <tki>: Pointer to a <TK_Info> structure where the data is stored.
     <name>: File name without ".tki" extension

   Remarks:
     The extension ".tki" is appended to the file name.
   
   Return:
     0 = ok, -1 = Error
   -------------------------------------------------------------------------- */

int TK_ReadInfo(TK_Info *tki, const char *name)
{
    char fn[500];
    StfData *f;
    int result = 0;
    memset(tki,0,sizeof(TK_Info));

    /* Open input file
       --------------- */
    strcat(strcpy(fn,name),".tki");
    if ((f = StfOpen(fn,"rt")) == NULL)
    {
	fprintf(stderr,"Cannot open %s for reading, error %d\n",fn,errno);
	return -1;
    }

    /* Read data from file
       ------------------- */
    result = ParseTKInfoFile(f,tki);

    /* Close input file
       ---------------- */
    StfClose(f);
    return result;
}




/* -----------------------------------------------------------------------------
   TK_WriteInfo() - Write a .tki file
   
   Description:
     This function writes the contents of a <TK_Info> structure into a file.
     
   Arguments:
     <tki>: Pointer to a <TK_Info> structure.
     <name>: File name without ".tki" extension

   Remarks:
     The extension ".tki" is appended to the file name.
   
   Return:
     0 = ok, -1 = Error
   -------------------------------------------------------------------------- */

int TK_WriteInfo(TK_Info *tki, const char *name)

{
    char fn[500];
    StfData *f;
    int result = 0;
    strcat(strcpy(fn,name),".tki");
    if ((f = StfOpen(fn,"wt")) == NULL)
    {
	fprintf(stderr,"Cannot open %s for writing, error %d\n",fn,errno);
	return -1;
    }

    StfWriteValue(f,"TKInfo","rec()");
    StfWriteInt(f,"TKInfo.NGen",tki->NGen);
    StfWriteInt(f,"TKInfo.Order",tki->Order);
    StfWriteInt(f,"TKInfo.Dim",tki->Dim);
    StfWriteVector(f,"TKInfo.Source",tki->Order,tki->Source);
    StfWriteVector(f,"TKInfo.Gen",tki->Order,tki->Gen);
    StfWriteInt(f,"TKInfo.NCf",tki->NCf);
    StfWriteVector(f,"TKInfo.CfIndexM",tki->NCf,tki->CfIndex[0]);
    StfWriteVector(f,"TKInfo.CfIndexN",tki->NCf,tki->CfIndex[1]);
    StfWriteVector(f,"TKInfo.BasisModeM",tki->NCf,tki->BasisMode[0]);
    StfWriteVector(f,"TKInfo.BasisModeN",tki->NCf,tki->BasisMode[1]);

    StfClose(f);
    MESSAGE(1,("Wrote %s: order=%d, NCf=%d\n",fn,tki->Order,tki->NCf));
    return result;
}

