/* ============================= C MeatAxe ==================================
   File:        $Id: tkond.c,v 1.9 1998/06/25 19:02:34 mringe Exp $
   Comment:     Condensation of tensor products.
   --------------------------------------------------------------------------
   Written by M. Wiegelmann.
   Revised by Michael Ringe.
   (C) Copyright 1998  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>


/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

static char TkiName[LAT_MAXBASENAME];	    /* Base name for .tki file */
static char ResultName[LAT_MAXBASENAME];    /* Base name for result */
static TK_Info TKInfo;			    /* Data from .tki file */
static Lat_Info InfoM, InfoN;		    /* Data from .cfinfo files */
static char AName[LAT_MAXBASENAME];    	    /* Left factor */
static char BName[LAT_MAXBASENAME];    	    /* Right factor */
static int NGen = 2;			    /* No. of generators for A and B */

static char *helptext[] = {
"SYNTAX",
"    tkond -g <NGen> <A> <B> <M> <N> <Info> <Result>",
"",
"ARGUMENTS",
"    <A>, <B>   Representations of G",
"    <M>, <N>   Restrictions to H < G",
"    <Info>     Base name of .tki file",
"    <Result>   Name for condensed representation",
"",
"FILES",
"    <M>.1, <M>.2, ...    i  Generators of left factor",
"    <N>.1, <N>.2, ...    i  Generators of right factor",
"    <Info>.q.1, ...      i  Basis matrices for constituents",
"    <Info>.p.1, ...      i  Projection matrices for constituents",
"    <Result>.1, ...      o  Condensed matrices",
"",
"DESCRIPTION",
"    The program calculates the condensed representation of the tensorproduct",
"    of <A> with <B>. It strongly relys on the chosen basis for all occuring",
"    matrices and on the information in the <Info>.tki file. Thus you have to",
"    run some precondensation programs before.",
"",
NULL};

static proginfo_t pinfo =
   { "tkond", "Condense tensor product", "$Revision: 1.9 $", helptext };



/* --------------------------------------------------------------------------
   Init() - Initialize everything

   Description:
     This function initializes global variables, processes command line
     options and arguments, and reads the .cfinfo and .tki files.

   Arguments:
     <argc>: Number of arguments
     <argv>: Command line arguments
   -------------------------------------------------------------------------- */

static void Init(int argc, char *argv[])

{
    int k;

    MtxInit();

    /* Process command line options
       ---------------------------- */
    MtxSetCommandLine(argc,argv,&pinfo);
    while ((k = MtxGetOpt("g:")) != OPT_END )
    {
        switch (k)    
        { 
	    case 'g':
		NGen = MtxGetInt();
		if (NGen < 1 || NGen > 1000)
		    MtxErrorExit("-g: %E",ERR_OPTION);
		break;
            default: 
		MtxErrorExit("Unknown option '%c'",k);
        }
    }

    /* Process command line arguments
       ------------------------------ */
    if (opt_ind != argc - 6)
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    strcpy(AName,argv[opt_ind]);
    strcpy(BName,argv[opt_ind+1]);
    strcpy(TkiName,argv[opt_ind+4]);
    strcpy(ResultName,argv[opt_ind+5]);

    /* Read info files
       --------------- */
    if (TK_ReadInfo(&TKInfo,TkiName) != 0)
	MtxErrorExit("Error reading %s.tki",TkiName);
    if (Lat_ReadInfo(&InfoM,argv[opt_ind + 2]) != 0)
	MtxErrorExit("Error reading %s.cfinfo",argv[opt_ind + 2]);
    if (Lat_ReadInfo(&InfoN,argv[opt_ind + 3]) != 0)
	MtxErrorExit("Error reading %s.cfinfo",argv[opt_ind + 3]);

    /* Some checks on info file data
       ----------------------------- */
    if (TKInfo.Dim <= 0)
	MtxErrorExit("No dimension found in .tki file - did you run precond?");
    if (InfoM.Field != InfoN.Field)
	MtxErrorExit("Different fields in .cfinfo files");
    if (InfoN.NGen != InfoM.NGen)
    {
	MtxErrorExit("Different number of generators in %s and %s",InfoM.BaseName,
	    InfoN.BaseName);
    }
}



/* --------------------------------------------------------------------------
   FirstRow() - Get first row in Q matrix

   Description:
     This function calculates the index of the first row in the Q matrix 
     which belongs to a given constituent.

   Arguments:
     <info>: Pointer to constituent information
     <cf>: Constituent index
     <k>: Additional index to distiguish isomorphic copies.

   Remarks:
     Both <cf> and <k> are 0-based. The return value is to be used as a row
     index and thus 1-based.

   Return:
     Index of first row belonging to the constituent.
   -------------------------------------------------------------------------- */

static int FirstRow(Lat_Info *info, int cf, int k)

{
    MTXCONTEXT(FirstRow);
    int ind = 0;
    int i;

    MTX_VERIFY(cf >= 0 && cf < info->NCf);
    MTX_VERIFY(k >= 0 && k < info->Cf[cf].mult);

    /* Constituents before <cf> consume dim*mult rows
       ---------------------------------------------- */
    for (i = 0; i < cf; ++i)
	ind += info->Cf[i].dim * info->Cf[i].mult;

    /* Copies of constituent <cf> before <k> consume dim rows
       ------------------------------------------------------ */
    ind += info->Cf[cf].dim * k;

    /* Return 1-based row index
       ------------------------ */
    return ind + 1;
}




/*----------------------------------------------------------------------------*/

void gemap(matrix_t *conma, matrix_t *q, matrix_t *mrow, matrix_t *nrow)

  /* the vektors from "q" are mapped under g*e */

{
    int j;
    long bcol = 1;


    /* For each irreducible constituent I
       ---------------------------------- */
    for (j = 0; j < TKInfo.NCf; ++j)
    {
	char pname[LAT_MAXBASENAME + 20];	/* P matrix file name */
	matrix_t *p;				/* P matrix */
	int cfm = TKInfo.CfIndex[0][j];		/* Index of constituent in M */
	int cfn = TKInfo.CfIndex[1][j];		/* Index of constituent in M */
	int d = InfoM.Cf[cfm].dim;		/* Dimension */
	int mj;
	
	/* Load the P matrix
	   ----------------- */
        sprintf(pname,"%s.p.%d",TkiName,j + 1);
        p = MatLoad(pname);

	/* For each copy of I in M
	   ----------------------- */
        for (mj = 0; mj < InfoM.Cf[cfm].mult; ++mj)
        {
	    long mstart = FirstRow(&InfoM,cfm,mj);
	    int nj;

	    /* For each copy of I in M
	       ----------------------- */
            for (nj = 0; nj < InfoN.Cf[cfn].mult; nj++)
            {   
		long nstart = FirstRow(&InfoN,cfn,nj);
                matrix_t *image = MatDup(q);
                matrix_t *mop = MatCut(mrow,1,mstart,-1,d);
                matrix_t *nop = MatCut(nrow,1,nstart,-1,d);
                matrix_t *op = MatTensor(mop,nop);
                MatFree(mop);
                MatFree(nop);
                MatMul(image, op);  /* operation */
                MatMul(image, p);   /* projection */
                MatFree(op);
		MatCopyRegion(conma,1,bcol,image,1,1,-1,-1);
                bcol += p->noc;
                MatFree(image);
            }
        }
        MatFree(p);
    }
}





/* --------------------------------------------------------------------------
   CondenseMat() - Condense one generator

   Description:

   Arguments:
     <gen>: Index of generator (starting with 0)
   -------------------------------------------------------------------------- */

static void CondenseMat(int gen)

{
    char resname[LAT_MAXBASENAME + 20];	/* Output file name */
    char mname[LAT_MAXBASENAME + 20];	/* Generator on M */
    char nname[LAT_MAXBASENAME + 20];	/* Generator on M */
    FILE *fptr;				/* Output file */
    int cf;				/* Constituent index */
    matrix_t *mmat, *nmat;		/* The generator on M, N */


    /* Make file names 
       --------------- */
    sprintf(resname,"%s.%d",ResultName,gen+1);
    sprintf(mname,"%s.%d",AName,gen+1);
    sprintf(nname,"%s.%d",BName,gen+1);
    MESSAGE(0,("Condensing %s x %s --> %s\n",mname,nname,resname));

    /* Load the generator on M and N
       ----------------------------- */
    mmat = MatLoad(mname);
    nmat = MatLoad(nname);

    /* Open the output file
       -------------------- */
    fptr = zwritehdr(resname,zfl,TKInfo.Dim,TKInfo.Dim);

    /* Main loop: for each constituent
       ------------------------------- */
    for (cf = 0; cf < TKInfo.NCf; ++cf)
    {
	int cfm = TKInfo.CfIndex[0][cf];    /* Index in M */
	int cfn = TKInfo.CfIndex[1][cf];    /* Index in N */
	matrix_t *q;			    /* Basis matrix */
	char qname[LAT_MAXBASENAME+20];
	int rownb = InfoM.Cf[cfm].dim;	/* Number of rows to extract */
	int mi;				/* Counter for copies of this const. */

        MESSAGE(1,("  Processing %s",Lat_CfName(&InfoM,cfm)));
        MESSAGE(1,(" x %s\n",Lat_CfName(&InfoN,cfn)));

	/* Load Q matrix
	   ------------- */
        sprintf(qname,"%s.q.%d",TkiName,cf+1);
        q = MatLoad(qname);

        for (mi = 0; mi < InfoM.Cf[cfm].mult; ++mi)
        {
	    matrix_t *mrow;
	    int ni;
	    int firstrow = FirstRow(&InfoM,cfm,mi);
	    mrow = MatCut(mmat,firstrow,1,rownb,-1);

            for (ni = 0; ni < InfoN.Cf[cfn].mult; ++ni)
            {
		matrix_t *nrow, *condmat;

		firstrow = FirstRow(&InfoN,cfn,ni);
		nrow = MatCut(nmat,firstrow,1,rownb,-1);
		condmat = MatAlloc(zfl,q->nor,TKInfo.Dim);
		if (condmat == NULL)
		{
		    MtxErrorExit("Cannot allocate %dx%d matrix",(int)q->nor,
			(int)TKInfo.Dim);
		}

	    	MESSAGE(2,(" %dx%d",mi,ni));
                gemap(condmat,q,mrow,nrow);
                
                /* write result */
                zsetlen(condmat->noc);
        	zwritevec(fptr,condmat->d,condmat->nor);
                MatFree(condmat);
                MatFree(nrow);
            }
	    MESSAGE(2,("\n"));
            MatFree(mrow);
        }   
        MatFree(q);
    }   

    /* Clean up
       -------- */
    fclose(fptr);
    MatFree(mmat);
    MatFree(nmat);
}





/* --------------------------------------------------------------------------
   main() - Program entry point
   -------------------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    int i;
    Init(argc,argv);
    for (i = 0; i < NGen; ++i)
	CondenseMat(i);
    MtxCleanup();
    return 0;
}  


