/* ============================= C MeatAxe ==================================
   File:        $Id: words.c,v 1.15 1998/09/17 17:03:56 mringe Exp $
   Comment:     The word generator.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */



#include "meataxe.h"
#include <string.h>
#include <stdlib.h>


/* NOTE: There are currently two versions of the word generator.
   The version can be selected at run-time, so each of the WgXXXX() functions
   is defined three times: The old version (suffix _OLD), the new versinm
   (suffix _NEW), and a wrapper function which calls either the new or the
   old version depending on a global flag.
*/
          


/* ------------- OLD WORD GENERATOR STARTS HERE ------------ */

/*{{{ --------------------------------------------------------------------------
!function WgInit "Initialize the word generator"
!section wgen
!synopsis 
    wgdata_t *WgInit(int ngen, const matrix_t *gen[]);
!arg ngen
    Number of generators.
!arg gen
    Pointer to array of generators.
!return
    Pointer to a new word generator data structure, or |NULL| on error.
!description
    This function initializes the word generator for a given matrix
    algebra. |ngen| is the number of generators, and |gen| is an array of 
    pointers to the generators. The generators must be square matrices of 
    the same dimension over the same field.

    On success, |WgInit()| returns a pointer to an internal data
    structure which can be used in subsequent calls to |WgMakeWord()|
    and |WgFree()|. If an error occurs, the return value is NULL.

    Note that the word generator does not make internal copies of the
    generators. The caller must assure that the generators are not 
    deleted or modified in any way as long as the word generator is
    used.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

wgdata_t *WgInit_OLD(int ngen, const matrix_t *gen[])

{
    MTXCONTEXT(WgInit);
    int i, k;
    long fl, dim;
    wgdata_t *b;

    /* Check the generators
       -------------------- */
    if (ngen < 1) 
    {
	MTXERROR1("Invalid number of generators (%d)",ngen);
	return NULL;
    }

    for (i = 0; i < ngen; ++i)
    {
	if (!MATRIX_IS_VALID(gen[i]))
	    return NULL;
    	if (gen[i]->fl != gen[0]->fl || gen[i]->nor != gen[0]->nor)
	{
	    MTXERROR1("%E",ERR_INCOMPAT);
	    return NULL;
	}
    	if (gen[i]->nor != gen[i]->noc)
	{
	    MTXERROR1("%E",ERR_NOTSQUARE);
	    return NULL;
	}
    }
    fl = gen[0]->fl;
    dim = gen[0]->nor;

    /* Create a new data structure
       ---------------------------  */
    if ((b = ALLOC(wgdata_t)) == NULL
	|| (b->Gen = (const matrix_t **)NALLOC(matrix_t*,ngen + 10)) == NULL
       )
    {
	MTXERROR("Cannot allocate word generator data: %S");
	return NULL;
    }

    b->max = ngen + 10;
    b->len = ngen;
    b->NGen = ngen;
    for (k = 0; k < ngen; ++k)
    	b->Gen[k] = (matrix_t *) gen[k];	/* BUG: casting away const! */
    return (void *) b;
}


/*{{{ --------------------------------------------------------------------------
!function WgFree "Terminate the word generator"
!section wgen
!synopsis 
    void WgFree(wgdata_t *b);
!arg n
    Pointer to word generator data.
!description
    This function terminates the word generator specified by |b| and cleans up 
    internal data structures. Note that the generators that were passed to 
    |WgInit()| are {\em not} freed.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

void WgFree_OLD(wgdata_t *b)

{
    MTXCONTEXT(WgFree);
    int k;

    /* Check the handle
       ---------------- */
    if (b == NULL)
    {
	MTXERROR1("%E",ERR_BADARG);
	return;
    }

    /* Free all matrices except the generators
       --------------------------------------- */
    for (k = b->NGen; k < b->len; ++k)
    	MatFree((matrix_t *)b->Gen[k]);	/* HACK: casting away const */
    free((void *)b->Gen);
    free(b);
}


/* ------------------------------------------------------------------
   mkbasis() - Calculate basis elements up to number <max>
   ------------------------------------------------------------------ */

#define MAXCP 30
static int cptab[MAXCP] =
      {1 ,1 ,2,2,3,3,2,1,1, 1,5,1,3,7,3,3,3,3,3,3,3,2,4,3,3,1,4,4,3,1};
static int mul[MAXCP] =
      {-1,-1,1,7,6,3,1,5,1, 0,2,5,1,4,7,3,5,1,4,7,0,6,3,9,1,4,3,2,7,1};
 
static int cp(int i)

{
    return i < MAXCP ? i - cptab[i] : (i * 87) % i;
}

static int mu(int i, int ngen)

{
    return (i < MAXCP ? mul[i] : (i * 23)) % ngen;
}


static int mkbasis(wgdata_t *b, int max)

{
    MTXCONTEXT(mkbasis);
    size_t i;

    if (max >= (int) b->max)
    {
    	matrix_t **x = NREALLOC((void*)b->Gen,matrix_t *,max+5);
    	if (x == NULL) 
	{
	    MTXERROR("Cannot grow word generator data: %S");
	    return -1;
	}
	b->Gen = (const matrix_t **)x;
	b->max = max + 5;
    }
    for (i = b->len; i <= (size_t) max; ++i)
	b->Gen[i] = MatMul(MatDup(b->Gen[cp(i)]),b->Gen[mu(i,b->NGen)]);
    if (max >= (int) b->len)
    	b->len = max + 1;
    return 0;
}


/* ------------------------------------------------------------------
   mapnumber() - Map word number (to get some special words first)
   ------------------------------------------------------------------ */

static long fpnum[MAXFP+1] = {-1,7,39,217,219,223,249};

static long mapnumber(long n)

{
    int i;
    if (n <= MAXFP && n >= 1)
	n = fpnum[n];
    else
	for (i = 1; i <= MAXFP; ++i)
	    if (n == fpnum[i]) n = i;
    return n;
}


/*{{{ --------------------------------------------------------------------------
!function WgMakeWord "Make a word"
!section wgen
!synopsis 
    matrix_t *WgMakeWord(wgdata_t *b, long n);
!arg b
    Pointer to word generator data.
!arg n
    Word number.
!return
    Matrix representation of word |n|, or |NULL| on error.
!description
    This function calculates a `random' element in a matrix algebra.
    The algebra, i.e., a set of generating matrices, must be specified with
    |WgInit()|, and |b| is the pointer returned by |WgInit()|.
    |n| is the number of the element to calculate. Refer to section
    \ref{GdSection:wgen} for more details.

    Here is an example demonstrating the usage of the word generator:
    \begin{verbatim}
	matrix_t *mat[3];
	wgdata_t *wg;
	matrix_t *word;
	long nul;
	...
	wg = WgInit(3,mat);
	word = WgMakeWord(wg,1833);
	nul = MatNullity__(word);
	printf("Word 1833 has nullity %ld\n",nul);
	WgFree(wg);
    \end{verbatim}
    This example initializes the word generator with three matrices,
    calculates the word number 1833 and prints out its nullity.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

matrix_t *WgMakeWord_OLD(wgdata_t *b, long n)

{
    MTXCONTEXT(WgMakeWord);
    long l;
    int i;
    matrix_t *w;

    /* Initialize
       ---------- */
    if (n <= 0)
    {
	MTXERROR1("Invalid word number %d",(int)n);
	return NULL;
    }
    n = mapnumber(n);
    w = NULL;

    /* Calculate the word
       ------------------ */
    for (i = 0, l = n; l != 0; ++i, l >>= 1)
    {
	if ((l % 2) == 0) continue;
	if (i >= (int) b->len) mkbasis(b,i);
	if (w != NULL)
	    MatAdd(w,b->Gen[i]);
	else
	    if ((w = MatDup(b->Gen[i])) == NULL) return NULL;
    }
    return w;
}


/*{{{ --------------------------------------------------------------------------
!function WgMakeFingerPrint "Calculate finger print"
!section wgen
!synopsis 
    void WgMakeFingerPrint(wgdata_t *b, long fp[]);
!arg b
    Pointer to word generator data.
!arg fp
    Buffer for the finger print (6 numbers).
!description
    This function calculates the `finger print' of a module, i.e. the nullities
    of the first 6 words. |fp| must be a pointer to an array of at least 6
    long integers.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

#if (MAXFP != 6)
	************* FEHLER *****************
#endif

void WgMakeFingerPrint(wgdata_t *b, long fp[])

{	
    int i;

    for (i = 1; i <= MAXFP; ++i)
    	fp[i-1] = MatNullity__(WgMakeWord(b,i));
}



/* ------------------------------------------------------------------
   basisname() - Symbolic representation of basis elements.
   ------------------------------------------------------------------ */

static const char *basisname(int i, int ngen)

{
    static char buf[50];
    char b[2];

    if (i < ngen)
    {
    	buf[0] = (char) (i + 'a');
    	buf[1] = 0;
    	return buf;
    }

    basisname(cp(i),ngen);
    b[0] = (char) (mu(i,ngen) + 'a');
    b[1] = 0;
    strcat(buf,b);
    return buf;
}




/*{{{ --------------------------------------------------------------------------
!function WgSymbolicName "Symbolic name of a word"
!section wgen
!synopsis 
    const char *WgSymbolicName(const wgdata_t *b, long n);
!arg b
    Pointer to word generator data.
!arg n
    Word number.
!return
    Symbolic name of the word |n|.
!description
    This function returns a symbolic representation of the word number |n| as
    a polynomial in the generators. The generators are named a, b, c\ldots

    The return value is a pointer to a static buffer which is overwritten on 
    each call.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

const char *WgSymbolicName_OLD(const wgdata_t *b, long n)

{
    MTXCONTEXT(WgSymbolicName);
    static char name[200];
    char *c = name;
    int i;
    long digit;

    /* Check the arguments
       ------------------- */
    if (n <= 0)
    {
    	MTXERROR1("Invalid word number %d",(int)n);
    	return NULL;
    }

    /* Translate the word number
       ------------------------- */
    n = mapnumber(n);

    /* Build the symbolic name
       ----------------------- */
    for (i = 0; n != 0; ++i)
    {	
	digit = n % 2;	/* Get next digit */
	n /= 2;
	if (digit == 0) continue;
	if (c != name) *c++ = '+';
	strcpy(c,basisname(i,b->NGen));
	while (*c != 0) ++c;
    }
    return name;
}


/* ------------- NEW WORD GENERATOR STARTS HERE ------------ */


#define MINLEN	5
#define MAXLEN	8

static int B0Tab[8][MAXLEN+1] = {
    {0,-1,-1,-1,-1}, {1,-1,-1,-1,-1}, {2,3,-1,-1,-1}, {5,4,-1,-1,-1},
    {7,9,-1,-1,-1}, {6,11,13,-1,-1}, {17,8,1,-1,-1}, {19,10,21,23,-1}
};

static int B1Tab[8][MAXLEN+1] = {
    {0,1,2,-1,-1}, {4,3,5,6,-1},{8,10,7,-1,-1},{12,9,14,11,-1},
    {13,16,15,18,-1}, {17,20,22,19,-1},{24,21,23,25,26,-1},{27,29,28,31,33,-1}
};

static int BitTab[238] = 
{
    0x07,0x27,0xD9,0xDB,0xDF,0xF9,0xE0,0x03,0x05,0x06,0x09,0x0A,0x0B,0x0C,
    0x0D,0x0E,0x0F,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,
    0x1C,0x1D,0x1E,0x1F,0x21,0x22,0x23,0x24,0x25,0x26,0x28,0x29,0x2A,0x2B,
    0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,
    0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,
    0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,0x50,0x51,0x52,0x53,0x54,0x55,0x56,
    0x57,0x58,0x59,0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,0x60,0x61,0x62,0x63,0x64,
    0x65,0x66,0x67,0x68,0x69,0x6A,0x6B,0x6C,0x6D,0x6E,0x6F,0x70,0x71,0x72,
    0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x7B,0x7C,0x7D,0x7E,0x81,0x82,
    0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,
    0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,
    0x9F,0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,
    0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,
    0xBB,0xBC,0xBD,0xBE,0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,
    0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,
    0xD8,0xDA,0xDC,0xDD,0xDE,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,
    0xEA,0xEB,0xEC,0xED,0xEE,0xF0,0xF1,0xF2,0xF3,0xF4,0xF6,0xF8,0xFA,0xFC 
};


#define RND(x) ((x) * 214013L + 2531011L)
#define Mod ((1 << (MAXLEN+1)) - 2)


static int CalcLen(int n2)
{
    if (n2 < 200)
	return 5;
    if (n2 < 2000)
	return 6;
    if (n2 < 20000)
	return 7;
    return MAXLEN;
}


static void MakeBufX2(int n2, int buf[8][MAXLEN+1])
{
    unsigned long r = n2;
    int i;
    int m[8];
    int len, mod;

    len = CalcLen(n2);
    mod = (1 << (len+1)) - 2;

    for (i = 0; i < 8; )
    {
	int mm, k;
	r = RND(r);
	mm = (r >> 16) % mod + 2;
	for (k = 0; k < i && m[k] != mm; ++k);
	if (k >= i)
	    m[i++] = mm;
    }
    for (i = 0; i < 8; ++i)
    {
	unsigned long x = m[i];
	int k = len;
	int mask = 1 << len;
	while ((x & mask) == 0)
	{
	    buf[i][k] = -1;
	    mask >>= 1;
	    --k;
	};
        buf[i][k] = -1;
	mask >>= 1;
	--k;
	while (k >= 0)
	{
	    int gg = (int)((r >> 16) * 2);
	    r = RND(r);
	    buf[i][k] = ((x & mask) == 0) ? gg : gg + 1;
	    mask >>= 1;
	    --k;
	}
    }
}


static void MakeBuf(int n2, int ngen, int buf[8][MAXLEN+1])

{
    int pos;

    if (n2 == 0)
	memcpy(buf,B0Tab,sizeof(B0Tab));
    else if (n2 == 1)
	memcpy(buf,B1Tab,sizeof(B0Tab));
    else
	MakeBufX2(n2,buf);
    for (pos = 0; pos < 8; ++pos)
    {
	int *x;
	for (x = buf[pos]; *x >= 0; ++x)
	    *x %= ngen;
    }
}


static const int *B(int n2, int pos, int ngen)

{
    static int buf[8][MAXLEN+1];
    static int lastn2 = -1;
 
    if (n2 != lastn2)
    {
	MakeBuf(n2,ngen,buf);
	lastn2 = n2;
    }
    return buf[pos];
}


static char *BasisName(int n2, int pos, int ngen)
{
    static char buf[MAXLEN+1];
    char *c = buf;
    const int *x = B(n2,pos,ngen);
    while (*x >= 0)
    {
	int i;
	*c++ = (char)(*x++ + 'a');
	for (i = 1; i < 9 && *x == x[-1]; ++i, ++x);
	if (i > 1)
	    *c++ = (char)(i + '0');
    }
    *c = 0;
    return buf;
}

const char *WgSymbolicName_NEW(const wgdata_t *b, long n)

{
    int i;
    static char name[8 * (MAXLEN+1)+1];
    char *c = name;
    MTXCONTEXT(WgSymbolicName);
    long n1, n2;

    MTX_VERIFY(n > 0);
    --n;
    n1 = BitTab[(int)(n % 238)];
    n2 = (int)(n / 238);

    for (i = 0; i < 8 && n1 != 0; ++i, n1 >>= 1)
    {
	const char *bn;
	if ((n1 % 2) == 0)
	    continue;
	bn = BasisName(n2,i,b->NGen);
	if (c != name)
	    *c++ = '+';
	strcpy(c,bn);
	while (*c != 0) ++c;
    }
    return name;
}




static void GenBasis(wgdata_t *b, int n2, int pos)
{
    const int *x;
    matrix_t *buf = NULL;
    MTXCONTEXT(GenBasis);

    if (b->Basis[pos] != NULL)
	MatFree(b->Basis[pos]);
    for (x = B(n2,pos,b->NGen); *x >= 0; ++x)
    {
	MTX_ASSERT(*x >= 0 && *x < b->NGen);
	if (buf == NULL)
	    buf = MatDup(b->Gen[*x]);
	else
	    MatMul(buf,b->Gen[*x]);
    }
    MTX_VERIFY(buf != NULL);
    b->Basis[pos] = buf;
    b->N2[pos] = n2;
}



matrix_t *WgMakeWord_NEW(wgdata_t *b, long n)

{
    matrix_t *w = NULL;
    int n1, n2;
    int i;
    MTXCONTEXT(WgMakeWord);

    MTX_VERIFY(n > 0);
    --n;
    n1 = BitTab[(int)(n % 238)];
    n2 = (int)(n / 238);
    for (i = 0; i < 8 && n1 != 0; ++i, n1 >>= 1)
    {
	if ((n1 % 2) == 0)
	    continue;
	if (b->N2[i] != n2)
	    GenBasis(b,n2,i);
	if (w == NULL)
	    w = MatDup(b->Basis[i]);
	else
	    MatAdd(w,b->Basis[i]);
    }
    return w;
}



wgdata_t *WgInit_NEW(int ngen, const matrix_t *gen[])

{
    MTXCONTEXT(WgInit);
    int i, k;
    long fl, dim;
    wgdata_t *b;

    /* Check the generators
       -------------------- */
    if (ngen < 1) 
    {
	MTXERROR1("Invalid number of generators (%d)",ngen);
	return NULL;
    }

    for (i = 0; i < ngen; ++i)
    {
	if (!MATRIX_IS_VALID(gen[i]))
	    return NULL;
    	if (gen[i]->fl != gen[0]->fl || gen[i]->nor != gen[0]->nor)
	{
	    MTXERROR1("%E",ERR_INCOMPAT);
	    return NULL;
	}
    	if (gen[i]->nor != gen[i]->noc)
	{
	    MTXERROR1("%E",ERR_NOTSQUARE);
	    return NULL;
	}
    }
    fl = gen[0]->fl;
    dim = gen[0]->nor;

    /* Create a new data structure
       ---------------------------  */
    if ((b = ALLOC(wgdata_t)) == NULL)
    {
	MTXERROR("Cannot allocate word generator data: %S");
	return NULL;
    }
    if ((b->Gen = (const matrix_t **)NALLOC(matrix_t*,ngen)) == NULL)
    {
	free(b);
	MTXERROR("Cannot allocate word generator data: %S");
	return NULL;
    }

    b->NGen = ngen;
    for (k = 0; k < ngen; ++k)
    	b->Gen[k] = gen[k];
    for (k = 0; k < 8; ++k)
    {
	b->Basis[k] = NULL;
	b->N2[k] = -1;
    }

    return b;
}

void WgFree_NEW(wgdata_t *b)

{
    MTXCONTEXT(WgFree);
    int k;

    /* Check the handle
       ---------------- */
    if (b == NULL)
    {
	MTXERROR1("%E",ERR_BADARG);
	return;
    }

    /* Free all basis matrices 
       ----------------------- */
    for (k = 0; k < 8; ++k)
    {
	if (b->Basis[k] != NULL)
    	    MatFree(b->Basis[k]);
    }
    free((void*)b->Gen);    /* HACK: casting away const */
    memset(b,0,sizeof(wgdata_t));
    free(b);
}


/* Wrapper functions which select the old or new version at run-time
*/

extern int UseNewWordGenerator;	
    /* Defined in init.c. Since this is a hack we don't put it into */
    /* meataxe.h. */

void WgFree(wgdata_t *b)
{
    if (UseNewWordGenerator)
	WgFree_NEW(b);
    else
	WgFree_OLD(b);
}

wgdata_t *WgInit(int ngen, const matrix_t *gen[])
{
    if (UseNewWordGenerator)
	return WgInit_NEW(ngen,gen);
    else
	return WgInit_OLD(ngen,gen);
}


matrix_t *WgMakeWord(wgdata_t *b, long n)
{
    if (UseNewWordGenerator)
	return WgMakeWord_NEW(b,n);
    else
	return WgMakeWord_OLD(b,n);
}

const char *WgSymbolicName(const wgdata_t *b, long n)

{
    if (UseNewWordGenerator)
	return WgSymbolicName_NEW(b,n);
    else
	return WgSymbolicName_OLD(b,n);
}
