/* ============================= C MeatAxe ==================================
   File:        $Id: ycomp.c,v 1.6 1998/09/17 17:03:57 mringe Exp $
   Comment:     Compare vector spaces
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <stdlib.h>
#include "meataxe.h"


static long *piv = NULL;
static long maxpiv;


/*{{{ --------------------------------------------------------------------------
!function IsSubspace "Vector space incidence relation"
!section reps
!synopsis 
    int IsSubspace(const matrix_t *u, const matrix_t *v, long ngen);
!arg u
    Generating set for the first space, $U$.
!arg v
    Pointer to a basis for the second space, $V$.
!arg ngen
    Number of vectors in |sub| to consider.
!return
    The function returns $1$ if $U\leq V$, $0$ if $U\not\leq V$, and $-1$ 
    on error.
!description
    This function compares two vector spaces $U,V\in\GF{q}^n$ and returns $1$,
    if $U\leq V$.
    The spaces to compare are given as matrices where the rows of |u| 
    (linearly) generate $U$ and the rows of |v| generate $V$.
    Thus, |u| and |v| must be matrices over the same field and with the same
    number of columns. Also, |v| must be in full echelon form while 
    there is no further restriction on |u|.

    In normal mode, |ngen| is 0. Then, the rows of |u| are checked one-by-one
    to see if they are in the span of the rows of |v|. If this test passes
    for each row of |u|, the return value is 1, otherwise it is 0. 
    There is no special return value for the cases $U=V$ or $U>V$.

    If |ngen| is different from zero, the function assumes that $U$ is 
    generated by the first |ngen| rows, and only this number of rows are
    checked. For example, $U$ may be a cyclic submodule, 
    which is generated from a single vector under the action of an algebra. 
    Then, one only needs to check if the generating vector, which 
    is supposed to be the first row of |u|, is contained in $V$.
!seealso 
-------------------------------------------------------------------------- }}}*/

int IsSubspace(const matrix_t *sub, const matrix_t *space, long ngen)
{
    MTXCONTEXT(IsSubspace);
    long nvec, spcdim, i;
    PTR tmp, y;

    /* Check the arguments
       ------------------- */
    if (!MATRIX_IS_VALID(sub) || !MATRIX_IS_VALID(space))
	return -1;
    if (sub->fl != space->fl || sub->noc != space->noc)
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return -1;
    }

    /* Decide how many vectors from <sub> we shall check
       ------------------------------------------------- */
    zsetfield(space->fl);
    zsetlen(space->noc);
    spcdim = space->nor;
    nvec = sub->nor;
    if (ngen > 0 && ngen < nvec)
	nvec = ngen;

    /* Build the pivot table
       --------------------- */
    if (piv == NULL || maxpiv < spcdim)
    {
	piv = NREALLOC(piv,long,spcdim + 1);
	if (piv == NULL)
	{
	    MTXERROR("Cannot allocate pivot table");
	    return -1;
	}
	maxpiv = spcdim;
    }
    if (zmkpivot(space->d,spcdim,piv) != 0) 
    {
	MTXERROR1("Cannot build pivot table: %E",ERR_NOTECH);
	return -1;
    }

    /* Allocate workspace 
       ------------------ */
    tmp = zalloc(1);
    if (tmp == NULL)
    {	
	MTXERROR("Cannot allocate workspace");
	return -1;
    }

    /* Check if vectors from <sub> are in the span of <space>
       ------------------------------------------------------ */
    for (i = nvec, y = sub->d; i > 0; --i, zadvance(&y,1))
    {
	FEL f;
	zmoverow(tmp,y);			/* Get next vector */
	zcleanrow(tmp,space->d,spcdim,piv);	/* Clean with basis */
        if (zfindpiv(tmp,&f) > 0) 
	    break;				/* Break the loop, return 0 */
    }

    /* Clean up and return the result
       ------------------------------ */
    free(tmp);
    return (i <= 0);
}



#if 0

      ============ OLD STUFF - TO BE REMOVED SOON ===============
/* ------------------------------------------------------------------
   spccomp() - Compare two vector spaces. If n == 0, m1 and m2 are
   full vector spaces. The spaces must be in echelon form.
   If n > 0, the first n rows of each matrix are considered to
   generate the whole space.

   Return value:
          -1  m1 < m2
	   0  m1 = m2
	   1  m1 > m2
	   9  else
	  -9  error
   ------------------------------------------------------------------ */

int spccomp(const matrix_t *m1, const matrix_t *m2, long n)

{
    MTXCONTEXT(spccomp);
    PTR tmp, y;
    long i;
    FEL f;
    long dim1 = m1->nor, dim2 = m2->nor;  /* Dimensions */
    long ngen1 = dim1, ngen2 = dim2;	  /* Number of generators */

    /* Check the arguments
       ------------------- */
    if (n < 0) 
	return MTXERROR("Invalid argment"), -9;
    if (m1->noc != m2->noc || m1->fl != m2->fl)
	return MTXERROR1("%E",ERR_INCOMPAT), -9;
    zsetfield(m1->fl);
    zsetlen(m1->noc);

    /* Set ngen1 and ngen2
       ------------------- */
    if (n > 0)
    {
	if (n < dim1) ngen1 = n;
	if (n < dim2) ngen2 = n;
    }

    /* Check if m1 <= m2. We don't try this, of
       course, if dim(m1) > dim(m2).
       ---------------------------------------- */
    if (dim1 <= dim2)
    {
	if (piv == NULL || maxpiv < dim2)
	{
	    if (piv != NULL) free(piv);
	    piv = NALLOC(long,dim2+1);
	    maxpiv = dim2;
	}
	if (zmkpivot(m2->d,dim2,piv) != 0) 
	    return MTXERROR1("Space 2: %E",ERR_NOTECH), -9;
    	tmp = zalloc((long)1);
	for (i = ngen1, y = m1->d; i > 0; --i, zadvance(&y,1))
	{
	    zmoverow(tmp,y);
	    zcleanrow(tmp,m2->d,dim2,piv);
	    if (zfindpiv(tmp,&f) > 0) break;
	}
	free(tmp);
	if (i == 0)
	{
	    if (dim1 == dim2) return 0;	/* m1 = m2 */
	    return -1;			/* m1 <= m2 */
	}
    }

    /* Now, check if m1 > m2.
       ---------------------- */
    if (dim1 > dim2)
    {
	if (piv == NULL || maxpiv < dim1)
	{
	    if (piv != NULL) free(piv);
	    piv = NALLOC(long,dim1+1);
	    maxpiv = dim1;
	}
	if (zmkpivot(m1->d,dim1,piv) != 0) 
	    return MTXERROR1("Space 1: %E",ERR_NOTECH), -9;
    	tmp = zalloc((long)1);
	for (i = ngen2, y = m2->d; i > 0; --i, zadvance(&y,1))
	{
	    zmoverow(tmp,y);
	    zcleanrow(tmp,m1->d,dim1,piv);
	    if (zfindpiv(tmp,&f) > 0) break;
	}
	free(tmp);
	if (i == 0)
	    return 1;			/* m1 > m2 */
    }

    /* Neither m1<=m2 nor m1>=m2.
       -------------------------- */
    return 9;	/* */
}


/* ------------------------------------------------------------------
   spcequal() - Check if m1==m2. m1 must be in echelon form, m2
	may be any matrix.
   Return value:  0  m1!=m2
		  1  m1=m2
		  -1 error
   ------------------------------------------------------------------ */

int spcequal(const matrix_t *m1, const matrix_t *m2)

{
    MTXCONTEXT(spcequal);
    PTR tmp, y;
    long dim, i;
    FEL f;

    /* Check the arguments
       ------------------- */
    if (m1->noc != m2->noc || m1->fl != m2->fl)
	return MTXERROR1("%E",ERR_INCOMPAT), -9;
    zsetfield(m1->fl);
    zsetlen(m1->noc);

    /* Compare the dimensions
       ---------------------- */
    if (m1->nor != m2->nor)
	return 0;

    /* Make the pivot table for m1
       --------------------------- */
    dim = m1->nor;
    if (piv == NULL || maxpiv < dim)
    {
	if (piv != NULL) free(piv);
	piv = NALLOC(long,dim+1);
	maxpiv = dim;
    }
    if (zmkpivot(m1->d,dim,piv) != 0)
	return MTXERROR1("Space 1: %E",ERR_NOTECH), -9;

    /* Clean the rows of m2 with m1
       ---------------------------- */
    tmp = zalloc((long)1);
    for (i = dim, y = m2->d; i > 0; --i, zadvance(&y,1))
    {
	zmoverow(tmp,y);
	zcleanrow(tmp,m1->d,dim,piv);
	if (zfindpiv(tmp,&f) > 0) break;
    }
    free(tmp);
    if (i == 0)
	return 1;			/* m1 = m2 */
    else
	return 0;			/* m1 != m2 */
}


/* ------------------------------------------------------------------
   spccontains() - Check if m1>=m2. m1 must be in echelon form, m2
	may be any matrix.
   Return value:  0  no
		  1  yes
		  -1 error
   ------------------------------------------------------------------ */

int spccontains(const matrix_t *m1, const matrix_t *m2)

{
    MTXCONTEXT(spccontains);
    PTR tmp, y;
    long dim, i;
    FEL f;

    /* Check the arguments
       ------------------- */
    if (m1->noc != m2->noc || m1->fl != m2->fl)
	return MTXERROR1("%E",ERR_INCOMPAT), -9;
    zsetfield(m1->fl);
    zsetlen(m1->noc);

    /* Compare the dimensions
       ---------------------- */
    if (m1->nor < m2->nor)
	return 0;

    /* Make the pivot table for m1
       --------------------------- */
    dim = m1->nor;
    if (piv == NULL || maxpiv < dim)
    {
	if (piv != NULL) free(piv);
	piv = NALLOC(long,dim+1);
	maxpiv = dim;
    }
    if (zmkpivot(m1->d,dim,piv) != 0)
	return MTXERROR1("m1: %E",ERR_NOTECH), -1;

    /* Clean the rows of m2 with m1
       ---------------------------- */
    tmp = zalloc((long)1);
    for (i = m2->nor, y = m2->d; i > 0; --i, zadvance(&y,1))
    {
	zmoverow(tmp,y);
	zcleanrow(tmp,m1->d,dim,piv);
	if (zfindpiv(tmp,&f) > 0) break;
    }
    free(tmp);
    if (i == 0)
	return 1;			/* m1 >= m2 */
    else
	return 0;			/* 'no' */
}

#endif

