/* ============================= C MeatAxe ==================================
   File:        $Id: ymatrix.c,v 1.10 1998/06/25 19:02:34 mringe Exp $
   Comment:     Matrix operations
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */



#include "meataxe.h"
#include <string.h>
#include <stdlib.h>


/*{{{ --------------------------------------------------------------------------
!structure matrix_t "Matrix data structure"
!section mat
!synopsis 
    typedef struct
    {
	long fl, nor, noc;
	PTR d;
    } matrix_t;
!description
The |matrix_t| structure represents a matrix ofer a finite field. |fl| is 
the field order, |nor| contains the number of rows, and |noc| the number 
of columns. |d| is a pointer to the matrix elements, organized as an array 
of rows of length |noc|.

Both |nor| and |noc| may be zero. In this case, |d| ist still a valid 
pointer, but the memory block it points to may have size zero.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!text mat
\index{matrix_t@{\tt matrix\protect\_t} data type}
In the {\MeatAxe}, matrices over finite fields are represented by |matrix_t|
structures. Besides the marks, each matrix carries the field order, the 
number of rows and the number of columns. There is no global field order
or row length as at the kernel layer.

Matrices can be created in many ways, including (i) calling |MatAlloc()|, 
(ii) making a copy of an existing matix with |MatDup()|, and (iii) by reading 
a matrix from a data file with |MatRead()| or |MatLoad()|.
The application is responsible for releasing matrices which are no longer
needed. Matrices may consume lage amounts of memory, so it always a good
idea to release a matrix as early as possible. There is only one possibility 
of releasing a matrix: calling |MatFree()|.
-------------------------------------------------------------------------- }}}*/



/*{{{ --------------------------------------------------------------------------
!function MatPrint "Print a matrix on stdout"
!section mat.io
!synopsis 
    void MatPrint(char *name, const matrix_t *m);
!arg name
    Name to print before the matrix, or |NULL|.
!arg m
    Pointer to the matrix.
!description
    This function prints a matrix on the standard output in a human-readable
    format. If |name| is not |NULL| the name followed by an equal sign is 
    printed before the matrix.
!seealso
-------------------------------------------------------------------------- }}}*/

void MatPrint(char *name, const matrix_t *m)

{
    MTXCONTEXT(MatPrint);
    PTR x;
    long i, k;

    if (!MATRIX_IS_VALID(m)) 
	return;

    zsetfield(m->fl);
    zsetlen(m->noc);
    x = m->d;
    if (name != NULL) printf("%s=\n",name);
    for (i = 1; i <= m->nor; ++i)
    {
	for (k = 1; k <= m->noc; ++k)
	    printf("%d",zextract(x,k));
	printf("\n");
	zadvance(&x,(long)1);
    }
}


/*{{{ --------------------------------------------------------------------------
!function MatAlloc "Create a new matrix"
!section mat.basic
!synopsis 
    matrix_t *MatAlloc(long fl, long nor, long noc);
!arg fl
    Field order.
!arg nor
    Number of rows.
!arg noc
    Number of columns.
!return
    Pointer to the new matrix or |NULL| on error.
!description
    This function creates a new matrix with \verb"nor" rows and \verb"noc" 
    columns over GF(\verb"fl"). The return value is a pointer to the 
    allocated |matrix_t| structure, or |NULL| if there is not enough memory. 

    To destroy a matrix, use \verb"MatFree()", {\em not} \verb"free()".
!seealso @MatAllocFree
-------------------------------------------------------------------------- }}}*/

matrix_t *MatAlloc(long fl, long nor, long noc)

{
    MTXCONTEXT(MatAlloc);
    matrix_t *m;

    if (fl == 0) 
	fl = zfl;	/* Default value (UNDOCUMENTED, NOT RECOMMENDED) */
    zsetfield(fl);
    zsetlen(noc);
    m = ALLOC(matrix_t);
    if (m == NULL)
    {
	MTXERROR("Cannot allocate matrix_t structure");
	return NULL;
    }
    m->id = T_MATRIX;
    m->d = zalloc(nor);
    if (m->d == NULL)
    {
	free(m);
	MTXERROR("Cannot allocate matrix data");
	return NULL;
    }
    m->fl = fl;
    m->nor = nor;
    m->noc = noc;
    return m;
}



/*{{{ --------------------------------------------------------------------------
!function MatFree "Delete a matrix"
!section mat.basic
!synopsis 
    void MatFree(matrix_t *m);
!arg m
    Pointer to the matrix.
!return
    $0$ on success, $-1$ on error.
!description
    This function frees a matrix which has beed created by |MatAlloc()|. This
    implies freeing the internal data buffer as well as the |matrix_t| structure
    itself.
!seealso @MatAllocFree
-------------------------------------------------------------------------- }}}*/

int MatFree(matrix_t *m)

{
    MTXCONTEXT(MatFree);
    if (!MATRIX_IS_VALID(m)) return -1;

    if (m->d != NULL)
	free(m->d);
    free(m);
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function MatId "Identity matrix"
!section mat.basic
!synopsis 
    matrix_t *MatId(long fl, long nor);
!arg fl
    Field order.
!arg nor
    Number of rows.
!return
    Pointer to the matrix, or |NULL| on error.
!description
    This function creates an identity matrix with |nor| nows over GF(|fl|).
!seealso @MatAllocFree
-------------------------------------------------------------------------- }}}*/

matrix_t *MatId(long fl, long nor)

{
    matrix_t *m;
    PTR x;
    long i;

    m = MatAlloc(fl,nor,nor);
    if (m == NULL) return NULL;

    /* Note: We assume that MatAlloc()
       fills the matrix with 0s */
    for (i = 1, x = m->d; i <= m->nor; ++i,zadvance(&x,1))
	zinsert(x,i,F_ONE);
    return m;
}




/*{{{ --------------------------------------------------------------------------
!function MatDup "Duplicate a matrix"
!section mat.basic
!synopsis 
    matrix_t *MatDup(const matrix_t *src);
!arg src
    Pointer to the matrix.
!return
    Pointer to a copy of |mat|, or |NULL| on error.
!description
    This function creates a copy of an existing matrix.
!seealso @MatAllocFree
-------------------------------------------------------------------------- }}}*/

matrix_t *MatDup(const matrix_t *src)

{
    MTXCONTEXT(madup);
    matrix_t *m;

    if (!MATRIX_IS_VALID(src)) 
	return NULL;
    m = MatAlloc(src->fl,src->nor,src->noc);
    if (m == NULL) 
    {
	MTXERROR("Cannot allocate matrix");
	return NULL;
    }
    memcpy(m->d,src->d,zsize(src->nor));
    return m;
}


/*{{{ --------------------------------------------------------------------------
!function MatRead "Read a matrix from a file"
!section mat.io
!synopsis 
    matrix_t *MatRead(FILE *f);
!arg f
    File to read from.
!return
    Pointer to the matrix, or |NULL| on error.
!description
    This function opens reads a matrix from a file.
!seealso MatLoad
-------------------------------------------------------------------------- }}}*/

matrix_t *MatRead(FILE *f)

{
    MTXCONTEXT(MatRead);
    matrix_t *m;
    long hdr[3];

    if (zreadlong(f,hdr,3) != 3) 
    {
	MTXERROR("Cannot read header");
	return NULL;
    }
    if (hdr[0] < 2) 
    {
	MTXERROR1("%E",ERR_NOTMATRIX);
	return NULL;
    }
    m = MatAlloc(hdr[0],hdr[1],hdr[2]);
    if (m == NULL) 
	return NULL;
    if (zreadvec(f,m->d,(size_t)m->nor) != (size_t)m->nor)
    {
	MatFree(m);
	return NULL;
    }
    return m;
}


/*{{{ --------------------------------------------------------------------------
!function MatWrite "Write a matrix to a file"
!section mat.io
!synopsis 
    int MatWrite(FILE *f, const matrix_t *mat);
!arg f
    Pointer to the file.
!arg mat
    Pointer to the matrix.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function writes a matrix to a file.
!seealso MatSave
-------------------------------------------------------------------------- }}}*/

int MatWrite(FILE *f, const matrix_t *mat)

{
    MTXCONTEXT(MatWrite);
    long hdr[3];

    if (!MATRIX_IS_VALID(mat)) 
	return -1;
    hdr[0] = mat->fl;
    hdr[1] = mat->nor;
    hdr[2] = mat->noc;
    if (zwritelong(f,hdr,3) != 3) 
    {
	MTXERROR("Cannot write header");
	return -1;
    }
    zsetfield(mat->fl);
    zsetlen(mat->noc);
    if (zwritevec(f,mat->d,(size_t)mat->nor) != (size_t)mat->nor)
	return -1;
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function MatLoad "Read a matrix from a file"
!section mat.io
!synopsis 
    matrix_t *MatLoad(const char *fn);
!arg fn
    File name.
!return
    Pointer to the matrix, or |NULL| on error.
!description
    This function opens a file, reads a single matrix, and closes the file.

    To read more than one matrix from a file, use |MatRead()|.
!seealso MatRead
-------------------------------------------------------------------------- }}}*/

matrix_t *MatLoad(const char *fn)

{
    FILE *f;
    matrix_t *m;

    if ((f = SFOpen(fn,FM_READ)) == NULL)
	return NULL;
    m = MatRead(f);
    fclose(f);
    return m;
}


/*{{{ --------------------------------------------------------------------------
!function MatSave "Write a matrix to a file"
!section mat.io
!synopsis 
    int MatSave(const matrix_t *mat, const char *fn);
!arg mat
    Pointer to the matrix.
!arg fn
    File name.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function opens a file, writes a matrix to the file, and closes the
    file. If a file with the specified name already exists, the old contents
    of the file are destroyd.

    To write more than one matrix to a file, use |MatWrite()|.
!seealso MatWrite
-------------------------------------------------------------------------- }}}*/

int MatSave(const matrix_t *mat, const char *fn)

{
    MTXCONTEXT(MatSave);
    FILE *f;
    int i;

    if (!MATRIX_IS_VALID(mat)) 
	return -1;
    if ((f = SFOpen(fn,FM_CREATE)) == NULL)
	MtxErrorExit("Cannot open %s: %S",fn);
    i = MatWrite(f,mat);
    fclose(f);
    if (i != 0)
	MTXERROR1("Cannot write matrix to %s",fn);
    return i;
}




/*{{{ --------------------------------------------------------------------------
!function MatMul "Multiply matrices"
!section mat.op
!synopsis 
    matrix_t *MatMul(matrix_t *dest, const matrix_t *src);
!arg dest
    Left factor.
!arg src
    Right factor.
!return
    The function returns |dest|, or |NULL| on error.
!description
    This function multiplies the matrix |dest| from the right by the matrix 
    |src|.
    The matrices must be compatible for multiplication, i.e. they must be over
    the same field, and the number of columns of |dest| must be equal to the 
    number f rows of |src|.
    The result of the multiplication is stored in |dest|, overwriting the
    original contents.
!seealso MatPower
-------------------------------------------------------------------------- }}}*/

matrix_t *MatMul(matrix_t *dest, const matrix_t *src)

{
    MTXCONTEXT(MatMul);
    PTR x, tmp, result;
    long i;

    /* Check the arguments
       ------------------- */
    if (!MATRIX_IS_VALID(src) || !MATRIX_IS_VALID(dest))
	return NULL;
    if (src->fl != dest->fl || src->nor != dest->noc)
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Matrix multiplication
       --------------------- */
    zsetfield(src->fl);
    zsetlen(src->noc);
    result = tmp = zalloc(dest->nor);
    if (result == NULL)
    {
	MTXERROR("Cannot allocate product");
	return NULL;
    }
    x = dest->d;
    for (i = dest->nor; i != 0; --i)
    {
	zsetlen(src->noc);
	zmaprow(x,src->d,src->nor,tmp);
	zadvance(&tmp,(long)1);
	zsetlen(dest->noc);
	zadvance(&x,(long)1);
    }
    free(dest->d);
    dest->d = result;
    dest->noc = src->noc;
    return dest;
}



/*{{{ --------------------------------------------------------------------------
!function MatTr "Transpose a matrix"
!section mat.op
!synopsis 
    matrix_t *MatTr(const matrix_t *mat);
!arg mst
    Pointer to the matrix.
!return
    Pointer to the transposed matrix or |NULL| on error.
!description
    This function transposes a matrix. A new matrix is allocated for the 
    result. The original matrix remains unchanged.
!seealso 
-------------------------------------------------------------------------- }}}*/

matrix_t *MatTr(const matrix_t *src)

{
    MTXCONTEXT(MatTr);
    PTR s, d;
    long i;
    matrix_t *dest;

    if (!MATRIX_IS_VALID(src))
	return NULL;
    dest = MatAlloc(src->fl,src->noc,src->nor);
    if (dest == NULL) return NULL;
    d = dest->d;
    s = src->d;

    /* Note: We assume that MatAlloc() fills the matrix with 0s */
    for (i = 1; i <= src->noc; ++i)
    {
	zsetlen(src->noc);
	zextractcol(s,src->nor,i,d);
	zsetlen(dest->noc);
	zadvance(&d,(long)1);
    }
    return dest;
}


/*{{{ --------------------------------------------------------------------------
!function MatInv "Invert a matrix"
!section mat.op
!synopsis 
    matrix_t *MatInv(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Inverse matrix or |NULL| on error.
!description
    This function calculates the inverse of a matrix. |mat| must be a 
    non-singular square matrix. The inverse matrix is returned in a newly
    allocated |matrix_t| structure, and the original matrix remains 
    unchanged.
!seealso  zmatinv
-------------------------------------------------------------------------- }}}*/

matrix_t *MatInv(const matrix_t *src)

{
    MTXCONTEXT(MatInv);
    PTR tmp = NULL;	/* Workspace */
    matrix_t *dest;

    if (!MATRIX_IS_VALID(src))
	return NULL;
    if (src->nor != src->noc)
    {
	MTXERROR1("%E",ERR_NOTSQUARE);
	return NULL;
    }
    dest = MatId(src->fl,src->nor);
    if (dest == NULL) 
	return NULL;

    /* Copy matrix into workspace
       -------------------------- */
    tmp = zalloc(src->nor);
    memcpy(tmp,src->d,zsize(src->nor));

    /* Inversion
       --------- */
    if (zmatinv(tmp,dest->d) != 0) 
    {
	MatFree(dest);
	return NULL;
    }
    return dest;
}


/*{{{ --------------------------------------------------------------------------
!function MatIsEchelon "Check if a matrix is in echelon form"
!section mat.op
!synopsis 
    int MatIsEchelon(const matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    1 if the matrix is in echelon form, 0 if not, -1 on error.
!description
    This function checks if a given matrix is in (full) echelon form. A matrix
    is in full echelon form, if each row contains a 1 as first non-zero element,
    and all the other rows have a 0 at this position.
!seealso @ech
-------------------------------------------------------------------------- }}}*/

int MatIsEchelon(const matrix_t *mat)

{
    PTR x, y;
    long i, k, piv;
    FEL f;

    zsetfield(mat->fl);
    zsetlen(mat->noc);
    x = mat->d;
    for (i = 1; i <= mat->nor; ++i)
    {	piv = zfindpiv(x,&f);
    	if (f != F_ONE) 
	    return 0;
    	y = mat->d;
    	for (k = 1; k <= mat->nor; ++k, zadvance(&y,(long)1))
    	{
	    if (k == i) 
		continue;
    	    if (zextract(y,piv) != F_ZERO) 
		return 0;
    	}
    	zadvance(&x,(long)1);
    }
    return 1;
}




/* ------------------------------------------------------------------
   matpower() - Calculate the <n>-th power of <mat> using the
	binary method.
   ------------------------------------------------------------------ */

static void matpwr_(long n, PTR inp, PTR out, PTR tmp2)

{
    PTR x, y;
    long i;
    int first = 1;

    while (n > 0)
    {	if (n % 2 == 1)
	{
	    if (first)
	    {
		memcpy(out,inp,zsize(znoc));
		first = 0;
	    }
	    else
	    {
		x = out;
		for (i = 1; i <= znoc; ++i)
		{
		    zmaprow(x,inp,znoc,tmp2);
		    zmoverow(x,tmp2);
		    zadvance(&x,(long)1);
		}
	    }
	}
	x = inp;
	y = tmp2;
	for (i = 1; i <= znoc; ++i)
	{
	    zmaprow(x,inp,znoc,y);
	    zadvance(&x,(long)1);
	    zadvance(&y,(long)1);
	}
	memcpy(inp,tmp2,zsize(znoc));
	n /= 2;
    }
}



/*{{{ --------------------------------------------------------------------------
!function MatPower "Power of a matrix"
!section mat.op
!synopsis 
    matrix_t *MatPower(const matrix_t *mat, long n);
!arg mat
    Pointer to the matrix.
!arg n
    Exponent.
!return
    |n|-th power of |mat|, or |NULL| on error.
!description
    This function calculates the $n$-th power of a matrix, using the binary
    method. This method is generally faster than multiplying the matrix $n$
    times by itself. On the other hand, a third matrix is temporarily created 
    in addition to the original matrix and the result matrix.
    The cases $n=0$ and $n=1$ are handled separately, avoiding unnecessary
    memory allocation and calculation. 

    Negative exponents are not allowed. To calculate a negative power, you 
    must first invert the matrix with |MatInv()| and then call |MatPower()|
    with the inverted matrix and a positive exponent.
!seealso MatMul MatInv
-------------------------------------------------------------------------- }}}*/

matrix_t *MatPower(const matrix_t *mat, long n)

{
    MTXCONTEXT(MatPower);
    matrix_t *result;
    PTR tmp, tmp2;

    /* Check the arguments
       ------------------- */
    if (!MATRIX_IS_VALID(mat))
	return NULL;
    if (mat->nor != mat->noc)
    {
	MTXERROR1("%E",ERR_NOTSQUARE);
	return NULL;
    }

    /* Handle special cases n = 0 and n = 1
       ------------------------------------ */
    if (n == 0) 
	return MatId(mat->fl,mat->nor);
    else if (n == 1)
	return MatDup(mat);

    zsetfield(mat->fl);
    zsetlen(mat->noc);
    tmp = zalloc(znoc);
    memcpy(tmp,mat->d,zsize(znoc));
    tmp2 = zalloc(znoc);
    result = MatAlloc(mat->fl,mat->nor,mat->noc);
    if (result == NULL)
    {
	free(tmp);
	free(tmp2);
	return NULL;
    }
    matpwr_(n,tmp,result->d,tmp2);
    free(tmp);
    free(tmp2);
    return result;
}


