/* ============================= C MeatAxe ==================================
   File:        $Id: yperm.c,v 1.10 1998/07/29 16:52:17 mringe Exp $
   Comment:     Operations with permutations
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <stdlib.h>
#include <string.h>
#include "meataxe.h"



/*{{{ --------------------------------------------------------------------------
!structure perm_t "Permutation data structure"
!section perm
!synopsis 
    typedef struct
    {
	long deg;
	long *d;
    } perm_t;
!description
    The |perm_t| structure represents a permutation. |deg| is the degree 
    and |d| is a pointer to the permutation proper, which is stored internally 
    as an array of long integers. To be more precise, |d[0]| contains the image 
    of $1$ under the permutation, and |d[n-1]| contains the image of $n$.
    This is the same sructure which is used by the kernel-level permutation
    functions.
-------------------------------------------------------------------------- }}}*/

/*{{{ --------------------------------------------------------------------------
!text perm
\index{permutations}
Permutations operate on the set $\{1,2,\ldots,n\}$, where $n$ is the
degree. A permutation is represented by the |perm_t| data structure.
This structure also carries the degree of the permutation. This is 
important because you cannot, for example, multiply permutations of 
different degrees.
-------------------------------------------------------------------------- }}}*/




/*{{{ --------------------------------------------------------------------------
!function PermPrint "Print a permutation"
!section perm.basic
!synopsis 
    int PermPrint(const char *name, const perm_t *perm);
!arg name
    Name to print before the permutation or |NULL| (see below).
!arg perm
    Pointer to the permutation.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function prints a permutation on the standard output using
    cycle notation. If |name| is not |NULL| the name followed by an
    equal sign is printed before the permutation. For example, the
    statement \verb:PermPrint("P",P);: may produce the following output:
    \begin{verbatim}
	P=(1 9)(2 3 6)(4 5 7)
    \end{verbatim}
    Fixed points are always suppressed in the output.
!seealso
-------------------------------------------------------------------------- }}}*/

#define SIZE(i) ((i)<9?2 : (i)<99?3 : (i)<999?4 : (i)<9999?5 : \
	(i)<99999?6 : (i)>999999?7 : (i)<9999999?8 : 9)
    
int PermPrint(const char *name, const perm_t *perm)

{
    MTXCONTEXT(PermPrint);
    long *p;
    int cycle = 1;
    int empty = 1;
    int count = 0, k;
    int i;

    /* Check arguments
       --------------- */
    if (!PERMUTATION_IS_VALID(perm))
	return -1;
  
    /* Print the name
       -------------- */
    if (name != NULL) 
	printf("%s=",name);

    /* Print the permutation
       --------------------- */
    p = perm->d - 1;
    while (1)
    {
	/* Find next cycle
	   --------------- */
	while (cycle <= perm->deg && p[cycle] < 0) ++cycle;
	if (cycle > perm->deg) break;	/* Done */
	i = cycle;

	     /* Check if it is a fixed point (we don't
	        print fixed points, GAP doesn't like them)
	        ----------------------------------------- */
	    if (p[i] == i)
	    {
		p[i] = -p[i];	/* Mark it as done */
		continue;
	    }

	    /* Print cycle
	       ----------- */
	    if ((count += SIZE(i)) > 77)
	    {
		printf("\n    (%ld",(long)i);
	        count = 5 + SIZE(i);
	    }
	    else
		printf("(%ld",(long)i);
	    empty = 0;

	    while (1)
	    {
		k = i;
		i = p[i];
		p[k] = -p[k];
		if (p[i] < 0) break;

	        if ((count += SIZE(i)) > 77)
	        {
		    printf(",\n     %ld",(long)i);
	            count = 4 + SIZE(i);
	        }
	        else
		    printf(",%ld",(long)i);
	    }
	    printf(")");
	    ++count;
    }
    if (empty) printf("()");
    if (name != NULL) printf("\n");
    for (i = 1; i <= perm->deg; ++i) p[i] = -p[i];
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function PermAlloc "Allocate a permutation"
!section perm.basic
!synopsis 
    perm_t *PermAlloc(long deg);
!arg deg
    Degree.
!return
    Pointer to the new permutation, or |NULL| on error.
!description
    This function creates a permutation of the degree |deg|. It returns 
    a pointer to the new permutation |NULL| on error. The new permutation
    is initialized to the identity.
!seealso @permall
-------------------------------------------------------------------------- }}}*/

perm_t *PermAlloc(long deg)

{
    MTXCONTEXT(PermAlloc);
    perm_t *p;
    int i;

    if (deg <= 0)
    {
	MTXERROR1("Invalid degree %ld",deg);
	return NULL;
    }
    p = ALLOC(perm_t);
    if (p == NULL)
    {
	MTXERROR("Cannot allocate perm_t structure");
	return  NULL;
    }
    p->id = T_PERM;
    p->deg = deg;
    p->d = NALLOC(long,deg);
    if (p == NULL)
    {
	free(p);
	MTXERROR("Cannot allocate permutation data");
	return NULL;
    }
    for (i = 0; i < deg; ++i)
	p->d[i] = i + 1;
    return p;
}



/*{{{ --------------------------------------------------------------------------
!function PermFree "Free a permutation"
!section perm.basic
!synopsis 
    int PermFree(perm_t *p);
!arg p
    Pointer to the permutation.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function deletes a permutation and returns the memory to the 
    system. Do not use |free()| on permutations because this would only 
    free the |perm_t| structure but not the data buffer.
!seealso @permall
-------------------------------------------------------------------------- }}}*/

int PermFree(perm_t *p)

{
    MTXCONTEXT(PermFree);

    /* Check arguments
       --------------- */
    if (!PERMUTATION_IS_VALID(p))
	return -1;
  
    free(p->d);
    free(p);
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function PermDup "Duplicate a permutation"
!section perm.basic
!synopsis 
    perm_t *PermDup(const perm_t *src);
!arg src
    Pointer to the permutation.
!return
    Pointer to a copy of |src| or |NULL| on error.
!description
    This function creates a copy of an existing permutation. 
!seealso
-------------------------------------------------------------------------- }}}*/

perm_t *PermDup(const perm_t *src)

{
    MTXCONTEXT(PermDup);
    perm_t *p;

    if (!PERMUTATION_IS_VALID(src))
	return NULL;
    p = PermAlloc(src->deg);
    if (p == NULL) 
	return NULL;
    memcpy(p->d,src->d,(size_t) src->deg * sizeof(long));
    return p;
}





/*{{{ --------------------------------------------------------------------------
!function PermRead "Read a permutation"
!section perm.io
!synopsis 
    perm_t *PermRead(FILE *f);
!arg f
    File to read from.
!return
    Pointer to the permutation read from |f|, or |NULL| on error.
!description
    This function reads a permutation from a file. |f| must be a pointer to an
    open file with read permission. 
    
    If a permutation was successfully read, the function returns a pointer to 
    a newly created |perm_t| object. The caller is responsible for deleting 
    this object as soon as the permutation is no longer needed.
!seealso PermLoad
-------------------------------------------------------------------------- }}}*/
 
perm_t *PermRead(FILE *f)

{
    MTXCONTEXT(PermRead);
    perm_t *p;
    long hdr[3];

    if (zreadlong(f,hdr,3) != 3) 
    {
	MTXERROR("Cannot read header");
	return NULL;
    }
    if (hdr[0] != -1) 
    {
	MTXERROR1("%E", ERR_NOTPERM);
	return NULL;
    }
    p = PermAlloc(hdr[1]);
    if (p == NULL) 
	return NULL;
    if (zreadlong(f,(long *)p->d,hdr[1]) != (size_t) hdr[1])
    {
	PermFree(p);
	MTXERROR("Cannot read permutation data");
	return NULL;
    }
    return p;
}



/*{{{ --------------------------------------------------------------------------
!function PermWrite "Write a permutation"
!section perm.io
!synopsis 
    int PermWrite(FILE *f, const perm_t *perm);
!arg f
    File to write to.
!arg perm
    Permutation to write.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function writes a permutation to a file. The file |f| must be open 
    for writing.
!seealso PermSave
-------------------------------------------------------------------------- }}}*/

int PermWrite(FILE *f, const perm_t *perm)

{
    MTXCONTEXT(PermWrite);
    long hdr[3];

    if (!PERMUTATION_IS_VALID(perm))
	return -1;
    hdr[0] = -1;
    hdr[1] = perm->deg;
    hdr[2] = 1;
    if (zwritelong(f,hdr,3) != 3) 
    {
	MTXERROR("Cannot write header");
	return -1;
    }
    if (zwritelong(f,(long *)perm->d,hdr[1]) != (size_t) hdr[1])
    {
	MTXERROR("Cannot write data");
	return -1;
    }
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!function PermLoad "Read a permutation"
!section perm.io
!synopsis 
    perm_t *PermLoad(char *fn);
!arg fn
    File name.
!return
    Pointer to the permutation read from the file, or |NULL| on error.
!description
    This function opens a file, reads a single permutation, and closes the 
    file. The return value is a pointer to the permutation or \verb"NULL" on 
    error. If the file contains more than one permutation, only the first one 
    is read.

    If a permutation was successfully read, the function returns a pointer to 
    a newly created |perm_t| object. The caller is responsible for deleting 
    this object as soon as the permutation is no longer needed.
!seealso PermRead
-------------------------------------------------------------------------- }}}*/

perm_t *PermLoad(char *fn)

{
    MTXCONTEXT(PermLoad);
    FILE *f;
    perm_t *p;

    if ((f = SFOpen(fn,FM_READ)) == NULL)
    {
	MTXERROR1("Cannot open %s",fn);
	return NULL;
    }
    p = PermRead(f);
    fclose(f);
    if (p == NULL) 
    {
	MTXERROR1("Cannot read permutation from %s",fn);
	return NULL;
    }
    return p;
}



/*{{{ --------------------------------------------------------------------------
!function PermSave "Write a permutation"
!section perm.io
!synopsis 
    int PermSave(const perm_t *perm, char *fn);
!arg perm
    Permutation to write.
!arg fn
    File name.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function creates a file, writes a single permutation to the file and
    closes the file. If a f ile with the specified name already exists, it's
    contents are destroyed.
!seealso PermWrite
-------------------------------------------------------------------------- }}}*/

int PermSave(const perm_t *perm, char *fn)

{
    MTXCONTEXT(PermSave);
    FILE *f;
    int result;

    if (!PERMUTATION_IS_VALID(perm))
	return -1;
    if ((f = SFOpen(fn,FM_CREATE)) == NULL)
    {
	MTXERROR1("Cannot open %s",fn);
	return -1;
    }
    result = PermWrite(f,perm);
    fclose(f);
    if (result != 0)
    {
	MTXERROR1("Cannot write permutation to %s",fn);
	return -1;
    }
    return result;
}


/*{{{ --------------------------------------------------------------------------
!function PermMul "Multiply permutations"
!section perm.op
!synopsis 
    perm_t *PermMul(perm_t *dest, const perm_t *src);
!arg dest
    Pointer to the first permutation.
!arg src
    Pointer to the second permutation.
!return
    The function returns |dest|, or |NULL| on error.
!description
    This function multiplies |dest| from the right by |src|. Both 
    permutations must have the same degree.
!seealso
-------------------------------------------------------------------------- }}}*/

perm_t *PermMul(perm_t *dest, const perm_t *src)

{
    MTXCONTEXT(PermMul);
    register long i;
    register long *d, *s;

    /* Check arguments 
       --------------- */
    if (!PERMUTATION_IS_VALID(dest) || !PERMUTATION_IS_VALID(src))
	return NULL;
    if (dest->deg != src->deg)
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Multiply
       -------- */
    d = dest->d;
    s = src->d - 1;
    for (i = dest->deg; i > 0; --i)
    {
	*d = s[*d];
	++d;
    }
    return dest;
}




/*{{{ --------------------------------------------------------------------------
!function PermOrder "Order of a permutation"
!section perm.op
!synopsis 
    long PermOrder(const perm_t *perm);
!arg perm
    Pointer to the permutation.
!return
    The function returns the order of |dest|, or $-1$ on error.
!description
    This function calculates the order of a permutation.
!seealso
-------------------------------------------------------------------------- }}}*/

long PermOrder(const perm_t *perm)

{
    MTXCONTEXT(PermOrder);
    long ord = 1, tord;
    long deg;
    long *p;
    long *end;
    register long *seed, *x;

    /* Check arguments 
       --------------- */
    if (!PERMUTATION_IS_VALID(perm))
	return -1;

    deg = perm->deg;
    p = perm->d - 1;
    end = p + (deg + 1);
    seed = p;

    /* Calculate the order by running through all orbits. To keep track of
       which points have been visited, we flip the sign of the corresponding
       entries in the permuatation. With this trick we don't need any 
       variable-sized buffer, but we must undo the changes afterwards.
       ---------------------------------------------------------------------- */

    while (1)
    {
	/* Find next starting point
	   ------------------------ */
	while (*++seed < 0 && seed != end);
	if (seed == end) 
	    break;	    /* Done! */

	/* Find orbit size
	   --------------- */
	tord = 0;
	x = seed;
	while (1)
	{
	    register long tmp = *x;
	    if (tmp < 0) 
		break;	    /* Done! */
	    *x = -tmp;
	    x = p + tmp;
	    ++tord;
	}
	MTX_ASSERT(tord > 0);

	/* Calculate the l.c.m of all orbit sizes
	   -------------------------------------- */
	ord = ord / gcd(ord,tord) * tord;
    }

    /* Now, all entries have their sign changed. Undo this.
       ---------------------------------------------------- */
    for (x = p + 1; x != end; ++x)
	*x = -*x;

    return ord;
}




/*{{{ --------------------------------------------------------------------------
!function PermPower "Power of a permutation"
!section perm.op
!synopsis 
    perm_t *PermPower(const perm_t *p, long n);
!arg perm
    Pointer to the permutation.
!arg n
    Exponent.
!return
    |n|-th power of |perm| or |NULL| on error.
!description
    This function calculates the \verb"n"-th power of a permutation.
    It allocates a new permutation on each call, leaving the original
    permutation intact. The caller is responsible for deleting the 
    result when it is no longer needed.

    $n$ must be positive or 0.
!seealso
-------------------------------------------------------------------------- }}}*/

perm_t *PermPower(const perm_t *p, long n)

{
    MTXCONTEXT(PermPower);
    perm_t *q;
    long *xp;
    long *xq;
    long i, k, l;

    /* Check arguments 
       --------------- */
    if (!PERMUTATION_IS_VALID(p))
	return NULL;
    if (n < 0)
    {
	MTXERROR1("Invalid exponent %ld < 0",n);
	return NULL;
    }

    /* Allocate a new permutation for the result
       ----------------------------------------- */
    q = PermAlloc(p->deg);
    if (q == NULL)
	return NULL;
    xp = p->d - 1;
    xq = q->d - 1;
    
    /* Calculate the n-th power
       ------------------------ */
    for (i = 1; i <= p->deg; ++i)
    {	
	for (k = i, l = n; l > 0; --l)
	    k = xp[k];
	xq[i] = k;
    }
    return q;
}




