/* ============================= C MeatAxe ==================================
   File:        $Id: ypoly.c,v 1.8 1998/06/25 19:02:34 mringe Exp $
   Comment:     Polynomials over finite fields
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


#define INITIAL_SIZE 10



/*{{{ --------------------------------------------------------------------------
!structure poly_t "Polynomial data structure"
!section pol
!synopsis 
    typedef struct
    {
	tMtxObjType id;
	long fl;
	long deg;
	size_t size;
	FEL *buf;
    }
    poly_t;
!description
    The |poly_t| structure represents a polynomial over a finite field. |fl| 
    is the field order, and |deg| is the degree of the polynomial. The 
    coefficients are stored in |buf| as an array of ${\tt deg}+1$ field
    elements. |buf[i]| contains the coefficient of $x^i$. The |size| member
    is used internally to keep track of the buffer's size.

    Polynomials need not be normalized. However, many polynomial functions 
    depend on the leading coefficient being non-zero.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function PolAlloc "Allocate a polynomial"
!section pol.basic
!synopsis 
    poly_t *PolAlloc(long fl, long degree);
!arg fl
    Field order.
!arg degree
    Degree of the polynomial.
!return
    Pointer to a new |poly_t| structure or |NULL| on error.
!description
    This function creates a new polynomial over GF(\verb"fl") and initialializes 
    its value to $x^n$, with $n=\verb"degree"$.
!seealso @polbasic
-------------------------------------------------------------------------- }}}*/

poly_t *PolAlloc(long fl, long degree)

{
    MTXCONTEXT(PolAlloc);
    poly_t *x;
    size_t s;
    int i;

    if (degree < 0) 
	degree = -1;
    if ((s = degree + 1) < 1) 
	s = 1;

    zsetfield(fl);
    if ((x = ALLOC(poly_t)) == NULL) 
    {
	MTXERROR("Cannot allocate polynomial");
	return NULL;
    }
    x->id = T_POLY;
    x->fl = fl;
    x->deg = degree;
    x->size = s;
    if ((x->buf = NALLOC(FEL,s)) == NULL)
    {
	free(x);
        MTXERROR("Cannot allocate polynomial data");
        return NULL;
    }
    for (i = 0; i < (int) s-1; ++i) 
	x->buf[i] = F_ZERO;
    x->buf[s-1] = F_ONE;
    return x;
}




/*{{{ --------------------------------------------------------------------------
!function PolFree "Free a polynomial"
!section pol.basic
!synopsis 
    void PolFree(poly_t *x);
!arg x
    Pointer to the polynomial.
!description
    This function frees a polynomial data structure and cleans up all internal
    data.
!seealso @polbasic
-------------------------------------------------------------------------- }}}*/

void PolFree(poly_t *x)

{
    MTXCONTEXT(PolFree);
    if (!POLYNOMIAL_IS_VALID(x))
	return;
    free(x->buf);
    free(x);
}



/*{{{ --------------------------------------------------------------------------
!function PolDup "Duplicate a polynomial"
!section pol.basic
!synopsis 
    poly_t *PolDup(const poly_t *x);
!arg x
    Pointer to the polynomial.
!return
    Pointer to a copy of |x|, or |NULL| on error.
!description
    This function creates a copy of an existing polynomial.
!seealso @polbasic
-------------------------------------------------------------------------- }}}*/

poly_t *PolDup(const poly_t *x)

{
    MTXCONTEXT(PolDup);
    poly_t *y;

    if (!POLYNOMIAL_IS_VALID(x))
	return NULL;
    if ((y = ALLOC(poly_t)) == NULL)
    {
	MTXERROR("Cannot allocate polynomial");
	return NULL;
    }
    y->id = T_POLY;
    y->buf = NALLOC(FEL,x->deg+1);
    if (y->buf == NULL)
    {
	free(y);
	MTXERROR("Cannot allocate polynomial data");
	return NULL;
    }
    memcpy(y->buf,x->buf,x->deg+1);
    y->fl = x->fl;
    y->deg = x->deg;
    y->size = x->deg + 1;
    return y;
}


/* ------------------------------------------------------------------
   resize() - Extend buffer.
   ------------------------------------------------------------------ */

static int resize(poly_t *p, size_t newdeg)

{
    int i;
    FEL *x;

    if (p->deg < (long) newdeg)
    {
	if (p->size < newdeg+1)	/* Allocate new buffer */
    	{
	    x = NALLOC(FEL,newdeg+1);
	    if (x == NULL) 
		return 1;
	    memcpy(x,p->buf,(p->deg+1)*sizeof(FEL));
	    free(p->buf);
	    p->buf = x;
	    p->size = newdeg + 1;
    	}
    	for (i = p->deg + 1; i <= (int) newdeg; ++i)
	    p->buf[i] = F_ZERO;
        p->deg = newdeg;
    }
    return 0;
}


/* ------------------------------------------------------------------
   normalize()
   ------------------------------------------------------------------ */

static void normalize(poly_t *p)

{
    int i = p->deg;

    while (i >= 0 && p->buf[i] == F_ZERO) 
	--i;
    p->deg = i;
}



/*{{{ --------------------------------------------------------------------------
!function PolAdd "Add polynomials"
!section pol.op
!synopsis 
    poly_t *PolAdd(poly_t *dest, const poly_t *src);
!arg dest
    Pointer to the first polynomial.
!arg src
    Pointer to the second polynomial.
!return
    The function returns |dest|, or |NULL| on error.
!description
    Thus function adds \verb"src" to \verb"dest" and returns \verb"dest".
    The polynomials must be over the same field. Otherwise, the return
    value is \verb"NULL".
!seealso
-------------------------------------------------------------------------- }}}*/

poly_t *PolAdd(poly_t *dest, const poly_t *src)

{
    MTXCONTEXT(PolAdd);
    FEL *s, *d;
    int i;

    if (!POLYNOMIAL_IS_VALID(src) || !POLYNOMIAL_IS_VALID(dest))
	return NULL;
    if (dest->fl != src->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }
    if ((i = src->deg) == -1) 
	return dest;	/* src = 0 */

    zsetfield(src->fl);
    if (resize(dest,i)) 
    {
	MTXERROR("Cannot resize: %S");
	return NULL;
    }
    s = src->buf;
    d = dest->buf;
    for (; i >= 0; --i)
    {
       *d = zadd(*d,*s++);
       ++d;
    }
    normalize(dest);
    return dest;
}




/*{{{ --------------------------------------------------------------------------
!function PolMul "Multiply polynomials"
!section pol.op
!synopsis 
    poly_t *PolMul(poly_t *dest, const poly_t *src);
!arg dest
    Pointer to the first polynomial.
!arg src
    Pointer to the second polynomial.
!return
    The function returns |dest|, or |NULL| on error.
!description
    Thus function multiplies \verb"src" by \verb"dest" and returns \verb"dest".
    The polynomials must be over the same field. Otherwise, the return
    value is \verb"NULL".
!seealso
-------------------------------------------------------------------------- }}}*/

poly_t *PolMul(poly_t *dest, const poly_t *src)

{
    MTXCONTEXT(PolMul);
    FEL *x, *y, *d, *s;
    int di, si;
    size_t xdeg;

    /* Check arguments
       --------------- */
    if (!POLYNOMIAL_IS_VALID(src) || !POLYNOMIAL_IS_VALID(dest))
	return NULL;
    if (dest->fl != src->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    /* Handle special cases: dest = 0, src = 0
       --------------------------------------- */
    if (dest->deg == -1)
	return dest;
    if (src->deg == -1)
    {
	dest->deg = -1;
	return dest;
    }

    
    d = dest->buf;
    s = src->buf;
    xdeg = src->deg + dest->deg;
    zsetfield(src->fl);

    /* Allocate memory for the result
       ------------------------------ */
    x = NALLOC(FEL,xdeg+1);
    if (x == NULL) 
    {
	MTXERROR("Cannot allocate result");
	return NULL;
    }
    for (di = xdeg, y = x; di >= 0; --di) 
	*y++ = F_ZERO;

    /* Multiply
       -------- */
    for (di = 0; di <= dest->deg; ++di)
    	for (si = 0; si <= src->deg; ++si)
	    x[si+di] = zadd(x[si+di],zmul(s[si],d[di]));

    /* Overwrite |dest| with the result
       -------------------------------- */
    free(dest->buf);
    dest->buf = x;
    dest->deg = xdeg;
    dest->size= xdeg+1;
    return dest;
}




/*{{{ --------------------------------------------------------------------------
!function PolDivMod "Polynomial division"
!section pol.op
!synopsis 
    poly_t *PolDivMod(poly_t *a, const poly_t *b);
!arg a
    Pointer to the first polynomial (numerator).
!arg src
    Pointer to the second polynomial (denominator).
!return
    The function returns the quotient, or |NULL| on error.
!description
    This function performs a polynomial division. Given two polynomials $a$ and
    $b\neq 0$, which must be over the same field, |PolDivMod()| finds two 
    polynomials $q$ and $r$ such that $a=qb+r$, and $\deg(r)<\deg(b)$.

    The quotient, $q$, is returned as the function result. This is a newly 
    allocated polynomial. The caller is responsible for deleting the quotient
    when it no longer needed.

    The remainder, $r$, is stored in |a| and replaces the original value. If you
    need to keep the value of |a| you must make a copy using |PulDup()| before
    calling |PolDivMod()|. |b| is not changed.
!seealso @polymod
-------------------------------------------------------------------------- }}}*/

poly_t *PolDivMod(poly_t *a, const poly_t *b)

{
    MTXCONTEXT(PolDivMod);
    poly_t *q;

    /* Check arguments
       --------------- */
    if (!POLYNOMIAL_IS_VALID(a) || !POLYNOMIAL_IS_VALID(b))
	return NULL;
    if (a->fl != b->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }
    zsetfield(a->fl);
    if (b->deg <= -1) 
    {
	MTXERROR1("%E",ERR_DIV0);
	return NULL;
    }
    if (a->deg < b->deg)
    {
	q = PolAlloc(a->fl,-1);	/* Trivial case: Quotient = 0 */
    }
    else
    {
	FEL lead = b->buf[b->deg];
	int i, k;

	if (lead == F_ZERO) 
	{
	    MTXERROR1("%E",ERR_DIV0);
	    return NULL;
	}
	q = PolAlloc(zfl,a->deg - b->deg);
	if (q == NULL) 
	{
	    MTXERROR("Cannot allocate result");
	    return NULL;
	}
	for (i = a->deg; i >= b->deg; --i)
	{
	    FEL qq = zneg(zdiv(a->buf[i],lead));
	    for (k = 0; k <= b->deg; ++k)
		a->buf[i-k] = zadd(a->buf[i-k],
		    zmul(qq,b->buf[b->deg - k]));
	    q->buf[i-b->deg] = zneg(qq);
	}
        normalize(a);
    }
    return q;
}




/*{{{ --------------------------------------------------------------------------
!function PolMod "Polynomial division"
!section pol.op
!synopsis 
    poly_t *PolMod(poly_t *a, const poly_t *b)
!arg a
    Pointer to the first polynomial (numerator).
!arg src
    Pointer to the second polynomial (denominator).
!return
    The function returns |a|, which has been reduced modulo |b|. On error
    the function returns |NULL|.
!description
    This function reduces |a| modulo |b|.
!seealso @polymod
-------------------------------------------------------------------------- }}}*/

poly_t *PolMod(poly_t *a, const poly_t *b)

{
    MTXCONTEXT(PolMod);


    /* Check arguments
       --------------- */
    if (!POLYNOMIAL_IS_VALID(a) || !POLYNOMIAL_IS_VALID(b))
	return NULL;
    if (a->fl != b->fl) 
    {
	MTXERROR1("%E",ERR_INCOMPAT);
	return NULL;
    }

    zsetfield(a->fl);
    if (b->deg <= -1) 
    {
	MTXERROR1("%E",ERR_DIV0);
	return NULL;
    }
    if (a->deg >= b->deg)
    {
	FEL lead = b->buf[b->deg];
	int i, k;

	if (lead == F_ZERO) 
	{
	    MTXERROR1("%E",ERR_DIV0);
	    return NULL;
	}
	for (i = a->deg; i >= b->deg; --i)
	{
	    FEL qq = zneg(zdiv(a->buf[i],lead));
	    for (k = 0; k <= b->deg; ++k)
		a->buf[i-k] = zadd(a->buf[i-k],
		    zmul(qq,b->buf[b->deg - k]));
	}
        normalize(a);
    }
    return a;
}



/*{{{ --------------------------------------------------------------------------
!function PolPrint "Print a polynomial"
!section pol.basic
!synopsis 
    void PolPrint(char *name, const poly_t *p);
!arg name
    Name to print before the polynomial or |NULL| (see below).
!arg p
    Pointer to the polynomial.
!return
    $0$ on success, $-1$ on error.
!description
    This function prints a polynomial on the standard output in a human-readable
    form. If |name| is not |NULL| the name followed by an equal sign is printed 
    before the polynomial. For example, the
    statement \verb:PolPrint("P",P);: may produce the following output:
    \begin{verbatim}
	P=3x^2+x+1
    \end{verbatim}
!seealso
-------------------------------------------------------------------------- }}}*/

void PolPrint(char *name, const poly_t *p)

{
    MTXCONTEXT(PolPrint);
    int i,flag = 0;

    if (!POLYNOMIAL_IS_VALID(p))
	return;
    if (name != NULL) 
	printf("%s=",name);
    zsetfield(p->fl);
    if (p->deg == -1) 
    {
	printf("0x^0");
    }
    for (i = p->deg; i >= 0; i--)
    {
	if (p->buf[i] != F_ZERO)
	{
	    if (flag) 
		printf("+");
	    if (p->buf[i] != F_ONE || i == 0)
		printf("%ld",zftoi(p->buf[i]));
	    if (i > 1) 
		printf("x^%d",i);
	    else if (i == 1) 
		printf("x");
	    flag=1;
       	}
    }
    if (name != NULL) printf("\n");
}


/* -----------------------------------------------------------------
   mktmp() - Allocate temporary workspace
   ----------------------------------------------------------------- */

static long tmpfl = 0;
static long tmpdeg = 0;
static PTR tmpvec = NULL;

static void mktmp(long fl, long deg)

{
    zsetfield(fl);
    if (deg > 0) zsetlen(deg+1);
    if (tmpfl != fl || tmpdeg < deg)
    {
	if (tmpvec != NULL) free(tmpvec);
	tmpvec = zalloc((long)1);
	tmpdeg = deg;
	tmpfl = fl;
    }
}



/*{{{ --------------------------------------------------------------------------
!function PolRead "Read a polynomial from a file"
!section pol.io
!synopsis 
    poly_t *PolRead(FILE *f);
!arg f
    File to read from.
!return
    Pointer to the plynomial, or |NULL| on error.
!description
    This function reads a polynomial from a file. If successful, the return 
    value is a pointer to a new |poly_t| object. The caller is responsible
    for deleting this polynomial with |PolFree()| when it is no longer needed.
!seealso
-------------------------------------------------------------------------- }}}*/

poly_t *PolRead(FILE *f)

{
    MTXCONTEXT(PolRead);
    poly_t *p;
    long hdr[3];
    
    if (zreadlong(f,hdr,3) != 3)
    {
	MTXERROR("Cannot read header");
	return NULL;
    }
    if (hdr[0] != -2)
    {
	MTXERROR1("No polynomial (type=%d)",(int)hdr[0]);
	return NULL;
    }
    mktmp(hdr[1],hdr[2]);
    if ((p = PolAlloc(hdr[1],hdr[2])) == NULL)
    {
	MTXERROR("Cannot create polynomial");
	return NULL;
    }
    if (p->deg > 0)
    {
	int i;
	if (zreadvec(f,tmpvec,1) != 1)
	{
	    PolFree(p);
	    MTXERROR("Cannot read data");
	    return NULL;
	}
	for (i = 0; i <= p->deg; ++i)
	    p->buf[i] = zextract(tmpvec,i+1);
    }
    return p;
}



/*{{{ --------------------------------------------------------------------------
!function PolWrite "Write a polynomial to a file"
!section pol.io
!synopsis 
    int PolWrite(FILE *f, const poly_t *p);
!arg f
    File to write to.
!arg p
    Pointer to the polynomial.
!return
    $0$ on success, $-1$ on error.
!description
    This function writes a polynomial to a file.
!seealso
-------------------------------------------------------------------------- }}}*/

int PolWrite(FILE *f, const poly_t *p)

{
    MTXCONTEXT(PolWrite);
    long hdr[3];

    if (!POLYNOMIAL_IS_VALID(p))
	return -1;
    mktmp(p->fl,p->deg);
    hdr[0] = -2;
    hdr[1] = p->fl;
    hdr[2] = p->deg;
    if (zwritelong(f,hdr,3) != 3) 
    {
	MTXERROR("Cannot write header");
	return -1;
    }
    if (p->deg >= 0)
    {
	int i;
    	for (i = 0; i <= p->deg; ++i)
	    zinsert(tmpvec,i+1,p->buf[i]);
        if (zwritevec(f,tmpvec,1) != 1)
	{
	    MTXERROR("Cannot write data");
	    return -1;
	}
    }
    return 0;
}

/*{{{ --------------------------------------------------------------------------
!function PolLoad "Read a polynomial"
!section pol.io
!synopsis 
    poly_t *PolLoad(char *fn)
!arg fn
    File name.
!return
    Pointer to the polynomial read from the file, or |NULL| on error.
!description
    This function opens a file, reads a single polynomial, and closes the 
    file. The return value is a pointer to the polynomial or \verb"NULL" on 
    error. If the file contains more than one polynomial, only the first one 
    is read.

    If a polynomial was successfully read, the function returns a pointer to 
    a newly created |poly_t| object. The caller is responsible for deleting 
    this object as soon as it no longer needed.
!seealso PolRead
-------------------------------------------------------------------------- }}}*/

poly_t *PolLoad(char *fn)

{
    MTXCONTEXT(PolLoad);
    FILE *f;
    poly_t *p;

    if ((f = SFOpen(fn,FM_READ)) == NULL)
    {
	MTXERROR1("Cannot open %s",fn);
	return NULL;
    }
    p = PolRead(f);
    fclose(f);
    if (p == NULL) 
    {
	MTXERROR1("Cannot read polynomial from %s",fn);
	return NULL;
    }
    return p;
}

/*{{{ --------------------------------------------------------------------------
!function PolSave "Write a polynomial"
!section pol.io
!synopsis 
    int PolSave(const poly_t *pol, char *fn);
!arg pol
    Polynomial to write.
!arg fn
    File name.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function creates a file, writes a single polynomial to the file and
    closes the file. If a f ile with the specified name already exists, it's
    contents are destroyed.
!seealso PolWrite
-------------------------------------------------------------------------- }}}*/

int PolSave(const poly_t *pol, char *fn)

{
    MTXCONTEXT(PolSave);
    FILE *f;
    int result;

    if (!POLYNOMIAL_IS_VALID(pol))
	return -1;
    if ((f = SFOpen(fn,FM_CREATE)) == NULL)
    {
	MTXERROR1("Cannot open %s",fn);
	return -1;
    }
    result = PolWrite(f,pol);
    fclose(f);
    if (result != 0)
    {
	MTXERROR1("Cannot write polynomial to %s",fn);
	return -1;
    }
    return result;
}


/*{{{ --------------------------------------------------------------------------
!function PolCompare "Compare polynomials"
!section pol.op
!synopsis 
    int PolCompare(const poly_t *a, const poly_t *b);
!arg a
    First polynomial.
!arg b
    Second polynomial.
!return
    The return value is $1$ if $a>b$, $-1$ if $a < b$, and $0$ if $a=b$.
    In case of error, the function returns $-2$.
!description
    This function compares two polynomials and returns $0$
    if the polynomials are equal, $-1$ if $a<b$, and $1$ if $a>b$.
    The ordering of polynomials is defined as follows.
    \begin{enumerate}
    \item If $a$ and $b$ are over different fields, the polynomials over 
	the larger field is greater. 
    \item Otherwise, if they have different degrees, the polynomial
	with the higher degree is greater. 
    \item If both field and degree are equal, the result of the comparison 
	depends on the internal representation of the field elements and 
	does not have any significance.
    \end{enumerate}
!seealso
-------------------------------------------------------------------------- }}}*/

int PolCompare(const poly_t *a, const poly_t *b)

{
    MTXCONTEXT(PolCompare);
    int i;

    /* Check the arguments
       ------------------- */
    if (!POLYNOMIAL_IS_VALID(a) || !POLYNOMIAL_IS_VALID(b))
	return -2;

    /* Check if the polynomials are over the same field.
       ------------------------------------------------- */
    if (a->fl > b->fl) 
	return 1;
    if (a->fl < b->fl) 
	return -1;

    /* Check if the polynomials have equal degrees.
       -------------------------------------------- */
    if (a->deg > b->deg) 
	return 1;
    if (a->deg < b->deg) 
	return -1;

    /* Compare the coefficients
       ------------------------ */
    for (i = a->deg; i >= 0; --i)
    {
	if (a->buf[i] > b->buf[i]) 
	    return 1;
	if (a->buf[i] < b->buf[i]) 
	    return -1;
    }

    /* The polynomials are equal!
       -------------------------- */
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function PolDerive "Derive a polynomials"
!section pol.op
!synopsis 
    poly_t *PolDerive(poly_t *pol);
!arg pos
    Pointer to the polynomial.
!return
    Returns |pol|.
!description
    This function derives a polynomial. Note that the derived polynomial is
    stored in |pol|, replacing the original polynomial. The following piece of
    code shows how to keep the original polynomial intact while calculating
    the derivative:
    \begin{verbatim}
	poly_t *pol, *der;
	...
	der = PolDerive(PolDup(pol));
    \end{verbatim}
!seealso
-------------------------------------------------------------------------- }}}*/

poly_t *PolDerive(poly_t *pol)

{
    MTXCONTEXT(PolDerive);
    int i, maxdeg = -1;
    register FEL *buf;
    FEL f = F_ZERO;

    /* Check the argument
       ------------------ */
    if (!POLYNOMIAL_IS_VALID(pol))
	return NULL;

    buf = pol->buf;
    zsetfield(pol->fl);
    for (i = 0; i < pol->deg; ++i)
    {
	f = zadd(f,F_ONE);
	buf[i] = zmul(buf[i+1],f);
	if (buf[i] != F_ZERO) 
	    maxdeg = i;
    }
    pol->deg = maxdeg;
    return pol;
}
      

