/* ============================= C MeatAxe ==================================
   File:        $Id: ysets.c,v 1.7 1998/06/25 19:02:34 mringe Exp $
   Comment:     Integer sets (= sorted lists)
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>


#define VERIFY_VALID_SET(s) MTX_VERIFY(s != NULL && s->id == T_SET)

/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static size_t init_max = 10;
static size_t blksize = 10;



/*{{{ --------------------------------------------------------------------------
!structure set_t "Integer set data structure"
!section sets
!synopsis 
    typedef struct
    {
	size_t len, max;
	long *buf;
    } set_t;
!description
    The type \verb"set_t" represents a set of (long) integers. Internally, the 
    set is stored as a sorted list. \verb"set_t" should be used if the number 
    of elements in the set is typically much less than the elements themselves.
    In other cases, when the number of elements is comparable to the elements 
    themselves, the \verb"bitstring_t" type may be more suitable.
-------------------------------------------------------------------------- }}}*/



/*{{{ --------------------------------------------------------------------------
!text sets
    The {\MeatAxe} has two data types that represent subsets of the natural
    numbers, |set_t| and |bitstring_t|.
-------------------------------------------------------------------------- }}}*/



/*{{{ --------------------------------------------------------------------------
!function SetPrint "Write a set to stdout"
!section sets
!synopsis 
    void SetPrint(char *name, const set_t *s);
!arg name
    Name to print before the set, or |NULL|.
!arg set
    Pointer to the set.
!description
    This function prints a set in human-readable format on the standard 
    output. If |name| is different from |NULL|, |name| and an `=' is printed
    before the set.
!seealso
-------------------------------------------------------------------------- }}}*/
    
void SetPrint(char *name, const set_t *s)

{
    MTXCONTEXT(SetPrint);
    int i;

    VERIFY_VALID_SET(s);
    if (name != NULL) 
	printf("%s=",name);
    printf("{");
    for (i = 0; i < (int) s->len; ++i)
    {
	if (i > 0) printf(",");
	printf("%ld",s->buf[i]);
    }
    printf("}");
    if (name != NULL) printf("\n");
    fflush(stdout);
}



/*{{{ --------------------------------------------------------------------------
!function SetAlloc "Create a new set"
!section sets
!synopsis 
    set_t *SetAlloc();
!return
    Pointer to the new set or |NULL| on error.
!description
    This function creates a set. The set is initially empty. To destroy a 
    set, use \verb"SetFree()", {\em not} \verb"free()".
!seealso @SetAllocFree
-------------------------------------------------------------------------- }}}*/

set_t *SetAlloc()

{
    MTXCONTEXT(SetAlloc);
    set_t *x;

    x = ALLOC(set_t);
    if (x == NULL)
    {
 	MTXERROR("Cannot allocate set");
	return NULL;
    }
    x->id = T_SET;
    x->len = 0;
    x->max = init_max;
    x->buf = NALLOC(long,init_max);
    if (x->buf == NULL)
    {
	free(x);
 	MTXERROR("Cannot allocate set data");
	return NULL;
    }

    return x;
}




/*{{{ --------------------------------------------------------------------------
!function SetFree "Destroy a set"
!section sets
!synopsis 
    void SetFree(set_t *x);
!arg x
    Pointer to the set.
!description
    This function frees an integer set. The argument must be a \verb"set_t" 
    structure which has previously been allocated with \verb"SetAlloc()". 
!seealso @SetAllocFree
-------------------------------------------------------------------------- }}}*/

void SetFree(set_t *x)

{
    MTXCONTEXT(SetFree);
    VERIFY_VALID_SET(x);

    free(x->buf);
    free(x);
}


/*{{{ --------------------------------------------------------------------------
!function SetDup "Duplicate a set"
!section sets
!synopsis 
    set_t *SetDup(const set_t *s);
!arg s
    Pointer to the set.
!return
    Pointer to a copy of the set, or |NULL| on error.
!description
    This function creates a copy of an existing set.
!seealso @SetAllocFree
-------------------------------------------------------------------------- }}}*/

set_t *SetDup(const set_t *s)

{
    MTXCONTEXT(SetDup);
    set_t *x;
    int i;

    VERIFY_VALID_SET(s);
    x = ALLOC(set_t);
    if (x == NULL)
    {
 	MTXERROR("Cannot allocate set");
	return NULL;
    }
    x->id = T_SET;
    x->max = x->len = s->len;
    x->buf = NALLOC(long,x->max);
    if (x->buf == NULL)
    {
	free(x);
 	MTXERROR("Cannot allocate set data");
	return NULL;
    }
    for (i = 0; i < (int) s->len; ++i)
	x->buf[i] = s->buf[i];
    return x;
}


/*{{{ --------------------------------------------------------------------------
!function SetInsert "Insert an element into a set"
!section sets
!synopsis 
    int SetInsert(set_t *set, long elem);
!arg set
    Pointer to the set.
!arg elem
    Number to insert.
!return
    $0$ on success, $-1$ on error.
!description
    This function inserts the number |elem| into |set|.
!seealso @SetOps
-------------------------------------------------------------------------- }}}*/

int SetInsert(set_t *set, long elem)

{
    MTXCONTEXT(SetInsert);
    int i,k;
    long *l,*m;

    VERIFY_VALID_SET(set);
    for (l = set->buf, i = set->len; i > 0 && *l < elem; --i, ++l);
    if (i > 0 && *l == elem) 
	return 0;	/* Ok, already in set */
    if (set->len >= set->max - 1)	/* Need extend the set? */
    {
	size_t newmax = set->max + blksize;
	long *newbuf = NREALLOC(set->buf,long,newmax);
	if (newbuf == NULL)
	{
	    MTXERROR("Cannot grow set");
	    return -1;
	}
	set->max = newmax;
	set->buf = newbuf;
    }
    for (m = set->buf+set->len-1, k = i; k > 0; --k,--m)
	m[1] = m[0];
    *l = elem;
    ++set->len;
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!function SetContains "Check if a number is in a set"
!section sets
!synopsis 
    int SetContains(const set_t *set, long elem);
!arg set
    Pointer to the set.
!arg elem
    Number to check.
!return
    1 if |set| contains |elem|, 0 else.
!description
    This function checks if |set| contains the number |elem|.
!seealso @SetOps
-------------------------------------------------------------------------- }}}*/

int SetContains(const set_t *set, long elem)

{
    MTXCONTEXT(SetContains);
    int i;
    long *l;
 
    VERIFY_VALID_SET(set);
    l = set->buf;
    for (i = set->len; i > 0 && *l < elem; --i, ++l);
    return (i > 0 && *l == elem);
}


/*{{{ --------------------------------------------------------------------------
!function SetRead "Read a set from a file"
!section sets
!synopsis 
    set_t *SetRead(FILE *f);
!arg f
    Pointer to the file.
!return
    Pointer to the set read from |f|, or |NULL| on error.
!description
    This function reads a set from a file.
!seealso @SetIo
-------------------------------------------------------------------------- }}}*/

set_t *SetRead(FILE *f)

{
    MTXCONTEXT(SetRead);
    set_t *x;
    long l;

    if (zreadlong(f,&l,1) != 1)
    {
	MTXERROR("Cannot read header");
	return NULL;
    }
    x = ALLOC(set_t);
    if (x == NULL)
    {
	MTXERROR("Cannot allocate set");
	return NULL;
    }
    x->id = T_SET;
    x->max = x->len = (size_t)l;
    x->buf = NALLOC(long,x->len);
    if (x->buf == NULL)
    {
	free(x);
	MTXERROR("Cannot allocate set data");
	return NULL;
    }
    if (zreadlong(f,x->buf,x->len) != x->len)
    {
	free(x->buf);
	free(x);
	MTXERROR("Cannot read set");
	return NULL;
    }
    return x;
}


/*{{{ --------------------------------------------------------------------------
!function SetWrite "Write a set to a file"
!section sets
!synopsis 
    int SetWrite(FILE *f,const set_t *set);
!arg f
    Pointer to the file.
!arg set
    Pointer to the set.
!return
    $0$ on success, $-1$ on error.
!description
    This function write a set to a file.
!seealso @SetIo
-------------------------------------------------------------------------- }}}*/

int SetWrite(FILE *f, const set_t *set)

{
    MTXCONTEXT(SetWrite);
    long l = set->len;

    VERIFY_VALID_SET(set);
    if (zwritelong(f,&l,1) != 1)
    {
	MTXERROR("Cannot write header");
	return -1;
    }
    if (zwritelong(f,set->buf,set->len) != set->len)
    {
	MTXERROR("Cannot write data");
	return -1;
    }
    return 0;
}


