/* ============================= C MeatAxe ==================================
   File:        $Id: zbitstr.c,v 1.5 1998/06/25 19:02:35 mringe Exp $
   Comment:     Bit string operations
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static unsigned char mask[8] = {1,2,4,8,16,32,64,128};
static unsigned char clmask[8] =
	{(unsigned char)~1,(unsigned char)~2,(unsigned char)~4,
	(unsigned char)~8,(unsigned char)~16,(unsigned char)~32,
        (unsigned char)~64,(unsigned char)~128};

static int matchcount[256] =
	{	0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
	};
size_t bs_size = (size_t) -1;
size_t bs_len = (size_t) -1;






/*{{{ --------------------------------------------------------------------------
!structure bitstring_t "Bit string data type"
!section bs
!synopsis 
    typedef struct { ... } bitstring_t;
!description
    The |bitstring_t| type represents a string of 0's and 1'. You
    may also think of it as a subset of $\{1,2,\ldots,N\}$, where
    each 1 in the bit string means that the subset contains the
    corresponding element.
    Unlike the |set_t| type (see section \ref{GdSection:sets}) bit
    strings are static objects which cannot change size.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!text bs
    The {\MeatAxe} has two data types that represent subsets of the natural
    numbers, |set_t| and |bitstring_t|.

    The |bitstring_t| type represents a string of 0's and 1's. You
    may also think of it as a subset of $\{1,2,\ldots,N\}$, where
    each 1 in the bit string means that the subset contains the
    corresponding element. Unlike the |set_t| type bit
    strings are static objects which cannot change size. However,
    operations like the intersection of two sets are much faster
    with |bitstring_t| than with |set_t| objects.

    The length of a bit strings is a global variable which can be set
    with |bs_setlen()|. You must call |bs_setlen()| at least once before
    creating a bit string. Since most bit string function assume that all
    bit strings have the same length, it is difficult to operate with
    bit strings of different sizes.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function bs_setlen "Set the bit string length"
!section bs
!synopsis 
    void bs_setlen(int l);
!arg l
    Bit string length.
!description
    This function sets the bit string length which is used by most bit string
    functions.
!seealso
-------------------------------------------------------------------------- }}}*/

void bs_setlen(int l)

{
    bs_len = l;
    bs_size = ((l-1) / 8) + 1;
}



/*{{{ --------------------------------------------------------------------------
!function bs_alloc "Create a bit string"
!section bs
!synopsis 
    bitstring_t *bs_alloc();
!return
    Pointer to the new bit string, or |NULL| on error.
!description
    This function creates a new bit string. Before yo can use |bs_alloc()|
    you must set the bit string length with |bs_setlen()|.
!seealso
-------------------------------------------------------------------------- }}}*/

bitstring_t *bs_alloc()

{
    MTXCONTEXT(bs_alloc);
    bitstring_t *n;

    n = (bitstring_t *) Smalloc(sizeof(bitstring_t) + bs_size - 1);
    if (n == NULL) 
    {
	MTXERROR("Cannot allocate bit string");
	return NULL;
    }
    n->id = T_BITS;
    memset(n->buf,0,(size_t)bs_size);
    return n;
}




/*{{{ --------------------------------------------------------------------------
!function bs_dup "Duplicate a bit string"
!section bs
!synopsis 
    bitstring_t *bs_dup(const bitstring_t *s)
!arg s
    Pointer to the bit string.
!return
    Pointer to the copy, or |NULL| on error.
!description
    This function creates an independent copy of an exiting bit tring.
!seealso
-------------------------------------------------------------------------- }}}*/

bitstring_t *bs_dup(const bitstring_t *s)

{
    MTXCONTEXT(bs_dup);
    bitstring_t *n;

    n = (bitstring_t *) Smalloc(sizeof(bitstring_t) + bs_size - 1);
    if (n == NULL) 
    {
	MTXERROR("Cannot allocate bit string");
	return NULL;
    }
    n->id = T_BITS;
    memcpy(n->buf,s->buf,(size_t)bs_size);
    return n;
}



/*{{{ --------------------------------------------------------------------------
!function bs_free "Free a bit string"
!section bs
!synopsis 
    void bs_free(bitstring_t *b);
!arg b
    Pointer to the bit string.
!description
    This function frees a bit string, releasing all associated memory.
!seealso
-------------------------------------------------------------------------- }}}*/

void bs_free(bitstring_t *b)

{
    free(b);
}



/*{{{ --------------------------------------------------------------------------
!function bs_reset "Clear a bit string"
!section bs
!synopsis 
    void bs_reset(bitstring_t *x);
!arg x
    Pointer to the bit string.
!description
    This function clears all bits in a bit string.
!seealso
-------------------------------------------------------------------------- }}}*/

void bs_reset(bitstring_t *x)

{
    memset(x->buf,0,bs_size);
}



/*{{{ --------------------------------------------------------------------------
!function bs_cmp "Compare two bit strings"
!section bs
!synopsis 
    int bs_cmp(const bitstring_t *a, const bitstring_t *b);
!arg a
    Pointer to the first bit string.
!arg b
    Pointer to the second bit string.
!return
    $0$ if the bit strings are equal. Otherwise the return value is different 
    from zero.
!description
    This function compares two bit strings and returns zero if they are equal.
!seealso
-------------------------------------------------------------------------- }}}*/

int bs_cmp(const bitstring_t *a, const bitstring_t *b)

{
    return memcmp(a->buf,b->buf,bs_size);
}




/*{{{ --------------------------------------------------------------------------
!function bs_cpy "Copy a bit string"
!section bs
!synopsis 
    void bs_cpy(bitstring_t *dest, const bitstring_t *src);
!arg dest
    Pointer to the destination bit string.
!arg src
    Pointer to the source bit string.
!description
    This function copy a bit string from |src| to |dest|. The bit strings must
    have the same size.
!seealso
-------------------------------------------------------------------------- }}}*/

void bs_cpy(bitstring_t *dest, const bitstring_t *src)

{	
    memcpy(dest->buf,src->buf,bs_size);
}




/*{{{ --------------------------------------------------------------------------
!function bs_set "Set one bit in a bit string"
!section bs
!synopsis 
    void bs_set(bitstring_t *b, int i);
!arg b
    Pointer to the bit string.
!arg i
    Index of bit to set (starting from 0).
!description
    This function sets a single bit in a bit string.
!seealso @bsbit
-------------------------------------------------------------------------- }}}*/

void bs_set(bitstring_t *b, int i)

{	
    b->buf[i / 8] |= mask[i % 8];
}

/*{{{ --------------------------------------------------------------------------
!function bs_clear "Clear one bit in a bit string"
!section bs
!synopsis 
    void bs_clear(bitstring_t *b, int i);
!arg b
    Pointer to the bit string.
!arg i
    Index of bit to clear (starting from 0).
!description
    This function clears a single bit in a bit string.
!seealso @bsbit
-------------------------------------------------------------------------- }}}*/

void bs_clear(bitstring_t *b, int i)

{	
    b->buf[i / 8] &= clmask[i % 8];
}



/*{{{ --------------------------------------------------------------------------
!function bs_test "Test a bit in a bit string"
!section bs
!synopsis 
    int bs_test(const bitstring_t *b, int i);
!arg b
    Pointer to the bit string.
!arg i
    Index of bit to test (starting from 0).
!return
    $1$ if the bit is set, $0$ otherwise.
!description
    This function tests a single bit in a bit string. The return value is
    $1$ id the bit is set and $0$ otherwise.
!seealso @bsbit
-------------------------------------------------------------------------- }}}*/

int bs_test(const bitstring_t *b, int i)

{	
    return ((b->buf[i / 8] & mask[i % 8]) != 0);
}



/*{{{ --------------------------------------------------------------------------
!function bs_read "Read a bit string from a file"
!section bs
!synopsis 
    bitstring_t *bs_read(FILE *f);
!arg f
    File to read from.
!return
    Pointer to the bit string, or |NULL| on error.
!description
    This function reads a bit string from a file. The bit string size must
    have been set before with \verb"bs_setlen()".
    The return value is a pointer to the bit string or \verb"NULL" on error.
!seealso bs_read
-------------------------------------------------------------------------- }}}*/

bitstring_t *bs_read(FILE *f)

{
    MTXCONTEXT(bs_read);
    bitstring_t *b;

    b = bs_alloc();
    if (fread(b->buf,sizeof(unsigned char),bs_size,f) != bs_size)
    {
	bs_free(b);
	MTXERROR("Cannot read bit string: %S");
	return NULL;
    }
    return b;
}



/*{{{ --------------------------------------------------------------------------
!function bs_write "Write a bit string to a file"
!section bs
!synopsis 
    int bs_write(FILE *f, const bitstring_t *b);
!arg f
    File to write to.
!arg b
    Bit string to write.
!return
    $0$ on success, $-1$ on error.
!description
    This function writes a bit string to a file. The bit string size must
    have been set previosly with \verb"bs_setlen()".
!seealso bs_read
-------------------------------------------------------------------------- }}}*/

int bs_write(FILE *f, const bitstring_t *b)

{
    MTXCONTEXT(bs_write);
    if (fwrite(b->buf,sizeof(unsigned char),bs_size,f) != bs_size)
    {
	MTXERROR("Cannot write bit string: %S");
	return -1;
    }
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function bs_and "Intersection of two bit strings"
!section bs
!synopsis 
    void bs_and(bitstring_t *dest, const bitstring_t *src);
!arg dest
    Destination bit string.
!arg src
    Source bit string, and-ed to |dest|.
!description
    This function computes the intersection of two bit strings, i.e., the
    logical `and'.
    The result is stored in \verb"dest" and overwrites the previous value.
!seealso 
-------------------------------------------------------------------------- }}}*/

void bs_and(bitstring_t *dest, const bitstring_t *src)

{
    register int i;
    /* gefaehrlich: buf liegt evtl nicht auf einer longword grenze*/
    register unsigned long *dl = (unsigned long *)dest->buf;
    register unsigned long *sl = (unsigned long *)src->buf;
    register unsigned char *d;
    register unsigned char *s;

    for (i = bs_size; i > sizeof(long); i -= sizeof(long))
	*dl++ &= *sl++;
    d = (unsigned char *) dl;
    s = (unsigned char *) sl;
    for (; i > 0; --i)
	*d++ &= *s++;
}


/*{{{ --------------------------------------------------------------------------
!function bs_or "Union of two bit strings"
!section bs
!synopsis 
    void bs_or(bitstring_t *dest, const bitstring_t *src);
!arg dest
    Destination bit string.
!arg src
    Source bit string, or-ed to |dest|.
!description
    This function computes the union of two bit strings.
    The result is stored in \verb"dest" and overwrites the previous value.
!seealso 
-------------------------------------------------------------------------- }}}*/

void bs_or(bitstring_t *dest, const bitstring_t *src)

{	register int i;
	register unsigned long *dl = (unsigned long *)dest->buf;
	register unsigned long *sl = (unsigned long *)src->buf;
	register unsigned char *d;
	register unsigned char *s;

	for (i = bs_size; i > sizeof(long); i -= sizeof(long))
		*dl++ |= *sl++;
	d = (unsigned char *) dl;
	s = (unsigned char *) sl;
	for (; i > 0; --i)
		*d++ |= *s++;
}


   
/*{{{ --------------------------------------------------------------------------
!function bs_minus "Bit string difference"
!section bs
!synopsis 
    void bs_minus(bitstring_t *dest, const bitstring_t *src);
!arg dest
    Destination bit string.
!arg src
    Source bit string, subtracted from |dest|.
!description
    \verb"bs_minus()" computes the (set theoretical) difference of two
    bit strings, i.e., each bit in the result is set if and only if
    it is set in \verb"dest" and cleared in \verb"src".
    \verb"src" and \verb"dest" must have have equal size. 
    
    The result is stored in \verb"dest" and overwrites the previous value.
!seealso 
-------------------------------------------------------------------------- }}}*/

void bs_minus(bitstring_t *dest, const bitstring_t *src)

{	
    register int i;
    register unsigned long *dl = (unsigned long *)dest->buf;
    register unsigned long *sl = (unsigned long *)src->buf;
    register unsigned char *d;
    register unsigned char *s;

    for (i = bs_size; i > sizeof(long); i -= sizeof(long))
	*dl++ &= ~*sl++;
    d = (unsigned char *) dl;
    s = (unsigned char *) sl;
    for (; i > 0; --i)
	*d++ &= ~*s++;
}



/*{{{ --------------------------------------------------------------------------
!function bs_match "Intersection count"
!section bs
!synopsis 
    int bs_match(const bitstring_t *x, const bitstring_t *y);
!arg x
    First bit string.
!arg y
    Second bit string.
!return
    $0$, $1$ or $2$  (see below).
!description
    This function compares two bit strings of equal size looking for
    bits which are set in both strings. The return value is 0 if there
    are no matching bits, 1 if there is exactly one matching bit, and
    2 if there are 2 or more matching bits.
!seealso 
-------------------------------------------------------------------------- }}}*/

int bs_match(const bitstring_t *x, const bitstring_t *y)

{
    register int i, k = 0;
    register unsigned char *d = (unsigned char *)x->buf;
    register unsigned char *s = (unsigned char *)y->buf;
    register unsigned char t;

    for (i = bs_size; i > 0 && k < 2; i--)
    {
	if ((t = (unsigned char)(*s++ & *d++)) != 0)
	    k += matchcount[t];
    }
    return k;
}


/*{{{ --------------------------------------------------------------------------
!function bs_issub "Bit string incidence relation"
!section bs
!synopsis 
    int bs_issub(const bitstring_t *a, const bitstring_t *b);
!arg a
    First bit string.
!arg b
    Second bit string.
!return
    $1$ if $a\subseteq b$, 0 otherwise.
!description
    This function checks, if one bit string is contained in a
    second bit string. By definition, $a\subseteq b$ if and only if every
    bit which is set in $a$ is also set in $b$.
!seealso 
-------------------------------------------------------------------------- }}}*/

int bs_issub(const bitstring_t *a, const bitstring_t *b)

{	
    register int i;
    register unsigned long *al = (unsigned long *)a->buf;
    register unsigned long *bl = (unsigned long *)b->buf;
    register unsigned char *ac;
    register unsigned char *bc;

    for (i = bs_size; i > sizeof(long); i -= sizeof(long))
    {	
	if ((*al ^ (*al & *bl)) != 0)
	    return 0;
	al++;
	bl++;
    }
    ac = (unsigned char *) al;
    bc = (unsigned char *) bl;
    for (; i > 0; --i)
    {	
	if ((*ac ^ (*ac & *bc)) != 0)
	    return 0;
	ac++;
	bc++;
    }
    return 1;
}


