/* ============================= C MeatAxe ==================================
   File:        $Id: zcf.c,v 1.3 1998/04/28 17:37:53 mringe Exp $
   Comment:     Change field.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static char *iname, *oname;
FILE *inp, *out;
long fl;	/* Field parameter of input file */
long fl2;	/* Field to convert to */
long nor, noc;	/* Parameters of input file */

static char *helptext[] = {
"SYNTAX",
"    zcf [-QV] <Field> <Input> <Output>",
"",
"OPTIONS",
"    -Q     Quiet, no messages",
"    -V     Verbose, more messages",
"",
"ARGUMENTS",
"    <Field>    The field order of the output matrix",
"    <Input>    The input file, a Permutation or matrix",
"    <Output>   The output matrix",
NULL};


static proginfo_t pinfo =
   { "zcf", "Change Format", "$Revision: 1.3 $", helptext };





/* ------------------------------------------------------------------
   checkfl() - Check if GF(fl) < GF(fl2)
   ------------------------------------------------------------------ */

static int checkfl()

{	
    long f;
	
    if (fl2 < 2) 
	MtxErrorExit("Invalid field order %d",(int)fl2);
    if (fl == -1) 
	return 3;
    if (fl == fl2) 
	MtxErrorExit("%s is already over GF(%d)",iname,(int)fl);
    else if (fl < fl2)
    {	
	for (f = fl2; f % fl == 0; f /= fl);
	if (f == 1) 
	    return 1;
	else 
	    MtxErrorExit("Cannot change from GF(%d) to GF(%d)",(int)fl,(int)fl2);
    }
    else if (fl > fl2)
    {	
	for (f = fl; f % fl2 == 0; f /= fl2);
	if (f == 1) 
	    return 2;
	else 
	    MtxErrorExit("Cannot change from GF(%d) to GF(%d)",(int)fl,(int)fl2);
    }
    return 0;
}


/* ------------------------------------------------------------------
   permmat() - Convert permutation -> matrix
   ------------------------------------------------------------------ */

void permmat()

{	PTR m2;
	long *p, i;

    /* Read the permutation
       -------------------- */
    p = NALLOC(long,nor);
    if (p == NULL) 
	MtxErrorExit(NULL);
    if (zreadlong(inp,p,nor) != (size_t) nor)
	MtxErrorExit(NULL);
    fclose(inp);
    --p;

    /* Allocate one row
       ---------------- */
    zsetfield(fl2);
    zsetlen(nor);
    m2 = zalloc(1);

    /* Create the output matrix
       ------------------------ */
    if ((out = zwritehdr(oname,fl2,nor,nor)) == NULL)
	MtxErrorExit(NULL);
    for (i = 1; i <= nor; ++i)
    {	
	zmulrow(m2,F_ZERO);
	zinsert(m2,p[i],F_ONE);
	zwritevec(out,m2,1);
    }
    fclose(out);
    MESSAGE(0,("Converted to GF(%ld)\n",fl2));
}




/* ------------------------------------------------------------------
   restrictfield() - Restrict to subfield
   ------------------------------------------------------------------ */

void restrictfield()

{	
    PTR m1, x;
    long j, k;
    FEL *fp, *tmp = NULL;

    checkfl();
    zsetfield(fl);
    zsetlen(noc);

    if ((m1 = zalloc(nor)) == NULL
	|| (tmp = NALLOC(FEL,nor*noc)) == NULL
	|| zreadvec(inp,m1,(size_t) nor) != (size_t) nor
       )
	MtxErrorExit(NULL);
    fclose(inp);

    fp = tmp;
    x = m1;
    for (j = 1; j <= nor; ++j)
    {	
	for (k = 1; k <= nor; ++k)
	    *fp++ = zrestrict(zextract(x,k),fl2);
	zadvance(&x,(long)1);
    }

    free(m1);

    zsetfield(fl2);
    zsetlen(noc);
    if ((out = zwritehdr(oname,fl2,nor,noc)) == NULL)
	MtxErrorExit(NULL);
    m1 = zalloc((long)1);

    fp = tmp;
    for (j = 1; j <= nor; ++j)
    {
	zmulrow(m1,F_ZERO);
	for (k = 1; k <= nor; ++k)
	    zinsert(m1,k,*fp++);
	zwritevec(out,m1,(long)1);
    }
    fclose(out);
    MESSAGE(0,("Restricted to GF(%ld)\n",fl2));
}



/* ------------------------------------------------------------------
   matmat() - Convert matrix -> matrix
   ------------------------------------------------------------------ */

void matmat()

{
    PTR m1, x;
    long j, k;
    FEL *fp, *tmp = NULL;

    checkfl();
    zsetfield(fl); 
    zsetlen(noc);
    if ((m1 = zalloc(nor)) == NULL
	|| (tmp = NALLOC(FEL,nor*noc)) == NULL
	|| zreadvec(inp,m1,nor) != (size_t) nor
       )
	MtxErrorExit(NULL);
    fclose(inp);

    fp = tmp;
    x = m1;
    for (j = 1; j <= nor; ++j)
    {	
	for (k = 1; k <= nor; ++k)
	    *fp++ = zextract(x,k);
	zadvance(&x,(long)1);
    }

    free(m1);
    zsetfield(fl2);
    zsetlen(noc);
    if ((out = zwritehdr(oname,fl2,nor,noc)) == NULL
	|| (m1 = zalloc(1)) == NULL
       )
	MtxErrorExit(NULL);

    fp = tmp;
    for (j = 1; j <= nor; ++j)
    {	
	zmulrow(m1,F_ZERO);
	for (k = 1; k <= nor; ++k)
	    zinsert(m1,k,zembed(*fp++,fl));
	zwritevec(out,m1,1);
    }
    fclose(out);
    MESSAGE(0,("Embedded into GF(%ld)\n",fl2));
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    int i;


    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("")) != OPT_END)
    {
    }
    if (opt_ind != argc - 3) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    iname = argv[opt_ind+1];
    oname = argv[opt_ind+2];
    fl2 = atol(argv[opt_ind+0]);
    if ((inp = zreadhdr(iname,&fl,&nor,&noc)) == NULL)
	MtxErrorExit(NULL);

    /* Convert
       ------- */
    switch (checkfl())
    {	case 1:	/* Embed into larger field */
	    matmat();
	    break;
	case 2:
	    restrictfield();
	    break;
	case 3:
	    permmat();
	    break;
	default:
	    MtxErrorExit("Cannot change from GF(%d) to GF(%d)",
	        (int)fl,(int)fl2);
    }
    MtxCleanup();
    return EXIT_OK;
}

