/* ============================= C MeatAxe ==================================
   File:        $Id: zcp.c,v 1.7 1998/04/28 17:37:56 mringe Exp $
   Comment:     Characteristic polynomial of a matrix
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"




/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

char *fname;			/* Matrix file name */
matrix_t *mat;			/* The matrix */
int opt_G = 0;			/* GAP output */
int opt_f = 0;			/* Factorization */
int opt_m = 0;			/* Minimal polynomial */
fpoly_t *cpol;			/* Char. polynomial (with -f) */

static char *helptext[] = {
    "SYNTAX",
    "    zcp [-GQVfm] <File>",
    "",
    "OPTIONS",
    "    -G   GAP output",
    "    -Q   Quiet, no messages",
    "    -V   Verbose, more messages",
    "    -m   Calculate the minimal polynomial",
    "    -f   Factor the polynomial",
    "",
    "FILES",
    "    <File>    i  A square matrix",
    NULL
    };

static proginfo_t pinfo =
   { "zcp", "Characteristic and Minimal Polynomial",
     "$Revision: 1.7 $", helptext };




/* ------------------------------------------------------------------
   init()
   ------------------------------------------------------------------ */

static void init(argc, argv)
int argc;
char *argv[];

{	
    int i;
    FILE *f;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("Gmf")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 'f': opt_f = 1; break;
	    case 'm': opt_m = 1; break;
	}
    }
    if (opt_ind != argc - 1) 
	MtxErrorExit("%s: %E",pinfo.name,ERR_NARGS);
    fname = argv[opt_ind];

    /* Read the matrix
       --------------- */
    f = SFOpen(fname,FM_READ);
    if (f == NULL) 
	MtxErrorExit(NULL);
    mat = MatRead(f);
    fclose(f);
    if (mat == NULL) 
	MtxErrorExit(NULL);
    if (mat->nor != mat->noc) 
	MtxErrorExit("%s: %E",fname,ERR_NOTSQUARE);
}


/* ------------------------------------------------------------------
   write_init() - Called once before factors are written
   write_factor() - Write one factor
   write_end() - Called once after all factors have been written
   ------------------------------------------------------------------ */

static int first;

static void write_init()

{
    first = 1;
    if (opt_m)
    {
    	if (opt_G)
	    printf("MeatAxe.MinPol:=[\n");
    	else
	    printf("MINIMAL POLYNOMIAL:\n");
    }
    else
    {
    	if (opt_G)
	    printf("MeatAxe.CharPol:=[\n");
    	else
	    printf("CHARACTERISTIC POLYNOMIAL:\n");
    }
}


static void write_end()

{
    int i;

    if (opt_G)
	printf("];\n");
    else
    {
	if (opt_f)
	{
	    for (i = 0; i < (int) cpol->len; ++i)
    	    {
		printf("( ");
		PolPrint(NULL,cpol->p[i]);
		printf(" )^%ld\n",cpol->e[i]);
    	    }
	}
    }
}



static void write_one(pol)
poly_t *pol;

{ 
    if (opt_G)
    {
	int i;
	if (!first)
	    printf(",\n");
	printf("[");
	for (i = 0; i < pol->deg; ++i)
	    printf("%s,",zftogap(pol->buf[i]));
	printf("%s]",zftogap(F_ONE));
    }
    else if (!opt_f)
    {
	PolPrint(NULL,pol);
	printf("\n");
    }
    else
    {
	fpoly_t *f = PolFactorization(pol);
	FpolMul(cpol,f);
	FpolFree(f);
    }
    first = 0;
}




/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{	
    poly_t *p;

    init(argc,argv);
    if (opt_f) cpol = FpolAlloc();
    write_init();
    if (opt_m)
    {
    	for (p = MinPolFactor(mat); p != NULL; p = MinPolFactor(NULL))
	    write_one(p);
    }
    else
    {
    	for (p = CharPolFactor(mat); p != NULL; p = CharPolFactor(NULL))
	    write_one(p);
    }
    write_end();
    MtxCleanup();
    return EXIT_OK;
}


