/* ============================= C MeatAxe ==================================
   File:        $Id: zcv.c,v 1.4 1998/04/28 17:38:00 mringe Exp $
   Comment:     Convert a matrix or permutation from ASCII (readable)
                format to binary format.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#define MAXLINE 4000	/* Max. input line size */

#include "meataxe.h"
#include <ctype.h>
#include <string.h>
#include <stdlib.h>




/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static int GrpLibFormat = 0;	/* File is in Group Library Format */
static long fl;
static int mod;
static long nor,noc;
static FILE *src;			/* Input file */
static FILE *out;			/* Output file */
static char lbuf[MAXLINE] = {0};	/* Input line buffer */
static char *lptr = lbuf;		/* Read pointer */
static char *inpname;
static char *outname = "P2";
static int MemberCount = 0;		/* Number of members */

static char *helptext[] = {
"SYNTAX",
"    zcv [<Inp> [<Out>]]",
"",
"FILES",
"    <Inp>    i  Text file. <Inp> = `=' means standard input.",
"                Default is `T1'.",
"    <Out>    o  Binary file. Dafault name is `P2'.",
NULL};


static proginfo_t pinfo =
   { "zcv", "Convert to Binary Format", "$Revision: 1.4 $", helptext };






/* ------------------------------------------------------------------
   readline() - Read next input line, skip empty lines and strip
   	comments. Returns 0 on success, 1 on end-of-file.
   ------------------------------------------------------------------ */

static int readline()

{
    char *c;
    int mt = 1;

    while (mt)
    {	lbuf[0] = 0;
	if (feof(src)) return 1;
	fgets(lbuf,sizeof(lbuf),src);
	if (ferror(src)) 
	    MtxErrorExit("Unexpected end of input file");
	for (c = lbuf; *c != 0 && *c != '#'; ++c)
		if (!isspace(*c)) mt = 0;
	*c = 0;
    }
    lptr = lbuf;
    return 0;
}


/* ------------------------------------------------------------------
   readlong() - Read integer number
   ------------------------------------------------------------------ */

static long readlong()

{
    long l;
    int minus = 0;

    while (!isdigit(*lptr) && *lptr != '-')
    {
	while (*lptr != 0 && !isdigit(*lptr) && *lptr != '-') ++lptr;
	if (*lptr == 0)
	{
	    if (readline()) 
		MtxErrorExit("Unexpected end of input file");
	}
    }
    if (*lptr == '-') { minus = 1; ++lptr; }
    if (!isdigit(*lptr))
	MtxErrorExit("%s: Bad file format",inpname);
    for (l = 0; isdigit(*lptr); )
    {	
	l *= 10;
	switch (*lptr)
	{   case '0': break;
	    case '1': l += 1; break;
	    case '2': l += 2; break;
	    case '3': l += 3; break;
	    case '4': l += 4; break;
	    case '5': l += 5; break;
	    case '6': l += 6; break;
	    case '7': l += 7; break;
	    case '8': l += 8; break;
	    case '9': l += 9; break;
	}
	++lptr;
	if (mod != 2 && mod != 5 && fl >= 2 && fl <= 9) break;
    }
    return minus ? -l : l;
}


/* ------------------------------------------------------------------
   WriteHeader() - Write a header consisting of three integers
   ------------------------------------------------------------------ */

static void WriteHeader(long a, long b, long c)

{
    long hdr[3];
    hdr[0] = a;
    hdr[1] = b;
    hdr[2] = c;
    if (zwritelong(out,hdr,3) != 3)
	MtxErrorExit(NULL);
}


/* ------------------------------------------------------------------
   convmatrix() - Convert matrix in fixed format (mode 1)
   ------------------------------------------------------------------ */

static void convmatrix()

{	
    long i, j;
    PTR m1;
    long val = 0;
    int inp;

    if (fl > 9) 
	MtxErrorExit("Unexpected end of input file");
    zsetfield(fl); zsetlen(noc);
    m1 = zalloc((long)1);
    WriteHeader(fl,nor,noc);
    MESSAGE(0,("%ldx%ld matrix over GF(%ld)\n",nor,noc,fl));
    for (i = 1; i <= nor; ++i)
    {	zmulrow(m1,F_ZERO);
	inp = 81;
	for (j = 1; j <= noc; ++j)
	{	if (inp >= 80)	/* read next line */
		{	memset(lbuf,0,sizeof(lbuf));
			if (readline()) 
			    MtxErrorExit("Unexpected end of input file");
			inp = 0;
		}
		switch (lbuf[inp++])
		{	case ' ':
			case '0': val = 0; break;
			case '1': val = 1; break;
			case '2': val = 2; break;
			case '3': val = 3; break;
			case '4': val = 4; break;
			case '5': val = 5; break;
			case '6': val = 6; break;
			case '7': val = 7; break;
			case '8': val = 8; break;
			default: MtxErrorExit("%s: Bad file format",inpname);
		}
		if (val > fl) MtxErrorExit("%s: Bad file format",inpname);
		zinsert(m1,j,zitof(val));
	}
	zwritevec(out,m1,1);
    }
}


/* ------------------------------------------------------------------
   conv23456() - Convert matrix in free format (mode 3,4,5,6)
   ------------------------------------------------------------------ */

static void conv3456()

{	long i, j;
        PTR m1;
	long val;

    MESSAGE(0,("%ldx%ld matrix over GF(%ld)\n",nor,noc,fl));
    zsetfield(fl); zsetlen(noc);
    m1 = zalloc((long)1);
    WriteHeader(fl,nor,noc);
    for (i = 1; i <= nor; ++i)
    {	zmulrow(m1,F_ZERO);
	for (j = 1; j <= noc; ++j)
	{
	    val = readlong();
	    if (mod == 5)
	    {	val %= zchar;
		if (val < 0) val += zchar;
	    }
	    zinsert(m1,j,zitof(val));
	}
	zwritevec(out,m1,1);
    }
}


/* ------------------------------------------------------------------
   ConvertMatrix() - Convert matrix (new format)
   ------------------------------------------------------------------ */

static void ConvertMatrix()

{
    long i, j;
    PTR m1;
    long val;

    MESSAGE(0,("%ldx%ld matrix over GF(%ld)\n",nor,noc,fl));
    zsetfield(fl); zsetlen(noc);
    m1 = zalloc((long)1);
    WriteHeader(fl,nor,noc);
    for (i = 1; i <= nor; ++i)
    {
    	zmulrow(m1,F_ZERO);
	for (j = 1; j <= noc; ++j)
	{
	    val = readlong();
	    zinsert(m1,j,zitof(val));
	}
	zwritevec(out,m1,1);
    }
}


/* ------------------------------------------------------------------
   ConvertIntMatrix() - Convert integer matrix
   ------------------------------------------------------------------ */

static void ConvertIntMatrix()

{
    long i, j;
    long *x;

    MESSAGE(0,("%ldx%ld integer matrix\n",nor,noc));
    x = NALLOC(long,noc);
    WriteHeader(-T_IMAT,nor,noc);
    for (i = 1; i <= nor; ++i)
    {
	for (j = 0; j < noc; ++j)
	    x[j] = readlong();
	zwritelong(out,x,noc);
    }
    free(x);
}


/* ------------------------------------------------------------------
   ConvertPermutation() - Convert permutation (new format)
   ------------------------------------------------------------------ */

static void ConvertPermutation()

{
    long i;
    long *buf;
    long kk;


    MESSAGE(0,("Permutation on %ld points\n",nor));
    buf = NALLOC(long,nor);
    if (buf == NULL) 
	MtxErrorExit("Cannot allocate permutation: %S");
    WriteHeader(-1,nor,1);

    for (i = 0; i < nor; ++i)
    {
	kk = readlong();
	buf[i] = kk;
    }
    if (zwritelong(out,buf,nor) != (size_t) nor)
	MtxErrorExit("Cannot write to %s: %S",outname);
}


/* ------------------------------------------------------------------
   ConvertPolynomial() - Convert polynomial (new format)
   ------------------------------------------------------------------ */

static void ConvertPolynomial()

{
    long i;
    poly_t *p;

    MESSAGE(0,("Polynomial of degree %ld over GF(%ld)\n",nor,fl));
    zsetfield(fl);
    p = PolAlloc(fl,nor);
    if ((out = SFOpen(outname,FM_CREATE)) == NULL)
	MtxErrorExit("Cannot open %s: %S",outname);
    for (i = 0; i <= nor; ++i)
    {
	long kk = readlong();
	p->buf[i] = zitof(kk);
    }
    PolWrite(out,p);
}


/* ------------------------------------------------------------------
   convperm() - Convert permutation (mode 2)
   ------------------------------------------------------------------ */

void convperm()		/* mode 2 */

{
    long i, val;
    PTR m1;

    MESSAGE(0,("%ldx%ld permutation matrix over GF(%ld)\n",nor,noc,fl));
    zsetfield(fl); zsetlen(noc);
    m1 = zalloc((long)1);
    WriteHeader(fl,nor,noc);
    for (i = 1; i <= nor; ++i)
    {	val = readlong();
	zmulrow(m1,F_ZERO);
	zinsert(m1,val,F_ONE);
	zwritevec(out,m1,1);
    }
}


/* ------------------------------------------------------------------
   conv1213() - Convert permutation (mode 12, 13)
   ------------------------------------------------------------------ */

void conv1213()		/* modes 12, 13 */

{
    long nper, i;
    long *buf;
    long kk = 0, f1, f2;


    MESSAGE(0,("Permutation on %ld points\n",nor));
    buf = NALLOC(long,nor);
    if (buf == NULL) 
	MtxErrorExit("Cannot allocate permutation: %S");
    WriteHeader(-fl,nor,noc);
    for (nper = noc; nper != 0; --nper)
    {	
	for (i = 0; i < nor; ++i)
	{
	    switch (mod)
	    {	case 12:
			kk = readlong();
			break;
		case 13:
			f1 = readlong();
			f2 = readlong();
			kk = (f1-1)*fl + f2 + 1;
			break;
	    }
	    buf[i] = kk;
	}
	if (zwritelong(out,buf,nor) != (size_t) nor)
	    MtxErrorExit("Cannot write %s: %S",outname);
    }
}


/* ------------------------------------------------------------------
   ReadHeader() - Read the next header. Returns 0 on end of file,
   1 on success.
   ------------------------------------------------------------------ */

static int ReadHeader(void)

{
    if (readline())
    	return 0;
    return 1;
}


/* ------------------------------------------------------------------
   Convert() - Convert one member.
   ------------------------------------------------------------------ */

static void Convert(void)

{
    static char sfl[20], smode[20], snor[20], snoc[20];
    char *c;

    /* Check for new file format
       ------------------------- */

    if ((c = strstr(lbuf,"MeatAxeFileInfo")) != NULL)
    {
	char *d = lbuf;
	for (c += 15; *c != 0 && *c != '"'; ++c);
	if (*c != '"') MtxErrorExit("%s: Bad file format",inpname);
	for (++c; *c != '"' && *c != 0; ++c)
	    *d++ = *c;
	*d = 0;
	GrpLibFormat = 1;
    }

    if (!strncmp(lbuf,"matrix",6))
    {
    	char *c;
    	fl = nor = noc = -1;
    	for (c=strtok(lbuf+6," \t\n"); c!=NULL; c=strtok(NULL," \t\n"))
    	{
    	    if (!strncmp(c,"field=",6)) fl = atol(c+6);
    	    else if (!strncmp(c,"nor=",4)) nor = atol(c+4);
    	    else if (!strncmp(c,"rows=",5)) nor = atol(c+5);
    	    else if (!strncmp(c,"noc=",4)) noc = atol(c+4);
    	    else if (!strncmp(c,"cols=",5)) noc = atol(c+5);
    	    else MtxErrorExit("%s: Bad file format",inpname);
    	}
    	if (nor == -1 || noc == -1 || fl == -1) MtxErrorExit("%s: Bad file format",inpname);
    	readline();
    	ConvertMatrix();
    	return;
    }
    if (!strncmp(lbuf,"integer matrix",14))
    {
    	char *c;
    	fl = nor = noc = -1;
    	for (c=strtok(lbuf+14," \t\n"); c!=NULL; c=strtok(NULL," \t\n"))
    	{
    	    if (!strncmp(c,"nor=",4)) nor = atol(c+4);
    	    else if (!strncmp(c,"rows=",5)) nor = atol(c+5);
    	    else if (!strncmp(c,"noc=",4)) noc = atol(c+4);
    	    else if (!strncmp(c,"cols=",5)) noc = atol(c+5);
    	    else MtxErrorExit("%s: Bad file format",inpname);
    	}
    	if (nor == -1 || noc == -1) MtxErrorExit("%s: Bad file format",inpname);
    	readline();
    	ConvertIntMatrix();
    	return;
    }
    else if (!strncmp(lbuf,"permutation",11))
    {
    	char *c;
    	fl = nor = -1;
    	noc = 1;
    	for (c=strtok(lbuf+11," \t\n"); c!=NULL; c=strtok(NULL," \t\n"))
    	{
    	    if (!strncmp(c,"degree=",7)) nor = atol(c+7);
    	    else if (!strncmp(c,"deg=",4)) nor = atol(c+4);
    	    else MtxErrorExit("%s: Bad file format",inpname);
    	}
    	if (nor == -1) MtxErrorExit("%s: Bad file format",inpname);
    	readline();
    	ConvertPermutation();
    	return;
    }
    else if (!strncmp(lbuf,"polynomial",10))
    {
    	char *c;
    	fl = nor = -1;
    	noc = 1;
    	for (c=strtok(lbuf+11," \t\n"); c!=NULL; c=strtok(NULL," \t\n"))
    	{
    	    if (!strncmp(c,"degree=",7)) nor = atol(c+7);
    	    else if (!strncmp(c,"deg=",4)) nor = atol(c+4);
    	    else if (!strncmp(c,"field=",6)) fl = atol(c+6);
    	    else MtxErrorExit("%s: Bad file format",inpname);
    	}
    	if (nor < 0 || fl < 2) MtxErrorExit("%s: Bad file format",inpname);
    	readline();
    	ConvertPolynomial();
    	return;
    }

    /* Use old file format
       ------------------- */
    strncpy(smode,lbuf,2); smode[2] = 0;
    strncpy(sfl,lbuf+2,6); sfl[6] = 0;
    strncpy(snor,lbuf+8,6); snor[6] = 0;
    strncpy(snoc,lbuf+14,6); snoc[6] = 0;
    if (sscanf(lbuf,"%d%ld%ld%ld",&mod,&fl,&nor,&noc) != 4)
    {
        sscanf(smode,"%d",&mod);
        sscanf(sfl,"%ld",&fl);
        sscanf(snor,"%ld",&nor);
        sscanf(snoc,"%ld",&noc);
    }

    if (mod != 1) readline();
    switch (mod)
    {
	case 1:
	    convmatrix();
	    break;
	case 2:
	    convperm();
	    break;
	case 3:
	case 4:
	case 5:
	case 6:
	    conv3456();
	    break;
	case 12:
	case 13:
	    conv1213();
	    break;
	default:
		MtxErrorExit("%s: Unknown mode",inpname);
    }
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{	
    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while (MtxGetOpt("") != OPT_END);

    switch (argc-opt_ind)
    {	case 2:
	    outname = argv[opt_ind + 1];
	case 1:
	    if (!strcmp(argv[opt_ind],"="))
	    {
		inpname = "[stdin]";
		src = stdin;
	    }
	    else
	    {
		inpname = argv[opt_ind];
		src = SFOpen(argv[opt_ind],FM_READ|FM_TEXT|FM_LIB);
		if (src == NULL)
		    MtxErrorExit("Cannot open %s: %S",inpname);
	    }
	    break;
	case 0:
	    inpname = "T1";
	    src = SFOpen("T1",FM_READ|FM_TEXT);
	    if (src == NULL)
    		MtxErrorExit("Cannot open %s: %S",inpname);
	    break;
	default:
	    MtxErrorExit("Bad usage, try '%s -help'",pinfo.name);
    }

    /* Open the output file
       -------------------- */
    out = SFOpen(outname,FM_CREATE);
    if (out == NULL)
	MtxErrorExit("Cannot open %s for output",outname);

    /* Convert all members
       ------------------- */
    while (ReadHeader())
    {
    	Convert();
	++MemberCount;
    }
    if (MemberCount == 0)
    	MESSAGE(0,("Warning: nothing found in %s",inpname));

    fclose(out);
    MtxCleanup();
    return (EXIT_OK);
}

