/* ============================= C MeatAxe ==================================
   File:        $Id: zfile.c,v 1.4 1998/04/28 17:38:10 mringe Exp $
   Comment:     Stream based i/o for simple data types.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


static long test = 1;


/*{{{ --------------------------------------------------------------------------
!function zreadlong "Read long integers"
!section kernel.io
!synopsis 
    size_t zreadlong(FILE *f, long* buf, size_t n);
!arg f
    File to read from.
!arg buf
    Pointer to buffer.
!arg n
    Number of integers to read.
!return
    Number of integers that were actually read. A return value different from
    |n| indicates an error.
!description
    This function reads |n| long integers from the file |f| into the array 
    |buf|. |buf| must point to a memory area of at least |n*sizeof(long)| 
    bytes and |f| must be open for reading. 
    
    This function expects that the numbers in the file are 4-byte integers 
    in `little-endian' format, i.e. the least significant byte first.
    Using a machine-independent data format makes {\MeatAxe} data files 
    more portable, but there are also some disadvantages:
    \begin{enumerate}
    \item 
	The conversion to and from machine-independent format involves several
	arithmetic operations for each number read/written.
    \item
	The highest number which can be read/written is $2^{32}-1$, even on
	64 bit machines.
    \end{enumerate}
!seealso zwritelong
-------------------------------------------------------------------------- }}}*/


size_t zreadlong(FILE *f, long* buf, size_t n)

{
    MTXCONTEXT(zreadlong);
    unsigned char a[4];
    unsigned nread;

    if (sizeof(long) == 4 && *(char *)&test == 1)
	nread = fread((char *)buf,sizeof(long),n,f);
    else
    {
    	for (nread = 0; nread < n; ++nread)
    	{
	    if (fread(a,1,4,f) != 4) break;
	    *buf = ((unsigned long)a[0]|((unsigned long)a[1] << 8)|
		((unsigned long)a[2] << 16)|((long)(char)a[3] << 24));
	    ++buf;
    	}
    }
    if (ferror(f))
	MTXERROR("Read failed: %S");
    return nread;
}



/*{{{ --------------------------------------------------------------------------
!function zwritelong "Write long integers"
!section kernel.io
!synopsis 
    size_t zwritelong(FILE *f, const long *buf, size_t n);
!arg f
    File to write to.
!arg buf
    Pointer to buffer.
!arg n
    Number of integers to write.
!return
    Number of integers that were actually written. A return value different 
    from |n| indicates an error.
!description
    This function writes |n| long integers from the the array |buf| to the 
    file |f|. |buf| must point to a memory area of at least |n*sizeof(long)| 
    bytes and |f| must be open for writing. The numbers are written in a 
    machine-independent format which can be read by |zreadlong()|.

!seealso zreadlong
-------------------------------------------------------------------------- }}}*/

size_t zwritelong(FILE *f, const long *buf, size_t n)

{
    unsigned char a[4];
    unsigned nwritten;

    if (sizeof(long) == 4 && *(char *)&test == 1)
	nwritten = fwrite((char *)buf,sizeof(long),n,f);
    else
    {
    	for (nwritten = 0; nwritten < n; ++nwritten)
    	{
    	    a[0] = (unsigned char) *buf;
    	    a[1] = (unsigned char) (*buf >> 8);
    	    a[2] = (unsigned char) (*buf >> 16);
    	    a[3] = (unsigned char) (*buf >> 24);
    	    if (fwrite(a,1,4,f) != 4) break;
	    ++buf;
    	}
    }
    return nwritten;
}


/*{{{ --------------------------------------------------------------------------
!function zreadvec "Read rows"
!section kernel.io
!synopsis 
    size_t zreadvec(FILE *f, PTR buf, size_t n);
!arg f
    Pointer to File.
!arg buf
    Pointer to data buffer.
!arg n
    Number of rows to read.
!return
    The function returns the number of rows that were successully read.
    A return value different from |n| indicates an error.
!description
    This function reads 1 or more rows from a file.
    The row size must have been set before.
!seealso zseek zwritevec
-------------------------------------------------------------------------- }}}*/

size_t zreadvec(FILE *f, PTR buf, size_t n)

{
    MTXCONTEXT(zreadvec);
    size_t i;
    register char *b = (char *) buf;

    for (i = 0; i < n; ++i)
    {
        if (fread(b,zrowsize_io,1,f) != 1) break;
	b += zrowsize;
    }
    if (ferror(f)) 
	MTXERROR("Read failed: %S");
    return i;
}


/*{{{ --------------------------------------------------------------------------
!function zwritevec "Write rows"
!section kernel.io
!synopsis 
    size_t zwritevec(FILE *f, PTR  buf, size_t n);
!arg f
    Pointer to File.
!arg buf
    Pointer to data buffer.
!arg n
    Number of rows to write.
!return
    The function returns the number of rows that were successully written.
    A return value different from |n| indicates an error.
!description
    This function writes 1 or more rows to a file.
    The row size must have been set before.
!seealso zseek zreadvec
-------------------------------------------------------------------------- }}}*/

size_t zwritevec(FILE *f, PTR  buf, size_t n)

{
    MTXCONTEXT(zwritevec);
    size_t i;
    register char *b = (char *) buf;

    for (i = 0; i < n; ++i)
    {
        if (fwrite(b,zrowsize_io,1,f) != 1) break;
	b += zrowsize;
    }
    if (ferror(f)) 
	MTXERROR("Write failed: %S");
    return i;
}


/*{{{ --------------------------------------------------------------------------
!function zseek "Open file and read header"
!section kernel.io
!synopsis 
    int zseek(FILE *f, long pos);
!arg f
    Pointer to File.
!arg pos
    Row number (1-based).
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function sets the read/write pointer of file \verb"f" to position 
    \verb"pos". I.e., the next |zreadvec()| or |zwritevec()| will start at row
    |pos|. Note that row numbers start with 1, not 0. If |pos| is different 
    from 1, the row size must have been set before with |zsetlen()|.
    
    You should always use |zseek()| rather than |fseek()| because |zseek()| 
    knows about {\MeatAxe} file headers and adjusts the file pointer 
    appropriately. 
!seealso zreadvec zwritevec
-------------------------------------------------------------------------- }}}*/

int zseek(FILE *f, long pos)

{
    MTXCONTEXT(zseek);

    long addr;

    if (zfl != -1)
	addr = zrowsize_io * (pos - 1) + 12;
    else
	addr = znoc * 4 * (pos - 1) + 12;
    if (SFSeek(f,addr) == -1)
    {
	MTXERROR("Seek failed: %S");
	return -1;
    }
    return 0;
}




/* ------------------------------------------------------------------
   swapbytes() - Swap bytes in a 4-byte block: ABCD -> DCBA
   ------------------------------------------------------------------ */

static void swapbytes(long *l)

{	char inp[4], out[4];

	memcpy(inp,l,4);
	out[0] = inp[3];
	out[1] = inp[2];
	out[2] = inp[1];
	out[3] = inp[0];
	memcpy(l,out,4);
}



/*{{{ --------------------------------------------------------------------------
!function zreadhdr "Open file and read header"
!section kernel.io
!synopsis 
    FILE *zreadhdr(const char *name, long *field, long *nrows, long *ncols);
!arg name
    File name.
!arg field
    Pointer to buffer for first header entry (usually the field order).
!arg nor
    Pointer to buffer for second header entry (usually the number of rows).
!arg noc
    Pointer to buffer for third header entry (usually the number of columns).
!return
    Pointer to open file, or |NULL| on error.
!description
    This function opens a data file for input and reads the file header
    (3 long integers). The exact meaning of the header values three numbers 
    depends on the file type. For a matrix they are `field order', `number 
    of rows' and `number of columns'.
!seealso zwritehdr
-------------------------------------------------------------------------- }}}*/

FILE *zreadhdr(const char *name, long *field, long *nrows, long *ncols)

{
    MTXCONTEXT(zreadhdr);
    FILE *fd;
    long header[3];

    /* Open the file
       ------------- */
    fd = SFOpen(name,FM_READ);
    if (fd == NULL)
	return NULL;

    /* Read the file header
       -------------------- */
    if (zreadlong(fd,header,3) != 3)
    {
	fclose(fd);
	MTXERROR1("%s: Error reading file header",name);
	return NULL;
    }

    /* Try to correct wrong byte ordering.
       ----------------------------------- */
    if (header[0] > 256 || header[1] < 0 || header[2] < 0)
    {	
	swapbytes(header);
	swapbytes(header+1);
	swapbytes(header+2);
	if (header[0] > 256 || header[1] < 0 || header[2] < 0)
	{
	    MTXERROR1("%s: Invalid header, possibly non-MeatAxe file",name);
    	    fclose(fd);
	    return NULL;
	}
    }

    /* Store header values in user-supplied buffers
       -------------------------------------------- */
    *field = header[0];
    *nrows = header[1];
    *ncols = header[2];

    return fd;
}



/*{{{ --------------------------------------------------------------------------
!function zwritehdr "Open file and write header"
!section kernel.io
!synopsis 
    FILE *zwritehdr(const char *name, long field, long nrows, long ncols):
!arg name
    File name.
!arg field
    First header entry (usually the field order).
!arg nor
    Second header entry (usually the number of rows).
!arg noc
    Third header entry (usually the number of columns).
!return
    Pointer to open file, or |NULL| on error.
!description
    This function opens a data file for input and writes the the file header.
    If the file does not exist, a new file is created. If the file exists it
    is overwritten.
!seealso zreadhdr
-------------------------------------------------------------------------- }}}*/

FILE *zwritehdr(const char *name, long field, long nrows, long ncols)

{
    MTXCONTEXT(zwritehdr);
    FILE *fd;
    long header[3];

    /* Open the file
       ------------- */
    fd = SFOpen(name,FM_CREATE);
    if (fd == NULL) 
	return NULL;

    /* Write the file header
       --------------------- */
    header[0] = (long) field;
    header[1] = (long) nrows;
    header[2]= (long) ncols;
    if (zwritelong(fd,header,3) != 3)
    {
	MTXERROR1("%s: Error writing file header",name);
	fclose(fd);
	return NULL;
    }

    return fd;
}



/*{{{ --------------------------------------------------------------------------
!text kernel.io

\paragraph{Binary file format}
Binary data files contain a sequence of objects. Each object consists of a 
12-byte header, followed by the data part. The data is stored in the internal 
format (see section \ref{GdSection:kernel.ff}) with the following 
differences:
\begin{itemize}
\item
    Packed rows (vectors) are truncated to their actual length. For
    example, a row with 17 entries over GF(125) on a machine with
    ${\tt sizeof(long)}=4$ would occupy 20 bytes in memory but only
    17 bytes in a data file.
\item
    All integers, including the file header, are stored as 32-bit
    numbers with least significant byte first.
\end{itemize}

The object header consists of three integers. The first one, $T$, specifies
which kind of object follows. The two other values, $R$ and $C$, carry
additional information about the object. Their exact meaning depends on
the value of $T$.
\begin{center}\begin{tabular}{|l|l|l|l|}
\hline
$T$ & Type of object	    & $R$ 		& $C$ \\
\hline
> 1 & Matrix over $\GF{T}$  & no.\ of rows	& no.\ of columns \\
-1  & Permutation	    & no.\ of points	& no.\ of permutations \\
-2  & Polynomial	    & field		& degree \\
-8  & Integer matrix	    & no.\ of rows	& no.\ of columns \\
\hline
\end{tabular}
\end{center}
-------------------------------------------------------------------------- }}}*/
