/* ============================= C MeatAxe ==================================
   File:        $Id: zgauss.c,v 1.2 1998/03/11 17:04:45 mringe Exp $
   Comment:     Gauss elimination and related functions.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>



/* ------------------------------------------------------------------
   zmkpivot() - Find the pivot columns of a matrix. <matrix> is a
	<nor> by <noc> matrix in echelon form. The pivot columns
	are stored in <piv>, beginning with piv[1]. Thus, <piv> must
	be large enough to hold at least <nor>+1 entries.

   Return value:
      0  = ok
      -1 = Error (out of memory or matrix not in echelon form)
   ------------------------------------------------------------------ */

/*{{{ --------------------------------------------------------------------------
!function zmkpivot "Build pivot table"
!section kernel.mat
!synopsis 
    int zmkpivot(PTR matrix, long nor, long *piv);
!arg matrix
    Pointer to the matrix.
!arg nor
    Number of rows.
!arg piv
    Pointer to pivot table buffer.
!return
    The return value is $0$ on success, or $-1$ on error.
!description
    This function is used to find the pivot elements of a matrix which is 
    already in semi-echelon form (see |zmkechelon()|). |nor| is the number
    of rows in \verb"matrix". The number of columns must have been set
    before with |zsetlen()|. |piv| must be a pointer to an array of at 
    least $|nor|+1$ long integers. The pivot positions are stored here, 
    beginning with \verb"piv[1]" for the first row.

    The pivot table is used by some functions like |zcleanrow()|
    which perform Gaussian elimination.
!seealso zmkechelon
-------------------------------------------------------------------------- }}}*/

int zmkpivot(PTR matrix, long nor, long *piv)

{
    MTXCONTEXT(zmkpivot);
    PTR x;
    long i;
    FEL f;
    static unsigned char *ispiv = NULL;
    static long maxnoc;

    if (ispiv == NULL || znoc > maxnoc)
    {
	if (ispiv != NULL) 
	    free(ispiv);
	ispiv = NALLOC(unsigned char,znoc+1);
	maxnoc = znoc;
    }
    for (i = 1; i <= znoc; ++i) 
	ispiv[(int)i] = 0;
    for (i=1, x=matrix; i<=nor; ++i, zadvance(&x,(long)1))
    {
	long pivi = zfindpiv(x,&f);
	if (pivi == 0 || ispiv[pivi])
	    return MTXERROR1("%E",ERR_NOTECH), -1;
	piv[i] = pivi;
	ispiv[pivi] = 1;
    }
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!function zcleanrow "Clean row"
!section kernel.rowop
!synopsis 
    int zcleanrow(PTR row, PTR matrix, long nor, const long *piv);
!arg row
    Pointer to row to be cleaned.
!arg matrix
    Matrix to clean with.
!arg nor
    Number of rows.
!arg piv
    Pivot table for |matrix|.
!return
    The return value is always $0$.
!description
    This function `cleans' a row with a matrix, i.e., it adds suitable multiples 
    of the rows of |matrix| to |row| such that all pivot positions in |row| are 
    zero. |nor| is the number of rows in |matrix|. The number of columns (for 
    both |matrix| and |row|) must have been set before with |zsetlen()|. 
    
    |piv| is a pointer to the pivot table of |matrix|, which can be generated 
    |zmkpivot()|.
!seealso zcleanrow2
-------------------------------------------------------------------------- }}}*/

int zcleanrow(PTR row, PTR matrix, long nor, const long *piv)

{
    long i;
    PTR x;

    for (i=1, x=matrix; i<=nor; ++i, zadvance(&x,(long)1))
    {
	FEL f = zextract(row,piv[i]);
	if (f == F_ZERO) 
	    continue;
	f = zdiv(f,zextract(x,piv[i]));
	zaddmulrow(row,x,zneg(f));
    }

    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function zcleanrow2 "Clean row"
!section kernel.rowop
!synopsis 
    int zcleanrow(PTR row, PTR matrix, long nor, const long *piv);
!arg row
    Pointer to row to be cleaned.
!arg matrix
    Matrix to clean with.
!arg nor
    Number of rows.
!arg piv
    Pivot table for |matrix|.
!return
    The return value is always $0$.
!description
    This function acts exactly as |zcleanrow()|. In addition, it stores the 
    operations performed in \verb"row2". \verb"row2" must be a row of at 
    least \verb"nor" entries. It must be initialized to zero by the caller, 
    or the results are not defined. 
    
    On return, \verb"row2" contains the coefficients by which the rows
    of \verb"matrix" were multiplied and then subtracted from \verb"row".
!seealso zcleanrow
-------------------------------------------------------------------------- }}}*/

int zcleanrow2(PTR row, PTR matrix, long nor, const long *piv, PTR row2)

{
    long i;
    PTR x;

    for (i=1, x=matrix; i<=nor; ++i, zadvance(&x,1))
    {
	FEL f = zextract(row,piv[i]);
	if (f == F_ZERO) 
	    continue;
	f = zdiv(f,zextract(x,piv[i]));
	zaddmulrow(row,x,zneg(f));
	zinsert(row2,i,f);
    }
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function zmkechelon "Reduce to echelon form"
!section kernel.mat
!synopsis 
    long zmkechelon(PTR matrix, long nor, long *piv);
!arg matrix
    Matrix to reduce.
!arg nor
    Number of rows.
!arg piv
    Pointer to pivot table buffer.
!return
    This function returns the rank of |matrix|.
!description
    This function converts a matrix to semi-echelon form and builds a pivot 
    table suitable for use with |zcleanrow()|. A matrix is said to be in
    \begin{itemize}
    \item {\em semi-echelon form} if each row contains a first
	  nonzero element (pivot element), and all entries below
	  this element are zero.
	  \index{semi-echelon form}
    \item {\em normalized semi-echelon form} if it is in semi-echelon form
	  and each pivot element is equal to 1.
	  \index{echelon form}
    \end{itemize}
    In the {\MeatAxe}, matrices are often stored in semi-echelon form.
    Many functions and programs expect their arguments to be in
    semi-echelon form.

    \verb"matrix" is a matrix of \verb"nor" rows. The number of columns
    must have been set with \verb"zsetlen()". \verb"piv" must be a pointer
    to an array of at least \verb"nor+1" long integers.
    On return, \verb"matrix" is in semi-echelon form. Zero rows are moved
    to the end, and the number of non-zero rows (i.e., the rank) is
    stored in \verb"piv[0]". The rest of the \verb"piv" array contains the
    pivot table, beginning with \verb"piv[1]" for the first row.
    The return value is the rank.
!seealso zmkpivot
-------------------------------------------------------------------------- }}}*/

long zmkechelon(PTR matrix, long nor, long *piv)

{
    PTR x, newrow;
    long i, rank;

    rank = 0;
    newrow = matrix;
    for (i = 1, x = matrix; i <= nor; ++i, zadvance(&x,(long)1))
    {
	long newpiv;
	FEL f;

	if (rank < i-1) 
	    zmoverow(newrow,x);
	zcleanrow(newrow,matrix,rank,piv);
	newpiv = zfindpiv(newrow,&f);
	if (newpiv != 0)
	{
	    ++rank;
	    piv[rank] = newpiv;
	    zadvance(&newrow,(long)1);
	}
    }
    piv[0] = rank;
    return rank;
}
