/* ============================= C MeatAxe ==================================
   File:        $Id: zmatinv.c,v 1.2 1998/03/11 17:04:47 mringe Exp $
   Comment:     Matrix inversion
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"


/*{{{ --------------------------------------------------------------------------
!function zmatinv "Matrix inversion"
!section kernel.mat
!synopsis 
    int zmatinv(PTR mat, PTR result);
!arg mat
    Pointer to the matrix.
!arg result
    Pointer to result buffer.
!return
    The function returns 0 on success and -1 on error.
!description
    This function calculates the inverse of a square matrix.
    This is a is a low-level, destructive matrix inversion which,
    unlike |matinv()|, does not allocate a temporary matrix.

    Both |mat| and |result| must be square matrices, and the row size must 
    be set by the caller. |result| need not be initialized. A return value of 
    0 indicates that the function was successfull and the inverse matrix is 
    now in |result|. If the matrix is singular, the function returns -1. In 
    any case the original matrix is destroyed.
!seealso MatInv
-------------------------------------------------------------------------- }}}*/

int zmatinv(PTR mat, PTR result)

{
    MTXCONTEXT(zmatinv);
    PTR xj1, xj2, xk1, xk2;
    FEL f1 = F_ZERO, f2;
    long j, k;

    /* Initialize result = identity matrix
       ----------------------------------- */
    for (j = 1, xj1 = result; j <= znoc; ++j, zadvance(&xj1,1))
    {
	zmulrow(xj1,F_ZERO);
	zinsert(xj1,j,F_ONE);
    }

    /* Matrix inversion
       ---------------- */
    xj1 = mat;
    xj2 = result;
    for (j = 1; j <= znoc; ++j)
    {

        for (xk1 = xj1, k = j;
	     k <= znoc && (f1 = zextract(xk1,j)) == F_ZERO;
	     ++k, zadvance(&xk1,(long)1));
	if (f1 == F_ZERO) 
	    return MTXERROR1("%E",ERR_DIV0), -1;
	if (k > j)	/* Swap rows */
	{
	    zswaprow(xk1,xj1);
	    xk2 = xj2;
	    zadvance(&xk2,k-j);
	    zswaprow(xk2,xj2);
	}
	f2 = zinv(f1);
	zmulrow(xj1,f2);
	zmulrow(xj2,f2);
	xk1 = mat;
	xk2 = result;
	for (k = 1; k <= znoc; ++k)
	{
	    if (k != j)
	    {
		f1 = zneg(zextract(xk1,j));
		zaddmulrow(xk1,xj1,f1);
		zaddmulrow(xk2,xj2,f1);
	    }
	    zadvance(&xk1,1);
	    zadvance(&xk2,1);
	}
	zadvance(&xj1,1);
	zadvance(&xj2,1);
    }
    return 0;
}

