/* ============================= C MeatAxe ==================================
   File:        $Id: zmo.c,v 1.3 1998/04/28 17:38:18 mringe Exp $
   Comment:     Make orbits under permutations
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>
#include <string.h>


#define MAXPERM 50		/* Max. number of permutations */
#define MAXLEV 100000
#define MAXNORB 20000


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

long *perm[MAXPERM];			/* Permutations */
long sstt[201], ors[MAXNORB+1];
long fl, nor;
char *orbname, *sizname;
int nperm = 0;

static char *helptext[] = {
"SYNTAX",
"    zmo [<Options>] <Perm1> <Perm2> <Orbits> <OrbSz>",
"    zmo [<Options>] [-g <#Perms>] <Perms> <Orbits> <OrbSz>",
"",
"OPTIONS",
"    -Q         Quiet, no messages",
"    -V         Verbose output",
"    -g         Set number of permutations.",
"    -s <Seed>  Set seed (default:1)",
"",
"FILES",
"    <Perm1,2>  i  Permutations",
"    <Orbits>   o  Orbits under the permutations (Permutation format)",
"    <OrbSz>    o  Orbit sizes (Permutation format)",
"",
NULL};

static proginfo_t pinfo =
   { "zmo", "Make Orbits", "$Revision: 1.3 $", helptext };




/* ------------------------------------------------------------------
   err() - Print error message and exit
   ------------------------------------------------------------------ */

static void err(int c)

{
    fprintf(stderr,"ZMO ERROR - ");
    switch (c)
    {
	case 'o':
	    fprintf(stderr,"MORE THAN %ld ORBITS\n",(long)MAXNORB);
	    break;
	}
	exit(EXIT_ERR);
}


/* ------------------------------------------------------------------
   ReadPerm() - Read a permutation
   ------------------------------------------------------------------ */

static void ReadPerm(int i, char *fn)

{
    MTXCONTEXT(ReadPerm);
    FILE *f;
    static char n1[200];
    static int first = 1;
    static long nor1;
    long nop;

    if ((f = zreadhdr(fn,&fl,&nor,&nop)) == NULL)
	MtxErrorExit(NULL);
    if (fl != -1)
	MtxErrorExit("%s: %E",fn,ERR_NOTPERM);
    MESSAGE(1,("Reading %s: Degree = %ld\n",fn,nor));
    if (first)
    {
	first = 0;
	strcpy(n1,fn);
	nor1 = nor;
    }
    else if (nor != nor1)
	MtxErrorExit("%s and %s: %E",fn,n1,ERR_INCOMPAT);
    if (nop > 1)
	fprintf(stderr,"%s: Only 1 of %ld permutations read\n",fn,nop);
    
    perm[i] = NALLOC(long,nor);
    if (perm[i] == NULL)
    	MtxErrorExit("%s: %E",pinfo.name,ERR_NOMEM);
    if (zreadlong(f,perm[i],nor) != (size_t) nor)
    {
	MTXERROR1("Cannot read %s",fn);
	MtxErrorExit(NULL);
    }
    fclose(f);
    perm[i] = perm[i] - 1;
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    long i, seed = 1, nout, siz, dis, norb;
    long *stk, *mout = NULL;
    long f1, lev;
    FILE *f;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("g:s:")) != OPT_END)
    {
	switch (i)
	{
	    case 'g':
		nperm = (int) MtxGetInt();
		if (nperm < 1 || nperm > MAXPERM || *opt_text_ptr != 0)
		    MtxErrorExit("-g: %E",ERR_OPTION);
		break;
	    case 's':
		seed = MtxGetInt();
		if (seed == GETINT_ERR || *opt_text_ptr != 0)
		    MtxErrorExit("-s: %E",ERR_OPTION);
		break;
	}
    }

    if (nperm == 0)	/* Two generators */
    {
    	if (opt_ind != argc-4) 
	    MtxErrorExit("%s: %E",pinfo.name,ERR_NARGS);
	nperm = 2;
	for (i = 0; i < 2; ++i)
	    ReadPerm(i,argv[opt_ind+i]);
    }
    else		/* N generators */
    {
	char n1[200];
    	if (opt_ind != argc-3)
	    MtxErrorExit("%s: %E",pinfo.name,ERR_NARGS);
	for (i = 0; i < nperm; ++i)
	{
	    sprintf(n1,"%s.%ld",argv[opt_ind],i+1);
	    ReadPerm(i,n1);
	}
    }
    orbname = argv[argc-2];
    sizname = argv[argc-1];

    if (seed < 0 || seed > nor)
	MtxErrorExit("Seed %d is out of range (1-%d)",(int)seed,(int)nor);
    if ((stk = NALLOC(long,MAXLEV)) == NULL
	|| (mout = NALLOC(long,250)) == NULL
       )
	MtxErrorExit(NULL);
    if ((f = zwritehdr(orbname,fl,nor,(long)1)) == NULL)
	MtxErrorExit(NULL);
    norb = 0;
    nout = 0;
    while (seed <= nor)
    {
	if (norb >= MAXNORB) 
	    err('o');
	f1 = perm[0][seed];
	if (f1 < 0) 
	    MtxErrorExit("Internal error: invalid seed %d",(int)f1);
	perm[0][seed] = -f1;	/* Mark as done */
	siz = 0;
	stk[lev = 1] = seed;	/* Put first point on stack */

	while (lev > 0)		/* Stack not empty */
	{
	    f1 = stk[lev--];	/* Take from stack */
	    if (nout == 250)	/* If buffer full, write out */
	    {	
		if (zwritelong(f,mout,250) != 250)
		    MtxErrorExit(NULL);
		nout = 0;
	    }
	    MESSAGE(2,("mout[%ld]=%ld\n",(long)nout,(long)f1));
	    mout[nout++] = f1;	/* Put into buffer */
	    ++siz;

	    /* Apply all permutations and collect new points
	       --------------------------------------------- */
	    for (i = 0; i < nperm; ++i)
	    {	
	    	register long f3, f5;
		f3 = perm[i][f1];
		if (f3 < 0) f3 = -f3;
		f5 = perm[0][f3];
		if (f5 <= 0) continue;	/* Already done, ignore */
		perm[0][f3] = -f5;	/* Mark as done */
		if (lev >= MAXLEV - 1)
		    MtxErrorExit("Stack overflow");
		stk[++lev] = f3;	/* Put on stack */
	    }
	}
	ors[norb++] = siz;

	/* find next unused point */
	for (seed = 1; seed <= nor && perm[0][seed] < 0; ++seed);
    }

    /* Flush buffer and close file
       --------------------------- */
    if (zwritelong(f,mout,nout) != (size_t)nout)
   	 MtxErrorExit(NULL);
    fclose(f);
 

    /* Write the orbit sizes file
       -------------------------- */
    MESSAGE(1,("Writing orbit sizes to %s\n",sizname));
    if ((f = zwritehdr(sizname,-T_IMAT,1,norb)) == NULL)
	MtxErrorExit(NULL);
    if (zwritelong(f,ors,norb) != (size_t)norb)
   	MtxErrorExit(NULL);
    fclose(f);

    /* Print orbit sizes
       ----------------- */
    if (msg_level >= 0)
    {
	dis = 0;
	for (i = 0; dis < 200 && i < norb; ++i)
	{
	    int k;
	    siz = ors[i];

	    for (k = 0; k < dis && ors[k] != siz; ++k);
	    if (k < dis) ++sstt[k];
	    else if (dis++ < 200)
	    {
		sstt[k] = 1;
		ors[k] = siz;
	    }
	}
	if (dis > 200)
	    printf("More than 200 distinct orbit sizes\n");
	else if (dis > 15)
	    printf("%ld distinct orbit sizes\n", (long) dis);
	else
	    for (i = 0; i < dis; ++i)
		printf("%6ld ORBIT%c OF SIZE %6ld\n",
		    sstt[i],sstt[i] > 1 ? 'S':' ',ors[i]);
    }
    MtxCleanup();
    return (EXIT_OK);
}
