/* ============================= C MeatAxe ==================================
   File:        $Id: zmu.c,v 1.4 1998/04/28 17:38:22 mringe Exp $
   Comment:     Multiply matrices or permutations.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>
#include <string.h>




/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

static char *helptext[] = {
"SYNTAX",
"    zmu [-r <Row>[.<#Rows>]] [-c <Col>[.<#Cols>]] <A> <B> <Result>",
"",
"OPTIONS",
"    -r and -c can be used for piecewise matrix multiplication.",
"    E.g., `-r1 -c2' selects the upper right part. By default,",
"    <#Rows> = <#Cols> = 2, i.e., the result is divided into",
"    four pieces.",
"",
"FILES",
"    <A> and <B> are the objects to be multiplied. Their product",
"    (A*B) is written to <Result>. Compatible data types are:",
"",
"        M(a,b) * M(b,c)                   = M(a,c)",
"        M(1,1) * M(a,b) = M(a,b) * M(1,1) = M(a,b)",
"        P(a) * P(b)                       = P(max {a,b})",
"        M(a,b) * P(b)                     = M(a,b)",
"        P(a) * M(a,b)                     = M(a,b)",
"",
"    M(a,b) means `a by b matrix' and P(a) `Permutation of degree a'",
NULL};

static proginfo_t pinfo =
   { "zmu", "Multiply", "$Revision: 1.4 $", helptext };


static char *aname, *bname, *cname;	/* File names */
static FILE *afile, *bfile, *cfile;	/* File handles */
static long fl1, fl2;			/* Field order */
static long nor1, noc1, nor2, noc2;	/* Matrix dimensions */
int nrows = 1, thisrow = 1;		/* Arguments to -r option */
int ncols = 1, thiscol = 1;		/* Arguments to -c option */




/* ------------------------------------------------------------------
   multpm() - Multiply permutation by matrix
   ------------------------------------------------------------------ */

static void multpm(void)

{
    PTR m2;
    long *p, i;

    if (nor1 != nor2) 
	MtxErrorExit("%s and %s are not compatible",aname,bname);

    /* Allocate memory and read the permutation
       ---------------------------------------- */
    p = NALLOC(long,nor1);
    if (zreadlong(afile,p,nor1) != (size_t) nor1)
	MtxErrorExit("Cannot read %s",aname);
    --p;
    zsetfield(fl2);
    zsetlen(noc2);
    m2 = zalloc((long) 1);

    /* Write out the rows in permuted order
       ------------------------------------ */
    if ((cfile = zwritehdr(cname,fl2,nor2,noc2)) == NULL)
	MtxErrorExit("Cannot open %s for output",cname);
    for (i = 1; i <= nor1; ++i)
    {	
	zseek(bfile,p[i]);
	zreadvec(bfile,m2,(long) 1);
	zwritevec(cfile,m2,(long)1);
    }
}


/* ------------------------------------------------------------------
   multmp() - Multiply matrix by permutation
   ------------------------------------------------------------------ */

static void multmp(void)

{
    PTR m1,m3;
    long *m2;

    if (noc1 != nor2) 
	MtxErrorExit("%s and %s are not compatible",aname,bname);

    zsetfield(fl1);
    zsetlen(noc1);
    m1 = zalloc((long) 1);
    m3 = zalloc((long) 1);
    m2 = NALLOC(long,nor2);
    if (zreadlong(bfile,m2,nor2) != (size_t) nor2)
	MtxErrorExit("Cannot write to %s: %E",bname,MtxErrNo);

    if ((cfile = zwritehdr(cname,fl1,nor1,noc1)) == NULL)
	MtxErrorExit("Cannot open %s for output: %S",cname);
    zsetlen(noc1);
    while (nor1--)
    {
        zreadvec(afile,m1,(long) 1);
	zpermrow(m1,m2,m3);
	zwritevec(cfile,m3,(long) 1);
    }
}



/* ------------------------------------------------------------------
   multsm() - Multiply scalar with matrix
   ------------------------------------------------------------------ */

static void multsm(FILE *s,FILE *m,long nor3,long noc3)

{
    PTR ms, mm;
    FEL f;
    long i;

    zsetfield(fl1);
    zsetlen(1);
    ms = zalloc(1);
    zreadvec(s,ms,1);
    f = zextract(ms,1);
    zsetlen(noc3);
    mm = zalloc((long)1);
	
    if ((cfile = zwritehdr(cname,fl1,nor3,noc3)) == NULL)
	MtxErrorExit("Cannot open %s for output: %S",cname);
    for (i = 1; i <= nor3; ++i)
    {	
	zreadvec(m,mm,(long)1);
	zmulrow(mm,f);
	zwritevec(cfile,mm,(long)1);
    }
}



/* ------------------------------------------------------------------
   multmm() - Multiply two matrices
   ------------------------------------------------------------------ */

static void multmm(void)

{
    PTR m1,m2,tmp,out,x;
    long row1, row2;	/* Range of rows */
    long col1, col2;	/* Range of columns */
    long i, k;

    if (fl1 != fl2)
	MtxErrorExit("%s and %s are not compatible",aname,bname);
    if (noc1 == 1 && nor1 == 1)
    {	multsm(afile,bfile,nor2,noc2);
	return;
    }
    else if (noc2 == 1 && nor2 == 1)
    {	multsm(bfile,afile,nor1,noc1);
	return;
    }
    if (noc1 != nor2) 
	MtxErrorExit("%s and %s are not compatible",aname,bname);
    if ((long) nrows > nor1 || (long) ncols > noc2) 
	MtxErrorExit("Matrix too small");
    row1 = (nor1 / nrows) * (thisrow - 1) + 1;
    row2 = (nor1 / nrows) * thisrow;
    col1 = (noc2 / ncols) * (thiscol - 1) + 1;
    col2 = (noc2 / ncols) * thiscol;

    /* First matrix
       ------------ */
    zsetfield(fl1);
    zsetlen(noc1);
    m1 = zalloc((long)1);
    zseek(afile,row1);

    /* Read second matrix
       ------------------ */
    zsetlen(noc2);
    tmp = zalloc((long)1);
    zsetlen(col2-col1+1);
    m2 = zalloc(nor2);
    out = zalloc((long)1);
    if (col2 - col1 + 1 < noc2)
    {
	x = m2;
	for (i = 1; i <= nor2; ++i)
	{
	    zsetlen(noc2);
	    zreadvec(bfile,tmp,(long)1);
	    for (k = col1; k <= col2; ++k)
		zinsert(x,k - col1 + 1,zextract(tmp,k));
	    zsetlen(col2-col1+1);
	    zadvance(&x,(long)1);
	}
    }
    else
    {
	zsetlen(noc2);
	zreadvec(bfile,m2,nor2);
    }

    /* Multiply and write result
       ------------------------- */
    if ((cfile = zwritehdr(cname,fl1,row2-row1+1,col2-col1+1)) == NULL)
	MtxErrorExit("Cannot open %s for output: %S",cname);
    for (i = 1; i <= row2 - row1 + 1; ++i)
    {
        zsetlen(noc1);
        zreadvec(afile,m1,(long)1);	/* Read one row */
	zsetlen(col2-col1+1);
	zmaprow(m1,m2,nor2,out);
	zwritevec(cfile,out,(long)1);
    }
}


/* ------------------------------------------------------------------
   multpp() - Multiply two permutations
   ------------------------------------------------------------------ */

static void multpp(void)

{
    long i;
    long *p1, *p2 = NULL;
    long maxdeg;

    if (fl1 != fl2) 
	MtxErrorExit("%s and %s are not compatible",aname,bname);
    if (fl1 != -1) 
	MtxErrorExit("Monomials are not supported");
    maxdeg = (nor1 > nor2) ? nor1 : nor2;

    /* Allocate memory and read the permutations
       ----------------------------------------- */
    if ((p1 = NALLOC(long,maxdeg)) == NULL ||
	(p2 = NALLOC(long,maxdeg)) == NULL)
	MtxErrorExit("Cannot allocate workspace");
    if (zreadlong(afile,p1,nor1) != (size_t) nor1)
	MtxErrorExit("Cannot read from %s",aname);
    if (zreadlong(bfile,p2,nor2) != (size_t) nor2)
	MtxErrorExit("Cannot read from %s",bname);
    --p1;
    --p2;

    /* If the permutations have different
       degrees, extend the smaller one.
       ---------------------------------- */
    for (i = nor1 + 1; i <= maxdeg; ++i)
	p1[i] = i;
    for (i = nor2 + 1; i <= maxdeg; ++i)
	p2[i] = i;

    /* Calculate the product
       --------------------- */
    for (i = 1; i <= maxdeg; ++i)
	p1[i] = p2[p1[i]];

	/* Write out the result
	   -------------------- */
    if ((cfile = zwritehdr(cname,fl1,maxdeg,(long)1)) == NULL)
	MtxErrorExit("Cannot open %s for output: %S",cname);
    if (zwritelong(cfile,p1+1,maxdeg) != (size_t) maxdeg)
	MtxErrorExit("Cannot write to %s: %E",MtxErrNo);
}



int main(int argc, char *argv[])

{   int i;


    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("r:c:")) != OPT_END)
    {
	switch (i)
	{
	    case 'c':
		thiscol = (int) MtxGetInt();
		if (*opt_text_ptr == '.')
		{	
		    ++opt_text_ptr;
		    ncols = (int) MtxGetInt();
		}
		else
			ncols = 2;
		if (*opt_text_ptr != 0 || thiscol < 1 || ncols < thiscol) 
		    MtxErrorExit("Invalid argument after -c");
		break;
	    case 'r':
		thisrow = (int) MtxGetInt();
		if (*opt_text_ptr == '.')
		{	
		    ++opt_text_ptr;
		    nrows = (int) MtxGetInt();
		}
		else
		    nrows = 2;
		if (*opt_text_ptr != 0 || thisrow < 1 || nrows < thisrow) 
		    MtxErrorExit("Invalid argument after -r");
		break;
	}
    }

    /* Set file names
       -------------- */
    switch (argc-opt_ind)
    {
	case 3:
	    aname = argv[opt_ind];
	    bname = argv[opt_ind+1];
	    cname = argv[opt_ind+2];
	    break;
	default:
	    MtxErrorExit("Bad number of arguments, try '%s -help'",pinfo.name);
    }
   if (!strcmp(aname,cname) || !strcmp(bname,cname))
       MtxErrorExit("Identical file names not allowed");

    /* Read file headers and call the 
       appropriate function for multiplying.
       ------------------------------------- */
    afile = zreadhdr(aname,&fl1,&nor1,&noc1);
    if (afile == NULL) 
	MtxErrorExit("Cannot open %s: %S",aname);
    bfile = zreadhdr(bname,&fl2,&nor2,&noc2);
    if (bfile == NULL) 
	MtxErrorExit("Cannot open %s: %S",bname);
    if (fl1 < 0 && fl2  < 0) multpp(); else
    if (fl1 > 1 && fl2 > 1) multmm(); else
    if (fl1 > 1 && fl2 == -1) multmp(); else
    if (fl1 == -1 && fl2  > 1) multpm(); else
	MtxErrorExit("%s and %s are not compatible",aname,bname);
    fclose(afile);
    fclose(bfile);
    fclose(cfile);
    MtxCleanup();
    return EXIT_OK;
}



