/* ============================= C MeatAxe ==================================
   File:        $Id: znullsp.c,v 1.2 1998/03/11 17:04:49 mringe Exp $
   Comment:     Calculate null-space
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include <stdlib.h>
#include "meataxe.h"


/*{{{ --------------------------------------------------------------------------
!function znullsp "Null space"
!section kernel.mat
!synopsis 
    long znullsp(PTR matrix, long nor, long *piv, PTR nsp);
!arg matrix
    Pointer to the matrix.
!arg nor
    Number of rows.
!arg piv
    Pointer to pivot table.
!arg nsp
    Pointer to result.
!return
    The function returns the dimension of the null-space, or $-1$ on error.
!description
    This function calculates the null-space of a matrix. The matrix is passed 
    as first argument, |nor| is the number of rows, |piv| must be a pointer 
    to an array of at least $|nor|+1$ integers, and |nsp| must be a pointer to 
    a square matrix of size |nor|.

    If the function is successfull (non-negative return value),
    \begin{list}{}{\itemsep 0mm\parsep 0mm}
    \item[-] \verb"matrix" is reduced to echelon form
    \item[-] \verb"nsp" contains the null-space in echelon form
    \item[-] \verb"piv" contains a pivot table for the null space, and
	    the null-space dimension is duplicated in \verb"piv[0]".
    \end{list}
!seealso 
-------------------------------------------------------------------------- }}}*/

long znullsp(PTR matrix, long nor, long *piv, PTR nsp)

{
    PTR x, y, a, b;
    int i;
    long noc = znoc;
    long dim;
    FEL f;

    /* Make identity matrix in nsp
       --------------------------- */
    zsetlen(nor);
    x = nsp;
    for (i = 1; i <= nor; ++i)
    {
	piv[i] = 0;
	zmulrow(x,F_ZERO);
	zinsert(x,i,F_ONE);
	zadvance(&x,(long)1);
    }

    /* Gaussian elimination
       -------------------- */
    x = matrix;
    y = nsp;
    for (i = 1; i <= nor; ++i)
    {
	PTR xx = matrix, yy = nsp;
	long k, p;

	for (k = 1; k < i; ++k)
	{
	    if ((p = piv[k]) != 0 && (f = zextract(x,p)) != F_ZERO)
	    {
		f = zneg(zdiv(f,zextract(xx,p)));
		zsetlen(noc);
		zaddmulrow(x,xx,f);
		zsetlen(nor);
		zaddmulrow(y,yy,f);
	    }
	    zsetlen(noc);
	    zadvance(&xx,(long)1);;
	    zsetlen(nor);
	    zadvance(&yy,(long)1);
	}
	zsetlen(noc);
	piv[i] = p = zfindpiv(x,&f);
	zsetlen(noc);
	zadvance(&x,(long)1);
	zsetlen(nor);
	zadvance(&y,(long)1);
    }

    /* Step 2: Reduce the null space to echelon form.
       ---------------------------------------------- */
    dim = 0;
    x = y = nsp;
    a = b = matrix;
    for (i = 1; i <= nor; ++i)
    {
	if (piv[i] == 0)
	{
	    zsetlen(nor);
	    if (y != x) zmoverow(y,x);
	    zcleanrow(y,nsp,dim,piv);
	    piv[++dim] = zfindpiv(y,&f);
	    zadvance(&y,1);
	}
	else
	{
	    zsetlen(noc);
	    if (b != a) zmoverow(b,a);
	    zadvance(&b,1);
	}
	zsetlen(nor);
	zadvance(&x,1);
	zsetlen(noc);
	zadvance(&a,1);
    }
    piv[0] = dim;
    return dim;
}

