/* ============================= C MeatAxe ==================================
   File:        $Id: zpr.c,v 1.3 1998/04/28 17:38:27 mringe Exp $
   Comment:     Print a matrix or permutaion.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static long HdrPos;
static long hdr[3];
static FILE *dest = stdout;	/* Output file */
static char *inpname = NULL;
static FILE *inpfile = NULL;
static long first = 1;		/* Output range (option -r) */
static long last = 10000000;

static char *helptext[] = {
"SYNTAX",
"    zpr [-Gs] [-r <Range>] [<Binfile> [<Textfile>]]",
"",
"OPTIONS",
"    -G   GAP output",
"    -r   Print a range of rows (or permutations). <Range> may be a",
"         single number, or of the form <First>-<Last>.",
"    -s   Print summary only\n",
"",
"FILES",
"    <Binfile>   i  A matrix or permutation in binary format",
"    <Textfile>  i  The output in text format (default: stdout)",
"",
NULL};

static proginfo_t pinfo =
   { "zpr", "Print Permutations Or Matrices", "$Revision: 1.3 $",
      helptext };



/* ------------------------------------------------------------------
   err() - Print error message and exit.
   ------------------------------------------------------------------ */

static void err(c)
int c;

{   fprintf(stderr,"ZPR ERROR - ");
    switch (c)
    {
	case 'f':
	    fprintf(stderr,"FILE I/O ERROR\n");
	    break;
	case 't':
	    fprintf(stderr,"CANNOT PRINT THIS FILE TYPE\n");
	    break;
    }
    exit(EXIT_ERR);
}


/* ------------------------------------------------------------------
   PrString(), PrLong() - Prettty printer
   ------------------------------------------------------------------ */

static void PrString(char *c)
{
    static int pos = 0;
    int l = strlen(c);

    if (l + pos >= 78)
    {
	fprintf(dest,"\n");
	pos = 0;
    }
    fprintf(dest,"%s",c);
    for (; *c != 0; ++c)
	if (*c == '\n') pos = 0; else ++pos;
}


static void PrLong(long l)

{
    char tmp[20];
    sprintf(tmp,"%ld",l);
    PrString(tmp);
}


/* ------------------------------------------------------------------
   prmatrix() - Print a matrix in standard format.
   ------------------------------------------------------------------ */

static void prmatrix()

{
    PTR m1;
    FEL f1;
    long loop1, j1;
    int md, mx, iv;

    last = hdr[1];
    first = 1;
    zsetfield(hdr[0]);
    zsetlen(hdr[2]);
    m1 = zalloc((long)1);

    if (hdr[0] < 10) {md = 1; mx = 80;}
    else if (hdr[0] < 100) {md = 3; mx = 25;}
    else if (hdr[0] < 1000) {md = 4; mx = 20;}
    else if (hdr[0] < 10000) {md = 5; mx = 15;}
    else {md = 6; mx = 12;}

    fprintf(dest,"matrix field=%ld rows=%ld cols=%ld\n",hdr[0],hdr[1],hdr[2]);
    for (loop1 = 1; loop1 <= last; ++loop1)
    {	
	if (zreadvec(inpfile,m1,1) != 1) 
	    MtxErrorExit("Cannot read %s: %E",inpname,MtxErrNo);
	if (loop1 < first) continue;
	iv = 1;
	for (j1 = 1; j1 <= hdr[2]; ++j1)
	{
	    f1 = zextract(m1,j1);
	    switch (md)
	    {	case 1: fprintf(dest,"%1ld",zftoi(f1)); break;
		case 2: fprintf(dest,"%2ld",zftoi(f1)); break;
		case 3: fprintf(dest,"%3ld",zftoi(f1)); break;
		case 4: fprintf(dest,"%4ld",zftoi(f1)); break;
		case 5: fprintf(dest,"%5ld",zftoi(f1)); break;
		case 6: fprintf(dest,"%6ld",zftoi(f1)); break;
	    }
	    if (iv++ >= mx)
	    {	fprintf(dest,"\n");
		iv = 1;
	    }
	}
	if (iv > 1)
		fprintf(dest,"\n");
    }
}


/* ------------------------------------------------------------------
   prgapmat() - Print a matrix in GAP format.
   ------------------------------------------------------------------ */

static void prgapmat()

{   PTR m1;
    FEL f1;
    FEL gen;
    long loop1, j1;
    int cnt, isprimefield;


    zsetfield(hdr[0]);
    zsetlen(hdr[2]);
    gen = zgen;		/* Generator */
    isprimefield = (zchar == zfl);

    m1 = zalloc((long)1);
    PrString("MeatAxe.Matrix := [\n");
    for (loop1 = 1; loop1 <= hdr[1]; ++loop1)
    {	
	if (zreadvec(inpfile,m1,1) != 1) 
	    MtxErrorExit("Cannot read %s: %E",inpname,MtxErrNo);
	cnt = 0;
	fprintf(dest,"[");
	for (j1 = 1; j1 <= hdr[2]; ++j1)
	{   if (cnt > 75)
	    {	fprintf(dest,"\n ");
		cnt = 0;
	    }
	    f1 = zextract(m1,j1);
	    if (isprimefield)
	    {   FEL f2=F_ZERO;
	   	long k=0;
	    	while (f2 != f1)
	    	{   f2 = zadd(f2,gen);
		    ++k;
		}
		fprintf(dest,"%ld",k);
		cnt += k>9999?5:k>999?4:k>99?3:k>9?2:1;
	    }
	    else
	    {   if (f1 == F_ZERO)
	    	{   fprintf(dest,"0*Z(%ld)",hdr[0]);
		    cnt += 5;
		    cnt += hdr[0]>9999?5:hdr[0]>999?4:hdr[0]>99?3:hdr[0]>9?2:1;
		}
		else
		{   FEL f2 = gen;
		    long k = 1;
		    while (f2 != f1)
		    {   f2 = zmul(f2,gen);
			++k;
		    }
		    fprintf(dest,"Z(%ld)^%ld",hdr[0],k);
		    cnt += 4;
		    cnt += hdr[0]>9999?5:hdr[0]>999?4:hdr[0]>99?3:hdr[0]>9?2:1;
		    cnt += k>9999?5:k>999?4:k>99?3:k>9?2:1;
		}
	    }
	    if (j1 < hdr[2])
	    {	fprintf(dest,",");
		++cnt;
	    }
	}
	fprintf(dest,"]");
	if (loop1 < hdr[1])
	    fprintf(dest,",");
	fprintf(dest,"\n");
    }
    fprintf(dest,"]");
    if (isprimefield)
	fprintf(dest,"*Z(%ld)",hdr[0]);
    fprintf(dest,";\n");
}


/* ------------------------------------------------------------------
   prgapperm() - Print a permutation in GAP format.
   ------------------------------------------------------------------ */

static void prgapperm()

{
    long i, pos;
    long *perm;

    /* Allocate memory for one permutation
      ------------------------------------ */
    perm = NALLOC(long,hdr[1]);
    if (perm == NULL) 
	MtxErrorExit("Cannot allocate work space: %S");

    PrString("MeatAxe.Perms := [\n");
    for (pos = 1; pos <= hdr[2]; ++pos)
    {
    	/* Read the next permutation
	   ------------------------- */
	if (zreadlong(inpfile,perm,hdr[1]) != (size_t) hdr[1])
	    MtxErrorExit("Cannot read %s: %E",inpname,MtxErrNo);

	/* Print it
	   -------- */
	PrString("    PermList([");
	for (i = 0; i < hdr[1]; ++i)
	{
	    if (i > 0)
	    	PrString(",");
	    PrLong(perm[i]);
	}
	PrString("])");
	if (pos < hdr[2]) PrString(",");
	PrString("\n");
    }
    PrString("];\n");
}



/* ------------------------------------------------------------------
   prgap() - Print a matrix or permutation in GAP format.
   ------------------------------------------------------------------ */

static void prgap()

{
    if (hdr[0] == -1)
	prgapperm();
    else if (hdr[0] >= 2)
	prgapmat();
    else
	err('t');
}



/* ------------------------------------------------------------------
   prpol() - Print a polynomial
   ------------------------------------------------------------------ */

static void prpol()

{
    int i;
    poly_t *p;
    
    SFSeek(inpfile,HdrPos);
    if ((p = PolRead(inpfile)) == NULL)
	MtxErrorExit("Cannot read %s: %E",inpname,MtxErrNo);
    

    fprintf(dest,"polynomial field=%ld degree=%ld\n",hdr[1],hdr[2]);
    for (i = 0; i <= p->deg; ++i)
    {
	PrLong(zftoi(p->buf[i]));
	PrString(" ");
    }
    PrString("\n");
}


/* ------------------------------------------------------------------
   prperm() - Print a permutation in standard format.
   ------------------------------------------------------------------ */

static void prperm()

{
    long f1, i, count;
    long *perm;

    perm = NALLOC(long,hdr[1]);
    if (perm == NULL) 
	MtxErrorExit("Cannot allocate work space");

    for (count = 1; count <= hdr[2]; ++count)
    {
        fprintf(dest,"permutation degree=%ld\n",hdr[1]);
	if (zreadlong(inpfile,perm,hdr[1]) != (size_t) hdr[1])
	    MtxErrorExit("Cannot read %s: %E",inpname,MtxErrNo);
	for (i = 0; i < hdr[1]; ++i)
	{
	    f1 = perm[i];
	    PrLong(f1);
	    PrString(" ");
	}
	PrString("\n");
    }
}

/* ------------------------------------------------------------------
   primat() - Print an integer matrix
   ------------------------------------------------------------------ */

static void primat()

{
    long k, i;
    long *row;

    row = NALLOC(long,hdr[2]);
    if (row == NULL) 
	MtxErrorExit("Cannot allocate work space: %S");

    fprintf(dest,"integer matrix rows=%ld cols=%ld\n",hdr[1],
	hdr[2]);
    for (i = 0; i < hdr[1]; ++i)
    {
	if (zreadlong(inpfile,row,hdr[2]) != (size_t) hdr[2])
    	    MtxErrorExit("Cannot read %s: %E",inpname,MtxErrNo);
    	for (k = 0; k < hdr[2]; ++k)
    	{
    	    PrLong(row[k]);
    	    PrString(" ");
    	}
    	PrString("\n");
    }
}



/* ------------------------------------------------------------------
   prmtx() - Print a matrix or permutation in standard format.
   ------------------------------------------------------------------ */

static void prmtx()

{	if (hdr[0] > 1)
		prmatrix();
	else if (hdr[0] == -1)
		prperm();
	else if (hdr[0] == -2)
		prpol();
	else if (hdr[0] == -T_IMAT)
		primat();
	else
		err('t');
}


/* ------------------------------------------------------------------
   setrange() - Set output range
   ------------------------------------------------------------------ */

static void setrange(char *c)

{	
    if ((first = MtxGetInt()) == GETINT_ERR) 
	MtxErrorExit("-r: %E",ERR_OPTION);
    if (*opt_text_ptr == 0)
    {	
	last = first;
    }
    else if (*opt_text_ptr == '-')
    {   
	++opt_text_ptr;
        if ((last = MtxGetInt()) == GETINT_ERR) 
	    MtxErrorExit("-r: %E",ERR_OPTION);
    }
    if (*opt_text_ptr != 0 || first < 1 || last < first)
       	MtxErrorExit("-r: %E",ERR_OPTION);
}

/* ------------------------------------------------------------------
   printsummary()
   ------------------------------------------------------------------ */

static void printsummary()

{
    printf("%s: ",inpname);
    if (hdr[0] == -1)
    {
	printf("%ld Permutation%s of degree %ld\n",hdr[2],
		hdr[2] == 1 ? "" : "s", hdr[1]);
	SFSeek(inpfile,ftell(inpfile) + hdr[1]*hdr[2]*4);
    }
    else if (hdr[0] >= 2)
    {
	PTR x;
	long l;
	printf("%ld x %ld matrix over GF(%ld)\n",hdr[1],hdr[2],hdr[0]);
	zsetfield(hdr[0]);
	zsetlen(hdr[2]);
	x = zalloc(1);
	for (l = hdr[1]; l > 0; --l)
	    zreadvec(inpfile,x,1);
	free(x);
    }
    else if (hdr[0] == -2)
    {
	printf("Polynomial of degree %ld over GF(%ld)\n",hdr[2],hdr[1]);
	zsetfield(hdr[1]);
	zsetlen(hdr[2]+1);
	SFSeek(inpfile,ftell(inpfile) + zsize(1));
    }
    else if (hdr[0] == -T_IMAT)
    {
	printf("%ld x %ld integer matrix\n",hdr[1],hdr[2]);
	SFSeek(inpfile,ftell(inpfile)+hdr[1]*hdr[2]*4);
    }
    else
	printf("Unknown file format.\n");
}

/* ------------------------------------------------------------------
   ReadHeader() - Read file header. Returns 1 on success, 0 on error
   or EOF.
   ------------------------------------------------------------------ */

static int ReadHeader(void)

{
    HdrPos = ftell(inpfile);
    if (feof(inpfile)) return 0;
    if (zreadlong(inpfile,hdr,3) != 3) return 0;
    /* Check the header */
    if (hdr[0] > 65536 || hdr[0] < -20 || hdr[1] < 0 || hdr[2] < 0)
    {
	MtxErrorExit("%s: %E (%d %d %d)\n",inpname,ERR_FILEFMT,
	    (int)hdr[0],(int)hdr[1],(int)hdr[2]);
    }
    return 1;
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    enum {MTX,GAP,SUMMARY} mode = MTX;
    int i;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("Gsr:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': mode = GAP; msg_level = -100; break;
	    case 's': mode = SUMMARY; break;
	    case 'r': setrange(opt_text); break;
	}
    }
    switch (argc - opt_ind)
    {  
    	case 2:
	    if ((dest = SFOpen(argv[opt_ind+1],FM_CREATE|FM_TEXT))
							== NULL)
	    {	perror(argv[opt_ind+1]);
		err('f');
	    }
	    /* no break here! */
        case 1:
	    inpname = argv[opt_ind];
	    if ((inpfile = SFOpen(inpname,FM_READ)) == NULL)
		MtxErrorExit("Cannot open %s: %S",inpname);
	    break;
    	default:
	    MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
    }

    while (ReadHeader())
    {
    	switch (mode)
    	{
	    case MTX: prmtx(); break;
	    case GAP: prgap(); break;
	    case SUMMARY: printsummary(); break;
    	}
    }
    fclose(inpfile);
    fclose(dest);
    MtxCleanup();
    return (EXIT_OK);
}


