/* ============================= C MeatAxe ==================================
   File:        $Id: zpseed.c,v 1.5 1998/04/28 17:38:28 mringe Exp $
   Comment:     Seed vector generator.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <limits.h>
#include <stdlib.h>



/*{{{ --------------------------------------------------------------------------
!text kernel.pseed

The seed vector generator is used to walk through the one-dimensional
subspaces of a given vector space $V$, the {\em seed space}. For each 
one-dimensional subspace $U<V$ the generator produces a representant
$u\in U$. These vectors are called {\em seed vectors}.

Once a basis for the seed space is fixed, each vector of the seed space 
can be identified by a natural number as follows. Let $b_1$,\ldots,$b_n$ 
be the basis and $v=\lambda_1b_1+\ldots\lambda_nb_n\in V$ a vector.
The coefficients $\lambda_i\in\GF{q}$ are identified with
positive integers in the usual way (see |zftoi()| for details) and
\[
    N(v):=\lambda_1+\lambda_2q+\lambda_3q^2+\ldots\lambda_nq^{n-1}
\]
is the `number of $v$'. Seed vectors are those vectors where the
leading coefficient is $1$.

To use the seed vector generator you must first call \verb"zpseed_init()"
to define the seed space. Then seed vectors can be generated by either
|zpseed_make()| or |zpseed_next()|.
-------------------------------------------------------------------------- }}}*/



/* ------------------------------------------------------------------
   Local data
   ------------------------------------------------------------------ */

static long dim;		/* Dimension of basis */
static PTR basis;		/* The basis vectors */
static FEL *coeff;		/* Coefficients */
static long lastnum;		/* Number of last seed vector */
static long maxnum;		/* Maximal valid vector number + 1 */



/*{{{ --------------------------------------------------------------------------
!function zpseed_vec "Seed vector buffer"
!section kernel.pseed
!synopsis
    extern PTR zpseed_vec;
!description
    |zpseed_vec| is maintained by the seed vector generator. After a successful
    call of |zpseed_make()| or |zpseed_next()|, the seed vector is stored in
    this variable.
!seealso @zpseed
-------------------------------------------------------------------------- }}}*/

PTR zpseed_vec = NULL;


/*{{{ --------------------------------------------------------------------------
!function zpseed_free "Clean up the seed vector generator"
!section kernel.pseed
!synopsis 
    void zpseed_free();
!description
    This function frees all resources used internally by the generator.
!seealso @zpseed
-------------------------------------------------------------------------- }}}*/

void zpseed_free()

{
    if (zpseed_vec != NULL) 
	free(zpseed_vec);
    if (coeff != NULL) 
	free(coeff);
}




/*{{{ --------------------------------------------------------------------------
!function zpseed_init "Initialize the seed vector generator"
!section kernel.pseed
!synopsis 
    int zpseed_init(long sdim, PTR seedbasis);
!arg sdim
    Dimension of seed space.
!arg seedbasis
    Basis for the seed space.
!return
    The function returns $0$ on success, and $-1$ on error.
!description
    This function initializes the seed vector generator. It must be called
    at least once before using |zpseed_make()| or |zpseed_next()|. |seedbasis|
    is a basis for the seed space, i.e., a matrix, and \verb"sdim" is the 
    number of basis vectors. Remember that the row size must have been set 
    before with \verb"zsetlen()".

    \verb"seedbasis" need not be in semi-echelon form. It is not checked that 
    the basis vectors are linearly independent. If they are not, you will get 
    seed vectors which are multiples of each other or different numbers will 
    produce the same vector.

    There is no internal buffer for the basis. \verb"zpseed_init()" just
    saves a pointer to the basis for internal use. Thus, you must not modify
    or free the basis as long as you use the seed vector generator.
!seealso @zpseed
-------------------------------------------------------------------------- }}}*/

int zpseed_init(long sdim, PTR seedbasis)

{
    MTXCONTEXT(zpseed_init);
    int i;

    zpseed_free();
    if ((zpseed_vec = zalloc(1)) == NULL
	|| (coeff = NALLOC(FEL,sdim)) == NULL
       )
	return MTXERROR("Cannot allocate work space"), -1;
    dim = sdim;
    basis = seedbasis;
    for (i = 0; i < (int) dim; ++i) 
	coeff[i] = F_ZERO;
    zmulrow(zpseed_vec,F_ZERO);
    lastnum = 0;
    maxnum = 2;
    for (i = 1; i < dim; ++i)
    {
	if (maxnum > LONG_MAX / zfl) 
	    break;
	maxnum *= zfl;
    }
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function zpseed_make "Make a seed vector"
!section kernel.pseed
!synopsis 
    long zpseed_make(long num);
!arg num
    Seed vector number.
!return
    Seed vector number, or $-1$ on error.
!description
    This function makes a seed vector given its number. The seed vector
    generator must have been initialized with \verb"zpseed_init()".
    The seed vector is stored in \verb"zpseed_vec" and its number is
    returned. A return value of -1 indicates an error.

    \verb"zpseed_make()" will not reject a number which corresponds to
    a non-normalized seed vector. See |zpseed_init()| for details.
!seealso @zpseed
-------------------------------------------------------------------------- }}}*/

long zpseed_make(long num)

{
    MTXCONTEXT(zpseed_make);
    int i;
    PTR x = basis;

    if (num < 0 || num >= maxnum)
	return MTXERROR1("Invalid seed vector number %d",(int)num), -1;
    lastnum = num;
    for (i = 0; i < (int) dim; ++i)
    {
	FEL co = zitof(num % zfl);
	num /= zfl;
	if (co != coeff[i])
	{
	    FEL delta = zsub(co,coeff[i]);
	    zaddmulrow(zpseed_vec,x,delta);
	    coeff[i] = co;
	}
	zadvance(&x,(long)1);
    }
    return lastnum;
}



/*{{{ --------------------------------------------------------------------------
!function zpseed_next "Make the next seed vector"
!section kernel.pseed
!synopsis 
    long zpseed_next();
!return
    Seed vector number, $-1$ on error.
!description
    This function calculates the next normalized seed vector and stores it 
    in |zpseed_vec|. Successive calls of |zpseed_next()| will step through 
    all normalized seed vectors. If the seed space is exhausted, 
    |zpseed_next()| returns $-1$.

    If you call \verb"zpseed_make(N)", \verb"zpseed_next()" will calculate
    with the next normalized seed vector after $N$.
    Seed |zpseed_init()| for details.
!seealso @zpseed
-------------------------------------------------------------------------- }}}*/

long zpseed_next()

{
    long w = lastnum + 1, x, i;

    for (x = 1; (i = w/x) >= zfl; x *= zfl);
    if (i != 1) 
	w = x * zfl;

    if (w >= maxnum) 
	return -1;	/* Keine Fehlerbedingung! */
    return zpseed_make(w);
}


