/* ============================= C MeatAxe ==================================
   File:        $Id: zquot.c,v 1.4 1998/04/28 17:38:32 mringe Exp $
   Comment:     Projection on quotient.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>


/* TODO: kernel functions should not allocate memory */


/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

static long *mypiv;		/* Pivot table */
static long *locn = NULL;	/* Table of non-pivot columns */
static long vdim, qdim, sdim;	/* Dimensions of V, subspace, quotient */
static PTR mysubsp;		/* Pointer to the subspace */



/*{{{ --------------------------------------------------------------------------
!function zquotinit "Prepare for projection on the quotient"
!section kernel.mat
!synopsis 
    int zquotinit(PTR subspace, long dim, long *piv);
!arg subspace
    Pointer to the list of vectors to project.
!arg dim
    Number of vetors in |space|.
!arg piv
    Pointer to pivot table.
!return
    $0$ on success, $-1$ on error.
!description
    This function is used to specify the subspace used by the quotient
    projection functions, |zquot()| and |zquotop()|. You must always call
    |zquot_init()| before |zquot()| or |zquotop()|.
    |subspace| is a pointer to a basis for the subspace, which must be in 
    semi-echelon form, and |dim| is the dimension of the subspace. |piv|
    must be a pointer to a pivot table for the basis. The pivot table
    can be calculated with |zmkpiv()|.
!seealso @zquot
-------------------------------------------------------------------------- }}}*/

int zquotinit(PTR subspace, long dim, long *piv)

{
    MTXCONTEXT(zquotinit);
    long i;
    char *ispiv;

    if (locn != NULL)
    {
	free(locn);
	locn = NULL;
    }
    mypiv = piv;
    vdim = znoc;
    sdim = dim;
    mysubsp = subspace;

    if ((ispiv = NALLOC(char,znoc+1)) == NULL 
	|| (locn = NALLOC(long,vdim-sdim+1)) == NULL
       )
       return MTXERROR("Cannot allocate work space"), -1;
    memset(ispiv,0,(size_t)(znoc+1));

    /* TODO: benuzte zmkpivot() */
    for (i = 1; i <= sdim; ++i)
    {
	if (ispiv[mypiv[i]])
	{
	    free(ispiv);
	    return MTXERROR1("%E",ERR_NOTECH), -1;
	}
	ispiv[mypiv[i]] = 1;
    }
    qdim = 0;
    for (i = 1; i <= znoc; ++i)
    {
	if (!ispiv[i])
	    locn[++qdim] = i;
    }
    free(ispiv);
    return 0;
}





/*{{{ --------------------------------------------------------------------------
!function zquot "Projection on the quotient"
!section kernel.mat
!synopsis 
    int zquot(PTR space, long dim, PTR quot)
!arg space
    Pointer to the list of vectors to project.
!arg dim
    Number of vetors in |space|.
!return
    $0$ on success, $-1$ on error.
!description
    Given two subspaces $U,W\leq V$, this function calculates the
    projection of $U$ on the quotient $V/W$. Before you can use
    \verb"zquot()", the subspace $W$ must be specified with
    \verb"zquotinit()".

    The projection depends on the basis for the subspace $W$. It is
    calculated in the following way:
    Let $V=F^{n\times n}$ and $w_1,\ldots,w_s$ be a basis for the subspace
    $W\leq V$. The basis, written as a matrix of row vectors, is assumed
    to be in semi-echelon form. By looking at the pivot columns we can
    construct the vectors $w_{s+1},\ldots,w_n$ by taking all vectors which
    have a exactly one 1 at any non-pivot position and are zero otherwise.
    Then, $(w_1,\ldots,w_s,w_{s+1},\ldots,w_n)$ is a basis for $V$ in
    semi-echelon form and defines the decomposition of any vector into
    subspace and quotient part.

    The result is stored in \verb"quot", which must be a matrix with
    \verb"dim" rows and $n-s$ columns. On return, the $i$-th row of
    \verb"quot" contains the projection of the $i$-th row of
    \verb"space". The result is not in echelon form and some of the 
    rows may be zero.
!seealso @zquot
-------------------------------------------------------------------------- }}}*/

int zquot(PTR space, long dim, PTR quot)

{
    long i, k;
    PTR qx, sx, tmp;

    qx = quot;
    zsetlen(vdim);
    sx = space;
    tmp = zalloc((long)1);

    for (i = 1; i <= dim; ++i)
    {
	zsetlen(vdim);
	zmoverow(tmp,sx);
	zadvance(&sx,1);
	zcleanrow(tmp,mysubsp,sdim,mypiv);
	zsetlen(qdim);
	zmulrow(qx,F_ZERO);
	for (k = 1; k <= qdim; ++k)
	    zinsert(qx,k,zextract(tmp,locn[k]));
	zadvance(&qx,1);
    }
    free(tmp);
    zsetlen(vdim);
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!function zquotop "Operation on the quotient"
!section kernel.mat
!synopsis 
    int zquotop(PTR matrix, PTR quot);
!arg matrix
    Pointer to the matrix
!arg quot
    Pointer to a buffer for the rsult.
!return
    $0$ on success, $-1$ on error.
!description
    After a subspace has been specified with \verb"zquotinit()", this
    function calculates the operation of any matrix on the quotient.
    Of course, this makes sense only if the subspace is invariant
    under the matrix, but this is not checked. If the subspace is
    not invariant, the results are undefined.

    \verb"matrix" must be a pointer to a square matrix with the
    same number of columns as the subspace. The result is a square
    matrix with the dimension of the quotient. This matrix is stored
    in |quot|.
!seealso @zquot
-------------------------------------------------------------------------- }}}*/

int zquotop(PTR matrix, PTR quot)

{
    long i, k;
    PTR qx, sx, tmp;

    qx = quot;
    zsetlen(vdim);
    tmp = zalloc((long)1);

    for (i = 1; i <= qdim; ++i)
    {
	zsetlen(vdim);
        sx = matrix;
	zadvance(&sx,locn[i]-1);
	zmoverow(tmp,sx);
	zcleanrow(tmp,mysubsp,sdim,mypiv);
	zsetlen(qdim);
	zmulrow(qx,F_ZERO);
	for (k = 1; k <= qdim; ++k)
	    zinsert(qx,k,zextract(tmp,locn[k]));
	zadvance(&qx,(long)1);
    }
    free(tmp);
    zsetlen(vdim);
    return 0;
}
