/* ============================= C MeatAxe ==================================
   File:        $Id: zsb.c,v 1.3 1998/04/28 17:38:34 mringe Exp $
   Comment:     Standard basis
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>

#include "meataxe.h"



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

enum {GENERATE,READ,READALL}
	smode = READ;		/* Seed mode */
long seedvecno = 0;		/* Selected seed vector (0 = all) */
long seedno = 0;		/* Current seed vector no */

char **genname;		/* File name of generators */
char *seedname;		/* Seed vector file name */
char *outname;		/* Output file name */
char *OpName = NULL;	/* Name of <op> file */
FILE *seedfile, *outfile;
long fl, nor;		/* Field and size of the generators */
long *piv;		/* Pivot table */
long *OpTable;		/* Operations, written to <Op> */
#define OPVEC(i) OpTable[2*(i)]
#define OPGEN(i) OpTable[2*(i)+1]
int ngen = 2;		/* Number of generators */
PTR *gen;		/* Generators */
PTR seed, A, B;		/* Workspace */
long seeddim;		/* Number of seed vectors in input file */
long nseed;		/* Number of seed vectors read */
long *cycldim = NULL;	/* Dimension of cyclic subspaces */
int opt_G = 0;		/* GAP output */
int FindGen = 0;	/* Find a generating vector */

static char *helptext[] = {
"SYNTAX",
"    zsb [<Options>] <Gen1> <Gen2> <Seed> <Basis> [<Op>]",
"    zsb [<Options>] [-g <#Gen>] <Gen> <Seed> <Basis> [<Op>]",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -V            Verbose, more messages.",
"    -G            Produce GAP output.",
"    -T <Seconds>  Set CPU time limit.",
"    -Q            Quiet, no messages.",
"    -g <#Gen>     Set number of generators.",
"    -1            Find a single generating vector.",
"    -s<Mode>      Select seed vector for spin-up:",
"                    -sg[<Num>]   Generate seed vectors (Default)",
"                    -sr[<Num>]   Read seed vectors one by one",
"                  <Num>, if specified, starts with this vector",
"",
"FILES",
"    <Gen1/2>   i  The generators (square matrices of equal size)",
"    <Seed>     i  Seed vectors (a matrix over the same field and",
"                  with the same row size as the generators)",
"    <Basis>    o  Standard basis",
"    <Op>       o  Description of <Basis> in terms of seed vectors",
"                  and generators (integer matrix)",
"",
"    If -g is used, Generators are read from <Gen>.1, <Gen>.2, ...",
NULL};

static proginfo_t pinfo =
   { "zsb", "Standard Basis", "$Revision: 1.3 $", helptext };



/* ------------------------------------------------------------------
   init()
   ------------------------------------------------------------------ */

static void init()

{
    int i;
    long fl2, nor2, noc2;
    FILE *f;

    MtxInit();

    /* Read generators
       --------------- */
    for (i = 0; i < ngen; ++i)
    {
	if ((f = zreadhdr(genname[i],&fl2,&nor2,&noc2)) == NULL)
	    MtxErrorExit(NULL);
	if (fl2 < 2) 
	    MtxErrorExit("%s: %E",genname[i],ERR_NOTMATRIX);
	if (nor2 != noc2) 
	    MtxErrorExit("%s: %E",genname[i],ERR_NOTSQUARE);
	if (i == 0)
	{
	    fl = fl2;
	    nor = nor2;
	}
	else if (fl != fl2 || nor != nor2)
	    MtxErrorExit("%s and %s: %E",genname[0],genname[i],ERR_INCOMPAT);
	zsetfield(fl);
	zsetlen(nor);
	gen[i] = zalloc(nor);
	zreadvec(f,gen[i],nor);
	fclose(f);
    }

    /* Allocate workspace
       ------------------ */
    A = zalloc(nor+1);
    B = zalloc(nor+1);
    seed = zalloc((long)1);
    piv = NALLOC(long,nor+2);
    OpTable = NALLOC(long,2*(nor + 2));

    /* Open seed vector file
       --------------------- */
    if ((seedfile = zreadhdr(seedname,&fl2,&seeddim,&noc2)) == NULL)
	MtxErrorExit(NULL);
    cycldim = NALLOC(long,seeddim);
    if (cycldim == NULL) 
	MtxErrorExit(NULL);
    if (fl2 < 2) 
	MtxErrorExit("%s: %E",seedname,ERR_NOTMATRIX);
    if (noc2 != nor || fl2 != fl)
	MtxErrorExit("%s and %s: %E",genname[0],seedname,ERR_INCOMPAT);
    nseed = 0;

    /* If a single seed vector has been selected in READ mode (-sr),
       seek to that vector */
    if (smode == READ && seedvecno > 0)
	zseek(seedfile,seedvecno);
}


/* ------------------------------------------------------------------
   init_args()
   ------------------------------------------------------------------ */

static void init_args(int argc, char **argv)

{
    int i;
    int opt_g = 0;

    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("G1g:s:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -1000; break;
	    case '1': FindGen = 1; break;
	    case 'g':
		ngen = MtxGetInt();
		opt_g = 1;
		if (ngen < 2 || ngen > 10000 || *opt_text_ptr != 0)
		    MtxErrorExit("-g: %E",ERR_OPTION);
		break;

            case 's':   /* Seed mode */
                switch (opt_text[0])
                {   case 'g': smode = GENERATE; ++opt_text_ptr; break;
                    case 'r': smode = READ; ++opt_text_ptr; break;
                    default: MtxErrorExit("-s: %E",ERR_OPTION); break; 
                }
                if (*opt_text_ptr != 0)
                {   seedvecno = MtxGetInt();
                    if (seedvecno <= 0) MtxErrorExit("-s: %E",ERR_OPTION);
                }
                if (*opt_text_ptr != 0) MtxErrorExit("-s: %E",ERR_OPTION);
                break;
	}
    }

    /* Allocate memory for generator pointer and name table
       ---------------------------------------------------- */
    genname = NALLOC(char *, ngen);
    gen = NALLOC(PTR, ngen);

    /* Set file names
       -------------- */
    if (opt_g)
    {
	char buf[200];
	switch (argc - opt_ind)
	{
	    case 4:
		OpName = argv[opt_ind + 3];
	    case 3:
    		seedname = argv[opt_ind + 1];
    		outname = argv[opt_ind + 2];
		break;
	    default:
	    	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
	}
	for (i = 0; i < ngen; ++i)
	{
	    sprintf(buf,"%s.%d",argv[opt_ind],i+1);
	    genname[i] = Smalloc(strlen(buf)+1);
	    strcpy(genname[i],buf);
	}
    }
    else	/* Two generators */
    {
	switch (argc - opt_ind)
	{
	    case 5:
		OpName = argv[opt_ind+4];
	    case 4:
		genname[0] = argv[opt_ind+0]; 
		genname[1] = argv[opt_ind+1]; 
    		seedname = argv[opt_ind+2];
    		outname = argv[opt_ind+3];
		break;
	    default:
	    	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
	}
    }
}




/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    long i, dim, src, olddim;
    PTR a, b, aget, bget;
    int g;
    FEL f;
    int finished = 0;
    PTR base;			/* Seed space basis for -sg */

    init_args(argc,argv);
    init();

    seedno = seedvecno - 1;

    if (smode == GENERATE)
    {
 	base = zalloc(seeddim);
	zreadvec(seedfile,base,seeddim);
	zpseed_init(seeddim,base);
    }

    a = A; b = B;
    dim = 0;

    /* Try until we have found a generating vector
       or there are no more seed vectors.
       ------------------------------------------- */
    while (!finished)
    {


	/* Get the next seed vector. Depending on the seed mode,
	   we read the next vector or call the seed generator
	   ----------------------------------------------------- */
	++seedno;
	if (smode == READ)
	{
	    if (seedno > seeddim)	/* End of vectors? */
		break;
	    zreadvec(seedfile,a,1);	/* Read vector */
	}
	else if (smode == GENERATE)
	{
	    if (seedno == seedvecno)	/* First time */
		seedno = zpseed_make(seedno);
	    else
		seedno = zpseed_next();
	    if (seedno < 0)		/* No more vectors */
		break;
	    zmoverow(a,zpseed_vec);	/* Copy vector */
	}
	zmoverow(b,a);

	/* Read next seed vector. Check if it is
	   contained in the subspace found so far.
  	   --------------------------------------- */
	cycldim[nseed] = 0;
	OPVEC(dim+1) = nseed + 1;
	OPGEN(dim+1) = 0;	/* It's a seed vector */
	zcleanrow(a,A,dim,piv);
	i = zfindpiv(a,&f);
	if (i == 0) continue;	/* Discard this vector */

	/* Initialize everything for spin-up.
   	   ---------------------------------- */
	olddim = dim;
	aget = a;
	bget = b;
	src = dim;
	zadvance(&a,(long)1);
	zadvance(&b,(long)1);
	piv[++dim] = i;
	g = 0;

	/* Spin up using the standard basis procedure
   	   ------------------------------------------ */
	while (dim < nor && aget != a)
	{
	    zmaprow(aget,gen[g],nor,a);
	    zcleanrow(a,A,dim,piv);
	    i = zfindpiv(a,&f);
	    if (i != 0)		/* We have a new vector */
	    {
		piv[++dim] = i;
		zmaprow(bget,gen[g],nor,b);
		zadvance(&a,(long)1);
		zadvance(&b,(long)1);
		OPVEC(dim) = src + 1;
		OPGEN(dim) = g + 1;
	    }
	    if (++g >= ngen)
	    {
		g = 0;
		zadvance(&aget,(long)1);
		zadvance(&bget,(long)1);
		++src;
	    }
	}

	/* Print a message for this seed vector
  	   ------------------------------------ */
	MESSAGE(0,("VECTOR %ld SPANS %ld\n",seedno,dim-olddim));
	cycldim[nseed++] = dim-olddim;

	/* If the vector did not span the whole space and the -1
           option was used, reset everything and start from zero.
           ------------------------------------------------------ */
	if (FindGen && dim < nor)
	{
    	    a = A; b = B;
    	    dim = 0;
	    MESSAGE(1,("The previous vector was not cyclic -- discarded\n"));
	}
    
	/* Are we done? */
	finished = (dim >= nor);
    }

    /* Done! Check if we have found a basis and print a
       warning message if not.
       ------------------------------------------------- */
    fclose(seedfile);
    if (dim < nor && msg_level >= 0)
	fprintf(stderr,"ZSB WARNING - SPAN IS ONLY %ld OF SPACE %ld\n",
	    dim,nor);
    if (nseed < seeddim && msg_level >= 0)
	fprintf(stderr,
	    "ZSB WARNING - ONLY %ld OF %ld SEED VECTORS WERE USED\n",
	    nseed,seeddim);

    /* Write out the result
       -------------------- */
    if (opt_G)
    {
	int k;

	printf("MeatAxe.Dimension := %ld;\n",dim);
	printf("MeatAxe.CyclicDimensions := [");
	for (k = 0; k < nseed; ++k)
	{
	    if (k > 0)
	    {
		printf(",");
		if (k % 10 == 0) printf("\n    ");
	    }
	    printf("%ld",cycldim[k]);
	}
	printf("];\n");
    }

    /* Write Standard basis
       -------------------- */
    if ((outfile = zwritehdr(outname,fl,dim,nor)) == NULL)
	MtxErrorExit(NULL);
    zwritevec(outfile,B,dim);
    fclose(outfile);

    /* Write <Op> file
       --------------- */
    if (OpName != NULL)
    {
    	if ((outfile = zwritehdr(OpName,-T_IMAT,dim,2)) == NULL)
	    MtxErrorExit(NULL);
	zwritelong(outfile,OpTable+2,2*dim);
    	fclose(outfile);
    }
    MtxCleanup();
    return (EXIT_OK);
}



