/* ============================= C MeatAxe ==================================
   File:        $Id: zsc.c,v 1.5 1998/07/29 16:52:20 mringe Exp $
   Comment:     Spin up with script.
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"

#include <string.h>
#include <stdlib.h>



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

char *GenName;		/* File name of generators */
char *SeedName;		/* Seed vector file name */
char *OutName = NULL;	/* Output file name (base name) */
char *OpName;		/* Name of <op> file */

FILE *seedfile, *outfile;
long fl, nor;		/* Field and size of the generators */
long *OpTable;		/* Operations, written to <Op> */
#define OPVEC(i) OpTable[2*(i)]
#define OPGEN(i) OpTable[2*(i)+1]
int ngen = 2;		/* Number of generators */
PTR *gen;		/* Generators */
PTR A;			/* Workspace */
long NOp;		/* Script size (number of rows in proj.) */

static char *helptext[] = {
"SYNTAX",
"    zsc [<Options>] <Gen> <Seed> <Op> [<Out>]",
"",
"OPTIONS",
"    -Q            Quiet, no messages.",
"    -V            Verbose, more messages.",
"    -T <Seconds>  Set CPU time limit.",
"    -Q            Quiet, no messages.",
"    -g <#Gen>     Set number of generators.",
"",
"FILES",
"    <Gen>   i  Generator name (base name, '.1', '.2' appended)",
"    <Seed>  i  Seed vectors",
"    <Op>    i  Skript file as produced by ZSB",
"    <Out>   o  Output file (base name). Full names are derived by appending",
"               the seed vector number. Defaults to <Seed>.",
"",
NULL};

static proginfo_t pinfo =
   { "zsc", "Spin up with script", "$Revision: 1.5 $", helptext };



/* ------------------------------------------------------------------
   InitArgs() - Kommandozeile interpretieren
   ------------------------------------------------------------------ */

static void InitArgs(int argc, char **argv)

{
    int i;

    /* Options
       ------- */
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("g:")) != OPT_END)
    {
	switch (i)
	{
	    case 'g':
		ngen = MtxGetInt();
		if (ngen < 2 || ngen > 1000 || *opt_text_ptr != 0)
		    MtxErrorExit("-g: %E",ERR_OPTION);
		break;
	}
    }

    /* Set file names
       -------------- */
    OutName = NULL;
    switch (argc - opt_ind)
    {
	case 4:
	    OutName = argv[opt_ind + 3];
	case 3:
    	    OpName = argv[opt_ind + 2];
    	    SeedName = argv[opt_ind + 1];
	    GenName = argv[opt_ind];
	    break;
	default:
	    MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    }
    if (OutName == NULL)
	OutName = SeedName;	/* Use default */


}


/* --------------------------------------------------------------------------
   ReadFiles() - Read input files

   Description:
     This function reads the generators and the script file. It also
     allocates a workspace for spin-up.
   -------------------------------------------------------------------------- */
	     
static void ReadFiles()

{
    int i;
    char fn[200];
    FILE *f;
    long noc, fl3;

    /* Read generators
      ----------------- */
    gen = NALLOC(PTR, ngen);
    for (i = 0; i < ngen; ++i)
    {
    	long fl2, nor2, noc2;

	/* Make file name, open file and read file header
	   ---------------------------------------------- */
	sprintf(fn,"%s.%d",GenName,i+1);	
	/* Read header */
	if ((f = zreadhdr(fn,&fl2,&nor2,&noc2)) == NULL)
	    MtxErrorExit(NULL);
	if (fl2 < 2) 
	    MtxErrorExit("%s: %E",fn,ERR_NOTMATRIX);
	MESSAGE(1,("reading %s: %ldx%ld matrix over GF(%ld)\n",
	    fn,nor2,noc2,fl2));
	if (nor2 != noc2) 
	    MtxErrorExit("%s: %E",fn,ERR_NOTSQUARE);

	/* Compatibility check
           ------------------- */
	if (i == 0)
	{
	    fl = fl2;
	    nor = nor2;
	}
	else if (fl != fl2 || nor != nor2)
	    MtxErrorExit("%s.1 and %s.%d: %E",GenName,GenName,i+1,ERR_INCOMPAT);

	/* Read matrix
	   ----------- */
	zsetfield(fl);
	zsetlen(nor);
	gen[i] = zalloc(nor);
	zreadvec(f,gen[i],nor);
	fclose(f);
    }

    /* Read operations
       --------------- */
    if ((f = zreadhdr(OpName,&fl3,&NOp,&noc)) == NULL)
	MtxErrorExit(NULL);
    if (fl3 != -T_IMAT)
	MtxErrorExit("%s: %E (bad file type)",OpName,ERR_FILEFMT);
    if (noc != 2)
	MtxErrorExit("%s: %E (bad number of columns)",OpName,ERR_FILEFMT);
    MESSAGE(1,("reading %s: %ld operations\n",OpName,NOp));
    OpTable = NALLOC(long,2*(NOp + 2));
    zreadlong(f,OpTable+2,2*NOp);
    fclose(f);

    /* Check the script for errors
       --------------------------- */
    if (OPGEN(1) != 0)
	MtxErrorExit("Illegal skript (does not start with first vector)");

    for (i = 2; i <= NOp; ++i)
    {
	if (OPGEN(i) == 0)
	    MtxErrorExit("Illegal skript (more than 1 seed vector)");
	if (OPGEN(i) < 1 || OPGEN(i) > ngen)
	    MtxErrorExit("Illegal skript (pos %d: generator out of range)",i);
	if (OPVEC(i) <= 0 || OPVEC(i) >= i)
	    MtxErrorExit("Illegal skript (pos %d: vector out of range)",i);
    }

    /* Allocate workspace
       ------------------ */
    A = zalloc(NOp+1);
}


/* ------------------------------------------------------------------
   Spinup() - Spin up the seed vector using the skript in OpTable.
   ------------------------------------------------------------------ */

static void Spinup()

{
    size_t i;
    PTR x;

    /* The seed vector is already at row 1, so we
       start with row 2.
       ------------------------------------------ */
    x = A;
    zadvance(&x,(long)1);

    /* Generate all remaining vectors (2...dim)
       ---------------------------------------- */
    for (i = 2; i <= (size_t) NOp; ++i, zadvance(&x,1))
    {
	PTR y, g;
	long opvec = OPVEC(i);

	/* Select the vector to map
	   ------------------------ */
    	y = A;
    	zadvance(&y,opvec-1);

	/* Select the generator to apply
	   ----------------------------- */
	g = gen[(int)OPGEN(i)-1];

	/* Multiply the vector by the generator
	   ------------------------------------ */
	zmulrow(x,F_ZERO);
	zmaprow(y,g,nor,x);
	MESSAGE(2,("%d = %ld * gen %ld\n",(int)i,opvec,OPGEN(i)));
    }
}


/* ------------------------------------------------------------------
   WriteBasis() - Write the basis in A to a file. <no> is the seed
   	vector number.
   ------------------------------------------------------------------ */

static void WriteBasis(long no)

{
    char fn[200];
    FILE *f;

    /* Make file name
       -------------- */
    sprintf(fn,"%s.%4.4ld",OutName,no);

    /* Write file
       ---------- */
    if ((f = zwritehdr(fn,fl,NOp,nor)) == NULL)
	MtxErrorExit(NULL);
    zwritevec(f,A,NOp);
    fclose(f);
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    FILE *f;
    long seedno;
    long fl3, nseed, noc3;

    InitArgs(argc,argv);
    ReadFiles();

    /* Open seed file
       -------------- */
    if ((f = zreadhdr(SeedName,&fl3,&nseed,&noc3)) == NULL)
	MtxErrorExit(NULL);
    if (fl3 != fl)
	MtxErrorExit("%s.1 and %s: %E",GenName,SeedName,ERR_INCOMPAT);
    if (fl3 < 2)
    	MtxErrorExit("%s: %E",SeedName,ERR_NOTMATRIX);
    MESSAGE(1,("%s: %ld seed vectors\n",SeedName,nseed));

    /* Loop through all seed vectors
       ----------------------------- */
    for (seedno = 1; seedno <= nseed; ++seedno)
    {
    	/* Read next vector into workspace
	   ------------------------------- */
	zreadvec(f,A,(long)1);
	MESSAGE(1,("Vector %ld\n",seedno));

	/* Spin up and write out the result
	   -------------------------------- */
	Spinup();
	WriteBasis(seedno);
    }

    fclose(f);
    MtxCleanup();
    return EXIT_OK;
}



