/* ============================= C MeatAxe ==================================
   File:        $Id: zsi.c,v 1.3 1998/04/28 17:38:38 mringe Exp $
   Comment:     Sum and intersection of two subspaces.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <stdlib.h>
#include <string.h>
#include "meataxe.h"



/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

long fl, nor1, nor2, noc;
long *piv;
FILE *afile, *bfile;
long sumdim, secdim;
PTR sum, sec, intersection;
char *aname, *bname, *sumname, *intname;

static char *helptext[] = {
"SYNTAX",
"    zsi [-GQV] <Space1> <Space2> <Sum> <Int>",
"    zsi -s [-GQV] <Space1> <Space2> <Sum>",
/*"    zsi -i [-GQV] <Space1> <Space2> <Int>",*/
"",
"OPTIONS",
"    -s     Calculate sum only",
/*"    -i     Calculate intersection only",*/
"    -G     GAP output",
"    -Q     Quiet, no messages",
"    -V     Verbose, more messages",
"",
"FILES",
"    <Spc1>    i  The first space, a matrix.",
"    <Spc2>    i  The second space, a matrix.",
"    <Sum>     o  The sum of the two spaces.",
"    <Int>     o  The intersection of the two spaces.",
NULL};

static proginfo_t pinfo =
   { "zsi", "Sum And Intersection", "$Revision: 1.3 $", helptext };


int opt_G = 0;
int opt_s = 0;

/* ------------------------------------------------------------------
   openinp() - Open input files
   ------------------------------------------------------------------ */

static void openinp()

{
    long fl1, fl2, noc1, noc2;

    if ((afile = zreadhdr(aname,&fl1,&nor1,&noc1)) == NULL)
	MtxErrorExit(NULL);
    MESSAGE(1,("%s: %ld vectors, row size = %ld\n",aname,nor1,noc1));
    if (fl1 < 2) 
	MtxErrorExit("%s: %E",aname,ERR_NOTMATRIX);
    if ((bfile = zreadhdr(bname,&fl2,&nor2,&noc2)) == NULL)
	MtxErrorExit(NULL);
    MESSAGE(1,("%s: %ld vectors, row size = %ld\n",bname,nor2,noc2));
    if (fl2 < 2) 
	MtxErrorExit("%s: %E",bname,ERR_NOTMATRIX);
    if (fl1 != fl2 || noc1 != noc2)
	MtxErrorExit("%s and %s: %E",aname,bname,ERR_INCOMPAT);

    fl = fl1;
    noc = noc1;
    zsetfield(fl); 
    zsetlen(noc);
}


/* ------------------------------------------------------------------
   sumonly() - Calculate the sum
   ------------------------------------------------------------------ */

static void sumonly()

{
    FILE *cfile;
    PTR a, b;
    long maxrows;

    /* Read first space
       ---------------- */
    openinp();
    maxrows = (noc < nor1 + nor2) ? noc : nor1 + nor2;
    a = zalloc(maxrows);
    zreadvec(afile,a,nor1);
    fclose(afile);

    /* Make echelon form of first space
       -------------------------------- */
    piv = NALLOC(long,maxrows+1);
    if (piv == NULL) 
	MtxErrorExit(NULL);
    nor1 = zmkechelon(a,nor1,piv);

    /* Clean vectors from 2nd space and extend first space
       --------------------------------------------------- */
    b = a;
    zadvance(&b,nor1);
    while (nor1 < maxrows && nor2 > 0)	/* Any vectors left? */
    {
	long nread;
	PTR x;

	/* Read as many vectors as possible
	   -------------------------------- */
	nread = maxrows - nor1;
	if (nread > nor2) 
	    nread = nor2;
        zreadvec(bfile,b,nread);
	nor2 -= nread;

	/* Clean the new vectors
	   --------------------- */
	for (x = b; nread > 0; --nread, zadvance(&x,(long)1))
	{
	    long p;
	    FEL f;
	    zcleanrow(x,a,nor1,piv);
	    if ((p = zfindpiv(x,&f)) > 0)  /* A new vector! */
	    {
		++nor1;
		piv[nor1] = p;
		if (x != b) zmoverow(b,x);
		zadvance(&b,1);
	    }
	}
    }
    fclose(bfile);

    MESSAGE(0,("SUM %ld\n",nor1));
    if ((cfile = zwritehdr(sumname,fl,nor1,noc)) == NULL)
	MtxErrorExit(NULL);
    zwritevec(cfile,a,nor1);
    fclose(cfile);
}


/* ------------------------------------------------------------------
   readfiles() - Read the input files, allocate tables
   ------------------------------------------------------------------ */

static void readfiles()

{
    PTR x;

    openinp();
    sum = zalloc(nor1+nor2);
    sec = zalloc(nor1+nor2);
    zreadvec(afile,sum,nor1);
    fclose(afile);
    x = sum;
    zadvance(&x,nor1);
    zreadvec(bfile,x,nor2);
    fclose(bfile);
    memcpy(sec,sum,zsize(nor1));
    piv = NALLOC(long,nor1+nor2+1);
    if (piv == NULL) 
	MtxErrorExit(NULL);
}


/* ------------------------------------------------------------------
   suminter() - Calculate sum and intersection (Zassenhaus algorithm)
   ------------------------------------------------------------------ */

static void suminter()

{
    long i, k, p = 0;
    PTR xi, xk, yi, yk;
    FEL f;

    /* Step 1
       ------ */
    xi = sum;
    yi = sec;
    for (i = 1; i <= nor1+nor2; )
    {
    	xk = xi;
    	yk = yi;
    	for (k = i; k <= nor1+nor2; ++k)
    	{	p = zfindpiv(xk,&f);
		if (p != 0) break;
		zadvance(&xk,(long)1);
		zadvance(&yk,(long)1);
	}
	if (p == 0) break;
	if (k != i)
	{	zswaprow(xk,xi);
		zswaprow(yk,yi);
	}
	xk = sum;
	yk = sec;
	for (k = 1; k < i; ++k)
	{	f = zneg(zextract(xi,piv[k]));
		zaddmulrow(xi,xk,f);
		zaddmulrow(yi,yk,f);
		zadvance(&xk,(long)1);
		zadvance(&yk,(long)1);
	}
	p = zfindpiv(xi,&f);
	if (p != 0)
	{	piv[i++] = p;
		zmulrow(xi,zinv(f));
		zmulrow(yi,zinv(f));
		zadvance(&xi,(long)1);
		zadvance(&yi,(long)1);
	}
    }
    sumdim = i-1;
    intersection = yi;
    MESSAGE(1,("Step 1 completed, sumdim = %ld\n",sumdim));

    /* Step 2
       ------ */
    while (1)
    {	if (i > nor1+nor2) break;
		yk = yi;
		for (k = i; k <= nor1+nor2; ++k)
		{	p = zfindpiv(yk,&f);
			if (p != 0) break;
			zadvance(&yk,(long)1);
		}
		if (p == 0) break;
		if (k != i)
			zswaprow(yk,yi);
		yk = intersection;
		for (k = sumdim+1; k < i; ++k)
		{	f = zneg(zextract(yi,piv[k]));
			zaddmulrow(yi,yk,f);
			zadvance(&yk,(long)1);
		}
		p = zfindpiv(yi,&f);
		if (p != 0)
		{	piv[i++] = p;
			zmulrow(yi,zinv(f));
			zadvance(&yi,(long)1);
		}
    }
    secdim = i - sumdim - 1;

    /* Print result
       ------------ */
    if (opt_G)
    {
        printf("MeatAxe.SumDimension := %ld;\n",sumdim);
        printf("MeatAxe.IntDimension := %ld;\n",secdim);
    }
    else
        MESSAGE(0,("SUM %ld INTERSECTION %ld\n",sumdim,secdim));
}



/* ------------------------------------------------------------------
   writefiles() - Write out the result
   ------------------------------------------------------------------ */

static void writefiles()

{	
    FILE *f;
    
    if ((f = zwritehdr(sumname,fl,sumdim,noc)) == NULL)
	MtxErrorExit(NULL);
    zwritevec(f,sum,sumdim);
    fclose(f);
    if ((f = zwritehdr(intname,fl,secdim,noc)) == NULL)
	MtxErrorExit(NULL);
    zwritevec(f,intersection,secdim);
    fclose(f);
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    int i;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("sG")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 's': opt_s = 1; break;
	}
    }
    if (opt_ind != argc - (opt_s ? 3 : 4)) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    aname = argv[opt_ind];
    bname = argv[opt_ind+1];
    sumname = argv[opt_ind+2];
    if (!opt_s) intname = argv[opt_ind+3];
	
    if (opt_s)
	sumonly();
    else
    {
	readfiles();
    	suminter();
    	writefiles();
    }
    MtxCleanup();
    return (EXIT_OK);
}

