/* ============================= C MeatAxe ==================================
   File:        $Id: zsm.c,v 1.8 1998/06/16 17:19:44 mringe Exp $
   Comment:     Small matrices package
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>




/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

char **args;
int arg_ind, nargs;
int ngen=2;
int opt_g = 0;
int opt_G = 0;
matrix_t *gen[MAXGEN];
perm_t *genp[MAXGEN];
int gentype;
static char *helptext[] = {
"SYNTAX",
"    zsm [-GQV] [-T <#Secs>] [-g <#Gen>] <Command> <Files>",
"",
"    zsm pwr<Num> <Mat|Perm> <Result>",
"      Raise a matrix or permutation to the <Num>-th power",
"",
"    zsm ro[<Num>] <Mat|Perm1> <Mat|Perm2>",
"      Calculate <Num> random orders. Default is <Num>=120.",
"",
"    zsm fp <Mat1> <Mat2>",
"       Calculate the standard fingerprint (6 Nullities).",
"",
"    zsm fp[<Start>-]<End> <Mat1> <Mat2>",
"      Calculate nullities from word number <Start> (default = 1) up",
"      to word number <End>.",
"",
"    zsm mw<Num> <Mat1> <Mat2> <Word> [<Nullspace>]",
"      Write word number <Num> to <Word> and its nullspace, if",
"      desired, to <Nullspace>. `mw<Num>id' makes <Num> * Id.",
"",
"OPTIONS",
"    -G    GAP output.",
"    -Q    Quiet, no messages.",
"    -V    Verbose, more messages.",
"    -T    Set CPU time limit.",
NULL
};

static proginfo_t pinfo =
   { "zsm", "Small Matrices Package", "$Revision: 1.8 $", helptext };




/* ------------------------------------------------------------------
   error() - Print error message and exit
   ------------------------------------------------------------------ */

static void error(char *fmt, ...)

{
    va_list vl;

    va_start(vl,fmt);
    fprintf(stderr,"zsm: ");
    vfprintf(stderr,fmt,vl);
    va_end(vl);
    fprintf(stderr,"\n");
    exit(EXIT_ERR);
}

/* ------------------------------------------------------------------
   filetype() - Get field parameter from file header
   ------------------------------------------------------------------ */

static long filetype(char *fn)

{
    long fl;
    FILE *f;

    f = SFOpen(fn,FM_READ);
    if (f == NULL) 
	MtxErrorExit("Cannot open %s: %S",fn);
    if (zreadlong(f,&fl,1) != 1) 
	MtxErrorExit("%s: Cannot read header: %S",fn);
    fclose(f);
    return fl;
}


/* ------------------------------------------------------------------
   readgens() - Read matrices or permutations
   ------------------------------------------------------------------ */

static void readgens(void)

{
    int i;
    char name[100];
    char name0[100];
    long ft;

    for (i = 0; i < ngen; ++i)
    {
	if (opt_g)
	{
	    if (arg_ind >= nargs)
		error("missing generator name");
	    sprintf(name,"%s.%d",args[arg_ind],i+1);
	}
	else
	{
	    if (arg_ind+i >= nargs)
		error("missing generator name");
	    strcpy(name,args[arg_ind+i]);
	}
	ft = filetype(name);
	if (i == 0)
	{   gentype = (ft >= 2) ? T_MATRIX : T_PERM;
	    strcpy(name0,name);
	}
	switch (gentype)
	{
	    case T_MATRIX:
		if (ft < 2) 
		    MtxErrorExit("%s: %E",name,ERR_NOTMATRIX);
		gen[i] = MatLoad(name);
	        if (gen[i]->nor != gen[i]->noc)
		    MtxErrorExit("%s: %E",name,ERR_NOTSQUARE);
	       	if (i > 0)
	    	{
		    if (gen[i]->fl != gen[0]->fl || gen[i]->nor != gen[0]->nor)
			MtxErrorExit("%s and %s: %E",name0,name,ERR_INCOMPAT);
	    	}
		break;
	    case T_PERM:
	    	if (filetype(name) != -1)
		    MtxErrorExit("%s: %E",name,ERR_NOTPERM);
	    	genp[i] = PermLoad(name);
	    	if (i > 0)
	    	{
		    if (genp[i]->deg != genp[0]->deg)
			MtxErrorExit("%s and %s: %E",name0,name,ERR_INCOMPAT);
	    	}
		break;
	}
    }
    if (opt_g)
	++arg_ind;
    else
	arg_ind += 2;
}


/* ------------------------------------------------------------------
   Makeword() - Calculate a word and its null-space
   ------------------------------------------------------------------ */

static void Makeword(long l)

{
    matrix_t *nsp, *word;
    wgdata_t *b;

    readgens();
    if (arg_ind >= nargs)
	error("Missing output file name");
    if (l >= 0)
    {
    	b = WgInit(ngen,(const matrix_t **)gen);
    	word = WgMakeWord(b,l);
    	MESSAGE(0,("WORD %ld (%s)",l,WgSymbolicName(b,l)));
    }
    else
    {
	long dim = gen[0]->nor, i;
	FEL f = zitof(-l);
	PTR x;
	word = MatAlloc(gen[0]->fl,dim,dim);
	for (x = word->d, i = 1; i <= dim; ++i, zadvance(&x,1))
	    zinsert(x,i,f);
    	MESSAGE(0,("WORD %ldid (%ld*Id)",-l,-l));
    }
    MatSave(word,args[arg_ind++]);
    if (arg_ind < nargs)
    {
	nsp = MatNullSpace__(word);
	MatSave(nsp,args[arg_ind++]);
	MESSAGE(0,(" HAS NULLITY %ld",nsp->nor));
    }
    MESSAGE(0,("\n"));
}


/* ------------------------------------------------------------------
   fingerprint() - Calculate nullities
   ------------------------------------------------------------------ */

static void fingerprint(long l1,long l2)

{
    wgdata_t *b;
    long w;
    int count = 0;

#ifdef HACK
int n[100];
int i;
memset(n,0,sizeof(n));
#endif

    readgens();
    b = WgInit(ngen,(const matrix_t **)gen);

    if (opt_G)
	printf("MeatAxe.Fingerprint := [");
    else
        printf("NUMBER NULLITY WORD\n");
    for (w = l1; w <= l2; ++w)
    {	
	matrix_t *word = WgMakeWord(b,w);
        if (opt_G)
	{
	    if (w > l1) printf(",");
	    if (++count % 5 == 0) printf("\n  ");
	    printf("[%ld,%ld]",w,MatNullity__(word));
	}
    	else
	{
	    long nul = MatNullity__(word);
#ifdef HACK
	    if (nul < 100)
		++n[nul];
#else
	    printf("%6ld %7ld %s\n",w,nul,WgSymbolicName(b,w));
#endif
	}
    }
    if (opt_G)
	printf("];\n");
#ifdef HACK
for (i = 0; i < 40; ++i) printf("%8d",n[i]);
#endif
}


/* ------------------------------------------------------------------
   std_fingerprint() - Standard fingerprint (6 nullities)
   ------------------------------------------------------------------ */

static void std_fingerprint(void)

{
    wgdata_t *b;
    int i;
    long fp[MAXFP];

    readgens();
    b = WgInit(ngen,(const matrix_t **)gen);
    WgMakeFingerPrint(b,fp);
    if (opt_G)
	printf("MeatAxe.Fingerprint := [ ");
    else
        printf("FINGERPRINT: ");
    for (i = 0; i < MAXFP; ++i)
    {
	if (opt_G)
	{
	    if (i > 0) printf(",");
	    printf("%ld",fp[i]);
	}
	else
	    printf("%3ld ",fp[i]);
    }

    if (opt_G)
	printf("];\n");
    else
        printf("\n");
}



/* ------------------------------------------------------------------
   random_orders() - Calculate random orders
   ------------------------------------------------------------------ */

static void random_orders(long n)

{
    matrix_t *m = NULL;
    perm_t *p = NULL;
    long count;

    readgens();
    if (gentype == T_MATRIX)
	m = MatDup(gen[0]);
    else
	p = PermDup(genp[0]);

    if (opt_G)
	printf("MeatAxe.RandomOrders := [");

    for (count = 0; count < n; ++count)
    {
	long o;
	if (gentype == T_MATRIX)
	    o = MatOrder(m);
	else
	    o = PermOrder(p);
	if (count % 12 == 0) printf("\n    ");
	printf("%4ld",o);
    	if (opt_G && count < n-1) printf(",");

	if (gentype == T_MATRIX)
	    MatMul(m,gen[MtxRandomInt(ngen)]);
	else
	    PermMul(p,genp[MtxRandomInt(ngen)]);
    }
    if (opt_G)
	printf("];\n");
    else
	printf("\n");
}



/* ------------------------------------------------------------------
   power() - Power of a matrix or permutation
   ------------------------------------------------------------------ */

static void power(long n)

{
    matrix_t *x, *y;
    perm_t *p, *q;

    if (filetype(args[0]) == -1)
    {
	p = PermLoad(args[0]);
	q = PermPower(p,n);
	PermSave(q,args[1]);
    }
    else
    {
	x = MatLoad(args[0]);
   	if (x->nor != x->noc)
	    MtxErrorExit("%s: %E",args[0],ERR_NOTSQUARE);
	y = MatPower(x,n);
	MatSave(y,args[1]);
    }
    arg_ind += 2;
}


/* ------------------------------------------------------------------
   mygetint() - Convert string to integer
   match() - Compare string with pattern
   ------------------------------------------------------------------ */

char *cp;

static int mygetint(long *l)

{	if (!isdigit(*cp))
		return 1;
	for (*l = 0; isdigit(*cp); ++cp)
		*l = 10 * *l + (*cp - '0');
	return 0;
}


static int match(char *pattern)

{	char *tmp = cp;

	while (*pattern != 0)
	{	switch (*pattern)
		{	case '$':
				if (*tmp != 0) return 1;
				break;
			case '#':
				if (!isdigit(*tmp)) return 1;
				while (isdigit(*tmp)) ++tmp;
				break;
			default:
				if (*tmp != *pattern) return 1;
				++tmp;
				break;
		}
		++pattern;
	}
	cp = tmp;
	return 0;
}




/* ------------------------------------------------------------------
   init() - Process command line options and arguments
   ------------------------------------------------------------------ */

static void init(int argc, char *argv[])

{   int i;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("Gg:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 'g':
		opt_g = 1;
		ngen = MtxGetInt();
		if (ngen < 1 || ngen > MAXGEN || *opt_text_ptr != 0)
		    MtxErrorExit("'-g': %E",ERR_OPTION);
		break;
	}
    }

    /* Skip command
       ------------ */
    if (opt_ind >= argc) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    cp = argv[opt_ind++];


    /* Set file names
       -------------- */
    args = argv + opt_ind;
    nargs = argc - opt_ind;
    arg_ind = 0;
}




/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
#if 0
wgdata_t *wg;
matrix_t *g[20];
int i;
for (i = 0; i < 20; ++i) g[i] = MatAlloc(2,10,10);

wg = WgInit(2,g);
for (i = 10000; i < 20000; i++)
{
char a[100];
printf("Word %5d=%s\n",i,WgSymbolicName(wg,i));
if (i % 20 == 0)
    gets(a);
}
exit(0);
#endif

    init(argc,argv);

    if (!match("pwr"))	/* pwr - Power */
    {	long l;
	if (mygetint(&l) || match("$") || nargs != 2)
	    error("Usage: `zsm pwr<Num> <Input> <Result>'");
	power(l);
    }
    else if (!match("ro"))	/* ro - Random orders */
    {	long l = 120;
	if ((match("$") && (mygetint(&l) || match("$"))))
	    error("Usage: `zsm ro<Num> <Gen1> <Gen2>'");
	random_orders(l);
    }
    else if (!match("fp"))	/* fp - Fingerprint */
    {
	if (!match("$"))
	    std_fingerprint();
	else
	{   long l1, l2;
	    if (mygetint(&l1))
	    	error("Usage: fp[<Num>[-<Num>]]");
	    if (!match("$"))
		fingerprint(1,l1);
	    else
	    {	if (match("-") || mygetint(&l2) || match("$"))
	    	    error("Usage: fp[<Num>[-<Num>]]");
		fingerprint(l1,l2);
	    }
	}
    }
    else if (!match("mw"))	/* mw - Make word */
    {	long l;

	if (mygetint(&l)) error("Usage: mw<Number>");
	if (!match("id$")) l = -l;
	else if (match("$")) error("Usage: mw<Number>");
	Makeword(l);
    }
    else 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    if (arg_ind != nargs)
	error("Extra arguments ignored: %s ...",args[arg_ind]);
    MtxCleanup();
    return EXIT_OK;
}


