/* ============================= C MeatAxe ==================================
   File:        $Id: zsp.c,v 1.4 1998/04/28 17:38:41 mringe Exp $
   Comment:     Spin up and split.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "meataxe.h"






/* ------------------------------------------------------------------
   Constants
   ------------------------------------------------------------------ */

#define EXIT_SPLIT	1  /* Return value if space was split */
#define EXIT_NOSPLIT	0  /* Return value if space was not split */




/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

enum {DONTKNOW,MATRIX,PERMUTATION}
	gentype = DONTKNOW;	/* Type of generators */
enum {GENERATE,READ,READALL}
	smode = GENERATE;	/* Seed mode */
int opt_G = 0;			/* GAP output */
int dosplit = 1;		/* Split after spin-up */
int output = 1;			/* Enable file ouput */
long seedvecno = 0;		/* Selected vector (0 = all) */
long iseed = 0;			/* Current vector */
long maxdim = -1;		/* Limit on susbspace dimension */
long maxtry = -1;		/* How often do we try to apply gens */
int nmat = 2;			/* Number of matrices to split */
int ngen = 2;			/* Number of generators */
char *genname[MAXGEN] = {"G1","G2"};
char *subname[MAXGEN] = {"P4","P5"};
char *quotname[MAXGEN] = {"P6","P7"};
char *seedname = "G3";
char *vname = "V";
FILE *seedfile;

PTR	gen[MAXGEN];		/* Matrices to split */
PTR	m3, 			/* Subspace */
	tmp1, tmp2;		/* Two extra rows */

long	nor,			/* Degree of representation */
	dim,			/* Dimension of invariant subspace */
	ssdim,			/* Dimension of seed vector space */
	fl;			/* Field order */

long *piv;		/* Pivot table for the subspace */
char *ins;		/* Used to find the insignificant columns */


/* ------------------------------------------------------------------
   help()
   ------------------------------------------------------------------ */

static char *helptext[] = {
"SYNTAX",
"    zsp [<Options>] [<G1> <G2> <Seed> [{<S1> <S2> <Q1> <Q2> | <Spc>}]]",
"    zsp [<Options>] -g <N>[.<M>] [<Gen> [<Seed> [{<Sub> <Quot> | <Spc>}]]]",
"",
"OPTIONS",
"    -s<Mode>      Select seed for spin-up:",
"                    -sg[<Num>]   Generate seed vectors (Default)",
"                    -sr[<Num>]   Read seed vectors one by one",
"                    -ss          Read a whole space and make its closure",
"                  <Num>, if specified, selects a single vector",
"    -n            Spin up and write out invariant subspace, but do not split",
"    -o            No output to files, messages only",
"    -V            Verbose",
"    -Q            Quiet, no messages",
"    -d <Dim>      Set limit on dimension of subspace",
"    -t <Num>      Assume subspace is invariant after <Num> tries",
"    -T <Maxtime>  Set limit on CPU time (in seconds)",
"    -g <N>[.<M>]  Read <M> matrices and treat the first <N> as generators",
"                  (default: <N>=<M>). Read matrices from <Gen>.i (i=1..<M>),",
"                  and write output to <Sub>.i and <Quot>.i.",
"",
"FILES",
"    <G1,2>   i  Generators: square matrices of equal size",
"    <Seed>   i  Seed vectors: a matrix with NOC(Seed)=NOC(G1)",
"    <S1,2>   o  Action on subspace",
"    <Q1,2>   o  Action on quotient",
"    <Spc>    o  Invariant subspace (with -n)",
NULL};

static proginfo_t pinfo =
   { "zsp", "Spin Up And Split", "$Revision: 1.4 $", helptext };



/* ------------------------------------------------------------------
   err() - Terminate program with error message
   ------------------------------------------------------------------ */

void err(int code)

{
    fprintf(stderr,"ZSP ERROR - ");
    switch (code)
    {
	case 'b':
	    fprintf(stderr,"INTERNAL ERROR\n");
	    break;
	case '0':
	    fprintf(stderr,"CANNOT SPIN UP ZERO VECTOR\n");
	    break;
	case 'v':
	    fprintf(stderr,"CANNOT MAKE THIS SEED VECTOR\n");
	    break;
	case 'S':
	    fprintf(stderr,
		  "SEED SPACE IS GREATER THAN DIMENSION LIMIT\n");
	    break;
	}
	exit(EXIT_ERR);
}



/* ------------------------------------------------------------------
   parseargs()
   ------------------------------------------------------------------ */

static void setname(char **name,int i,char *template)

{
    char *c;

    c = Smalloc(strlen(template) + 10);
    sprintf(c,"%s.%d",template,i+1);
    name[i] = c;
}



void parseargs(int argc, char *argv[])

{   int k, j;
    int opt_g = 0;


    /* Parse command line
       ------------------ */
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((k = MtxGetOpt("ionvGt:d:g:s:")) != OPT_END)
    {   switch (k)
	{ 
	    case 'o':
		output = 0;
		break;
	    case 'G':
		msg_level = -100;
		opt_G = 1;
		break;
	    case 'n':
		dosplit = 0;
		break;
	    case 't':
		maxtry = MtxGetInt();
		if (maxtry <= 0 || *opt_text_ptr != 0)
		    MtxErrorExit("Bad argument after -t");
		break;
	    case 'd':
		maxdim = MtxGetInt();
		if (maxdim <= 0 || *opt_text_ptr != 0)
		    MtxErrorExit("-d: %E",ERR_OPTION);
		break;
	    case 'g':
		opt_g = 1;
		nmat = (int) MtxGetInt();
		if (*opt_text_ptr == '.')
		{	++opt_text_ptr;
			ngen = (int) MtxGetInt();
		}
		else
		    ngen = nmat;
		if (ngen > nmat || ngen < 1 || nmat > MAXGEN ||
				*opt_text_ptr != 0)
		    MtxErrorExit("-g: %E",ERR_OPTION);
		break;
	    case 's':	/* Seed mode */
		switch (opt_text[0])
		{   
		    case 'g': smode = GENERATE; ++opt_text_ptr; break;
		    case 'r': smode = READ; ++opt_text_ptr; break;
		    case 's': smode = READALL; ++opt_text_ptr; break;
		    default: break; /* Allow -s<Num> for -sg<Num> */
		}
		if (smode == GENERATE || smode == READ)
		{   
		    if (*opt_text_ptr != 0)
		    {	
			seedvecno = MtxGetInt();
			if (seedvecno <= 0) 
			    MtxErrorExit("-s: %E",ERR_OPTION);
		    }
		}
		if (*opt_text_ptr != 0) 
		    MtxErrorExit("-s: %E",ERR_OPTION);
		break;
	}
    }

    /* Set file names
       -------------- */
    k = opt_ind;	/* Index of first file arg */
    if (!opt_g)		/* Old Syntax */
    {	switch (argc - k)
	{   case 0:
		break;
	    case 7:
		quotname[0] = argv[k+5];
		quotname[1] = argv[k+6];
	    case 5:
		subname[0] = argv[k+3];
		subname[1] = argv[k+4];
		if (!dosplit) 
		    MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
	    case 4:
		if (!dosplit)
			vname = argv[k+3];
	    case 3:
		genname[0] = argv[k+0];
		genname[1] = argv[k+1];
		seedname = argv[k+2];
		break;
	    default:
		MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
	}
    }
    else		/* The -g option was used */
    {
    	for (j = 0; j < nmat; ++j)
	{
	    setname(genname,j,"G");
	    setname(subname,j,"S");
	    setname(quotname,j,"Q");
	}
	seedname = "G0";
	switch (argc - k)
	{   case 4:
		if (!dosplit) 
		    MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
		for (j = 0; j < nmat; ++j)
		    setname(quotname,j,argv[k+3]);
	    case 3:
		if (!dosplit)
		    vname = argv[k+2];
		else
		    for (j = 0; j < nmat; ++j)
		        setname(subname,j,argv[k+2]);
	    case 2:
		seedname = argv[k+1];
	    case 1:
		for (j = 0; j < nmat; ++j)
		    setname(genname,j,argv[k]);
	    case 0:
		break;
	    default:
		MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
	}
    }
}




/* ------------------------------------------------------------------
   init() - Read input files and initialize
   ------------------------------------------------------------------ */

static char wm1[] =
   "zsp: Warning: Cannot calculate quotient for Permutations\n";
static char wm2[] =
   "zsp: Warning: Only one permutation read from %s\n";

void init(void)

{
    long fl2, noc, nor2, noc2;
    int j;
    FILE *f;

    /* Open the <Seed> file
       -------------------- */
    if (MSG2) printf("Open seed file (%s)\n",seedname);
    if ((seedfile = zreadhdr(seedname,&fl,&ssdim,&noc)) == NULL)
	MtxErrorExit(NULL);
    if (MSG1) printf("Found %ld seed vectors\n",ssdim);
    if (fl < 2) 
	MtxErrorExit("%s: %E",seedname,ERR_NOTMATRIX);
    nor = noc;


    /* Read the generators
       ------------------- */
    for (j = 0; j < nmat; ++j)	/* Read the generators */
    {
	if ((f = zreadhdr(genname[j],&fl2,&nor2,&noc2)) == NULL)
	    MtxErrorExit(NULL);
	if (fl2 == -1)	/* Permutations */
	{
	    if (gentype == MATRIX)
		MtxErrorExit("%s: %E",genname[j],ERR_NOTPERM);
	    if (gentype == DONTKNOW && dosplit) 
		fprintf(stderr,wm1);
	    gentype = PERMUTATION;
	    if (noc2 > 1)
		fprintf(stderr,wm2,genname[j]);
	    if (nor2 != noc)
		MtxErrorExit("%s and %s: %E",genname[j],seedname,ERR_INCOMPAT);
	    gen[j] = (PTR) NALLOC(long,noc);
	    if (gen[j] == NULL) 
		MtxErrorExit(NULL);
	    if (zreadlong(f,(long *)(gen[j]),noc) != (size_t) noc)
		MtxErrorExit(NULL);
	}
	else
	{
	    if (gentype == PERMUTATION)
		MtxErrorExit("%s: %E",genname[j],ERR_NOTMATRIX);
	    gentype = MATRIX;
	    if (fl2 != fl || noc2 != noc)
		MtxErrorExit("%s and %s: %E",genname[j],seedname,ERR_INCOMPAT);
	    if (noc2 != nor2) 
		MtxErrorExit("%s: %E",genname[j],ERR_NOTSQUARE);
	    zsetfield(fl);
	    zsetlen(nor);
	    gen[j] = zalloc(nor);
	    zreadvec(f,gen[j],nor);
	}
	fclose(f);
    }

    /* Set default value of maxdim
       --------------------------- */
    if (maxdim == -1) 
	maxdim = nor - 1; /* We test for dim > maxdim! */
    if (smode == READALL && maxdim < ssdim) 
	err('S');

    /* Allocate memory
       --------------- */
    zsetfield(fl);
    zsetlen(nor);
    m3 = zalloc(maxdim + 2);
    tmp1 = zalloc((long)1);
    tmp2 = zalloc((long)1);
    piv = NALLOC(long,nor+1);
    ins = NALLOC(char,nor+1);
    if (piv == NULL || ins == NULL)
	MtxErrorExit(NULL);
}




/* ------------------------------------------------------------------
   do_split() - Calculate the action of g1 and g2 on subspace and
	quotient.
   ------------------------------------------------------------------ */

void do_split(void)

{
    int g;
    long j, qdim;
    PTR xa, quotop;
    FILE *ofile;
    FEL dummy;


    /* Calculate the action on the subspace
       ------------------------------------ */
    MESSAGE(1,("Splitting..."));
    for (g = 0; g < nmat; ++g)
    {	
	if ((ofile = zwritehdr(subname[g],fl,dim,dim)) == NULL)
	    MtxErrorExit(NULL);
        MESSAGE(2,("%s...",subname[g]));
	xa = m3;
	for (j = 1; j <= dim; ++j)
	{
	    zsetlen(dim);
	    zmulrow(tmp2,F_ZERO);

	    /* Map the j-th basis vector
               ------------------------- */
	    zsetlen(nor);
	    switch (gentype)
	    {
	    	case MATRIX:
		    zmaprow(xa,gen[g],nor,tmp1);
		    break;
	    	case PERMUTATION:
		    zpermrow(xa,(long *)gen[g],tmp1);
		    break;
		case DONTKNOW:
		    err('b');
	    }
	    zadvance(&xa,(long)1);

	    /* Clean it with the basis and store the
               coefficients in tmp2.
               ------------------------- */
	    zcleanrow2(tmp1,m3,dim,piv,tmp2);

	    /* Check if the row is clean
	       ------------------------- */
	    if (zfindpiv(tmp1,&dummy) != 0)
	        fprintf(stderr,"WARNING: SUBSPACE IS NOT INVARIANT\n");

	    /* Write one row of output
	       ----------------------- */
	    zsetlen(dim);
		zwritevec(ofile,tmp2,(long)1);
	}
	fclose(ofile);
    }


    /* Calculate the action on the quotient
       ------------------------------------ */
    if (gentype == PERMUTATION) return;
    qdim = nor - dim;		/* Dimension of the quotient */
    zsetlen(qdim);
    quotop = zalloc(qdim);
    zsetlen(nor);
    zquotinit(m3,dim,piv);
    for (g = 0; g < nmat; ++g)
    {
        MESSAGE(2,("%s...",quotname[g]));
	if ((ofile = zwritehdr(quotname[g],fl,qdim,qdim)) == NULL)
	    MtxErrorExit(NULL);
    	zsetlen(nor);
	zquotop(gen[g],quotop);
	zsetlen(qdim);
	zwritevec(ofile,quotop,qdim);
	fclose(ofile);
    }
    MESSAGE(1,("Done\n"));
    zsetlen(nor);
    free(quotop);
}


/* ------------------------------------------------------------------
   wrmod() - Write out the submodule
   ------------------------------------------------------------------ */

void wrmod(long cnt)

{
    char buf[200];
    FILE *ofile;

    if (smode == READALL)
	strcpy(buf,vname);
    else
	sprintf(buf,"%s.%4.4ld",vname,cnt);
    if ((ofile = zwritehdr(buf,fl,dim,nor)) == NULL)
	MtxErrorExit(NULL);
    zwritevec(ofile,m3,dim);
    fclose(ofile);
}



/* ------------------------------------------------------------------
   spin_up() - Spin up the seed vector (s).

   returns 1 if a proper subspace was found, 0 otherwise
   ------------------------------------------------------------------ */

int spin_up(void)

{
    int g;
    long count, nmu, pv;
    PTR xput, xget;
    FEL mk;
    int message = 1;

    xput = m3;
    dim = 0;

    /* If smode == READALL, echelonize the seed space.
       ----------------------------------------------- */
    if (smode == READALL)
    {
	dim = zmkechelon(m3,ssdim,piv);
	if (dim <= 0) err('0');
	MESSAGE(0,("SEED HAS DIMENSION %ld\n",dim));
	--dim;
	xput = m3;
	zadvance(&xput,dim);
    }

    /* Spin up
       ------- */
    g = -1;
    xget = m3;
    count = nmu = 0;
    while (1)
    {
	pv = zfindpiv(xput,&mk);
	if (pv != 0)
	{
	    count = 0;
	    message = 1;
	    piv[(int)++dim] = pv;
	    zadvance(&xput,(long)1);
	    if (dim > maxdim)
		return 0; /* Spans whole Space */
	}
    	if (message)
	{
	    message = 0;
	    if (MSG3)
	    	printf("dim = %ld, stack = %ld\n",dim,dim-nmu);
	    else if (MSG2 && dim % 10 == 0)
	    	printf("dim = %ld, stack = %ld\n",dim,dim-nmu);
	    else if (MSG1 && dim % 100 == 0)
		printf("dim=%ld\n",dim);
	}

	/* Apply the next generator
	   ------------------------ */
	if (++g >= ngen)
	{
	    g = 0;
	    zadvance(&xget,(long)1);
	    ++count;
	    if (++nmu >= dim ||
		(maxtry > 0 && count >= maxtry))
	    {
		if (dim == 0) err('0');
		if (dim < nor)
		    return 1;	/* Split! */
		else
		    return 0;    /* Spans whole space */
	    }
	    message = 1;
	}
	switch (gentype)
	{
	    case MATRIX:
		zmaprow(xget,gen[g],nor,xput);
		break;
	    case PERMUTATION:
		zpermrow(xget,(long *)gen[g],xput);
		break;
	    case DONTKNOW:
		err('b');
		break;
	}

	/* Clean the image with the basis.
	   ------------------------------- */
	zcleanrow(xput,m3,dim,piv);
    }
}




/* ------------------------------------------------------------------
   try_split()
   ------------------------------------------------------------------ */

void try_split(void)

{
    int hassplit;

    hassplit = spin_up();
    if (opt_G)
	printf("MeatAxe.SplitDimensions := [%ld,%ld];\n",dim,nor-dim);

    if (!dosplit)
    {
	if (smode != READALL)
	    MESSAGE(0,("VECTOR %5ld: ",iseed));
	if (!hassplit)
	{
	    if (dim < nor)
		MESSAGE(0,("DIMENSION LIMIT EXCEEDED\n"));
	    else
		MESSAGE(0,("SPANS WHOLE SPACE\n"));
	}
	else
	    MESSAGE(0,("SUB=%6ld QUOT=%6ld\n",dim,nor-dim));
	if (output)
	    wrmod(iseed);
    }
    else
    {	if (hassplit)
	{
	    if (smode != READALL)
	        MESSAGE(0,("VECTOR %ld SPLITS: ",iseed));
	    MESSAGE(0,("SUBSPACE %ld  QUOTIENT %ld\n",dim,nor-dim));
	    if (output)
		do_split();
	    exit(EXIT_OK);
	}
    }
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    PTR base;

    MtxInit();
    parseargs(argc,argv);
    init();

    switch (smode)
    {
	case GENERATE:
	    base = zalloc(ssdim);
	    zreadvec(seedfile,base,ssdim);
	    zpseed_init(ssdim,base);
	    if (seedvecno == 0)
	    {
		iseed = zpseed_make((long)1);
		while (iseed >= 0)
		{
		    zmoverow(m3,zpseed_vec);
		    try_split();
		    iseed = zpseed_next();
		}
	    }
	    else
	    {
		iseed = zpseed_make(seedvecno);
		if (iseed < 0) err('v');
		zmoverow(m3,zpseed_vec);
		try_split();
	    }
	    break;
	case READ:
	    if (seedvecno == 0)
	    {
		for (iseed = 1; iseed <= ssdim; ++iseed)
		{
		    zreadvec(seedfile,m3,(long)1);
		    try_split();
		}
	    }
	    else
	    {	if (seedvecno > ssdim) err('v');
		zseek(seedfile,seedvecno);
		zreadvec(seedfile,m3,(long)1);
		iseed = seedvecno;
		try_split();
	    }
	    break;

	case READALL:
	    zreadvec(seedfile,m3,ssdim);
	    try_split();
	    break;
    }

    /* The representation was not split. Print an
       appropriate message depending on the options
       used and exit with the right exit code.
       --------------------------------------------- */
    if (!dosplit)
	return EXIT_OK;
    if (seedvecno == 0 && smode != READALL)
	MESSAGE(0,("EVERY SEED VECTOR "));
    MESSAGE(0,("SPANS WHOLE SPACE\n"));
    MtxCleanup();
    return EXIT_NOSPLIT;
}

