/* ============================= C MeatAxe ==================================
   File:        $Id: zsy.c,v 1.3 1998/04/28 17:38:43 mringe Exp $
   Comment:     Symmetric tensor product.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>
#include <string.h>




/* ------------------------------------------------------------------
   Global Data
   ------------------------------------------------------------------ */

static int opt_G = 0;		/* GAP output */
static char *iname, *oname;	/* File names */
static FILE *ofile;
enum {M_E2,M_E3,M_E4,M_S2,M_M3} mode;
static long fl;		/* Field */
static long nor, noc;	/* Input sizes */
static long nor2, noc2;	/* Output sizes */
static PTR m1;		/* Input */
static PTR m2;		/* One row of the output matrix */
static PTR *row;	/* Pointers to the rows of the input matrix */


static char *helptext[] = {
"SYNTAX",
"    zsy [-GQV] <Mode> <Inp> <Out>",
"",
"    Valid modes are",
"        e2       Antisymmetric (exterior) square",
"        e3       Antisymmetric (exterior) cube",
"        e4       Antisymmetric fourth power",
"        s2       Symmetric square",
"        m3       Tensor cube, mixed symmetry",
"",
"OPTIONS",
"    -G     GAP output",
"    -Q     Quiet, no messages",
"    -V     Verbose, more messages",
"",
"FILES",
"    <Inp>   i  The input matrix (must be square)",
"    <Out>   o  The output matrix",
"",
NULL};

static proginfo_t pinfo =
   { "zsy", "Symmetrized Tensor Product",
     "$Revision: 1.3 $", helptext };



/* ------------------------------------------------------------------
   parseargs() - Process command line arguments
   ------------------------------------------------------------------ */

static void parseargs(int argc, char **argv)

{
    int i;
    char *c;

    /* Parse command line
       ------------------ */
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((i = MtxGetOpt("G")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	}
    }
    if (opt_ind != argc - 3) 
	MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
    iname = argv[opt_ind+1];
    oname = argv[opt_ind+2];

    /* Process the <Mode> argument
       --------------------------- */
    c = argv[opt_ind];
    if (!strcmp(c,"e2")) 
	mode = M_E2;
    else if (!strcmp(c,"e3")) mode = M_E3;
    else if (!strcmp(c,"e4")) mode = M_E4;
    else if (!strcmp(c,"s2")) mode = M_S2;
    else if (!strcmp(c,"m3")) mode = M_M3;
    else 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
}



/* ------------------------------------------------------------------
   init() - Initialize everything, read input files, allocate memory
   ------------------------------------------------------------------ */

static char wm1[] =
    "zsy: Warning: %s contains more than one permutation";

static void init()

{
    long i;
    FILE *f;

    /* Read input
       ---------- */
    if ((f = zreadhdr(iname,&fl,&nor,&noc)) == NULL)
	MtxErrorExit(NULL);
    if (mode != M_E2 && mode != M_S2 && mode != M_E3 && fl < 2)
	MtxErrorExit("%s: %E",iname,ERR_NOTMATRIX);
    if (mode == M_M3 && nor != noc)
	MtxErrorExit("%s: %E",iname,ERR_NOTSQUARE);

    if (fl >= 2)	/* Matrix */
    {
	zsetfield(fl); zsetlen(noc);
	m1 = zalloc(nor);
	zreadvec(f,m1,nor);

	/* Set up pointers to the rows of the input matrix
   	   ----------------------------------------------- */
	row = NALLOC(PTR,nor+1);
	if (row == NULL)
	    MtxErrorExit(NULL);
	for (i = 1; i <= nor; ++i)
	{	
	    row[i] = m1;
	    zadvance(&m1,(long)1);
	}
    }
    else		/* Permutation */
    {	
	if (noc != 1) 
	    fprintf(stderr,wm1,iname);
	noc = nor;
	m1 = (PTR) NALLOC(long,nor);
	if (m1 == NULL) 
	    MtxErrorExit(NULL);
	if (zreadlong(f,(long*) m1,nor) != (size_t) nor)
	    MtxErrorExit(NULL);
    }
    fclose(f);

    /* Calculate nor2 and noc2
       ----------------------- */
    switch (mode)
    {	case M_S2:
	    nor2 = (nor * (nor+1)) / 2;
	    noc2 = (noc * (noc+1)) / 2;
	    break;
	case M_E2:
	    nor2 = (nor * (nor-1)) / 2;
	    noc2 = (noc * (noc-1)) / 2;
	    break;
	case M_E3:
	    nor2 = nor*(nor-1)/2*(nor-2)/3;
	    noc2 = noc*(noc-1)/2*(noc-2)/3;
	    break;
	case M_E4:
	    nor2 = nor*(nor-1)/2*(nor-2)/3*(nor-3)/4;
	    noc2 = noc*(noc-1)/2*(noc-2)/3*(noc-3)/4;
	    break;
	case M_M3:
	    noc2 = nor2 = nor*(nor-1)*(nor+1)/3;
	    break;
	default:
	    MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    }
    if (nor2 < 0 || noc2 <= 0)
	MtxErrorExit("%s: Matrix too small",pinfo.name);


    /* Allocate the output buffer
       -------------------------- */
    if (fl >= 2)
    {
	MESSAGE(0,("Output is %ld x %ld\n",nor2,noc2));
	zsetlen(noc2);
	if ((m2 = zalloc(1)) == NULL)
	    MtxErrorExit(NULL);
    }
    else
    {	
	MESSAGE(0,("Output has degree %ld\n",nor2));
	fflush(stdout);
	noc2 = 1;
    	m2 = (PTR) NALLOC(long,nor2);
	if (m2 == NULL) 
	    MtxErrorExit(NULL);
    }
}



/* ------------------------------------------------------------------
   zs2() - Symmetric square
   ------------------------------------------------------------------ */

static void zs2()

{   long i1, i2, j1, j2, j3;
    FEL f11, f12, f21, f22;
    FEL w1,w2,f1,f2;

    MESSAGE(1,("Mode S2, part 1\n"));
    for (i1 = 1; i1 < nor; ++i1)
    {	
	for (i2 = i1 + 1; i2 <= nor; ++i2)
	{   
	    zmulrow(m2,F_ZERO);
	    j3 = 0;
	    for (j1 = 1; j1 < noc; ++ j1)
	    {	
		f11 = zextract(row[i1],j1);
		f21 = zextract(row[i2],j1);
		for (j2 = j1+1; j2 <= noc; ++j2)
		{   
		    ++j3;
		    f12 = zextract(row[i1],j2);
		    f22 = zextract(row[i2],j2);
		    w1 = zmul(f11,f22);
		    w2 = zmul(f12,f21);
		    zinsert(m2,j3,zadd(w1,w2));
		}
	    }
	    for (j2 = 1; j2 <= noc; ++j2)
	    {	f1 = zextract(row[i1],j2);
		f2 = zextract(row[i2],j2);
		++j3;
		zinsert(m2,j3,zmul(f1,f2));
	    }
	    zwritevec(ofile,m2,1);
	}
    }

    MESSAGE(1,("Mode S2, part 2\n"));
    for (i1 = 1; i1 <= nor; ++i1)
    {	j3 = 0;
	zmulrow(m2,F_ZERO);
	for (j1 = 1; j1 < noc; ++j1)
	{   f1 = zextract(row[i1],j1);
	    for (j2 = j1+1; j2 <= noc; ++j2)
	    {	++j3;
		f2 = zextract(row[i1],j2);
		w2 = zmul(f1,f2);
		zinsert(m2,j3,zadd(w2,w2));
	    }
	}
	for (j2 = 1; j2 <= noc; ++j2)
	{   f1 = zextract(row[i1],j2);
	    ++j3;
	    zinsert(m2,j3,zmul(f1,f1));
	}
	zwritevec(ofile,m2,1);
    }
}


/* ------------------------------------------------------------------
   zs2p() - Antisymmetric square (permutations)
   ------------------------------------------------------------------ */

static long maps2(long i, long k)

{	
    if (i <= k)
	return ((k*(k-1))/2 + i);
    else
	return ((i*(i-1))/2 + k);
}



static void zs2p()

{	
    long *p1, *p2;
    register long i, k;

    p1 = (long *) m1 - 1;
    p2 = (long *) m2 - 1;

    for (i = 1; i <= nor; ++i)
    {	
	for (k = 1; k <= i; ++k)
	{	
	    p2[maps2(i,k)] = maps2(p1[i],p1[k]);
	}
    }
    zwritelong(ofile,(long*) m2,nor2);
}



/* ------------------------------------------------------------------
   ze2p() - Antisymmetric square (permutations)
   ------------------------------------------------------------------ */

static long mape2(long i, long k)

{
    if (i < k)
	return (((k-1)*(k-2))/2 + i);
    else
	return (((i-1)*(i-2))/2 + k);
}



static void ze2p()

{
    long *p1, *p2;
    register long i, k;

    p1 = (long *) m1 - 1;
    p2 = (long *) m2 - 1;

    for (i = 1; i <= nor; ++i)
    {	
	for (k = 1; k < i; ++k)
	{	
	    p2[mape2(i,k)] = mape2(p1[i],p1[k]);
	}
    }
    zwritelong(ofile,(long*) m2,nor2);
}




/* ------------------------------------------------------------------
   ze2() - Antisymmetric square
   ------------------------------------------------------------------ */

static void ze2()

{
    long i1, i2, j1, j2, j3;
    FEL f12, f22, w1, w2, w3;

    for (i1 = 1; i1 < nor; ++i1)
    {	
        MESSAGE(1,("i1 = %ld\n",i1)); 
	for (i2 = i1+1; i2 <= nor; ++i2)
	{
            MESSAGE(2,("i2 = %ld\n",i2)); 
	    zmulrow(m2,F_ZERO);
	    j3 = 1;
	    for (j1 = 1; j1 < noc; ++j1)
	    {	register FEL f11, f21;
		f11 = zextract(row[i1],j1);
		f21 = zextract(row[i2],j1);
		for (j2 = j1+1; j2 <= noc; ++j2)
		{
		    f12 = zextract(row[i1],j2);
		    f22 = zextract(row[i2],j2);
		    w1 = zmul(f11,f22);
		    w2 = zmul(f12,f21);
		    w3 = zsub(w1,w2);
		    zinsert(m2,j3,w3);
		    ++j3;
		}
	    }
	    zwritevec(ofile,m2,1);
	}
    }
}


/* ------------------------------------------------------------------
   ze3() - Antisymmetric cube
   ------------------------------------------------------------------ */

static void ze3()

{   FEL f11,f12,f13,f21,f22,f23,f31,f32,f33;
    FEL e,g12,g13,g23;
    long i1, i2, i3, j1, j2, j3, jins;

    for (i1 = 1; i1 < nor-1; ++i1)
    {	
   	MESSAGE(1,("i1 = %ld\n",i1)); 
	for (i2 = i1+1; i2 < nor; ++i2)
	{   
   	    MESSAGE(2,("i2 = %ld\n",i2)); 
	    for (i3 = i2+1; i3 <= nor; ++i3)
	    {	
   	       	MESSAGE(3,("i3 = %ld\n",i3)); 
		zmulrow(m2,F_ZERO);
		jins = 1;
		for (j1 = 1; j1 < noc-1; ++j1)
		{   f11 = zextract(row[i1],j1);
		    f21 = zextract(row[i2],j1);
		    f31 = zextract(row[i3],j1);
		    for (j2 = j1+1; j2 < noc; ++j2)
		    {	f12 = zextract(row[i1],j2);
			f22 = zextract(row[i2],j2);
			f32 = zextract(row[i3],j2);
			g12 = zsub(zmul(f11,f22),zmul(f21,f12));
			g13 = zsub(zmul(f31,f12),zmul(f11,f32));
			g23 = zsub(zmul(f21,f32),zmul(f31,f22));
			for (j3 = j2+1; j3 <= noc; ++j3)
			{   f13 =zextract(row[i1],j3);
			    f23 =zextract(row[i2],j3);
			    f33 =zextract(row[i3],j3);
			    e = zadd(zadd(zmul(g12,f33),zmul(g13,f23)),
			    	zmul(g23,f13));
			    zinsert(m2,jins,e);
			    ++jins;
			}
		    }
		}
		zwritevec(ofile,m2,1);
	    }
	}
    }
}


/* ------------------------------------------------------------------
   ze3p() - Antisymmetric cube (permutations)
   ------------------------------------------------------------------ */

#define SWAP(x,y) {tmp=x; x=y; y=tmp;}

static long mape3(long i, long k, long l)

{	
    register long tmp;
    if (i < k) SWAP(i,k);
    if (i < l) SWAP(i,l);
    if (k < l) SWAP(k,l);
    return ((i-1)*(i-2)/2*(i-3)/3 + (k-1)*(k-2)/2 + l);
}



static void ze3p()

{
    long *p1, *p2;
    register long i, k, l;

    p1 = (long *) m1 - 1;
    p2 = (long *) m2 - 1;

    for (i = 3; i <= nor; ++i)
    {	
	for (k = 2; k < i; ++k)
	{
	    for (l = 1; l < k; ++l)
		p2[mape3(i,k,l)] = mape3(p1[i],p1[k],p1[l]);
	}
    }
    zwritelong(ofile,(long*) m2,nor2);
}



/* ------------------------------------------------------------------
   ze4() - Antisymmetric fourth power
   ------------------------------------------------------------------ */

static void ze4()

{   FEL f11,f12,f13,f14,f21,f22,f23,f24,f31,f32,f33,f34,f41,f42,f43,f44;
    FEL e,g12,g13,g14,g23,g24,g34,g123,g124,g134,g234;
    long i1, i2, i3, i4, j1, j2, j3, j4, jins;

    for (i1 = 1; i1 < nor-1; ++i1)
    {	
        MESSAGE(1,("i1 = %ld\n",i1)); 
	for (i2 = i1+1; i2 < nor; ++i2)
	{   
            MESSAGE(2,("i2 = %ld\n",i2)); 
	    for (i3 = i2+1; i3 <= nor; ++i3)
	    {   
            	MESSAGE(3,("i3 = %ld\n",i3)); 
		for (i4 = i3+1; i4 <= nor; ++i4)
	    	{   zmulrow(m2,F_ZERO);
		    jins = 1;

		    for (j1 = 1; j1 < noc-1; ++j1)
		    {   f11 = zextract(row[i1],j1);
		    	f21 = zextract(row[i2],j1);
		    	f31 = zextract(row[i3],j1);
		    	f41 = zextract(row[i4],j1);

		    	for (j2 = j1+1; j2 < noc; ++j2)
		    	{   f12 = zextract(row[i1],j2);
			    f22 = zextract(row[i2],j2);
			    f32 = zextract(row[i3],j2);
			    f42 = zextract(row[i4],j2);

			    g12 = zsub(zmul(f11,f22),zmul(f21,f12));
			    g13 = zsub(zmul(f11,f32),zmul(f31,f12));
			    g14 = zsub(zmul(f11,f42),zmul(f41,f12));
			    g23 = zsub(zmul(f21,f32),zmul(f31,f22));
			    g24 = zsub(zmul(f21,f42),zmul(f41,f22));
			    g34 = zsub(zmul(f31,f42),zmul(f41,f32));

			    for (j3 = j2+1; j3 <= noc; ++j3)
			    {   f13 =zextract(row[i1],j3);
			    	f23 =zextract(row[i2],j3);
			    	f33 =zextract(row[i3],j3);
			    	f43 =zextract(row[i4],j3);

				g123 = zmul(f13,g23);
				g123 = zsub(g123,zmul(f23,g13));
				g123 = zadd(g123,zmul(f33,g12));
				g124 = zmul(f13,g24);
				g124 = zsub(g124,zmul(f23,g14));
				g124 = zadd(g124,zmul(f43,g12));
				g134 = zmul(f13,g34);
				g134 = zsub(g134,zmul(f33,g14));
				g134 = zadd(g134,zmul(f43,g13));
				g234 = zmul(f23,g34);
				g234 = zsub(g234,zmul(f33,g24));
				g234 = zadd(g234,zmul(f43,g23));

			    	for (j4 = j3+1; j4 <= noc; ++j4)
				{   f14 =zextract(row[i1],j4);
				    f24 =zextract(row[i2],j4);
				    f34 =zextract(row[i3],j4);
				    f44 =zextract(row[i4],j4);

				    e = zmul(f24,g134);
				    e = zsub(e,zmul(f14,g234));
				    e = zadd(e,zmul(f44,g123));
				    e = zsub(e,zmul(f34,g124));
			    	    zinsert(m2,jins,e);
			    	    ++jins;
				}
			    }
			}
		    }
		    zwritevec(ofile,m2,1);
		}
	    }
	}
    }
}


/* ------------------------------------------------------------------
   zm3() - (2,1) part of tensor cube
   ------------------------------------------------------------------ */

static void addtorow(long i1, long i2, long i3, FEL f)

{
    MTXCONTEXT(addtorow);
    long pos;

    if (i1 >= i2 || i1 > i3)
    {
	MTXERROR3("Invalid index (%d,%d,%d)",(int)i1,(int)i2,(int)i3);
	MtxErrorExit(NULL);
    }

    pos = (i2 - i1 - 1) * (nor - i1 + 1) + i3 - i1 + 1;
    for (; i1 > 1; --i1)
	pos += (nor-i1+1)*(nor-i1+2);
    zinsert(m2,pos,zadd(zextract(m2,pos),f));
}



static void mapvector(long i1, long i2, long i3)

{   long k1, k2, k3;
    FEL f1, f2, f3;
    long l1, l2;

    for (k1 = 1; k1 <= nor; ++k1)
    {   
	f1 = zextract(row[i1],k1);
	for (k2 = 1; k2 <= nor; ++k2)
	{   
	    if (k2 == k1) continue;
	    f2 = zmul(f1,zextract(row[i2],k2));
	    if (k2 < k1)
	    {	
		l1 = k2;
		l2 = k1;
		f2 = zadd(F_ZERO,f2);
	    }
	    else
	    {	
		l1 = k1;
		l2 = k2;
	    }
	    for (k3 = 1; k3 <= nor; ++k3)
	    {	
		f3 = zmul(f2,zextract(row[i3],k3));
		if (k3 < l1)
		{	
		    f3 = zneg(f3);
		    addtorow(k3,l1,l2,f3);
		    addtorow(k3,l2,l1,f3);
		}
		else
		{	addtorow(l1,l2,k3,f3);
		}
	    }
	}
    }
}


static void zm3()

{
    long i,j,k;

    for (i = 1; i < nor; ++i)
    {	
	for (j = i + 1; j <= nor; ++j)
	{
	    for (k = i; k <= nor; ++k)
	    {	
		zmulrow(m2,F_ZERO);
		mapvector(i,j,k);
	    	zwritevec(ofile,m2,1);
	    }
	}
    }
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    MtxInit();
    parseargs(argc,argv);
    init();
    ofile = zwritehdr(oname,fl,nor2,(fl >= 2) ? noc2 : 1);
    if (ofile == NULL)
	MtxErrorExit("Cannot create output file %s",oname);

    switch (mode)
    {	
	case M_S2: fl >= 2 ? zs2() : zs2p(); break;
	case M_E2: fl >= 2 ? ze2() : ze2p(); break;
	case M_E3: fl >= 2 ? ze3() : ze3p(); break;
	case M_E4: ze4(); break;
	case M_M3: zm3(); break;
	default:
	    MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
	    break;
    }
    fclose(ofile);
    MtxCleanup();
    return EXIT_OK;
}


