/* ============================= C MeatAxe ==================================
   File:        $Id: zte.c,v 1.3 1998/04/28 17:38:45 mringe Exp $
   Comment:     Tensor product of two matrices or permutations.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <stdlib.h>
#include "meataxe.h"


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

long fl, nor1, noc1, nor2, noc2;
char *aname, *bname, *cname;	/* File names */
FILE *afile, *bfile;

static char *helptext[] = {
"SYNTAX",
"    zte [-T <MaxTime>] <A> <B> <Result>",
"",
"OPTIONS",
"    -T <NSeconds>    Set CPU time limit in seconds.",
"",
"FILES",
"    <A>       i  A matrix or permutation",
"    <B>       i  A matrix or permutation",
"    <Result>  o  The tensor product (A*B)",
NULL};

static proginfo_t pinfo =
   { "zte", "Tensor Product", "$Revision: 1.3 $", helptext };





/* ------------------------------------------------------------------
   tensormatrices() - Calculate the tensor product of two matrices
   ------------------------------------------------------------------ */

static void tensormatrices()

{
    PTR m1, m2, m3, x1, x2;
    long nor3, noc3;
    long j1,j2,j3,i1,i2;
    FEL f1,f2,f3;
    FILE *f;

    zsetfield(fl);
    zsetlen(noc1);
    m1 = zalloc(nor1);
    if (zreadvec(afile,m1,nor1) != (size_t)nor1) 
	MtxErrorExit(NULL);
    zsetlen(noc2);
    m2 = zalloc(nor2);
    if (zreadvec(bfile,m2,nor2) != (size_t)nor2) 
	MtxErrorExit(NULL);
    nor3 = nor1 * nor2;
    noc3 = noc1 * noc2;
    MESSAGE(1,("Computing matrix tensor product:"));
    MESSAGE(1,("(%ld,%ld)*(%ld,%lx)=(%ld,%ld)\n", 
	nor1,noc1,nor2,noc2,nor3,noc3));
    zsetlen(noc3);
    m3 = zalloc((long)1);
    if ((f = zwritehdr(cname,fl,nor3,noc3)) == NULL)
	MtxErrorExit(NULL);
    x1 = m1;
    for (i1 = 1; i1 <= nor1; ++i1)
    {
		
	x2 = m2;
	for (i2 = 1; i2 <= nor2; ++i2)
	{	
	    j3 = 1;
	    zmulrow(m3,F_ZERO);
	    for (j1 = 1; j1 <= noc1; ++j1)
	    {	
		f1 = zextract(x1,j1);
		for (j2 = 1; j2 <= noc2; ++j2)
		{	
		    f2 = zextract(x2,j2);
		    f3 = zmul(f1,f2);
		    zinsert(m3,j3,f3);
		    ++j3;
		}
	    }
	    zsetlen(noc3);
	    zwritevec(f,m3,(long)1);
	    zsetlen(noc2);
	    zadvance(&x2,(long)1);
	}
	zsetlen(noc1);
	zadvance(&x1,(long)1);
    }
    fclose(f);
}


/* ------------------------------------------------------------------
    tensorperms() - Calculate the tensor product of two permutations.
   ------------------------------------------------------------------ */

static void tensorperms(void)

{
    MTXCONTEXT(tensorperms);
    long nor3;
    long i, k, ii, kk, l;
    long *m1, *m2;
    FILE *f;
    long *p1, *p2;

    /* Read permutations
       ----------------- */
    m1 = NALLOC(long,nor1);
    m2 = NALLOC(long,nor2);
    if (m1 == NULL || m2 == NULL) 
	MtxErrorExit(NULL);
    p1 = m1-1;
    p2 = m2-1;
    if (zreadlong(afile,m1,nor1) != (size_t) nor1)
	MtxErrorExit(NULL);
    if (zreadlong(bfile,m2,nor2) != (size_t) nor2)
	MtxErrorExit(NULL);

    /* Open output file
       ---------------- */
    nor3 = nor1 * nor2;
    if ((f = zwritehdr(cname,(long)-1,nor3,(long)1)) == NULL)
	MtxErrorExit(NULL);
    MESSAGE(1,("Computing permutation tensor product:"));
    MESSAGE(1,("%ld*%ld=%ld\n", nor1,nor2,nor3));

    /* Calculate the tensor product
       ---------------------------- */
    for (i = 1; i <= nor1; ++i)
    {
	ii = p1[i] - 1;
	for (k = 1; k <= nor2; ++k)
	{
	    kk = p2[k];
	    l = ii * nor2 + kk;
	    if (zwritelong(f,&l,1) != 1)
	    {
		MTXERROR1("Cannot write result to %s",cname);
		MtxErrorExit(NULL);
	    }
	}
    }
    fclose(f);
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    long fl2;

    /* Parse command line
       ------------------ */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while (MtxGetOpt("") != OPT_END)
    {
    }
    if (opt_ind != argc-3) 
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    aname = argv[opt_ind];
    bname = argv[opt_ind+1];
    cname = argv[opt_ind+2];

    if ((afile = zreadhdr(aname,&fl,&nor1,&noc1)) == NULL)
	MtxErrorExit(NULL);
    if ((bfile = zreadhdr(bname,&fl2,&nor2,&noc2)) == NULL)
	MtxErrorExit(NULL);
    if (fl2 != fl)
	MtxErrorExit("%s and %s: %E",aname,bname,ERR_INCOMPAT);
    if (fl == -1)
	tensorperms();
    else
	tensormatrices();
    fclose(afile);
    fclose(bfile);
    MtxCleanup();
    return EXIT_OK;
}

