/* ============================= C MeatAxe ==================================
   File:        $Id: ztm.c,v 1.4 1998/04/28 17:38:47 mringe Exp $
   Comment:     Map vector under tensor product.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>
#include <string.h>


/* ------------------------------------------------------------------
   Variables
   ------------------------------------------------------------------ */

static char *helptext[] = {
"SYNTAX",
"    ztm <A> <B> <Vectors> <Result>",
"",
"FILES",
"    <A> and <B> are square matrices of dimension m and n, resp.",
"    <Vectors> is a matrix with mn columns and any number of rows.",
"    The result has the same size as <Vectors>.\n",
NULL};

static proginfo_t pinfo =
   { "ztm", "Tensor Multiply", "$Revision: 1.4 $", helptext };

static char *AName, *BName;	/* Matrices */
static char *VName, *RName;	/* Vectors, Result */



/* ------------------------------------------------------------------
   VectorToMatrix() - Convert vector to matrix
   ------------------------------------------------------------------ */

matrix_t *VectorToMatrix(PTR vec, long fl, long nrows, long ncol)

{
    matrix_t *mat;
    PTR d;
    long i, k;

    mat = MatAlloc(fl, nrows, ncol);
    d = mat->d;
    zsetlen(ncol);
    for (i = 0; i < nrows; i++) 
    {
        for (k = 1; k <= ncol; k++) 
        {
            FEL f = zextract(vec,i*ncol+k);
            zinsert(d,k,f);
        }
        zadvance(&d,(long)1);
    }
    return mat;
}




/* ------------------------------------------------------------------
   MatrixToVector() - Convert matrix to vector
   ------------------------------------------------------------------ */

void MatrixToVector(matrix_t *mat, PTR vec)

{
    PTR y;
    long i, k, nrows, ncol, l;

    ncol = mat->noc;
    nrows = mat->nor;

    /* Clear the vector 
       ---------------- */
    zsetlen(ncol * nrows);
    zmulrow(vec,F_ZERO);

    /* Convert
       ------- */
    zsetlen(ncol);
    y = mat->d;
    l = 1;
    for (i=0; i < nrows; i++) 
    {
        for (k=1; k <= ncol; k++) 
        {
            FEL f = zextract(y,k);
            zinsert(vec,l++,f);
        }
        zadvance(&y,(long)1);
    }
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{  
    matrix_t *mat1=NULL, *mat2=NULL, *mat1tr;
    FILE *f, *vecfile, *resultfile;
    PTR tmp;
    long i, ssdim, fl, nocol;

    /* Initialize everything, get options
       ---------------------------------- */
    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while (MtxGetOpt("") != OPT_END);
    if ((i = opt_ind) != argc-4)
	MtxErrorExit("%s: %E",pinfo.name,ERR_BADUSAGE);
    AName = argv[opt_ind];
    BName = argv[opt_ind+1];
    VName = argv[opt_ind+2];
    RName = argv[opt_ind+3];

    /* Read matrices
       ------------- */
    if ((f = SFOpen(AName,FM_READ)) == NULL ||
        (mat1 = MatRead(f)) == NULL)
	MtxErrorExit(NULL);
    fclose(f);
    if ((f = SFOpen(BName,FM_READ)) == NULL ||
        (mat2 = MatRead(f)) == NULL)
	MtxErrorExit(NULL);
    fclose(f);

    /* Test if matrices are square and over the same field
       --------------------------------------------------- */
    if (mat1->nor != mat1->noc)
	MtxErrorExit("%s: %E",AName,ERR_NOTSQUARE);
    if (mat2->nor != mat2->noc)
	MtxErrorExit("%s: %E",BName,ERR_NOTSQUARE);
    if (mat1->fl != mat2->fl)
	MtxErrorExit("%s and %s: %E",AName,BName,ERR_INCOMPAT);
    MESSAGE(1,("%s: %ldx%ld matrix over GF(%ld)\n",
	AName,mat1->nor,mat1->noc,mat1->fl));
    MESSAGE(1,("%s: %ldx%ld matrix over GF(%ld)\n",
	BName,mat2->nor,mat2->noc,mat2->fl));

    /* Open the <Vectors> file and check the header
       -------------------------------------------- */
    if ((vecfile = zreadhdr(VName,&fl,&ssdim,&nocol)) == NULL)
	MtxErrorExit(NULL);
    if (fl < 2) 
	MtxErrorExit("%s: %E",VName,ERR_NOTMATRIX);
    MESSAGE(1,("%s: %ldx%ld matrix over GF(%ld)\n",VName,ssdim,
	nocol,fl));
    if (fl != mat1->fl || nocol != mat1->nor * mat2->nor)
	MtxErrorExit("%s x %s and %s: %E",AName,BName,VName,ERR_INCOMPAT);
    

    /* Write beginning of output file
       ------------------------------ */
    if ((resultfile = zwritehdr(RName,mat1->fl,ssdim,nocol)) == NULL)
        MtxErrorExit(NULL);

    /* Transpose first matrix 
       ---------------------- */
    mat1tr = MatTr(mat1);
    MatFree(mat1);

    /* Allocate buffer for one vector
       ------------------------------ */
    zsetlen(nocol);
    tmp = zalloc((long)1);

    /* Process <Vectors> one by one
       ---------------------------- */
    for ( i=1; i <= ssdim; ++i)
    {
	matrix_t *mat3, *newmat;

        /* Read on evector and convert to matrix
           ------------------------------------- */
        zsetlen(nocol);
        zreadvec(vecfile,tmp,1);

        /* Turn vector into matrix
           ----------------------- */
        mat3 = VectorToMatrix(tmp,mat1tr->fl,mat1tr->nor,mat2->nor);
       
        /* Multiply
           -------- */
	newmat = MatDup(mat1tr);
	MatMul(newmat,mat3);
	MatMul(newmat,mat2);

        /* Turn matrix into vector and write out
           ------------------------------------- */
        MatrixToVector(newmat,tmp);
        zsetlen(nocol);
        zwritevec(resultfile,tmp,1);

        /* Free memory
           ----------- */
        MatFree(mat3);
        MatFree(newmat);
    }
   
    /* Close files
       ---------- */
    fclose(vecfile);
    fclose(resultfile);
    MtxCleanup();
    return 0;
}


