/* ============================= C MeatAxe ==================================
   File:        $Id: zzz.c,v 1.8 1998/06/25 19:02:35 mringe Exp $
   Comment:     Finite field arithmetic and common functions. `Small' version
		for field orders q <= 256. Originally based on the `hprout.c'
		written by Klaus Lux.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


/* ------------------------------------------------------------------
   Gobal data
   ------------------------------------------------------------------ */

char zzzversion[] = "conway-qadic-1.0";
typedef unsigned char BYTE;
BYTE
	tmult[256][256],
	tadd[256][256],
	taddinv[256],
	tmultinv[256];

static BYTE tffirst[256][2],
	textract[8][256],
	tnull[8][256],
	tinsert[8][256];
static long embedord[4];	/* Subfield orders */
static BYTE embed[4][16];	/* Embeddings of subfields */
static BYTE restrict[4][256];	/* Restrictions to subfields */
static size_t MPB = 0;	/* No. of marks per byte */
static size_t LPR = 0;	/* Long ints per row */



/*{{{ --------------------------------------------------------------------------
!variable zchar  "Characteristic"
!section kernel.ff
!synopsis
    extern long zchar;
!description
    The |zchar| variable contains the characteristic of the current field.
    Like |zfl|, this variable may be used anywhere, but it must not
    be modified directly.
!seealso  zfl zgen
-------------------------------------------------------------------------- }}}*/

long zchar = 0;


/*{{{ --------------------------------------------------------------------------
!variable zfl  "Field order"
!section kernel.ff
!synopsis
    extern long zfl;
!description
    This variable contains the order of the current field. |zfl| max be used
    in expressiond but must never modified directly. To change the current 
    field, use |zsetfield()|.
!seealso  zfl zgen
-------------------------------------------------------------------------- }}}*/

long zfl = -1;


/*{{{ --------------------------------------------------------------------------
!variable zgen  "Field generator"
!section kernel.ff
!synopsis
    extern FEL zgen;
!description
    This variable contains a genrator for the multiplicative group of the
    current field.
!seealso  zfl zgen
-------------------------------------------------------------------------- }}}*/

FEL zgen = 0;



/*{{{ --------------------------------------------------------------------------
!variable znoc  "Current row size"
!section kernel.rowop
!synopsis
    extern long znoc;
!description
    This variable contains the current number of columns used for all
    row operations. |znoc| is updated automatically when the row
    size is changed with |zsetlen()|. 
!seealso zsetlen zrowsize
-------------------------------------------------------------------------- }}}*/

long znoc = 0;


/*{{{ --------------------------------------------------------------------------
!variable zrowsize  "Current row size"
!section kernel.rowop
!synopsis
    extern size_t zrowsize;
    extern size_t zrowsize_io;
!description
    \verb"zrowsize" is the current row size, i.e., the number of bytes
    occupied by one row. This is always a multiple of \verb"sizeof(long)".
    \verb"zrowsize_io" contains the `true' size of a row, i.e., the number
    of bytes that are actually used. This is also the row size in data
    files.
!seealso  znoc zsetlen
-------------------------------------------------------------------------- }}}*/

size_t zrowsize = 0;	/* Row size in memory */
size_t zrowsize_io = 0;	/* Row size for file i/o */


/* ------------------------------------------------------------------
   Argument checking (for debbugging)
   ------------------------------------------------------------------ */

#if defined(_DEBUG)
#define CHECKRANGE(x,lo,hi) if ((x)<(lo)||(x)>(hi)) {\
	MtxErrorExit("ZZZ RANGE CHECK ERROR: %d <= %d <= %d ?\n",\
	(int)(lo),(int)(x),(int)(hi));}
#else
#define CHECKRANGE(x,lo,hi)
#endif

#define CHECKFEL(x) CHECKRANGE(x,0,zfl-1)
#define CHECKCOL(x)  CHECKRANGE(x,1,znoc)




/*{{{ --------------------------------------------------------------------------
!function zadd "Finite field addition"
!section kernel.ff
!synopsis
    FEL zadd(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| + |b| $
!description
    This function returns the sum of two field elements. Before you can
    use this function, the field must have been selected with |zsetfield()|.
    The arguments are not checked. If either argument is not in the allowed 
    range the result is undefined and the program may crash. 

    |zadd()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/

/*{{{ --------------------------------------------------------------------------
!function zsub "Finite field subtraction"
!section kernel.ff
!synopsis
    FEL zsub(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| - |b| $
!description
    This function returns the difference of two field elements. Before you can
    use this function, the field must have been selected with |zsetfield()|.
    The arguments are not checked. If either argument is not in the allowed 
    range the result is undefined and the program may crash. 

    |zsub()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/

/*{{{ --------------------------------------------------------------------------
!function zmul "Finite field multiplication"
!section kernel.ff
!synopsis
    FEL zmul(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| \cdot |b| $
!description
    This function returns the product of two field elements. Before you can
    use this function, the field must have been selected with |zsetfield()|.
    The arguments are not checked. If either argument is not in the allowed 
    range the result is undefined and the program may crash. 

    |zmul()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function zdiv "Finite field division"
!section kernel.ff
!synopsis
    FEL zdiv(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| / |b| $
!description
    This function returns the quotient of two field elements. Before you can
    use this function, the field must have been selected with |zsetfield()|.
    The arguments are not checked. If either argument is not in the allowed 
    range, or if |b| is zero, the result is undefined and the program may 
    crash. 

    |zdiv()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function zinv "Finite field inversion"
!section kernel.ff
!synopsis
    FEL zinv(FEL a);
!arg a
    Field element.
!return
    $1/ |a|$
!description
    This function returns the multiplicative inverse of a field element. 
    Before you can use this function, the field must have been selected with 
    |zsetfield()|. The argument is not checked. If |a| is not a valid field 
    element or zero the result is undefined and the program may crash. 

    |zinv()| may be implemented as a macro. In this case, it is guaranteed
    that the argument is evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function zneg "Additive inverse"
!section kernel.ff
!synopsis
    FEL zneg(FEL a);
!arg a
    Field element.
!return
    $- |a| $
!description
    This function returns the additive inverse of a field element. Before 
    you can use this function, the field must have been selected with 
    |zsetfield()|. The argument is not checked. If |a| is not a valid field 
    element the result is undefined and the program may crash. 

    |zneg()| may be implemented as a macro. In this case, it is guaranteed
    that the argument is evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/





/*{{{ --------------------------------------------------------------------------
!text kernel.rowop
    \paragraph{Internal data format}
    A (row) vector is stored in memory as an array of bytes. The actual
    size of a row depends on the number of marks in the row and on the
    field order, because several field elements may be packed into one
    byte. However, the size is always a multiple of {\tt sizeof(long)}. 
    Thus, there may be unused bytes at the end of a row. The contents of 
    these extra bytes (and even the contents of unused bits in partially 
    used bytes) is not defined. For this reason, memory must be initialized 
    before it is used, or else some functions as `compare rows' may fail.

    Packing of field elements occurs for field order $q\leq16$. Let $q$ be 
    the field order and $m$ the largest natural number with $q^m\leq 256$, 
    i.e. $m := \lfloor 8/\log_2 q \rfloor$. Then, $m$ elements are packed 
    into one byte using the packing function
    \[
	    \mbox{pack}(k_0,\ldots,k_{m-1}) = \sum_{i=0}^{m-1} k_iq^i
    \]
    Packing of field elements is used exclusively for rows (vectors),
    not for polynomials or other data types.


    \paragraph{Matrices}
    A matrix is stored as a sequence of rows with no extra space between
    them. The rows and columns of a matrix are numbered $1,2,\ldots n$.
    Column vectors may be stored as matrices with one column but this is
    very inefficient because each mark of the vector will occupy 4 Bytes.

    \paragraph{Accessing rows and matrices}
    Because of packing you cannot access the marks of a row with the usual 
    bracket operator. Instead, the {\MeatAxe} library provides functions 
    which store and extract marks. A second difference is that the first 
    entry in a vector has index 1, not 0 as in C arrays.

    Packed rows must be initialized before they are used. \verb"zinsert()"
    (and many other row operations) will produce strange results if used
    with uninitialized rows. Memory is initialized automatically in the
    following cases: allocation with \verb"zalloc()", copying a row with
    \verb"zmoverow()", reading a row from a file. A row can be initialized
    manually by multiplication with zero: \verb"zmulrow(ptr,F_ZERO)".
    The latter works even if no field has been selected. 
!seealso  zextract zfindpiv
-------------------------------------------------------------------------- }}}*/



/*{{{ --------------------------------------------------------------------------
!function zinsert  "Insert a mark into a row"
!section kernel.rowop
!synopsis
    void zinsert(PTR row, long col, FEL mark);
!arg row
    Pointer to the row.
!arg col
    Insert positions.
!arg mark
    Value to insert.
!description
    \verb"zinsert()" inserts the field element \verb"mark" at position
    \verb"col" into \verb"row". Before this function can be used, a
    field must be selected with \verb"zsetlen()". Once a field is
    selected, \verb"zinsert()" works with any row, regardless of the
    current row size. So, if you are working with two rows of different
    size, you do not have to call \verb"zsetlen()" prior to each
    \verb"zinsert()". On the other hand, this means that there
    is no protection against writing beyond the end of a row. The caller
    must assure that \verb"row" points to a row of at least \verb"col"
    columns. Otherwise the program may crash or produce undefined results.
!seealso  zextract
-------------------------------------------------------------------------- }}}*/

void zinsert(PTR row, long col, FEL mark)

{
    register BYTE *loc = (BYTE *)row + ((col - 1) / MPB);
    register long idx = (col - 1) % MPB;

    CHECKFEL(mark);
    *loc = (BYTE) (tnull[idx][*loc] + tinsert[idx][mark]);
}



/*{{{ --------------------------------------------------------------------------
!function zextract  "Extract a mark from a row"
!section kernel.rowop
!synopsis
    FEL zextract(PTR row, long col);
!arg row
    Pointer to the row.
!arg col
    Index of mark to extract.
!return
    |col|-th entry of |row|.
!description
    This function returns the entry at position \verb"col" of a
    row.  Note that column indexes start with 1, not 0.
    Like \verb"zinsert()", this function
    does not depend on the current row size. Reading beyond the end of
    a row will probably not produce an error, but the result is undefined.
!seealso zinsert 
-------------------------------------------------------------------------- }}}*/

FEL zextract(PTR row, long col)

{
    register BYTE *loc = (BYTE *)row + ((col - 1) / MPB);
    return (textract[(col - 1) % MPB][*loc]);
}



/*{{{ --------------------------------------------------------------------------
!function zfindpiv "Find pivot column"
!section kernel.rowop
!synopsis
    long zfindpiv(PTR row, FEL *mark);
!arg row
    Pointer to the row.
!return
    Index of the first non-zero entry in |row|, or $0$ it alle entries 
    are zero.
!description
    This function scans the vector \verb"row" and finds the first non-zero 
    mark. The mark is stored into \verb"mark" and its position (counting 
    from 1) is returned. If the whole vector is zero, \verb"zfindpiv()" 
    returns 0 and leaves \verb"mark" unchanged.
!seealso
-------------------------------------------------------------------------- }}}*/

long zfindpiv(PTR row, FEL *mark)

{
    register long *l = (long *) row;
    register long idx;
    register BYTE *m;

    for (idx = 0; idx < (long) LPR && *l == 0; ++idx, ++l);
    if (idx == (long)LPR) return 0;
    idx = idx * sizeof(long) * MPB + 1;
    m = (BYTE *)l;
    while (*m == 0)
    {	++m;
	idx += MPB;
    }
    idx += tffirst[*m][1];
    if (idx <= znoc)
    {	*mark = tffirst[*m][0];
	return idx;
    }
    return 0;
}


/* -----------------------------------------------------------------
   OpenTableFile()
   ------------------------------------------------------------------ */

static FILE *OpenTableFile(long fl)

{   
    MTXCONTEXT(OpenTableFile);
    char fn[250];
    char cmd[250];
    FILE *fd;

    /* Versuche die Datei zu oeffenen
       ------------------------------ */
    sprintf(fn,"p%3.3ld.zzz",(long)fl);
    if ((fd = SFOpen(fn,FM_READ|FM_LIB|FM_NOERROR)) != NULL)
	return fd;

    /* Rufe maketab auf
       ---------------- */
    sprintf(cmd,"%s/maketab -Q %ld",MeatAxeBin,fl);
    if (system(cmd) != 0)
	MTXERROR1("Unable to run %s/maketab",MeatAxeBin);
    fd = SFOpen(fn,FM_READ|FM_LIB);
    return fd;
}


/*{{{ --------------------------------------------------------------------------
!function zsetfield "Set the field"
!section kernel.ff
!synopsis
    int zsetfield(long field);
!arg field
    Field order.
!return
    The function returns 0 on success, and $-1$ otherwise.
!description
    This function sets the current field to \GF{{\tt field}} and initializes
    the field arithmetic. Most kernel functions require that a field has
    been selected before they are used. 
    
    \verb"zsetfield()'" reads an arithmetic table file into memory. If the
    file does not exist, \verb"zsetlen()" tries to generate one by executing 
    the \verb"maketab" program.
!seealso zsetlen
-------------------------------------------------------------------------- }}}*/

int zsetfield(long field)

{
    FILE *fd;
    long hdr[5];

    if (field == zfl || field < 2) return 0;
    fd = OpenTableFile(field);
    if (fd == NULL)
	MtxErrorExit("Cannot open table file for GF(%d)",(int)field);

    /* Read header, perform some checks
       -------------------------------- */
    if (zreadlong(fd,hdr,5) != 5)
	MtxErrorExit("ERROR READING TABLE FILE HEADER");
    if (hdr[2] != field || hdr[1] < 0 || hdr[1] > field ||
        hdr[0] <= 1 || hdr[2] % hdr[0] != 0 || hdr[3] < 1 || hdr[3] > 8)
    {	
	MtxErrorExit("Table file is corrupted");
    }
    zchar = hdr[0];
    zgen = (FEL) hdr[1];
    MPB = hdr[3];
    if (hdr[4] != (long) ZZZVERSION)
    {
	MtxErrorExit("Bad table file version: expected %d, found %d",
	    (int)ZZZVERSION,(int)hdr[4]);
    }

    /* Read tables
       ----------- */
    if (fread(tmult,4,sizeof(tmult)/4,fd) != sizeof(tmult)/4 ||
        fread(tadd,4,sizeof(tadd)/4,fd) != sizeof(tadd)/4 ||
        fread(tffirst,1,sizeof(tffirst),fd) != sizeof(tffirst) ||
        fread(textract,1,sizeof(textract),fd) != sizeof(textract) ||
        fread(taddinv,1,sizeof(taddinv),fd) != sizeof(taddinv) ||
        fread(tmultinv,1,sizeof(tmultinv),fd) != sizeof(tmultinv) ||
        fread(tnull,1,sizeof(tnull),fd) != sizeof(tnull) ||
        fread(tinsert,1,sizeof(tinsert),fd) != sizeof(tinsert) ||
        zreadlong(fd,embedord,4) != 4 ||
        fread(embed,16,4,fd) != 4 ||
        fread(restrict,256,4,fd) != 4
       )
    {
        MtxErrorExit("ERROR READING TABLE FILE");
    }
    fclose(fd);
    zfl = field;
    zsetlen(zfl);
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!function zsetlen "Set the row length"
!section kernel.rowop
!synopsis
    int zsetlen(long ncols);
!arg ncols
    Number of columns per row.
!return
    The function returns 0 on success, and $-1$ otherwise.
!description
    This function sets the current row size to \verb"ncols". The row size
    is used for low-level row operations, {e.g.} \verb"zaddrow()".
!seealso zsetfield
-------------------------------------------------------------------------- }}}*/

int zsetlen(long ncols)

{ 
    znoc = ncols;
    zrowsize_io = (znoc-1) / MPB + 1;		/* Used bytes per row */
    LPR = (zrowsize_io-1) / (sizeof(long)) + 1;	/* Ints per row */
    zrowsize = LPR * sizeof(long);	/* Total bytes per row */
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function zembed "Embed a subfield"
!section kernel.ff
!synopsis 
    FEL zembed(FEL a, long subfield);
!arg a
    Field element.
!arg subfield
    Subfield order.
!return
    |a|, embedded into the current field.
!description
    This function embeds an element of a subfield into the currently
    selected field. |subfield| must be a divisor of the current field order.
!text kernel.ff
    \paragraph{Embedding of subfields}
    In the {\MeatAxe} there is a `standard' generator for each finite
    field. The generator for the field currently in use is available in 
    the \verb"zgen" variable. Thus, if $a$ and $a'$ are the {\MeatAxe} 
    generators of $\GF{q}$ and $\GF{q'}$, respectively, and $q'=q^n$, there 
    is a `standard' embedding of $\GF{q}$ into $\GF{q'}$ defined by 
    $a\mapsto (a')^n$. However, field elements which are identified 
    under this embedding are usually not represented by the same number. 
    For this reason there are two functions, \verb"zembed()" and 
    \verb"zrestrict()", which provide the embedding of subfields into the 
    current field. Note that the {\MeatAxe} is not well suited for 
    calculations involving different fields at the same time because the
    arithmetic uses lookup tables which are specific to each field

    Here is a short example. The following code converts a vector over \GF{3}
    to \GF{27}. The {\MeatAxe} cannot handle two fields at the same time, so 
    it is necessary to unpack the row over \GF{3}, change to \GF{27} and pack 
    the embedded elements into a new row.
    \begin{verbatim}
	PTR row1, row2;
	FEL buf[10];
	...
	zsetlen(3,10);
	for (i = 1; i <= 10; ++i) buf[i-1] = zextract(row1,i);
	zsetlen(27,10);
	for (i = 1; i <= 10; ++i) zinsert(row2,i,zembed(buf[i-1],3));
    \end{verbatim}
!seealso zrestrict
-------------------------------------------------------------------------- }}}*/

FEL zembed(FEL a, long subfield)

{   int i;

    if (subfield == zfl) 
	return a;
    for (i = 0; embedord[i] != subfield && i < 4; ++i);
    if (i >= 4)
	MtxErrorExit("Cannot embed GF(%d) into GF(%d)",(int)subfield,(int)zfl);
    return embed[i][a];
}



/*{{{ --------------------------------------------------------------------------
!function zrestrict "Restrict to a subfield"
!section kernel.ff
!synopsis 
    FEL zrestrict(FEL a, long subfield);
!arg a
    Field element.
!arg subfield
    Subfield order.
!return
    The restricted field element.
!description
    This function restricts a field element to a subfield. The return value 
    represents the same element as |a| but with respect to the subfield.
    Note that you cannot use the return value of \verb"zrestrict()"
    (for field arithmetic, say) until you change to the subfield
    with \verb"zsetlen(subfield,...)".

    \verb"a" must be of the form \verb"zembed(b,subfield)". Otherwise
    the result is undefined. For example, restricting from \GF{8} to
    \GF{2} works only with the prime field elements, i.e., $0$ and $1$.
!seealso zembed
-------------------------------------------------------------------------- }}}*/

FEL zrestrict(FEL a, long subfield)

{
    int i;

    if (subfield == zfl) 
	return a;
    for (i = 0; embedord[i] != subfield && i < 4; ++i);
    if (i >= 4)
	MtxErrorExit("Cannot embed GF(%d) into GF(%d)",(int)subfield,(int)zfl);
    return restrict[i][a];
}



/*{{{ --------------------------------------------------------------------------
!function zaddrow "Add two rows"
!section kernel.rowop
!synopsis
    PTR zaddrow(PTR dest, PTR src);
!arg dest
    The row to add to.
!arg src
    The row to add.
!return
    The function always returns |dest|.
!description
    This function adds \verb"src" to \verb"dest". The row size must have been
    set with |zsetlen()|.
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/

PTR zaddrow(PTR dest, PTR src)

{
    register long i;

    if (zchar == 2)	/* characteristic 2 is simple... */
    {	register long *l1 = (long *) dest;
	register long *l2 = (long *) src;
	for (i = LPR; i != 0; --i)
	{
	    register long x = *l2++;
	    if (x != 0) *l1 ^= x;
	    l1++;
	}
    }
    else		/* any other characteristic */
    {	register BYTE *p1 = dest;
	register BYTE *p2 = src;
	for (i = zrowsize_io; i != 0; --i)
	{
	    register int x = *p2++;
	    if (x != 0) *p1 = tadd[*p1][x];
	    p1++;
	}
    }
    return dest;
}



/*{{{ --------------------------------------------------------------------------
!function zmulrow "Multiply a row by a coefficient"
!section kernel.rowop
!synopsis
    PTR zmulrow(PTR row, FEL mark);
!arg row
    Pointer to the row.
!arg mark
    Coefficient.
!return
    The function always returns |row|.
!description
    This function multiplies |row| by the scalar |mark| using the current 
    row size and returns |row|. 
    
    Multiplying a row with zero |F_ZERO| initializes the memory. 
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/

PTR zmulrow(PTR row, FEL mark)

{
    register BYTE *m;
    register BYTE *multab;
    register long i;

    CHECKFEL(mark);
    if (mark == F_ZERO)
    {
        register long *l = (long *)row;
        for (i = LPR; i > 0; --i) *l++ = 0;
    }
    else if (mark != F_ONE)
    {
	multab = tmult[mark];
	m = row;
	for (i = zrowsize_io; i != 0; --i)
	{
	    register int x = *m;
	    if (x != 0) *m = multab[x];
	    ++m;
	}
    }
    return row;
}


/*{{{ --------------------------------------------------------------------------
!function zaddmulrow "Add a multiple of a row"
!section kernel.rowop
!synopsis
    PTR zaddmulrow(PTR dest, PTR src, FEL f);
!arg dest
    The vector to add to.
!arg dest
    The vector to add.
!arg f
    Coefficient.
!return
    The function always returns |dest|.
!description
    This function adds a multiple of \verb"src" to \verb"dest".
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/

PTR zaddmulrow(PTR dest, PTR src, FEL f)

{
    register long i;
    register BYTE *p1, *p2, *multab;

    CHECKFEL(f);
    if (f == F_ZERO) return dest;
    if (f == F_ONE) return zaddrow(dest,src);
    multab = tmult[f];
    p1 = dest;
    p2 = src;
    for (i = zrowsize_io; i != 0; --i)
    {
	*p1 = tadd[*p1][multab[*p2++]];
	++p1;
    }
    return dest;
}


/*{{{ --------------------------------------------------------------------------
!function zmaprow "Multiply a vector by a matrix"
!section kernel.rowop
!synopsis
    PTR zmaprow(PTR row, PTR matrix, long nor, PTR result);
!arg row
    The vector.
!arg matrix
    The matrix.
!arg nor
    Number of rows in |matrix|.
!arg result
    Pointer to a buffer where the result will be stored.
!return
    The function always returns |result|.
!description
    This function multiplies the vector \verb"row" from the right by
    the matrix \verb"mat" and stores the result into \verb"result".
    \verb"nor" is the number of rows of the matrix which must coincide
    with the size (number of columns) of the vector. The number of
    columns in both \verb"mat" and \verb"result" is determined by the
    current row size.

    {\em Note:} \verb"result" and \verb"row" must not overlap. Otherwise
    the result is undefined.
!seealso 
-------------------------------------------------------------------------- }}}*/

PTR zmaprow(PTR row, PTR matrix, long nor, PTR result)

{
    register int i;
    register FEL f;
    BYTE *m = (BYTE *) matrix;
    register long *l = (long *)result;

    for (i = LPR; i > 0; --i) *l++ = 0;

    if (zfl == 2)	/* GF(2) is a special case */
    {
	register long *x1 = (long *) matrix;
	register BYTE *r = (BYTE *) row;
	register BYTE mask;

	for (i = nor; i > 0; ++r)
    	{
	    if (*r == 0)	/* Skip zero bytes */
	    {
		x1 += 8 * LPR;
		i -= 8;
		continue;
	    }

	    for (mask = 0x80; mask != 0 && i > 0; --i, mask >>= 1)
	    {
	        if ((*r & mask) != 0)
	        {
		    register long *x2 = (long *)result;
		    register long k;
		    for (k = LPR; k; --k)
			*x2++ ^= *x1++;
	    	}
	    	else
		    x1 += LPR;	/* Skip that row */
	    }
	}
    }
    else		/* Any other field */
    {
	register BYTE *brow = (BYTE *) row;
	register int pos = 0;

	for (i = nor; i > 0; --i)
	{
	    f = textract[pos][*brow];
	    if (++pos == (int) MPB)
	    {
		pos = 0;
		++brow;
	    }
	    if (f != F_ZERO)
	    {
		register BYTE *v = m;
		register BYTE *r = result;
		register long k = zrowsize_io;
		if (f == F_ONE)
		{
		    for (; k != 0; --k)
		    {
			*r = tadd[*r][*v++];
			++r;
		    }
		}
		else
		{
		    register BYTE *multab = tmult[f];
		    for (; k != 0; --k)
		    {
			*r = tadd[*r][multab[*v++]];
			++r;
		    }
		}
	    }
	    m += zrowsize;		/* next row */
	}
    }
    return result;
}



/*{{{ --------------------------------------------------------------------------
!function zitof "Convert integer to field element"
!section kernel.ff
!synopsis
    FEL zitof(long l);
!arg l
    The integer.
!return
    The field element corresponding to |l|.
!description
    This function converts an integer to a field element using the
    same mapping as explained with |zftoi()|. Both functions are inverse
    in the sense that the expression \verb"f == zitof(zftoi(f))" is always 
    true. |zitof()| should be used whenever field elements are converted with 
    |scanf()|.
!seealso zftoi
-------------------------------------------------------------------------- }}}*/

FEL zitof(long l)

{
    register long f;
    f = l % zfl;
    if (f < 0) f += zfl;
    return (FEL) f;
}






/*{{{ --------------------------------------------------------------------------
!function zftoi "Convert field element to integer"
!section kernel.ff
!synopsis
    long zftoi(FEL f);
!arg f
    The field element.
!return
    The number corresponding to |f|.
!description
    This function converts a field element to an integer, using a `canonical' 
    representation of field elements as integers which may be different from 
    the internal representation. In particular, the prime field is mapped on
    $\{0,\ldots,p-1\}$. See section \ref{GdSection:kernel.ff} for details. 
    \verb"zftoi()" should be used whenever field elements are 
    converted with |printf()|.
!seealso zitof
-------------------------------------------------------------------------- }}}*/

long zftoi(FEL f)

{	
    return (long) f;
}




/*{{{ --------------------------------------------------------------------------
!function zextractcol "Extract one column of a matrix"
!section kernel.rowop
!synopsis 
    PTR zextractcol(PTR mat,long nor,long col,PTR result);
!arg mat
    Pointer to the matrix.
!arg nor
    Number of rows in matrix.
!arg col
    Column to extract (starting with 1).
!arg result
    Pointer to buffer for the extracted column.
!return
    The function returns |result|.
!description
    This function extracts one column out of a matrix and stores it as
    a row vector in |result|. The number of columns of the matrix must have 
    been set with |zsetlen()|. |nor| is the number of rows in the matrix. The 
    result is a row with |nor| entries, i.e., the length of |result| must be at 
    least |nor|. |mat| and |result| must not overlap, or ther esult is 
    undefined.
-------------------------------------------------------------------------- }}}*/

PTR zextractcol(PTR mat,long nor,long col,PTR result)

{
    register BYTE *x = (BYTE *)mat + ((col - 1) / MPB);
    register BYTE *extab = textract[(col - 1) % MPB];
    register BYTE a = 0;
    register int ind = 0;
    register BYTE *y = result;
    register long count;
    
    for (count = nor; count > 0; --count)
    {
	a += tinsert[ind][extab[*x]];
	if (++ind == (int) MPB)
	{
	    *y++ = a;
	    a = 0;
	    ind = 0;
	}
	x += zrowsize;
    }
    if (ind != 0) *y = a;
    return result;
}




/*{{{ --------------------------------------------------------------------------
!structure FEL "Finite field element"
!section kernel.ff
!synopsis
    typedef unsigned char FEL;
!description
    The |FEL| data type represents an element of the finite field. The actual
    definition of |FEL| depends on the kernel that was used in compiling the 
    {\MeatAxe}. The `standard' kernel for fields up to $\GF{256}$ defines a
    |FEL| as an unsigned character (which is assumed to be 8 bits wide).
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!text kernel
The kernel 
\begin{itemize}
\item provides the finite field arithmetic and low-level vector
    operations, and
\item provides a platform-independent interface to operating
    system services, including the interaction with the user.
\end{itemize}
There are two kernels available: one for small fields (up to 256) and
one for larger fields (up to $2^{16}$). The kernel is selected at
compile time. Kernel functions are optimized for high speed but they 
lack some features of the higher layers levels like error checking and 
debugging capabilities. Most kernel services are simple, but there are 
some more complex functions like Gaussian elimination, which have been 
implemented in the kernel for performance reasons. 

Most of the kernel functions have names beginning with `z'. Kernel functions 
generally do not allocate memory dynamically and do not check their arguments. 
The kernel cannot operate simultaneously with different finite fields, because
there is a global row size and a global field order which must be maintained 
by the higher layers.


\paragraph{Simple Data types}
%- - - - - - - - - - - - - - -
\index{FEL@{\tt FEL} data type}
\index{PTR@{\tt PTR} data type}
The kernel defines two basic data types
\GdBeginItemList
\GdItem{FEL}{A finite field element}
\GdItem{PTR}{A pointer to a vector (row). {\tt PTR} may be defined as
	{\tt \*FEL}, but this is not mandatory. However, it must be
	possible to dereference and to increment a {\tt PTR}.}
\GdEndItemList
The kernel also defines two constants:
\GdBeginItemList
\GdItem{F\_ZERO}{The zero element of the current field.}
\index{F_ZERO@{\tt F\_ZERO} constant}
\GdItem{F\_ONE}{The unit element of the current field.}
\index{F_ONE@{\tt F\_ONE} constant}
\GdEndItemList
Depending on which kernel you are using, \verb|F_ZERO| and \verb|F_ONE| need 
not be constants. They may be defined as variables or even function calls.


!text kernel.ff
In the `small' kernel, field elements of GF$(q)$ are represented
by the numbers $0,1,\ldots,q-1$. The field is defined by its
Conway polynomial $p(x)$, a polynomial of degree $n$ over $\Z_p[x]$,
where $q=p^n$. Thus, we have a one-to-one correspondence of field
elements $a\in{\rm GF}(q)$ and polynomials $f_a(x)\in\Z_p[x]$ of
degree less than $n$. By treating $\Z_p$ as a subset of $\Z$ ---
actually, on the computer, elements of $\Z_p$ are represented
by integers --- this is also a polynomial over $\Z$. Now,
calculate $f_a(p)$ giving the number of the field element $a$.
It follows that the elements of the prime field are represented by
$0$, $1$, \ldots, $p-1$. The number 0 represents the zero element,
and 1 represents the unit element.

The big version uses always 2 bytes (an {\tt unsigned short}) for
field elements. Non-zero elements are stored as their logarithms with
respect to a fixed generator, i.e. the unit element is represented by
the number 0. The zero element is represented by
the special value {\tt 0xFFFF}.

As a consequence of the different representations of field elements
in the small and big version, there are some rules which should be
respected by all programs:
\begin{itemize}
\item	Never assign numbers to variables of type {\tt FEL} or pass 
	numbers to functions expecting an argument of type {\tt FEL}.
\item	Never perform integer arithmetic on variables of type {\tt FEL}.
\item	Never {\tt printf} or {\tt scanf} variables of type {\tt FEL}.
\item	Never use the literals $0$ or $1$ where the zero and unit element 
	of the field is intended. Instead you should use the \verb|F_ZERO| 
	and \verb|F_ONE| constants define in |meataxe.h|.
\item	Do not cast an integer to |FEL| or vice versa. Instead use
	{\tt zitof()} and {\tt zftoi()}.
\end{itemize}

\paragraph{Converting between finite field elements and integers}
The {\MeatAxe} defines a `standard' correspondance between the elements of 
$\GF{q}$ and the set $\{0,1,\ldots,q\}$. For prime fields, the mapping 
is defined by assigning the integer $1$ to the unit element of the field.
For non-prime fields, each $a\in\GF{p^n}$ is represented -- modulo the 
ideal generated by the Conway polynomial of degree $n$ -- by a unique
polynomial $f_a(x)\in \GF{p}[x]$ with $\deg(f_a)<n$ . Using the standard
embedding of $\GF{p}$ into $\Z$ we can consider $f_a$ as a polynomial
over $\Z$. Then, the number assigned to $a$ if $f_a(p)$.

The mapping between $\GF{q}$ and $\{0,1,\ldots,q\}$ is provided by two
functions, |zitof()| and |zftoi()|. Since the actual numeric 
representation of field elements depends on the kernel, you cannot
convert a |FEL| to an |int| simply by type casting.
-------------------------------------------------------------------------- }}}*/
