/* ============================= C MeatAxe ==================================
   File:        $Id: zzz2.c,v 1.4 1998/04/28 17:38:58 mringe Exp $
   Comment:     Basic row operations.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>

#define LPR (zrowsize/sizeof(long))


/*{{{ --------------------------------------------------------------------------
!function zalloc "Allocate memory and initialize"
!section kernel.rowop
!synopsis 
    PTR zalloc(long nrows);
!arg nrows
    Number of rows.
!return
    Returns a pointer to the memory block or NULL on error.
!description
   This function allocates a block of memory for a vector or a
   matrix. Memory is initialized to zero. |nrows| ist the number 
   of rows to allocate.

   Field order and row size must have been set with <zsetfield()>
   and |zsetlen()|, respectively.
   Unlike |malloc()| on some systems, this function does not fail
   if |nrows| is zero. In this case, a `zero-length' block is
   allocated.
!seealso zsize
-------------------------------------------------------------------------- }}}*/

PTR zalloc(long nrows)

{
    PTR p;
    char *q;
    register long i;
    size_t req = (size_t)(zrowsize*nrows);

    if ((p = (PTR) Smalloc(req)) == NULL)
    {
	fprintf(stderr,"zalloc(): Request=%ld Bytes\n",(long)req);
	MtxErrorExit("OUT OF MEMORY");
    }

    /* Initialize */
    q = (char *) p;
    for (i = nrows; i > 0; --i)
    {
	zmulrow((PTR) q,F_ZERO);
	q += zrowsize;
    }
    return p;
}




/*{{{ --------------------------------------------------------------------------
!function zsize "Size of a row or matrix"
!section kernel.rowop
!synopsis 
    size_t zsize(long nrows);
!arg nrows
    Number of rows.
!description
    This function returns the number of bytes occupied by \verb"nrows"
    rows of the current row size. This is the number of bytes that
    |zalloc(nrows)| would allocate. The row size must have been specified
    previously with \verb"zsetlen()". 
!seealso zalloc
-------------------------------------------------------------------------- }}}*/

size_t zsize(long nrows)

{	
    return ((size_t)(zrowsize * nrows));
}




/*{{{ --------------------------------------------------------------------------
!function zmoverow "Copy a row"
!section kernel.rowop
!synopsis 
    void zmoverow(PTR dest, PTR src);
!arg dest
    Pointer to the destination.
!arg src
    Pointer to the source.
!description
    This function copies the contents of one row to another row.
    As with all row operations, the row length must have been set before 
    with |zsetlen()|.
!seealso zswaprow
-------------------------------------------------------------------------- }}}*/

void zmoverow(PTR dest, PTR src)

{
    register long *s = (long *) src, *d = (long *) dest, i;
    for (i = LPR; i > 0; --i)
   	 *d++ = *s++;
}



/*{{{ --------------------------------------------------------------------------
!function zswaprow "Exchange two rows"
!section kernel.rowop
!synopsis 
    void zswaprow(PTR dest, PTR src);
!arg dest
    Pointer to the first row
!arg src
    Pointer to the second row
!description
    This function exchanges the contents of two rows. As with all row 
    operations, the row length must have been set before with |zsetlen()|.
!seealso zmoverow
-------------------------------------------------------------------------- }}}*/

void zswaprow(PTR dest, PTR src)

{	register long *p1 = (long *) src, *p2 = (long *) dest, i, l;

	for (i = LPR; i > 0; --i)
	{	l = *p1;
		*p1++ = *p2;
		*p2++ = l;
	}
}



/*{{{ --------------------------------------------------------------------------
!function zcmprow "Compare two rows"
!section kernel.rowop
!synopsis 
    int zcmprow(PTR dest, PTR src);
!arg dest
    Pointer to the first row
!arg src
    Pointer to the second row
!return
    The function returns 0 if the two rows are identical, and 1 otherwise.
!description
    This function compares two rows. As with all row operations, the row 
    length must have been set before with |zsetlen()|.
    |zcmprow()| checks only for equality and does not define an ordering
    of rows.
!seealso
-------------------------------------------------------------------------- }}}*/

int zcmprow(PTR dest, PTR src)

{	
    register long k;
    register long *d = (long *) dest, *s = (long *) src;

    for (k = LPR; k > 0; --k)
    {
	if (*d++ != *s++) 
	    return 1;
    }
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!function zadvance "Advance a row pointer"
!section kernel.rowop
!synopsis 
    void zadvance(PTR *ptr, long nrows);
!arg ptr
    Pointer to the |PTR| variable to increment.
!description
    This function increments a pointer by |nrows| rows using the current row 
    size. This functions is typically used to step through the rows of a 
    matrix. Before it can be used, the row size must have been set with 
    |zsetlen()|. Here is an example:
    \begin{verbatim}
	PTR mat;
	zsetlen(100);
	mat = zalloc(100);
	zadvance(&mat,49);
    \end{verbatim}
    After this sequence, \verb"mat" points to the 50th row of the 
    $100\times 100$ matrix |mat|.

    Note that the first argument is not a |PTR| but a |*PTR|, i.e.,
    a pointer to a |PTR|! Thus you will always need the |&| operator
    as in the example.
!seealso
-------------------------------------------------------------------------- }}}*/

void zadvance(PTR *ptr, long nrows)

{
    *(long **)ptr += nrows * LPR;
}



/*{{{ --------------------------------------------------------------------------
!function zpermrow "Multiply a vector by a permutation"
!section kernel.rowop
!synopsis 
    PTR zpermrow(PTR row, long *perm, PTR result);
!return
    The function returns |result|.
!description
    This function multiplies the vector |row| from the right with the 
    permutation |perm| and stores the result into |result|. Multiplication 
    of vectors by permutations is defined as follows: If the permutation 
    maps $i$ to $k$, then the $i$-ith mark of the vector is stored in the 
    $k$-th position of the result. 

    {\em Note:} |result| and |row| must not overlap. Otherwise the result
    is undefined.
!seealso
-------------------------------------------------------------------------- }}}*/

PTR zpermrow(PTR row, long *perm, PTR result)

{
    register FEL f;
    register long i;
    register long *p = (long *)perm;

    for (i = 1; i <= znoc; ++i)
    {
	f = zextract(row,i);
	zinsert(result,*p++,f);
    }
    return result;
}

